#!/usr/bin/perl

package MY_PACKAGE;
use strict;
sub help;
sub parse_bits;
use Getopt::Long;

sub usage { return <<EOF;

Usage: $0             Manipulate FPGA directly or program flash
       $0 <card id>   interactive mode
       $0 <card id> r <regi id>       Read regi
       $0 <card id> w <regi id> <val> Write regi (val in HEX, no '0x' prefix)
       $0 <card id> f <page>          Reload FPGA from <page>
       $0 <card id> p <page> <file>   Program RBF <file> into <page>

EOF
;
}

my $help;
GetOptions("help|h"          => \$help);
die usage if $help;


my $amroot = ($< == 0);

if(@ARGV <= 1) {
    print "Welcome to fpga.pl, by John Jacobsen for LBNL/IceCube.\n";
    print "This program lets you manipulate the FPGA directly.\n";
    if(! $amroot) {
        warn "Note: You must be root to write FPGA registers or program\n".
            "the DOR flash.\n";
    }

}

sub lasterr { my $l = `cat /proc/driver/domhub/lasterr`; chomp $l; return $l; }

my $le = lasterr;
if($le !~ /no error occurred since last query/) { 
    print "WARNING: lasterr proc file: $le\n";
}

my $card = shift;

while(!defined $card || ($card ne "all") &&
      ($card !~ /\d+/ || $card < 0 || $card > 15)) {
    print "Enter card ID (0..7) [0]: ";
    $card = <STDIN>; chomp $card;
    $card = 0 if $card eq "";
}

my $cmdline;
if(@ARGV > 0) {
    $cmdline = join " ", @ARGV;
}

my $regi; my $val;

my @pds;

if($card eq "all") {
    foreach my $card (0..7) {
	my $pd = "/proc/driver/domhub/card$card";
	push @pds, $pd if(-f "$pd/fpga-regs");
    }
    die "No cards found... did you insmod the driver?\n" if(scalar @pds == 0);
} else {
    $pds[0] = "/proc/driver/domhub/card$card";
    die "Can't find $pds[0]/fpga-regs.  Did you insmod the driver?\n" unless -f "$pds[0]/fpga-regs";
}

sub prime_for_read {
    my $pd   = shift;
    my $regi = shift;
    open PRIME, ">$pd/fpga-regs" || die "Can't open $pd/fpga-regs: $!\n";
    print PRIME "r $regi\n";
    close PRIME;
}

sub read_proc_file {
    my $pd = shift;
    open READ, "$pd/fpga-regs" || die "Can't open $pd/fpga-regs: $!\n";
    my $pf = <READ>; chomp $pf;
    close READ;
    return $pf;
}


sub write_regi {
    my $pd   = shift;
    my $regi = shift;
    my $val  = shift;
    open PRIME, ">$pd/fpga-regs" || die "Can't open $pd/fpga-regs: $!\n";
    print PRIME "w $regi $val";
    close PRIME;
}

sub load_fpga {
    my $pd   = shift;
    my $page = shift;
    open FPGA, ">$pd/fpga" || die "Can't open $pd/fpga: $!\n";
    print FPGA "reload $page\n";
    close FPGA;
}

sub program_flash {
    my $pd   = shift;
    my $page = shift;
    my $filename = shift;
    if(! -f $filename) {
	print "Can't find $filename.  Flash NOT programmed.\n";
	return;
    }
    open FLASH, ">$pd/flash$page" || do {
	print STDERR "Can't open $pd/flash$page: $!\n";
	return;
    };
    print FLASH "flash $page program\n" || die "Can't write flash page $page: $!\n";
    close FLASH;
    my $result = `cat $filename 2>&1 > $pd/flash$page`; chomp $result;
    die "Got '$result' while programming flash (flash locked?) FAIL.\n" 
	unless $result eq "";
    print "Programmed $filename into flash page $page.\n";
}

my $lastcmd;
while(1) {
    my $cmd;
    if(defined $cmdline) {
	$cmd = $cmdline;
    } else {
	print "Enter command: ";
	$cmd = <STDIN>; chomp $cmd;
    }
    if($cmd eq "" && $lastcmd =~ /r (\d+)/) {
	$cmd = $lastcmd;
    } else {
	$lastcmd = $cmd;
    }
    if($cmd =~ /r (\d+)/) {
	if($card eq "all") {
	    warn "Read mode for all not supported.\n";
	    exit if defined $cmdline;
	    next;
	}
	if(! $amroot) {
	    warn "Can't read registers if you aren't root, sorry.\n";
	    exit if defined $cmdline;
	    next;
	}
	$regi = $1;
	prime_for_read $pds[0], $regi;
	$val = read_proc_file $pds[0];
	if($val =~ /FPGA on card $card: register $regi = (\S+) \((\S+)\)/) {
	    my $bits = parse_bits($1);
	    print "CARD $card REG $regi -> $2 (decimal $1; bits $bits).\n";
	} else {
	    print "Unparseable proc file for card $card, regi $regi: $val.\n";
	}
    } elsif($cmd =~ /w (\d+) (\S+)/) {
        if($card eq "all") {
            warn "Write mode for all not supported.\n";
            next;
        }
        if(! $amroot) {
            warn "Can't write registers if you aren't root, sorry.\n";
	    exit if defined $cmdline;
            next;
        }
	$regi = $1;
	$val  = $2;
	printf "Writing 0x%s to register %d on $pds[0].\n", $val, $regi;
	write_regi($pds[0], $regi, $val);
	print "Done.\n";
    } elsif($cmd =~ /f (\d+)/) {
	if($1 < 0 || $1 > 3) {
	    print "Can't load FPGA: invalid page argument ($1).\n";
	} else {
	    foreach my $pd (@pds) {
		load_fpga $pd, $1;
		print "Loaded FPGA for $pd, page $1.\n";
# It seems one needs a delay after loading, else machine will
# crash (PCI core in firmware needs time to initialize???)
		sleep 2;
	    }
	    my $le = lasterr;
	    if($le !~ /no error occurred since last query/) {
		print "ERROR: $le\n";
	    } else {
		print "Done.\n";
	    }
	}
    } elsif($cmd =~ /p (\d+) (\S+)/) {
        if(! $amroot) {
            warn "Can't program DOR flash if you aren't root, sorry.\n";
            next;
        }
	if($1 < 0 || $1 > 3) {
	    print "Can't program flash page $1.\n";
	} else {
            foreach my $pd (@pds) {
		program_flash $pds[0], $1, $2;
		print "Programmed $2 into flash page $1 on $pd.\n";
	    }
	    print "Done.\n";
	}
    } elsif($cmd eq "q") {
	print "Have a nice day.\n";
	exit;
    } else {
	print help;
	print "New command: " unless defined $cmdline;
    }
    if(defined $cmdline) {
	last;
    } 
}

sub help {
    return <<EOF;

Valid commands:  
 * Read a register:   r X   (X is register number) 
 * Write a register:  w X Y (Y is a hex value, without leading '0x'
   Reload the FPGA:   f Z   (Z is flash page number)
 * Program flash:     p Z filename (Z is flash page for filename)
   Quit:              q     

   Typing return repeats the last command.
 * == must be root to perform operation
EOF
			   ;
}

sub parse_bits { # Pull out bits from 32 bit value 
    my $val = shift;
    return "none" if(!defined $val || $val == 0);
    my $retval = "";
    my $gotone = 0;
    foreach my $bit (0..31) {
	if($val & 0x01) {
	    $retval .= ", " if $gotone;
	    $retval .= $bit;
	    $gotone++;
	}
	$val >>= 1;
    }
    return $retval;
}
	
__END__

