#!/usr/bin/perl

package MY_PACKAGE;
use strict;
use Getopt::Long;
sub frev;

sub usage { return <<EOF;

Usage: $0 -rev0 <rev0.rbf> -rev1 <rev1.rbf> -page <page>
       $0 -pci <pci_xyz.rbf> -page 3
EOF
;
}

my $help; my $rev0; my $rev1; my $pci; my $page;

GetOptions("help|h"    => \$help,
	   "rev0|0=s"  => \$rev0,
	   "rev1|1=s"  => \$rev1,
	   "pci|i=s"   => \$pci,
	   "page|p=i"  => \$page) || die usage;


die "You must be root to run $0.\n" unless $< == 0;

die "Usage: $0 <firmware>\n" unless defined $rev0 || defined $rev1 || defined $pci;
die "Can't find file: $rev0\n" if(defined $rev0 && ! -f $rev0);
die "Can't find file: $rev1\n" if(defined $rev1 && ! -f $rev1);
die "Can't find file: $pci\n"  if(defined $pci  && ! -f $pci);

if(defined $pci && $page != 3) {
    die "PCI firmware can only go on page 3!\n";
}

$page = 1 unless defined $page;

die "Page must be 1 or 2.\n" unless $page eq "1" || $page eq "2" || $page eq "3";
die "Page 3 install must be PCI only!!!\n" if $page eq "3" && !(defined $pci);

my @cards  = </proc/driver/domhub/card*>;
my %revs;
my %fw;
my $cstr   = "";
for(my $i=0; $i < @cards; $i++) {
    if($cards[$i] =~ /card(\d+)/) {
	my $card = $1;
	$cstr .= "$card";
	$revs{$card} = `cat $cards[$i]/rev`; chomp $revs{$card};
	if($revs{$card} == 1) {
	    die "Card $card needs DOR Rev 1 (or PCI core)!\n" unless defined $rev1 || defined $pci;
	    my $fw = (defined $pci)?$pci:$rev1;
	    print "Card $card is DOR Rev 1. Page $page <- $fw\n";
	    $fw{$card} = $fw;
	} elsif($revs{$card} == 0) {
	    next if defined $pci;
            die "Card $card needs DOR Rev 0!\n" unless defined $rev0;
            print "Card $card is DOR Rev 0. Page $page <- $rev0\n";
	    $fw{$card} = $rev0;
        } else {
	    die "Card $card has unknown DOR revision ($revs{$card}).\n";
	}
    }
    $cstr .= ", " if $i != @cards-1;
}

die "No DOR cards installed, or did you forget to insmod the driver?\n"
    unless @cards > 0;

print "Proceed with firmware installation? ";
my $yn = <STDIN>;
die "Ok.\n" unless $yn =~ /^y/i;

my $numinstalled = 0;

sub unprotectFlash {
    my $card   = shift; die unless defined $card;
    my $protpf = shift; die unless defined $protpf;
    print "Unprotecting flash on card $card...\n";
    open PROT, ">$protpf" or die "Can't open $protpf: $!\n";
    print PROT "unprotect\n";
    close PROT;
}

sub protectFlash {
    my $card   = shift; die unless defined $card;
    my $protpf = shift; die unless defined $protpf;
    print "Protecting flash on card $card...\n";
    open PROT, ">$protpf" or die "Can't open $protpf: $!\n";
    print PROT "protect\n";
    close PROT;
}

sub installFlashPage {
    my $cardproc = shift; die unless defined $cardproc;
    my $card     = shift; die unless defined $card;
    my $page     = shift; die unless defined $page;
    my $pf       = shift; die unless defined $pf;
    my $protpf   = shift; die unless defined $protpf;
    my $fw       = shift; die unless defined $fw;
    my $rev      = $revs{$card};
    my $frev     = frev $cardproc;
    if($frev =~ /corrupt/i) {
	print "WARNING!!! Card $card has corrupt FPGA, skipping...\n";
	next;
    }
    next unless defined $fw;
    $|++;
    open FLASH, ">$pf";
    # print "cmd... ";
    if(defined $pci && $page == 3) {
	unprotectFlash($card, $protpf);
    }
    print "INIT($card) ";
    print FLASH "flash $page program\n";
    close FLASH;

    print "PGRM($card) ";
    print `cat $fw > $pf`;

    if(defined $pci && $page == 3) {
	protectFlash($card, $protpf);
    }

    print "DONE($card). ";
}

sub reloadFPGA {
    my $cardproc = shift; die unless defined $cardproc;
    my $card = shift;     die unless defined $card;
    my $page = shift;     die unless defined $page;
    print "Card $card ($cardproc) reload...";
    my $ff = "$cardproc/fpga";
    open FPGA, ">$ff" or die "Can't open $ff: $!\n";
    print FPGA "reload $page\n";
    close FPGA;
    print "Done.\n";
}

print "Programming all cards...\n";

my @kids;
foreach my $cardproc (@cards) {
    $cardproc =~ /card(\d+)/;
    my $card = $1;
    die unless defined $card;
    my $pf = "$cardproc/flash$page";
    my $protpf = "$cardproc/flash-protect";
    my $fw = $fw{$card};
    next unless defined $fw;
    print "$fw -> $pf\n";
    my $kid = fork;
    die "Couldn't fork: $!\n" unless defined $kid;
    if($kid == 0) { # I'm the subprocess
	select undef, undef, undef, 0.3;
	installFlashPage($cardproc, $card, $page, $pf, $protpf, $fw);
	exit;
    } else {
	push @kids, $kid;
    }
    $numinstalled++;
}


use POSIX ":sys_wait_h";
my $kid;
while(1) {
    $kid = waitpid(-1, &WNOHANG);
    if($kid == -1) {
	last;
    } else {
	select(undef,undef,undef,0.01);
    }
}

print "\n\nProgramming done.\n";

print "Reloading FPGA on all cards...\n";

foreach my $cardproc (@cards) {
    $cardproc =~ /card(\d+)/;
    my $card = $1;
    die unless defined $card;
    reloadFPGA($cardproc, $card, $page);
}    

print "New configuration:\n".`fvers.pl`;

if(!defined $pci && $numinstalled > 0) {
    my $dfpfile = "/usr/local/dor/conf/default_fpga_page";
    print <<EOF;
Creating default FPGA configuration file so that page $page
is automatically loaded by the driver when you reboot....

EOF
    ;
    open DEFPAGE, ">$dfpfile" || die "Can't write $dfpfile: $!\n";
    print DEFPAGE "$page\n";
    close DEFPAGE;
}

print "Done: $numinstalled cards updated.\n";
exit;

sub frev {
    my $card = shift;
    my $rev = `cat $card/rev`; chomp $rev;
    my @fpga = `cat $card/fpga`;
    my $fw;
    for(@fpga) {
	if(/FREV/) {
	    if(/FREV\s+0x([0-9a-f][0-9a-f])0(\w)(\w\w)(\w\w)/) {
		my $pcirev = hex($1);
		$fw = $2.($rev==1?$3:hex($3)).chr(hex($4));
		return $fw;
	    } else {
		return "CORRUPT($_)\n";
	    }
	}
    }
}


__END__

