/* New, better implementation of circular I/O queues, based on the 
   ackq and retx queues.

   John Jacobsen, NPX Designs, Inc. jacobsen@npxdesigns.com
   May, 2004 */


#ifndef __DH_IOQ__
#define __DH_IOQ__

#define NIOQBITS      7
#define NIOQENTRIES   (1<<NIOQBITS)
#define IOQMASK       (NIOQENTRIES-1)

/** Message data type */
typedef struct dhimsg_t {
  u32 nbytes;
  int synfin; /* Set to 1 to indicate end of message; else set to 0 */
  unsigned char data[DH_MAX_SW_PKTSIZ];
} ioqMsg;

typedef struct ioqTypeSTruct {
  ioqMsg ioq[NIOQENTRIES];
  unsigned ioqHead, ioqTail;
  wait_queue_head_t wq;
} ioQueue;

static inline int getIOQHead(ioQueue *q) { return q->ioqHead&IOQMASK; }
static inline int getIOQTail(ioQueue *q) { return q->ioqTail&IOQMASK; }

static inline void putIOQ(ioQueue *q, ioqMsg *m) {
  /* Could be more efficient here by copying only significant data bytes rather than 
     whole message, but memcpy should be fast */
  memcpy(&(q->ioq[getIOQHead(q)]), m, sizeof(ioqMsg));
  q->ioqHead++;
}

static inline int putIOQ_from_user(ioQueue *q, const char * data, int nbytes, int synfin) {
  /* WARNING: Can sleep (if copy_from_user page faults) */
  ioqMsg * m = &(q->ioq[getIOQHead(q)]);
  m->nbytes = nbytes <= DH_MAX_SW_PKTSIZ ? nbytes : DH_MAX_SW_PKTSIZ;
  if(nbytes > DH_MAX_SW_PKTSIZ) 
    printk(KERN_WARNING "dh: putIOQ_from_user: bounds check failed!\n");
  m->synfin = synfin;
  int ret = copy_from_user(m->data, data, m->nbytes);
  q->ioqHead++;
  return ret;
}

static inline int ioQisEmpty(ioQueue *q) { return q->ioqHead == q->ioqTail; }

static inline ioqMsg * getIOQ(ioQueue *q) {
  /* You have to use the message pointed to by ret soon, if it gets stale 
     (e.g, waiting for copy_to_user) then you have a race.  Best to copy
     the resulting packet into another buffer before copy_to_user */

  if(ioQisEmpty(q)) return NULL;
  ioqMsg * ret = &(q->ioq[getIOQTail(q)]);
  q->ioqTail++;
  return ret;
}

static inline unsigned nextIOQEntrySize(ioQueue *q) {
  return q->ioq[getIOQTail(q)].nbytes;
}

static inline int nIOQEntries(ioQueue *q) { return q->ioqHead - q->ioqTail; }
static inline int nIOQSlots(ioQueue *q) { return NIOQENTRIES - nIOQEntries(q); }
static inline int ioQisFull(ioQueue *q) { return NIOQENTRIES == nIOQEntries(q); }
static inline void emptyIOQ(ioQueue *q) { q->ioqHead = q->ioqTail = 0; }

static inline int ioQhasSynfin(ioQueue *q) { 
  /* Linear search for syn/fin (complete message).  Could tag queue when found,
     but this is simple and not our performance bottleneck */
  unsigned i;
  for(i=q->ioqTail; i < q->ioqHead; i++) {
    //printk(KERN_INFO "ioQhasSynfin: tail=%d head=%d synfin[i]=%d\n", 
    //q->ioqTail, q->ioqHead, q->ioq[i&IOQMASK].synfin);
    if(q->ioq[i&IOQMASK].synfin) return 1;
  }
  return 0;
}

static ioQueue * newIOQ(void) {
  ioQueue * q = (ioQueue *) kmalloc(sizeof(ioQueue), GFP_ATOMIC);
  if(q != NULL) {
    emptyIOQ(q);
    init_waitqueue_head(&q->wq);
  }
  return q;
}

#endif
