/*  dh.c -- Linux device driver for the IceCube DOM Hub
    $Revision: 1.381 $ -- see version.h for official release tag.

    John Jacobsen, NPX Designs, Inc., for LBNL/IceCube
    jacobsen@npxdesigns.com
    Started November, 2002
    Last updated $Date: 2007/07/10 15:17:47 $

    This is the device driver for the DOM Hub, the communications
    interface to the Digital Optical Module (DOM) sensors in the
    IceCube experiment.  The DOM Hub is a Linux PC with up to 8
    custom DOM Readout (DOR) cards.  This driver allows one to control
    the DOR cards and communicate through them to up to 64 DOMs per
    DOM Hub.

    Refer to Rubini and Corbet, "Linux Device Drivers," published
    by O'Reilly (2nd Ed.), and to the DOR Driver Documentation on 
    the Icecube Docushare site:
    http://docushare.icecube.wisc.edu/docushare/dsweb/View/Collection-511

*/

#include <linux/init.h>
#include <linux/module.h>  /* Includes linux/config.h needed for PCI */
#ifndef KERNEL2_4
#include <linux/kernel.h>
#endif
#include <linux/pci.h>
#include <linux/delay.h>   /* For udelay() */
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/wait.h>
#include <linux/timer.h>
#include <linux/poll.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>   /* For /proc I/O */
#include <linux/interrupt.h> /* For tasklets */
#include <linux/mm.h>        /* For kmalloc flags */
#ifndef KERNEL2_4
#include <linux/jiffies.h>   /* For time_before */
#endif
#include <asm/semaphore.h>   /* Semaphore to protect concurrent time calibrations */
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/msr.h>

#include "dh.h"
#include "dh_tcalib.h" /* Time calibration stuff shared w/ tcaltest.c */
#include "dhswpkt.h"
#include "ackq.h"
#include "retx.h"
#include "ioq.h"
#include "dbgq.h"
#include "domstats.h"
#include "dhbench.h" /* Driver benchmarking */
#include "version.h"

/* ----------- GLOBALS, FUNCTION PROTOTYPES, CONSTANTS -------------- */

#ifndef CONFIG_PCI
#  error "Driver dh needs PCI support to work with DOM Hub/DOR cards."
#endif

MODULE_AUTHOR      ("Dr. John E. Jacobsen, NPX Designs, Inc., npxdesigns.com"
                    "\nCopyright (C) 2002-2006 Lawrence Berkeley National Laboratory.");
MODULE_DESCRIPTION ("IceCube DOMHub driver");
MODULE_LICENSE     ("GPL");
MODULE_SUPPORTED_DEVICE("IceCube DOM Hub Readout Adapter (DOR) Card.");

/* Clock select variables */
#define DH_CLKSEL_INTERNAL 0
#define DH_CLKSEL_EXTERNAL 1
int dh_clksel[DH_MAX_CARDS];         /* Store indicater of whether to use internal or external
		    			  clock.  Set to default in dh_init(). */
int dh_excluded[DH_MAX_CARDS];

/* Module parameter variable declarations */
static int dh_mode           = DH_NORMAL;      /* Indicates which mode driver is running in */
int  parm_rx_dma             = 1;              /* RX DMA (1) or no (0) */
int  parm_tx_dma             = 1;              /* TX " " */

int  parm_block              = 0;              /* Blocking mode -- 0 (nonblock) by default */
int  parm_fpga_page          = 0;              /* Load this page by default (0->use pwr on
						     default page 0 without reload) */
int  parm_unlock_pages       = 0;              /* Unlocks writes to DOR flash page 0 */
int  parm_auto_reset         = 1;              /* Do automatic comms reset if hardware timeout */
int  parm_fw_version_check   = 0;              /* 0=Short circuit firmware version check -- 
						    for testing only... [JEJ - removed, RPM takes
						    care of this now */
int  parm_verbose            = 0;              /* Verbose level: 0=silent(except errors)
						    1=quiet (default), >1=noisy */
int  parm_clock_source       = DH_CLKSEL_INTERNAL; /* Default clock source */
char * parm_exclude          = "";             /* List of devices to exclude, e.g., "023457" */

#ifdef KERNEL2_4
# define INTPARM(p) MODULE_PARM(p, "i")
# define CHRPARM(p) MODULE_PARM(p, "s")
#else
# define INTPARM(p) module_param(p, int, S_IRUGO)
# define CHRPARM(p) module_param(p, charp, S_IRUGO)
#endif

INTPARM(parm_block);
INTPARM(parm_rx_dma);
INTPARM(parm_tx_dma);
INTPARM(parm_fpga_page);
INTPARM(parm_unlock_pages);
INTPARM(parm_verbose);
INTPARM(parm_auto_reset);
INTPARM(parm_fw_version_check);
CHRPARM(parm_exclude);
INTPARM(parm_clock_source);

#ifdef KERNEL2_4
MODULE_PARM_DESC   (parm_block, 
		    "1 for blocking mode, 0 for nonblock (default)");
MODULE_PARM_DESC   (parm_rx_dma,
		    "1 for RX DMA (default), 0 for normal FIFO drain");
MODULE_PARM_DESC   (parm_tx_dma,
                    "1 for TX DMA, 0 for normal FIFO fill (default)");
MODULE_PARM_DESC   (parm_fpga_page,
		    "0 for default, 1..3 for alternate flash pages to load at insmod time");
MODULE_PARM_DESC   (parm_unlock_pages,
                    "Use 1 to allow flash page unlock.  DO NOT USE unless you know EXACTLY what"
		    " you are doing.");
MODULE_PARM_DESC   (parm_verbose,
		    "verbose level, 0=silent, 1=quiet (default), >1=noisy");
MODULE_PARM_DESC   (parm_auto_reset, "0=turn off automatic comms reset after hardware timeout");
MODULE_PARM_DESC   (parm_fw_version_check, "0=turn off firmware version check");
MODULE_PARM_DESC   (parm_exclude, "List of DOR cards to exclude from the driver");
MODULE_PARM_DESC   (parm_clock_source, "Clock source select (0=internal 1=external)");
#endif

#define v()    (parm_verbose > 0)    /* Anything but quiet */
#define vprintk     v() && printk

#define DH_MEGA_COMMENTS
#undef DH_MEGA_COMMENTS
#ifdef DH_MEGA_COMMENTS
# define vv()   (parm_verbose > 1)   /* Somewhat noisy */
# define vvv()  (parm_verbose > 2)   /* Really noisy */
# define vvprintk   vv() && printk
# define vvvprintk vvv() && printk
#else
# define vv() (0)
# define vvv() (0)
# define vvprintk(...)
# define vvvprintk(...)
#endif

#define  DH_TEST_INC_PAT /* Turn this on if you want to verify that rec'vd packets
			    match the readwrite incrementing test pattern in 3 
			    different code sections in the driver */
#undef   DH_TEST_INC_PAT

/* Turn this on to only have interrupt handler handle one message
   at a time */
#define DH_SHORT_INTERRUPTS
#undef  DH_SHORT_INTERRUPTS
#define SKIP_RETX
#undef  SKIP_RETX

#define DH_CLOSETIMEOUT (HZ*2)  /* Wait 2 seconds to drain queues before dying on close */
#define DH_OPENTIMEOUT  (HZ*30) /* Wait 30 seconds for open to succeed */

#define DH_SIMULATE_TIMEOUTS
#undef  DH_SIMULATE_TIMEOUTS
#define DH_MSGS_PER_TIMEOUT 1000

int dh_do_sim_timeout[DH_DOMHUB_DOMS];
int dh_comms_reset_in_progress[DH_DOMHUB_DOMS];

#define pprintk(fmt,args...) /* Makes printk easy to comment out when done debugging */
//#define pprintk printk  // Choose this one if you want to have all the debug statements 

#define USERSPACE_BUF   1 /* Indicates that a buffer belongs to user space */
#define KERNELSPACE_BUF 2 /* Indicates that a buffer belongs to kernel space */

/* Modes for dh_drain_fifo; LOUD reports contents to syslog */
#define DH_QUIET        0
#define DH_LOUD         1 

#define DH_STATE_RXINT  DH_BITN(0)  /* CTRL  10000 */
#define DH_STATE_TXINT  DH_BITN(1)  /* CTRL  20000 */
#define DH_STATE_READ   DH_BITN(2)  /* CTRL  40000 */
#define DH_STATE_WRITE  DH_BITN(3)  /* CTRL  80000 */
#define DH_STATE_TCALIB DH_BITN(4)  /* CTRL 100000 */

/* Comment this out to get rid of diagnostic messages with each 
    write to an FPGA register */
#define DH_VERBOSE_WRITES 3000
#undef  DH_VERBOSE_WRITES

/* Macro for printing printable/unprintable bytes */
#define DH_PR_BYTE(b) (((b)>31)&&((b)<127)?b:'X')

/* Macro for test pattern to read/write SRAM */
#define DH_SHIFT_INVERT(i) ( (((u32) i) & 1) ? ( 0xFFFFFFFF - (1<<((i)-1)%32)) : ( 1<<((i)%32)) )

/* Macro for testing whether timeout has exceeded (or jiffies rolled over) */
#ifdef KERNEL2_4
# define WITHIN_TOUT(timeout,jiff_0) ((jiffies-jiff_0) < timeout && (jiffies >= jiff_0))
#else
# define WITHIN_TOUT(timeout,jiff_0) time_before(jiffies, jiff_0+((unsigned long) timeout))
#endif

/* Input and output message queues - must be initialized and deallocated for each DOM */
static ioQueue  * inq[DH_DOMHUB_DOMS];
static ioQueue  * outq[DH_DOMHUB_DOMS];
static dbgQueue * dbgq[DH_DOMHUB_DOMS];
static ackQueue ackq[DH_DOMHUB_DOMS];

/* Retxb is an array of buffers, also treated as an array of statically allocated linked lists
   to make for easy ordered traversal */

static rtxMsg retxb[DH_DOMHUB_DOMS][NRETX];
static rtxMsg *retxbhd[DH_DOMHUB_DOMS]; /* Head of linked list */
int           nretxb[DH_DOMHUB_DOMS]; /* Num in list */
int           nretxb_bytes[DH_DOMHUB_DOMS];
int           nretxdiag[DH_DOMHUB_DOMS];
int           retx_triggered[DH_DOMHUB_DOMS]; /* Tag RETX in progress */

#define MAXRETXPRINTKS 10000

static rtxQueue retxq[DH_DOMHUB_DOMS];
static wait_queue_head_t retxbwq[DH_DOMHUB_DOMS];
static wait_queue_head_t ackqwq[DH_DOMHUB_DOMS];
static wait_queue_head_t domstatwq[DH_DOMHUB_DOMS];


static int dh_panic = 0; /* Set this nonzero to disable all interrupts in an emergency. */

static ioqMsg * dh_tx_scratch_msg[DH_DOMHUB_DOMS];
static ioqMsg * dh_rx_scratch_msg[DH_DOMHUB_DOMS];
static u32 * txpkt[DH_DOMHUB_DOMS];
static u32 * rxpkt[DH_DOMHUB_DOMS];

static struct dsStruct {
  DomStatsPkt stats;
  int have_new;
} domstats[DH_DOMHUB_DOMS];

static unsigned short txseq[DH_DOMHUB_DOMS];
static unsigned short rxseq[DH_DOMHUB_DOMS];
static unsigned txfree[DH_DOMHUB_DOMS];

#define DH_DO_POLLUTE
#ifdef  DH_DO_POLLUTE
/* Code for "polluting" (packet-corrupting) proc file */
static int dh_pollute[DH_DOMHUB_DOMS];
#endif

/* ReTX timers for each card */
static struct timer_list dh_timer_struct[DH_MAX_CARDS];
/* Basic time scale for RETXes, IC/CI, etc. (see kernel timer code for details) */
#define DH_TIMER_MSEC 200                                 /* msec */
#define DH_TIMER_JIF ((DH_TIMER_MSEC * HZ)/1000)          /* Convert to jiffies */
#define DH_TIMER_JIFFIES (DH_TIMER_JIF==0?1:DH_TIMER_JIF) /*At least 1 jif*/

#define DH_TIMER_RESTART_JIF  (DH_TIMER_JIFFIES/4)
#define DH_TIMER_UNSTICK_JIF  (DH_TIMER_JIFFIES/2)
#define DH_TIMER_RETX_JIF     DH_TIMER_JIFFIES

enum {
  DH_STATE_START,    /* Start */
  DH_STATE_WAITIC,   /* Wait for IC */
  DH_STATE_FLUSHIC,  /* Wait for extra ICs */
  DH_STATE_SENDCI,   /* Got IC, send CI */
  DH_STATE_SENDCI_NOWAIT, /* Got CI, send CI and go to connecting... */
  DH_STATE_WAITCI,   /* Sent CI, wait for CI */
  DH_STATE_CONNECTING, /* Last transition before connected - resets/enables */
  DH_STATE_CONNECTED /* Connected */
};

static int dh_conn_state[DH_DOMHUB_DOMS];
static int dh_conn_w_jif[DH_DOMHUB_DOMS];
static int dh_conn_flush_jif[DH_DOMHUB_DOMS];
static int dh_conn_startic_jif[DH_DOMHUB_DOMS];
static int dh_conn_waitci_jif[DH_DOMHUB_DOMS];
static int dh_w_cis_sent[DH_DOMHUB_DOMS];
static int dh_w_cis_reqd[DH_DOMHUB_DOMS];
#define MIN_W_JIFS 0 /* If non-zero, hang around in STATE_W until this time has elapsed
			to make sure no ICs arrive */
#define MIN_FLUSH_JIFS HZ/50   /* Wait 20 msec to make sure ICs are done */
#define MIN_WAITIC_JIF HZ      /* Wait up to 1 sec for first IC before TX IC */
#define MIN_WAITCI_JIF 2*HZ    /* Wait up to 2 seconds for CI to arrive before sending IC */
#define NUM_W_CIS 10  /* Default is to make sure at least 10 W-states pass 
			 before going to CONNECTED */

static wait_queue_head_t connectwq[DH_DOMHUB_DOMS];

/* Variables for comms reset state */
#define DH_COMMS_W_DOMC   1 /* Waiting for DOMC, bit 8+ich */
#define DH_COMMS_W_DOMS   2 /* Waiting for DOMS, bit 0+ich */
#define DH_COMMS_READY    3 /* Communications reset done */
static int dh_comms_state[DH_DOMHUB_DOMS];
static wait_queue_head_t comms_reset_wq[DH_DOMHUB_DOMS];

/* Wait queue for block-on-open */
static wait_queue_head_t device_in_use_wq[DH_DOMHUB_DOMS];

static struct     dh_tcalib_t * dh_tcalib_recs[DH_DOMHUB_DOMS]; /* Stores time calibration data*/
/* Protect dh_tcalib_recs from races between proc callbacks and interrupt handlers */
static int        dh_tcalib_avail[DH_DOMHUB_DOMS];

/* Tag dis-allowed firmware */
static int dh_firmware_mismatch[DH_MAX_CARDS];

/* Wait queues if device is needed and dh_opened false for that device */
static wait_queue_head_t file_open_waiter[DH_DOMHUB_DOMS];

/* Marks every opened device file as opened or no */
static int dh_opened[DH_DOMHUB_DOMS]; 
static int dh_is_communicating[DH_DOMHUB_DOMS];

/* Count number of messages read from each device */
static unsigned long dh_msgsread[DH_DOMHUB_DOMS]; 
static unsigned long dh_pktsread[DH_DOMHUB_DOMS];
static unsigned long dh_acksread[DH_DOMHUB_DOMS];
static unsigned long dh_badhdr[DH_DOMHUB_DOMS];
static unsigned long dh_badseq[DH_DOMHUB_DOMS];
static unsigned long dh_ctrlsent[DH_DOMHUB_DOMS];
static unsigned long dh_ctrlrecv[DH_DOMHUB_DOMS];
static unsigned long dh_nhdtouts[DH_DOMHUB_DOMS];
static unsigned long dh_connects[DH_DOMHUB_DOMS];
static unsigned long dh_icsent[DH_DOMHUB_DOMS];
static unsigned long dh_icrecv[DH_DOMHUB_DOMS];
static unsigned long dh_cisent[DH_DOMHUB_DOMS];
static unsigned long dh_cirecv[DH_DOMHUB_DOMS];
static unsigned long long dh_bytesread[DH_DOMHUB_DOMS];

/* Count number of messages written to each device */
static unsigned long dh_msgswritten[DH_DOMHUB_DOMS]; 
static unsigned long dh_pktswritten[DH_DOMHUB_DOMS];
static unsigned long dh_ackwritten[DH_DOMHUB_DOMS];
static unsigned long dh_resent[DH_DOMHUB_DOMS];
static unsigned long long dh_byteswritten[DH_DOMHUB_DOMS];

static unsigned long dh_total_msgs_read;    /* Count total messages read from driver */
static unsigned long dh_total_msgs_written; /* Count total messages written to driver */


/* Name string length for fancy proc dir entries */
#define PDE_NAME_LEN 80

/* These are the file names for the various proc dir entries */
#define DH_TOP_DIR          "driver/domhub"
#define DH_DRIVERSTAT       "status"
#define DH_LASTERR          "lasterr"
#define DH_REVISION         "revision"
#define DH_PWRALL           "pwrall"
#define DH_BUFSIZ           "bufsiz"
#define DH_BLOCKING         "blocking"
#define DH_BENCH            "bench"
#define DH_VERBOSE          "verbose"
#define DH_CARD_FPGA        "fpga"
#define DH_CARD_FPGA_REGS   "fpga-regs"
#define DH_CARD_PCI         "pci"
#define DH_CARD_1PPS        "1pps"
#define DH_CARD_CLKSEL      "clksel"  
#define DH_CARD_SYNCGPS     "syncgps"
#define DH_CARD_REV         "rev"
#define DH_CARD_TEST_LOG    "test-log"
#define DH_CARD_FLPROTECT   "flash-protect"
#define DH_CARD_PWR_CHECK   "pwr_check"
#define DH_CARD_AUTODAC     "autodac"
#define DH_CARD_DACMAX      "dacmax"
#define DH_CARD_MINCLEV     "minclev"
#define DH_CARD_MAXCLEV     "maxclev"
#define DH_CARD_RDELAY      "rdelay"
#define DH_CARD_SDELAY      "sdelay"
#define DH_CARD_THRESH      "thresh"
#define DH_CARD_MINVOLTS    "minvolts"
#define DH_CARD_MAXVOLTS    "maxvolts"
#define DH_CARD_MINCUR      "mincur"
#define DH_CARD_MAXCUR      "maxcur"
#define DH_PAIR_PWR         "pwr"
#define DH_PAIR_IS_PLUGGED  "is-plugged"
#define DH_PAIR_CURRENT     "current"
#define DH_PAIR_VOLTAGE     "voltage"
#define DH_PAIR_PWR_CHECK   "pwr_check"
#define DH_DOM_ID           "id"
#define DH_DOM_SOFTBOOT    "softboot"
#define DH_DOM_ISNOTCB      "is-not-configboot"
#define DH_DOM_ISCOMM       "is-communicating"
#define DH_DOM_ISCONN       "is-connected"
#define DH_DOM_TCALIB       "tcalib"
#define DH_DOM_COMSTAT      "comstat"
#define DH_DOM_DOMSTAT      "domstat"
#define DH_DOM_BENCH        "bench"
#define DH_DOM_POLLUTE      "pollute"
#define DH_DOM_RXDT         "rxdt"
#define DH_PSEUDO_INTR      "pseudo-interrupts"


/* fancy_pde_entry is a structure which contains a struct
    proc_dir_entry (which is what the kernel needs to implement /proc
    files) but which also contains its file name (char string) and
    whether or not it was successfully registered with the
    kernel/filesystem, and, therefore, should be removed at clean-up
    time 
*/
struct fancy_pde_entry {
  struct proc_dir_entry *pde;
  char name[PDE_NAME_LEN];
  int regi;
};

/* Top level proc directory, under which all the dh proc files live */
static struct fancy_pde_entry fpde_top_dir;

static struct fancy_pde_entry 
  fpde_driverstat,                    /* Driver status snapshot */
  fpde_revision,                      /* Version string */
  fpde_lasterr,                       /* Last error which occured */
  fpde_pseudo_intr,                   /* Pseudo-interrupt file for testing intr handler */
  fpde_pwrall,                        /* Power on/off ALL channels on the Hub */
  fpde_bufsiz,                        /* Buffer size: min to read, max to write */
  fpde_blocking,                      /* Proc file to select default block behav. for dev files */
  fpde_bench,                         /* Driver time profiling statistics */
  fpde_verbose,                       /* Verbose level proc file */
  fpde_card_dirs[DH_MAX_CARDS],       /* Card-wise directories */
  fpde_card_fpga[DH_MAX_CARDS],       /* FPGAs: write to load, read to get status */
  fpde_card_fpga_regs[DH_MAX_CARDS],  /* Read/write FPGA registers */ 
  fpde_card_flprotect[DH_MAX_CARDS],  /* Flash protect/unprotect */
  fpde_card_pci[DH_MAX_CARDS],        /* PCI configuration register display */
  fpde_card_1pps[DH_MAX_CARDS],       /* For initialization of DOR timer: setup/success */
  fpde_card_clksel[DH_MAX_CARDS],     /* Select internal/external timer */
  fpde_card_syncgps[DH_MAX_CARDS],    /* GPS 1pps/ASCII time string/DOR clock sync function */
  fpde_card_rev[DH_MAX_CARDS],        /* DOR Revision number */
  fpde_card_test_log[DH_MAX_CARDS],   /* dor-test test log */
  fpde_card_pwr_check[DH_MAX_CARDS],  /* Check power on all channels */
  fpde_card_autodac[DH_MAX_CARDS],    /* Comms params... */
  fpde_card_dacmax[DH_MAX_CARDS],
  fpde_card_minclev[DH_MAX_CARDS],
  fpde_card_maxclev[DH_MAX_CARDS],
  fpde_card_rdelay[DH_MAX_CARDS],
  fpde_card_sdelay[DH_MAX_CARDS],
  fpde_card_thresh[DH_MAX_CARDS],
  fpde_card_minvolts[DH_MAX_CARDS],
  fpde_card_maxvolts[DH_MAX_CARDS],
  fpde_card_mincur[DH_MAX_CARDS],
  fpde_card_maxcur[DH_MAX_CARDS],
  fpde_card_flash[DH_MAX_CARDS][DH_FLASH_PAGES],        /* For R/W flash pages */
  fpde_pair_dirs[DH_MAX_CARDS][DH_PAIRS_PER_CARD],      /* Dirs. for DOM pairs */
  fpde_pair_pwr[DH_MAX_CARDS][DH_PAIRS_PER_CARD],       /* Power on/off for pairs */
  fpde_pair_pwr_check[DH_MAX_CARDS][DH_PAIRS_PER_CARD], /* Power check for pairs */
  fpde_pair_plugged[DH_MAX_CARDS][DH_PAIRS_PER_CARD],   /* Show cable status */
  fpde_pair_current[DH_MAX_CARDS][DH_PAIRS_PER_CARD],   /* Cable current draw */
  fpde_pair_voltage[DH_MAX_CARDS][DH_PAIRS_PER_CARD],   /* Cable voltage */
  fpde_dom_dirs[DH_MAX_CARDS][DH_PAIRS_PER_CARD][DH_DOMS_PER_PAIR],   /* Dirs. for DOMs */
  fpde_dom_tcalib[DH_MAX_CARDS][DH_PAIRS_PER_CARD][DH_DOMS_PER_PAIR], /* time calib */
  fpde_dom_comstat[DH_MAX_CARDS][DH_PAIRS_PER_CARD][DH_DOMS_PER_PAIR], /* DOM status 
									  (DOR message counters)*/
  fpde_dom_domstat[DH_MAX_CARDS][DH_PAIRS_PER_CARD][DH_DOMS_PER_PAIR], /* DOM status 
									  (from DOM itself) */
  fpde_dom_id[DH_MAX_CARDS][DH_PAIRS_PER_CARD][DH_DOMS_PER_PAIR], /* DOM ID */
  fpde_dom_bench[DH_MAX_CARDS][DH_PAIRS_PER_CARD][DH_DOMS_PER_PAIR], /* time profiling */
  fpde_dom_pollute[DH_MAX_CARDS][DH_PAIRS_PER_CARD][DH_DOMS_PER_PAIR], /* TX packet pollution */
  fpde_dom_rxdt[DH_MAX_CARDS][DH_PAIRS_PER_CARD][DH_DOMS_PER_PAIR], /* RX delta-Ts */
  fpde_dom_softboot[DH_MAX_CARDS][DH_PAIRS_PER_CARD][DH_DOMS_PER_PAIR], /* DOM softboot */
  fpde_dom_iscomm[DH_MAX_CARDS][DH_PAIRS_PER_CARD][DH_DOMS_PER_PAIR], /* DOM communicating ok/not ok */
  fpde_dom_isconn[DH_MAX_CARDS][DH_PAIRS_PER_CARD][DH_DOMS_PER_PAIR], /* DOM "connected" */
  fpde_dom_isnotcb[DH_MAX_CARDS][DH_PAIRS_PER_CARD][DH_DOMS_PER_PAIR]; /* DOM in configboot */
/* Struct to pass the card number or other data to proc callbacks -  */
struct fpde_data {
  int cardnum; /* 0..7 */
  int pairnum; /* 0..3 */
  int domnum; /* 0 or 1 */
};

/* Same as fpde_data, but for flash operations */
struct flash_data {
  int cardnum; /* 0..7 */
  int flashnum; /* 0..3 */
  unsigned long fpos;
};

#define REQUEST_STR_LEN  512

/* Data for FPGA register read/write */
struct proc_fpga_data { 
  int cardnum;
  unsigned char sim_regs[DH_NUM_REGIS];
  int register_to_read;
  int read_request_given;
  char request_str[REQUEST_STR_LEN];
};

/* Write register copies (all driver modes). */
static u32 dh_write_regi_copies[DH_MAX_CARDS][DH_MAX_REGIS];

/* Simulated read register values (sim_* mode only) */
static u32 dh_sim_read_regi[DH_MAX_CARDS][DH_MAX_REGIS];

/* Variable to save power state of simulated wire pairs */
static int dh_sim_pwr_state[DH_MAX_CARDS][DH_PAIRS_PER_CARD];

/* State variable for timer initialization */
int dh_utc_timer_status[DH_MAX_CARDS];
#define DH_UTC_TIMER_IDLE    0
#define DH_UTC_TIMER_ERR    -1
#define DH_UTC_TIMER_START   1
#define DH_UTC_TIMER_LOADED  2
#define DH_UTC_TIMER_FETCHED 3
#define MAX_UTC_TIMER_LOAD_100MSEC_DELAYS 14

/* Delay, in msec, between wire pair power on operations, per Phil Robl @ PSL */
#define DH_PWRON_DELAY_MSEC 30

/* Structure for various wire pair-specific quantities */
struct dh_wp_struct {
  struct semaphore softboot_sem;
};
static struct dh_wp_struct dh_wirepairs[DH_MAX_CARDS][DH_PAIRS_PER_CARD];

#define PCI_CONFIG_SPACE_SIZE 256
/* Structure for storing data specific to PCI for each card */
struct dh_pci_dev {
  int found;    /* Indicate if PCI interface finds this card */
  struct pci_dev *dev; /* The Linux PCI data structure */
  unsigned char config_space[PCI_CONFIG_SPACE_SIZE];
  unsigned long rstart;
  unsigned long rend;
  unsigned long rflags;
  int irq; /* The (possibly shared) IRQ number for this card */
  int rev; /* The revision id of the card (rev0 or rev1) */
  int is_io; /* 1 if IO space, 0 if memory-mapped */
  int num_open;
  struct semaphore open_close_sem;
  int icard;

  /* Spinlock to protect ENTIRE state of card */
  spinlock_t card_lock;

  /* DMA address for mappings in TX DMA */
  dma_addr_t tx_dma_target_addr;

  /* Comms parameters - defaults in dh.h */
  int autodac, dacmax, minclev, maxclev, rdelay, sdelay, thresh;
  int minvolts, maxvolts, mincur, maxcur;
};

/* Store PCI data for each card here */
static struct dh_pci_dev dh_pci_dors[DH_MAX_CARDS];

#define dh_card_lock(icard, flags) spin_lock_irqsave(&dh_pci_dors[icard].card_lock, flags)
#define dh_card_unlock(icard, flags) spin_unlock_irqrestore(&dh_pci_dors[icard].card_lock, flags)

/* Because of the way we are measuring time differences, we need a separate
   set of bench_recs for each DOM (no races) */
static unsigned dtbench;
#define JIFFIES_TO_CALIB 10
static dh_bench_rec_t calbench0, calbench1, calbench2, jifbench, cpuint_bench[NR_CPUS],
  gread_bench, gwrite_bench, rxbench, txbench, rxnqbench, txdqbench, outslbench, inslbench,
  rxwhile, rxuncook, txdmabench, rxdmabench,
  dmarx_bench[DH_DOMHUB_DOMS], dmatx_bench[DH_DOMHUB_DOMS], 
  intrx_bench[DH_DOMHUB_DOMS], inttx_bench[DH_DOMHUB_DOMS], 
  read_bench[DH_DOMHUB_DOMS], write_bench[DH_DOMHUB_DOMS];

/** Last error code for lasterr proc file */
static int dh_lasterr = DH_ERR_NONE;

struct dh_domvars_struct {  /* Timer values for unblocking RX FIFOs ("dead DOMs") */
  unsigned long nonemptyRXjif;
  unsigned long rxdt;
  int had_rx_data;
} dh_domvars[DH_DOMHUB_DOMS];

/* A few function prototypes */
static int dh_minor_to_domno(int minor);
static void dh_do_recv_message(int icard);
static inline void dh_enable_msg_sndbl_intr_chan(int icard, int ich);

/* ----------------------------- FUNCTIONS ----------------------------- */

/************* Various register operation functions ***************/

static inline u32 dh_read_regi(int icard, int ireg) {
  /** Reads value of the PCI read register ireg on card icard. 
      Assumes PCI device register space, etc. is already found/available.
      \return the value read, or a bogus value if simulation mode.  */
  /* Get it from PCI register space. */
  return inl(dh_pci_dors[icard].rstart + ireg*4);
}


static inline void dh_write_regi_nosave(int icard, int ireg, u32 value) {
  /** Write value to register specified by ireg on device icard, without
      saving the new register value.  Ignores lock on the register.  Only
      call in DH_NORMAL mode.  Only use in I/O space. */
#ifdef DH_VERBOSE_WRITES
  static unsigned long numwritten = 0;
#endif
  outl(value, dh_pci_dors[icard].rstart+ireg*4);
#ifdef DH_VERBOSE_WRITES
  if(numwritten++ < DH_VERBOSE_WRITES) {
    printk(KERN_WARNING "DH_WRITE_REGI_NOSAVE(card %d, reg %d (%s), "
	   "value 0x%08lx)\n",\
	   icard, ireg, dh_reginame[ireg],\
	   (long)value);
  }
#endif
}


static inline void dh_write_regi(int icard, int ireg, u32 value) {
  /** Write value to register specified by ireg on device icard.
      Store a copy of the current register value.  */
#ifdef DH_VERBOSE_WRITES
  static unsigned long numwritten = 0;
#endif

  dh_write_regi_copies[icard][ireg] = value;
  outl(value, dh_pci_dors[icard].rstart+ireg*4);
#ifdef DH_VERBOSE_WRITES
  if(numwritten++ < DH_VERBOSE_WRITES) {
    printk(KERN_WARNING "DH_WRITE_REGI(card %d, reg %d (%s), value 0x%08lx)\n",\
	   icard, ireg, dh_reginame[ireg],\
	   (long)value);
  }
#endif
}


static inline void dh_set_write_regi_bits(int icard, int ireg, u32 mask) {
  /** Sets the bits given by mask in register ireg on card icard.
      Use existing write register copy and send the new value down using
      dh_write_regi(). */
  u32 val;
  unsigned long flags;
  pprintk(KERN_INFO "dh_set_write_regi_bits: card %d reg %d mask %lx.\n",\
	 icard, ireg, (unsigned long) mask);

  val = dh_write_regi_copies[icard][ireg];
  val |= mask;
  dh_write_regi(icard, ireg, val);
}

static inline void dh_set_write_regi_bits_nosave(int icard, int ireg, u32 mask) {
  u32 val;
  val = dh_write_regi_copies[icard][ireg];
  val |= mask;
  dh_write_regi_nosave(icard, ireg, val);
}


static inline void dh_clear_write_regi_copy_bits(int icard, int ireg, u32 mask) {
  /** Clears the bits given by mask in software copy of register ireg on card icard. */
  dh_write_regi_copies[icard][ireg] &= ~mask;
}


static inline void dh_clear_write_regi_bits(int icard, int ireg, u32 mask) {
  /** Clears the bits given by mask in register ireg on card icard.
      Use existing write register copy and send the new value down using
      dh_write_regi(). */
  u32 val = dh_write_regi_copies[icard][ireg];
  val &= ~mask;
  dh_write_regi(icard, ireg, val);
}

static inline void dh_clear_write_regi_bits_nosave(int icard, int ireg, u32 mask) {
  u32 val = dh_write_regi_copies[icard][ireg];
  val &= ~mask;
  dh_write_regi_nosave(icard, ireg, val);
}


static inline int dh_test_read_regi_bit(int icard, int ireg, int whichbit) {
  /** Tests bit given by whichbit (0..31) in register ireg on card icard. 
      \return the bit status (0 or 1) */
  u32 mask = 1 << whichbit;
  if(dh_mode != DH_NORMAL) {
    return (dh_sim_read_regi[icard][ireg] & mask) ? 1 : 0;
  } else {
    return 0;
  }
}

/**************** Code for LEDs *******************/
#define NUM_LED_BITS      8
#define MAX_LED_VAL       ((1U<<NUM_LED_BITS)-1)
#define FREV_MASK         0xF000F000
#define CTRL_MASK         (1U<<20)

static int dh_leds_enabled = 0;

static void dh_enable_leds(int icard) {
  dh_set_write_regi_bits(icard, DH_RG_CTRL, CTRL_MASK);
  dh_leds_enabled = 1;
}

static void dh_disable_leds(int icard) {
  dh_clear_write_regi_bits(icard, DH_RG_CTRL, CTRL_MASK);
  dh_leds_enabled = 0;
}

typedef enum { REPLACE, SET, CLEAR } LED_ACTION;

static inline void dh_set_led_value(int icard, unsigned val, LED_ACTION action) {

  if(! dh_leds_enabled) {
    printk(KERN_WARNING "dh_set_led_value: LEDs not enabled!\n");
    return;
  }

  if(val > MAX_LED_VAL) {
    printk(KERN_WARNING "dh_set_led_value: requested value %u > max %u -- will use LSBs\n",
           val, MAX_LED_VAL);
  }

  u32 led_lo = val & 0x0F;
  u32 led_hi = (val>>4) & 0x0F;

  u32 rv = 0;
  switch(action) {
  case REPLACE:
    rv = (~FREV_MASK) & dh_write_regi_copies[icard][DH_RG_FREV];
    rv |= led_lo << 12;
    rv |= led_hi << 28;
    dh_write_regi(icard, DH_RG_FREV, rv);
    break;
  case SET:
    rv |= led_lo << 12;
    rv |= led_hi << 28;
    dh_set_write_regi_bits(icard, DH_RG_FREV, rv);
    break;
  case CLEAR:
    rv |= led_lo << 12;
    rv |= led_hi << 28;
    dh_clear_write_regi_bits(icard, DH_RG_FREV, rv);
    break;
  default:
    printk(KERN_WARNING "dh_set_led_value: unknown mode.\n");
    break;
  }
}
/**************************************************/

static void dh_zero_write_regi_copies_card(int icard) {
  /** Zero out FPGA write register copies, at startup time and when
      reloading FPGA */
  int ireg;
  for(ireg=0;ireg < DH_MAX_REGIS; ireg++) {
    dh_write_regi_copies[icard][ireg] = 0;
  }
}


static void dh_zero_write_regi_copies(void) {
  /** Zero out FPGA write register copies, at startup time and when
      reloading FPGA */
  int icard;
  for(icard=0;icard < DH_MAX_CARDS; icard++) {
    dh_zero_write_regi_copies_card(icard);
  }
}


static char dh_read_flash_byte(int icard, int ipage, unsigned long pos) {
  /** Read a single byte from flash page card at position pos. */
  u32 r;
  pprintk(KERN_INFO "dh_read_flash_byte: card %d page %d pos %ld.\n",\
	  icard,ipage,pos);

  if(dh_pci_dors[icard].rev==0) {
    /* Page selects high order address bits (18 & 19) */
    r = (ipage & 0x03) << 18;
    /* Pos is the position in the page to write to - register bits 0..17 */
    r |= (pos & 0x3FFFF);
    /* write enable bit is NOT set, nor are data bits */
  }
  else {
    /* rev1: page is now split across bits 21 and 19 (bits 18 & 19 resp)...
     * pos is now 1 bit longer...
     */
    r = ((ipage & 2) << 20) | ((ipage & 1) << 19) | (pos & 0x7FFFF);
  }

  /* Now tell flash where we want to read from (address bits set) */
  dh_write_regi(icard, DH_RG_FLASH, r);

  /* Get the data back */
  r = dh_read_regi(icard, DH_RG_FLASH);
  return (r >> 24) & 0xFF;
}


/************** End of register operation functions ***************/


/******************** Miscellaneous Functions *********************/


static void dh_jsleep(int j) { 
  /** sleep ("busy-sleep") for j jiffies */
  unsigned long start_jiffies;
  unsigned long end_jiffies;

  start_jiffies = jiffies;
  end_jiffies = start_jiffies + j;
  pprintk(KERN_INFO "dh: Waiting from %ld to %ld jiffies...\n",start_jiffies,end_jiffies);
  /* Wait until enough time elapses, or jiffies overflows */
  while(jiffies < start_jiffies + j && jiffies >= start_jiffies) 
    schedule();
}


static u32 dh_hex_str_to_u32(char *s, int length) {
  /** Convert string s to an unsigned 32 bit word, 'cause sscanf seems to be broken.
      \param s      The string to convert
      \param length Size in bytes allocated to s
      \return The converted value (unsigned 32 bit word) */
  int i;
  u32 value;
  value = 0;
  for(i=0;i<length;i++) {
    pprintk(KERN_INFO "i %d s[i] %d\n",i,s[i]);
    if(s[i] == '\0') break;
    if(s[i] >= '0' && s[i] <= '9') {
      value *= 16;
      value += (s[i]-'0');
    } else if(s[i] >= 'a' && s[i] <= 'f') {
      value *= 16;
      value += ((s[i]-'a') + 10);
    } else if(s[i] >= 'A' && s[i] <= 'F') {
      value *= 16;
      value += ((s[i]-'A') + 10);
    } else {
      break;
    }
    pprintk(KERN_INFO "value = %d (0x%x)\n",value,value);
  }
  return value;
}


static u32 dh_dec_str_to_u32(char *s, int length) {
  /** Convert string s to an unsigned 32 bit word, 'cause sscanf seems to be broken.
      \param s      The string to convert
      \param length Size in bytes allocated to s
      \return The converted value (unsigned 32 bit word) */
  int i;
  u32 value;
  value = 0;
  for(i=0;i<length;i++) {
    pprintk(KERN_INFO "i %d s[i] %d\n",i,s[i]);
    if(s[i] == '\0') break;
    if(s[i] >= '0' && s[i] <= '9') {
      value *= 10;
      value += (s[i]-'0');
    } else {
      break;
    }
    pprintk(KERN_INFO "value = %d (0x%x)\n",value,value);
  }
  return value;
}


static void dh_printbuf(char *buf, int count) {
  /** Print contents of buf to system log */
  int i;
  for(i=0;i<count;i++) {
  //printk(KERN_INFO "%dth byte: 0x%02x ('%c')\n",i, buf[i], DH_PR_BYTE(buf[i]));
    printk(KERN_INFO "%dth byte: 0x%02x\n",i, buf[i]);
  }
}


static int dh_minor_to_cardno(int minor) {
  /** Convert minor number (0..127) to card no. (0..7) */
  return minor / (DH_PAIRS_PER_CARD * DH_DOMS_PER_PAIR);
}


static int dh_minor_to_wireno(int minor) {
  /** Convert minor number (0..127 to wire pair no. (0..3) */
  return (minor / DH_DOMS_PER_PAIR) % DH_PAIRS_PER_CARD;
}


static int dh_minor_to_domno(int minor) {
  /** Convert minor number (0..127 to DOM no. (0..1) */
  return minor % DH_DOMS_PER_PAIR;
}


static inline int dh_get_minor_from_card_pair_and_dom(int icard, int ipair, int idom) {
  /** Convert (card,pair,dom) number to minor number; 
      as always, high card numbers > 63 means simulation files simc?w?d?. */
  return 
      icard * DH_PAIRS_PER_CARD * DH_DOMS_PER_PAIR
    + ipair * DH_DOMS_PER_PAIR
    + idom;
}


static inline int dh_get_minor_from_card_and_channel(int icard, int ich) {
  /** Convert (pair, channel) numbers to minor number; ich is from 0 to 7 */
  return (icard * DH_PAIRS_PER_CARD*DH_DOMS_PER_PAIR) + ich;
}


static void dh_determine_excluded_cards(char *exclstr, int *excl, int n) {
  /** Fill list of cards to exclude from normal driver operations (so that e.g. Arthur can use dt
      on some of them */
  int ic;
  int id;
  for(id=0; id<n; id++) { /* Set default -- not excluded */
    excl[id] = 0;
  }
  for(ic=0; ; ic++) {
    if(exclstr[ic] == '\0') break;
    id = exclstr[ic]-'0';
    if(id >= 0 && id < DH_MAX_CARDS) {
      vprintk(KERN_INFO "dh: Will exclude card %d from the driver.\n", id);
      excl[id] = 1;
    } else {
      printk(KERN_WARNING "dh: Can't exclude invalid card %c.\n", exclstr[ic]);
    }
  }
}


static unsigned char dh_write_flash_byte(int icard, int ipage, unsigned long pos, unsigned char b) {
  /** Write a single byte b into flash page ipage on card icard at position pos. 
      For simplicity, there are some magic numbers here.  This function has to change
      if the API for the flash register changes (see DOR-API v. 0.6 or above) 
      @return the character written if success, else its bitwise complement */

  u32 r; /* Temporary register copy */
  int i;
  int failed;

  char b_back; /* Value read back from flash for comparison */
             /* AAA, write enable bit, AA */ 
  /* See datasheet for flash mem and Kalle's API doc: */
  static u32 prepA = (0xAAA | DH_BITN(20) | 0xAA << 24); 
             /* 555, write enable bit, 55 */
  static u32 prepB = (0x555 | DH_BITN(20) | 0x55 << 24); 
  static u32 prepC = (0xAAA | DH_BITN(20) | 0xA0 << 24); 

  /* Signal flash to do a write */
  dh_write_regi(icard, DH_RG_FLASH, prepA);
  dh_write_regi(icard, DH_RG_FLASH, prepB);
  dh_write_regi(icard, DH_RG_FLASH, prepC);

  if(dh_pci_dors[icard].rev==0) {
    /* Page selects high order address bits (18 & 19) */
    r = (ipage & 0x03) << 18;
    /* Pos is the position in the page to write to - register bits 0..17 */
    r |= (pos & 0x3FFFF);
  }
  else {
    /* rev1: page bits are shifted and split to bits 21 and 19,
     * pos is now an extra bit long...
     */
    r = ((ipage & 2) << 20) | ((ipage & 1) << 19) | (pos & 0x7FFFF);
  }

  /* Set write enable bit */
  r |= DH_BITN(20);
  /* Set data bits 24..31 based on b */
  r |= (b << 24);

  dh_write_regi(icard, DH_RG_FLASH, r);

  failed = 1;

  for(i=0; i < 100; i++) {
    if(( dh_read_regi(icard,DH_RG_FLASH) ^ r) & DH_FLASH_RDY) {
      pprintk(KERN_INFO "dh_write_flash_byte: Waiting for bits to change.\n");
      //dh_jsleep(1);
    } else {
      failed = 0;
      break;
    }
  }

  if(failed) {
    printk(KERN_WARNING "dh_write_flash_byte: Write to card %d page %d failed (flash timeout).\n",
	   icard, ipage);
    dh_lasterr = DH_ERR_FLASHTIMEOUT;
    return ~b; /* Guaranteed to cause error in caller */
  }

  /* Write is done.  Now read it back.  Address is already set, no need to write address. */

  r = dh_read_regi(icard, DH_RG_FLASH);
  pprintk(KERN_INFO "dh_write_flash_byte: flash query gave 0x%08lx.\n", \
	 (unsigned long) r);
  b_back = (r >> 24) & 0xFF;
  return b_back;
}


static void dh_flash_reset(int icard) {
  /* Issue reset signal to the flash */
  dh_write_regi(icard, DH_RG_FLASH, DH_FLASH_FL_RESET);
  udelay(100);
  dh_write_regi(icard, DH_RG_FLASH, 0);
  udelay(100);
}


static int dh_flash_erase_sector(int icard, int isector) {
  /* Erase a 64kb flash sector isector on DOR card icard */
  u32 secaddr, fl_dq6_1, fl_dq6_2, fl_dq2_1, fl_dq2_2;
  int ipoll, ok;
# define MAX_FLASH_POLLS 1000000
  vprintk(KERN_INFO "dh_flash_erase_sector: erasing flash sector %d.\n", isector);

  if( (isector > 15 && dh_pci_dors[icard].rev==0) ||
      (isector > 31 && dh_pci_dors[icard].rev!=0)) {
    printk(KERN_WARNING "dh_flash_erase_sector: bad sector number.\n");
    dh_lasterr = DH_ERR_BADFLASHSECT;
    return 1;
  }

  if(dh_pci_dors[icard].rev==0) {
    secaddr = ((u32) isector) << 16;
  }
  else {
    secaddr = ((isector & 0xf) << 16) | ((isector & 0x10)<<17);
  }
 
  dh_write_regi(icard, DH_RG_FLASH, 0xAA100AAAL);
  dh_write_regi(icard, DH_RG_FLASH, 0x55100555L);
  dh_write_regi(icard, DH_RG_FLASH, 0x80100AAAL);
  dh_write_regi(icard, DH_RG_FLASH, 0xAA100AAAL);
  dh_write_regi(icard, DH_RG_FLASH, 0x55100555L);
  dh_write_regi(icard, DH_RG_FLASH, 0x30100000L | secaddr);
  
  /* Wait for sector erase to finish */
  //do {
  ok = 0;
  for(ipoll = 0; ipoll < MAX_FLASH_POLLS; ipoll++) {
    fl_dq6_1 = dh_read_regi(icard, DH_RG_FLASH) & 0x40000000L;
    fl_dq6_2 = dh_read_regi(icard, DH_RG_FLASH) & 0x40000000L;
    fl_dq2_1 = dh_read_regi(icard, DH_RG_FLASH) & 0x04000000L;
    fl_dq2_2 = dh_read_regi(icard, DH_RG_FLASH) & 0x04000000L;
    if((fl_dq6_1!=fl_dq6_2) || (fl_dq2_1!=fl_dq2_2)) {
      pprintk(KERN_INFO "dh_flash_erase_sector: waiting till toggle bits change...\n");
      /* Keep going */
      schedule();
    } else {
      ok = 1;
      break;
    }
  }
  //} while((fl_dq6_1!=fl_dq6_2) || (fl_dq2_1!=fl_dq2_2));
  if(!ok) {
    printk(KERN_WARNING
           "dh_flash_erase_sector: Erasure of sector %d on card %d failed.\n",
           isector, icard);
    return 1;
  }

  pprintk(KERN_INFO "dh_flash_erase_sector: waiting for ready bit to come high....\n"); 
  /* Check ready bit */
  ok = 0;
  for(ipoll=0; ipoll < MAX_FLASH_POLLS; ipoll++) {
    if(dh_read_regi(icard, DH_RG_FLASH) & DH_FLASH_RDY) {
      ok = 1;
      break;
    } else {
      schedule();
    }
  }
    //while(! (dh_read_regi(icard, DH_RG_FLASH) & DH_FLASH_RDY)) {
    //schedule();
    //}

  if(ok) {
    dh_write_regi(icard, DH_RG_FLASH, 0);
    return 0;
  } else {
    printk(KERN_WARNING 
	   "dh_flash_erase_sector: Erasure of sector %d on card %d failed.\n",
	   isector, icard);
    return 1;
  }
}

/* sectors per page is a bit complicated -- by the fact
 * that we support two revisions of the cards and by the
 * fact that page 3 on the rev1 is restricted to the first
 * two sectors so that we don't have to mess with the 4Kb
 * sectors at the top -- here we just return the sectors
 * allocated to each page, flash_sector_count below will
 * return how many sectors are actually used in the page...
 */
static inline int flash_sectors_per_page(int icard) {
   return (dh_pci_dors[icard].rev==0) ? 4 : 8;
}

static inline int flash_sector_count(int icard, int ipage) {
   /* keep old rev0 behaviour... */
   if (dh_pci_dors[icard].rev==0) return 4;

   /* only page 3 has 2 sectors used on rev1... */
   return (ipage<3) ? 8 : 2;
}


static int dh_flash_erase_page(int icard, int ipage) {
  /* Erase a page on the flash (4 sectors) */
  int isector;
  const int sectors_per_page = flash_sectors_per_page(icard);
  const int nsectors = flash_sector_count(icard, ipage);

  vprintk(KERN_INFO "dh_flash_erase_page: erasing flash page %d on card %d.\n",
	 ipage, icard);
  for(isector=0; isector < nsectors; isector++) {
    if(dh_flash_erase_sector(icard, ipage*sectors_per_page + isector)) {
      printk(KERN_WARNING "dh_flash_erase_page: erasure of flash page %d on card %d failed!\n",
	     ipage, icard);
      return 1;
    }
  }
  return 0;
}

#define DH_PROTECT 1
#define DH_UNPROTECT 2

static int dh_rev0_flash_page_protect_unprotect(int icard, int ipage, int mode) {
  int isec, sector;
  u32 secadd, flreg, modemask, adrmask, donemask;
# define MAX_PROT_TRIES 25 /* c.f. kalle's code */
  int ipoll;
  const int sectors_per_page = flash_sectors_per_page(icard);
  const int nsectors = flash_sector_count(icard, ipage);

  if(mode == DH_UNPROTECT) {
    modemask = DH_FL_SECTOR_UNPROT;
    adrmask  = DH_FL_UNPROT_ADR;
    donemask = DH_FL_UNPROT_DONE;
  } else if(mode == DH_PROTECT) {
    modemask = DH_FL_SECTOR_PROT;
    adrmask  = DH_FL_PROT_ADR;
    donemask = DH_FL_PROT_DONE;
  } else {
    printk(KERN_WARNING "dh_rev0_flash_page_protect_unprotect: Bad mode argument (%d).\n",
	   mode);
    return 1;
  }

  /* Note: DH_FLASH_FL_RESET is zero, DH_FLASH_FL_RESET_VID is one: */
  dh_write_regi(icard, DH_RG_FLASH, DH_FLASH_FL_RESET_VID);
  dh_jsleep(1); /* at least 1 usec required here */

  for(isec=0; isec < nsectors; isec++) {
    sector = ipage*sectors_per_page + isec;

    dh_jsleep(10); /* c.f. Kalle 'delay' function */
    if (dh_pci_dors[icard].rev==0) {
      secadd = (sector << 16) | adrmask;
    }
    else {
      secadd = ((sector & 0xF) << 16) | ((sector & 0x10) << 17) | adrmask;
    }
    vprintk(KERN_INFO "dh_rev0_flash_page_protect_unprotect: %sing card %d page %d sector %d.\n",
	    mode==DH_PROTECT?"protect":"unprotect", icard, ipage, sector);
    pprintk(KERN_INFO "secadd 0x%08lx\n", (unsigned long) secadd);

    for(ipoll=0; ipoll<MAX_PROT_TRIES; ipoll++) {
      flreg  = DH_FLASH_FL_RESET_VID | modemask | DH_FL_WE | secadd;
      dh_write_regi(icard, DH_RG_FLASH, flreg);
      dh_jsleep(10); /* delay(1) in kalle but he has an extra write/delay @ beginning */
      flreg  = DH_FLASH_FL_RESET_VID | DH_FL_SECTOR_VER | DH_FL_WE | secadd;
      dh_write_regi(icard, DH_RG_FLASH, flreg);
      if((dh_read_regi(icard, DH_RG_FLASH) & DH_BITN(24)) == donemask) {
	pprintk(KERN_INFO "done at iteration %d.\n", ipoll);
	break;
      } else {
	if(ipoll == MAX_PROT_TRIES-1) {
	  printk(KERN_WARNING "dh_rev0_flash_page_protect_unprotect: %s of card %d page %d sector %d"
		 " failed.\n", mode==DH_PROTECT?"protect":"unprotect", icard, ipage, sector);
	  dh_write_regi(icard, DH_RG_FLASH, 0); /* pull down reset pin */
	  return 1;
	}
      }
    }
  }

  dh_write_regi(icard, DH_RG_FLASH, 0); /* pull down reset pin */
  dh_jsleep(1);
  flreg = DH_FL_RESET_CMD | DH_FL_WE ;
  /* | secadd;  Secadd bogus ?? but it's how kalle
     does it */
  dh_write_regi(icard, DH_RG_FLASH, flreg);
  dh_write_regi(icard, DH_RG_FLASH, 0);
  
  return 0;
}

static inline int dh_flash_sector_size(int icard) {
   /* so far, we're always 64k sectors -- though there are
    * 2x8kb, 1x16kb and 1x32kb sectors at the top of flash,
    * we ignore that detail for now...
    */
   return 64*1024;
}

static inline int dh_flash_page_size(int icard, int ipage) {
   return flash_sector_count(icard, ipage)*dh_flash_sector_size(icard);
}

static int dh_flash_page_isempty(int icard, int ipage) {
  /* Make sure flash has been erased correctly */
  int i;
  unsigned char b;

  for(i=0; i<dh_flash_page_size(icard, ipage); i++) {
    b = dh_read_flash_byte(icard, ipage, i);
    if(b != 0xFF) {
      printk(KERN_WARNING "dh_flash_page_isempty: byte at card %d page %d position %d not empty (%x).\n",
	     icard, ipage, i, (int) b);
      dh_lasterr = DH_ERR_FLASHNOTEMPTY;
      return 0;
    }
  }
  return 1;
}

/* Arthur either wrote or contributed heavily to the next few routines: */
static inline int cardRevision(int icard) { 
  return dh_pci_dors[icard].rev;
}

static inline int flashSectorsPerPage(int icard) { 
  return (cardRevision(icard)==0) ? 4 : 8;
}

static inline int flashSectorCount(int icard, int ipage) {
  /* keep old rev0 behaviour... */
  if (cardRevision(icard)==0) return 4;

  /* only page 3 has 2 sectors used... */
  return (ipage<3) ? 8 : 2;
}

static inline int flashNumSectors(void) {
  return 35;
}

static void flashReadReset(int icard) {
  dh_write_regi(icard, DH_RG_FLASH, DH_FL_RESET_CMD | DH_FL_WE);
  // outl( 0xf0 << 24 | FL_WE, dor + FLASH);
}

/* (ALJ) returns address in FLASH chip format */
static int flashSectorToAddr(int sector, int *ier) {
  *ier = 0;
  int ret;
  if (sector<32) {
    ret = ((sector&0xf)<<16) | ((sector&0x10)<<17);
  } else if (sector>=flashNumSectors()) {
    printk(KERN_WARNING "dh: flashSectorToAddr: Warning!!! Invalid sector %d.\n", sector);
    *ier = 1;
    return 0;
  } else {
    /* sectors 31..34 are special */
    int szs[] = { 16*1024, 4*1024, 4*1024, 8*1024 };
    int i;
    ret=flashSectorToAddr(31, ier);
    for (i=0; i<sector-31; i++) {
      ret+=szs[i]<<1;
    }
  }
  vvprintk(KERN_INFO "flashSectorToAddr(%d)=0x%08x (%d)\n", sector, ret, ret);
  return ret;
}


static int flashProtectSector(int dor, int sector) {
  int ipoll, ier;
  int addr = flashSectorToAddr(sector, &ier);
  if(ier) return 1;

  /* RP* = VID */

  dh_write_regi(dor, DH_RG_FLASH, DH_FLASH_FL_RESET_VID);
  //outl(FL_NRP, dor+FLASH);

  for (ipoll=0; ipoll<25; ipoll++) {
    /* write 60h, address=block address
     * A0=VIL, A1=VIH, A6VIL
     */
    dh_write_regi(dor, DH_RG_FLASH, 
		  DH_FL_SECTOR_PROT | DH_FLASH_FL_RESET_VID | DH_FL_WE 
		  | addr | (0x2<<1));
    //outl((0x60<<24) | FL_NRP | FL_WE | addr | (0x2<<1), dor+FLASH);

    /* wait 100us... */
    //usleep(100);
    udelay(100);

    /* write 40h, address=block address
     * A0=VIL, A1=VIH, A6=VIL
     */
    dh_write_regi(dor, DH_RG_FLASH,
                  DH_FL_SECTOR_VER | DH_FLASH_FL_RESET_VID | DH_FL_WE 
		  | addr | (0x2<<1));
    //outl( (0x40<<24) | FL_NRP | FL_WE | addr | (0x2<<1), dor+FLASH);

    /* wait 4us
     */
    //usleep(4);
    udelay(4);

    /* Read data -- expect 0x01... */
    u32 flash = dh_read_regi(dor, DH_RG_FLASH);
    vvprintk(KERN_INFO "After readback, FLASH=0x%08lx\n", (unsigned long) flash);
    if(((flash>>24) & 0xFF) == 0x01) break;
    //if ( ((inl(dor+FLASH)>>24) & 0xff) == 0x01) break;
  }

  /* RP* = VIH */
  dh_write_regi(dor, DH_RG_FLASH, 0);
  //outl(0, dor+FLASH);

  /* issue read/reset command */
  flashReadReset(dor);

  return (ipoll==25) ? 1 : 0;
}

static int flashUnprotectChip(int dor) {
/* we can only unprotect the whole chip... Shamelessly stolen from Arthur Jones */
/* Adapted to support old DOR Rev 0 cards */
  int i, isector=0;
  int ier;

  if(cardRevision(dor)==0) return dh_rev0_flash_page_protect_unprotect(dor, 0, DH_UNPROTECT);

  /* protect all blocks... */
  for (i=0; i<flashNumSectors(); i++) {
    if (flashProtectSector(dor, i)) {
      printk(KERN_WARNING "dh: flashUnprotectChip(card=%d): couldn't protect sector %d.\n", 
	     dor, i);
      return 1;
    }
  }

  /* RP* = VID */
  dh_write_regi(dor, DH_RG_FLASH, DH_FLASH_FL_RESET_VID);
  //outl(FL_NRP, dor+FLASH);

  /* write 60h any address with:
   * A0=VIL, A1=VIH, A6=VIH
   */
  dh_write_regi(dor, DH_RG_FLASH,
		DH_FL_SECTOR_PROT | DH_FLASH_FL_RESET_VID | DH_FL_WE | (0x42 << 1));
  //outl( (0x60 << 24) | FL_NRP | FL_WE | (0x42 << 1), dor+FLASH);

  for (i=0; i<1000 && isector<flashNumSectors(); i++) {
    /* write 60h any address with:
     * A0=VIL, A1=VIH, A6=VIH
     */
    dh_write_regi(dor, DH_RG_FLASH,
		  DH_FL_SECTOR_PROT | DH_FLASH_FL_RESET_VID | DH_FL_WE | (0x42 << 1));
    //outl( (0x60 << 24) | FL_NRP | FL_WE | (0x42 << 1), dor+FLASH);

    /* wait 10ms */
    set_current_state(TASK_INTERRUPTIBLE); 
    schedule_timeout(10*HZ/1000);
    //usleep(10*1000);

    while (isector<flashNumSectors()) {
      const int addr = flashSectorToAddr(isector, &ier);
      if(ier) return 1;
      vvprintk(KERN_INFO "Verifying address %d...\n", addr);
      /* write 40h address=isector block address...
       * A0=VIL, A1=VIH, A6=VIH
       */
      dh_write_regi(dor, DH_RG_FLASH,
		    DH_FL_SECTOR_VER | DH_FLASH_FL_RESET_VID | DH_FL_WE | addr | (0x42 << 1));
      //outl( (0x40 << 24) | FL_NRP | FL_WE | addr | (0x42 << 1), dor+FLASH);

      /* wait 4us */
      udelay(4);
      //usleep(4);

      /* read data, address=isector block address... */
      //if ( (inl(dor+FLASH)&(0xff<<24)) == 0x00) {
      u32 flash = dh_read_regi(dor, DH_RG_FLASH);
      if( (flash&(0xFF<<24)) == 0x00) {
	isector++;
      } else {
	break;
      }
    }
  }

  /* RP* = VIH */
  dh_write_regi(dor, DH_RG_FLASH, 0);
  //outl(0, dor+FLASH);

  /* issue read/reset... */
  flashReadReset(dor);

  return isector<flashNumSectors();
}

static int flashProtectPage(int dor, int page) { /* ALJ */
  int sector;
  const int spp = flashSectorsPerPage(dor);
  for (sector = 0; sector < flashSectorCount(dor, page); sector++) {
    if (flashProtectSector(dor, page*spp + sector)) {
      printk(KERN_INFO "flashProtectPage: can't protect sector %d\n",
	     page*spp + sector);
      return 1;
    }
  }

  return 0;
}

/* protect the flash chip, rev0 -> page 0, rev 1 -> page 3 and
 * LOGSECTOR (sector 30)
 * Shamelessly stolen from Arthur Jones
 */
static int flashProtectChip(int dor) {
  if (cardRevision(dor)==0) return dh_rev0_flash_page_protect_unprotect(dor, 0, DH_PROTECT);
  /* Rev 1 */
  if (flashProtectPage(dor, 3)) {
    printk(KERN_WARNING "dh: flashProtectChip: unable to protect page 3\n");
    return 1;
  }
  
  if (flashProtectSector(dor, DH_LOGSECTOR)) {
    printk(KERN_WARNING "dh: flashProtectChip: unable to protect log sector\n");
    return 1;
  }
  
  return 0;
}


static int dh_firmware_version_check(int icard) {
  /** Return 1 if firmware loaded on card icard is not in the list
      of allowed firmware versions.  Return 0 if the version is ok. */
  int iver, comms_fw_ok;
  u32 frev;
  pprintk(KERN_INFO "dh_firmware_version_check: Checking firmware for card %d.\n", icard);
  comms_fw_ok = 0;
  for(iver = 0; iver < DH_N_FIRMWARE_OK; iver++) {
    pprintk(KERN_INFO "dh_firmware_version_check: Comparing with %08lx...\n",\
	   (unsigned long) dh_firmware_ok[iver]);
    frev = dh_read_regi(icard, DH_RG_FREV);
    if((frev & 0xFFFFFF) == dh_firmware_ok[iver]) {
      pprintk(KERN_INFO "dh_firmware_version_check: Got good firmware (%08lx).\n",\
	     (unsigned long) frev);
      comms_fw_ok = 1;
      break;
    } else {
      pprintk(KERN_INFO "dh_firmware_version_check: %08lx doesn't match card %d.\n",\
	     (unsigned long) dh_firmware_ok[iver], icard);
    }
  }

  /* Check PCI firmware for Rev 1 cards */
  if(dh_pci_dors[icard].rev > 0 && parm_fw_version_check) {
    u32 frev = dh_read_regi(icard, DH_RG_FREV);
    char pci = frev >> 24;
    if(pci != DH_PCI_REV) {
      printk(KERN_WARNING "dh_firmware_version_check: PCI firmware version %d for card %d "
	     "is out of date.  You will only be able to program "
	     "DOR flash or load new FPGA designs with this card.  You will not be able to "
	     "power on DOMs or communicate with them.\n", pci, icard);
      dh_lasterr = DH_ERR_UNSUPPFWARE;
      return 1;
    }
  }

  if(! comms_fw_ok && parm_fw_version_check) {
    printk(KERN_WARNING "dh_firmware_version_check: Firmware version %08lx for card %d is NOT"
	   " supported by this version of the DOR driver.  You will only be able to program"
	   " DOR flash or load new FPGA designs with this card.  You will not be able to"
	   " power on DOMs or communicate with them.\n", (unsigned long) (frev&0xFFFFFF), icard);
    dh_lasterr = DH_ERR_UNSUPPFWARE;
    return 1;
  }

  return 0;
}


static inline int dh_tx_fifo_is_empty(int icard, int ich) {
  /** \return 1 if transmit FIFO is EMPTY (used for debugging) */
  u32 ttsic;
  if(ich < 0 || ich >= DH_DOMS_PER_CARD) return 1;
  ttsic = dh_read_regi(icard, DH_RG_TTSIC);
  if(! (ttsic & (DH_BITN(ich)))) return 0;
  return 1;
}


static inline int dh_tx_fifo_almost_empty(int icard, int ich) {
  /** \return 1 if transmit FIFO is ALMOST EMPTY (has room for a message) */
  rmb();
  u32 ttsic = dh_read_regi(icard, DH_RG_TTSIC);
  rmb();
  int aef = (ttsic & DH_BITN(ich+8))?1:0;
  vvvprintk(KERN_INFO "dh_tx_fifo_almost_empty(%lu): TTSIC=0x%08lx, returning %d.\n", 
	    jiffies&0xFFFF, (unsigned long) ttsic, aef);
  return aef;
}


static int dh_tx_fifo_hasroom(int icard, int ich, int msgsize) {
  /* Thanks to Arthur for the code template.
     Have to decrement txfree by msgsize when a message is added to TX FIFO. */
  int minor = (icard * DH_DOMS_PER_CARD) + ich;
  int flag;
#undef  USE_EMPTY
#define USE_EMPTY
#ifdef  USE_EMPTY
  flag = dh_tx_fifo_is_empty(icard, ich);
  if(flag) txfree[minor] = DH_TX_FIFO_DEPTH;
#else /* Use almost-empty */
  flag = dh_tx_fifo_almost_empty(icard, ich);
  if(flag) txfree[minor] = DH_TX_FIFO_DEPTH-DH_TX_AEF_SIZE; /* This many (or more) are free when
							      fifo_almost_empty occurs */
#endif
  /* FIXME */
  int packetized_bytes = ((msgsize+3)/4 * 4) + 4;
  int can_add = packetized_bytes <= txfree[minor];
  vvvprintk(KERN_INFO "dh_tx_fifo_hasroom(%lu): %d bytes in TX FIFO (%s), %s add %d.\n",
	    jiffies&0xFFFF,txfree[minor], flag?"EMPTY":"not empty", 
	    can_add?"can":"cannot", msgsize);
  return can_add;
}


/* static inline int dh_tx_fifo_not_full(int icard, int ich) { */
/*   /\** \return 1 if transmit FIFO has room for a message *\/ */
/*   u32 ttsic; */

/*   ttsic = dh_read_regi(icard, DH_RG_TTSIC); */
/*   //if(! ttsic & (DH_BITN(8 + ich))) return 0; */
/*   return (ttsic & DH_BITN(ich)) ? 1 : 0;  /\* Bits 0..7 are really FIFO empty bits *\/ */
/*   /\* Eventually build in knowledge of FIFO state (how many bytes already in) *\/ */
/* } */


static inline int dh_dom_rx_fifo_isfull(int icard, int ich) {
  /* Return 1 if the *DOM* rx fifo is full */
  u32 doms; 
  if(ich < 0 || ich >= DH_DOMS_PER_CARD) return 0;
  doms = dh_read_regi(icard, DH_RG_DOMS);
  if(doms & DH_BITN(ich+16)) return 1;
  return 0;
}

static inline int dh_rx_fifo_has_message(int icard, int ich) {
  return (dh_read_regi(icard, DH_RG_RTSIC) & DH_BITN(16+ich)) ? 1 : 0;
}

static inline int dh_rx_interrupt_is_enabled(int icard, int ich) {
  u32 rtsic = dh_read_regi(icard, DH_RG_RTSIC);
  u32 inten = dh_read_regi(icard, DH_RG_INTEN);
  if(! (rtsic & DH_BITN(ich+24))) return 0;
  if(! (inten & DH_BITN(16))) return 0;
  if(! (inten & DH_BITN(9))) return 0;
  return 1;
}

static inline int dh_rx_fifo_is_empty(int icard, int ich) {
  /** \return 1 if receive FIFO has a message from the DOM */

  u32 rtsic = dh_read_regi(icard, DH_RG_RTSIC);
  if(! (rtsic & (DH_BITN(ich)))) return 0;
  return 1;
}


static inline void dh_init_intr(int icard) {
  /** Disable all interrupts to begin with */
  dh_write_regi(icard, DH_RG_INTEN, 0);
}


static inline void dh_enable_global_intr(int icard) {
  /** This has to be on before any interrupts can occur on the card. */
  dh_set_write_regi_bits(icard, DH_RG_INTEN, DH_INTEN_GLOBAL_INT_EN);
}


static inline void dh_disable_global_intr(int icard) {
  /** Turning this off turns off all interrupts on the card. */
  dh_clear_write_regi_bits(icard, DH_RG_INTEN, DH_INTEN_GLOBAL_INT_EN);
}


static inline void dh_enable_msg_sndbl_intr(int icard) {
  /** Tell hardware to call interrupt handler if TX FIFO has room for another message */
  pprintk(KERN_INFO "dh_enable_msg_sndbl_intr.\n");
  dh_set_write_regi_bits(icard, DH_RG_INTEN, DH_INTEN_ON_MSG_SNDBL);
}


static inline void dh_enable_msg_sndbl_intr_chan(int icard, int ich) {
  /** Enable TX interrupts, including TX_INT_EN_0..7 */
  pprintk(KERN_INFO "dh_enable_msg_sndbl_intr_chan.\n");
  vvprintk(KERN_INFO "TX enable(card=%d)\n", icard);
  dh_set_write_regi_bits(icard, DH_RG_INTEN, DH_INTEN_ON_MSG_SNDBL);
  dh_set_write_regi_bits(icard, DH_RG_TTSIC, DH_BITN(24+ich));
}


static inline void dh_disable_msg_sndbl_intr(int icard) {
  /** Tell hardware to disable interrupt which happens if TX FIFO has room for another message */
  vvprintk(KERN_INFO "TX disable(card=%d)\n", icard);
  dh_clear_write_regi_bits(icard, DH_RG_INTEN, DH_INTEN_ON_MSG_SNDBL);
  dh_clear_write_regi_bits(icard, DH_RG_TTSIC, DH_TTSIC_INT_EN_0__7);
}

static inline void dh_disable_msg_sndbl_intr_chan(int icard, int ich) {
  /** Tell hardware to disable interrupt which happens if TX FIFO has room 
      for another message, single channel only */
  pprintk(KERN_INFO "dh_disable_msg_sndbl_intr_chan.\n");
  dh_clear_write_regi_bits(icard, DH_RG_TTSIC, DH_BITN(24+ich));
}


static inline void dh_enable_msg_recvd_intr(int icard) {
  /** Tell hardware to enable interrupt which would occur if RX FIFO has a message from DOM */
  pprintk(KERN_INFO "dh_enable_msg_recvd_intr.\n");
  dh_set_write_regi_bits(icard, DH_RG_INTEN, DH_INTEN_ON_MSG_RCVD);
}


static inline void dh_enable_msg_recvd_intr_chan(int icard, int ich) {
  /** Tell hardware to enable interrupt which would occur if RX FIFO has a message from DOM;
      Also include required RTSIC bits.  Note precise handling of RTSIC bits, to avoid
      clearing of interrupts signaled on R/W/C bits 16..23. */
  u32 rtsic, mask;
  unsigned long flags;

  vvprintk(KERN_INFO "dh_enable_msg_recvd_intr_chan(icard=%d, ich=%d)\n", icard, ich);
  dh_set_write_regi_bits(icard, DH_RG_INTEN, DH_INTEN_ON_MSG_RCVD);

  //dh_set_write_regi_bits(icard, DH_RG_RTSIC, DH_BITN(24+ich));  
  rtsic = dh_read_regi(icard, DH_RG_RTSIC);
  vvprintk(KERN_INFO "dh_enable_msg_recvd_intr_chan: read RTSIC 0x%lx\n",(unsigned long)rtsic);
  mask = (0xFF000000 & rtsic) | DH_BITN(ich+24);
  vvprintk(KERN_INFO "dh_enable_msg_recvd_intr_chan: mask is 0x%lx\n",(unsigned long) mask);
  /* Implicit lock in dh_write_regi is redundant, could remove */
  dh_write_regi(icard, DH_RG_RTSIC, mask);
}


static inline void dh_disable_msg_recvd_intr_chan(int icard, int ich) {
  /** Tell hardware to enable interrupt which would occur if RX FIFO has a message from DOM;
      Also include required RTSIC bits.  Note precise handling of RTSIC bits, to avoid
      clearing of interrupts signaled on R/W/C bits 16..23. */
  u32 rtsic, mask;
  unsigned long flags;

  pprintk(KERN_INFO "dh_disable_msg_recvd_intr_chan.\n");
  rtsic = dh_read_regi(icard, DH_RG_RTSIC);
  mask = (0xFF000000 & rtsic) &  (~DH_BITN(ich+24));
  dh_write_regi(icard, DH_RG_RTSIC, mask);
}

static inline void dh_disable_msg_recvd_intr(int icard) {
  /** Tell hardware to disable interrupt which would occur if RX FIFO has a message from DOM */
  pprintk(KERN_INFO "dh_disable_msg_recvd_intr.\n");

  dh_clear_write_regi_bits(icard, DH_RG_INTEN, DH_INTEN_ON_MSG_RCVD);

  /* The next step is subtle - you want to clear the highest 8 bits,
     you want to explicitly NOT SET bits 16-23 at this time; 
     the low 16 bits are read-only; so, we write a zero: */
  /* Recently re-added */
  dh_write_regi(icard, DH_RG_RTSIC, 0);

}


static void dh_discard_message_from_rx_fifo(int icard, int ich) {
  /** If we don't have room for a message from the DOM in our input
      queue, we must drain the appropriate FIFO. */
  u32 word;
  u32 bytecnt;
  int nwords,iword;

  /* First word has byte count -- convert to little-endian (portability) */
  bytecnt = __le32_to_cpu(0xFFFF & dh_read_regi(icard, DH_RG_MBF0 + ich));
  
  nwords = bytecnt / 4; /* Worry about remainder later */
  
  /* Read nwords out of RX FIFO */
  for(iword = 0; iword < nwords; iword++) {
    word = dh_read_regi(icard, DH_RG_MBF0 + ich);
  }
  
  if(bytecnt % 4) { /* One more word to read */
    word = dh_read_regi(icard, DH_RG_MBF0 + ich);
  }  
}


static void dh_clear_message_received(int icard, int ich) {
  /** Clear RTSIC message received bit. */
  u32 rtsic, mask;
  unsigned long flags;
  if(ich < 0 || ich >= DH_DOMS_PER_CARD) return;
  /* It's a little confusing, but you have to write a 1 to clear the 1;
     save interrupt enable/disable bits. */
  rtsic = dh_read_regi(icard, DH_RG_RTSIC);
  mask = (0xFF000000 & rtsic) | DH_BITN(ich+16);
  dh_write_regi(icard, DH_RG_RTSIC, mask);
}



/******************************* ReTX Buffer Functions ***********************************/

static inline int dh_retxb_bytes(int minor) { return nretxb_bytes[minor]; }

static inline int dh_retxb_past_bytelimit(int minor, int nbytes) { 
  return (dh_retxb_bytes(minor)+nbytes) > DH_MAX_RETXB_BYTES;
}

static inline unsigned dh_num_retxb(int minor) { return nretxb[minor]; }

static inline int dh_retx_buf_not_full(int minor) { return nretxb[minor] == NRETX ? 0 : 1; }

static inline int dh_retx_buf_isempty(int minor) { return nretxb[minor] == 0 ? 1 : 0; }


static void dh_empty_retxb(int minor) {
  int ipos;
  retxbhd[minor] = NULL;
  nretxb[minor]  = 0;
  for(ipos=0; ipos < NRETX; ipos++) {
    retxb[minor][ipos].pending = 0;
    retxb[minor][ipos].next    = NULL;
  }
  nretxb_bytes[minor] = 0;
}


static int dh_retx_have_timeout(int minor, unsigned long timeout, unsigned long now) {
  /** See if any retx buffers have timed out */
#ifndef SKIP_RETX
  rtxMsg * head = retxbhd[minor];
  while(head) {
    unsigned long dt = now - head->jiffies; /* jiffies overflow is ok, dt will be > timeout */
    if(dt > timeout) return 1;
    head = head->next;
  }
#endif
  return 0;
}

static void dh_queue_from_retx(int minor, unsigned long timeout, unsigned long now, 
			       int icard, int ich) {
  /** Move timed out packets from retxb to retxq, so long as retxq isn't full.
      Move them in the order of oldest sequence number to youngest. 
      Do this by traversing linked list, in which they are sorted */
#ifndef SKIP_RETX
  rtxMsg * head = retxbhd[minor];

  /* Dump retx messages into retxq until we traverse list or retxq is full */
  vvvprintk(KERN_INFO "dh_queue_from_retx(minor=%d): entered\n", minor);
  if(!head) {
    vvvprintk(KERN_INFO "dh_queue_from_retx(minor=%d): head is null.\n", minor);
  }
  while(head && !rtxQisFull(&retxq[minor])) {
    vvvprintk(KERN_INFO "dh_queue_from_retx(minor=%d): list not empty and retxq not full\n", 
	      minor);
    unsigned long dt = now - head->jiffies; /* jiffies overflow is ok, dt will be > timeout */
    if(dt > timeout) {
      /* Send the sucker down */
      head->jiffies = jiffies;
      head->nretxs++;
      retx_triggered[minor] = 1;
      vvprintk("dh_queue_from_retx(minor=%d): retx_triggered TRUE, won't outq->TX\n",
	       minor);
      putRetxQ(&retxq[minor], head);
      dh_enable_msg_sndbl_intr_chan(icard, ich);

      /* Only show a certain number of these, else a dead DOM will fill up the domhub's disk. */
      if(nretxdiag[minor] < MAXRETXPRINTKS) {
	vprintk(KERN_INFO "RETXB(%lu, minor=%d, dt=%lu, tmax=%lu): SEQN(%d) -> retxq.\n",
		jiffies&0xFFFF, minor, dt, timeout, head->seqn);
	nretxdiag[minor]++;
	if(nretxdiag[minor] == MAXRETXPRINTKS) {
	  vprintk(KERN_INFO "!!! Not showing any more RETX messages for minor %d !!!\n", minor);
	}
      }
    } else {
      break; /* No point in continuing, the rest are too fresh to continue */
    }
    head = head->next;
  }
#endif
}


static inline void dh_ack_deletes_retx(int minor, unsigned short seqn) {
  /* Old method was brain-dead inefficient; just traverse list and pick out
     packet that way.  Generally what we are deleting will be the head anyways. */
#ifndef SKIP_RETX
  rtxMsg * head = retxbhd[minor];
  rtxMsg * it;

  vvprintk(KERN_INFO "DH(%lu): Checking ACK, seqn=%d.\n", jiffies&0xFFFF, seqn);

  if(!head) {
    vprintk(KERN_WARNING "dh_ack_deletes_retx: DOM (minor=%d)"
	   " Ack'ed nonexistent packet (empty list).\n", minor);
    return;
  }

  it = NULL;
  if(head->seqn == seqn) { /* First guy in list is what we want */
    retxbhd[minor] = head->next;
    it = head;
  } else {
    while(head->next) {
      if(head->next->seqn == seqn) {
	it = head->next;
	head->next = head->next->next;
	break;
      }
      head = head->next;
    }
  }

  if(it) { 
    vvprintk(KERN_INFO "DH(%lu): ACK deleted seqn %d from retxb.\n", jiffies&0xFFFF, seqn);
    
    nretxb[minor]--;
    if(nretxb[minor]<=0) {
      retx_triggered[minor] = 0;
      vvprintk(KERN_INFO "DH(%lu): minor=%d, retx_triggered FALSE, outq->TX is OK now.\n",
	       jiffies&0xFFFF, minor);
    }
    nretxb_bytes[minor] -= it->len;
    vvprintk(KERN_INFO "DH: nretxb_bytes now == %d.\n", nretxb_bytes[minor]);
    it->pending = 0;
    it->nretxs  = 0;
    wake_up_interruptible(&retxbwq[minor]);
  } else {
    vprintk(KERN_WARNING "dh_ack_deletes_retx: DOM (minor=%d)"
           " Ack'ed nonexistent packet (SEQN=%d not in list).\n", minor, seqn);
  }
#endif
}


static inline void dh_retx_buf_add(int minor, unsigned char * data, int len, 
				   unsigned short seqn, unsigned short flags, unsigned long time) {
  /* Better call dh_retx_buf_not_full first */
  /* Adds to the END of the list */
  /* Assumes this is the most recent packet and that all the others in the list are older */

  /* First find the record we're going to use */
#ifndef SKIP_RETX
  int ipos, picked;
  picked = -1;
  for(ipos=0; ipos < NRETX; ipos++) {    /* Just pick the first empty spot */
    if(retxb[minor][ipos].pending == 0) {
      picked = ipos;
      break;
    }
  }
  if(picked == -1) {
    printk(KERN_ERR "dh_retx_buf_add: No room for message, forgot to check w/ "
	   "dh_retx_buf_not_full first!!!\n");
    return;
  }
  /* retxb[minor][picked] is now the place where we're going to put this data packet */
  retxb[minor][picked].len     = len;
  retxb[minor][picked].seqn    = seqn;
  retxb[minor][picked].jiffies = time;
  retxb[minor][picked].flags   = flags;
  retxb[minor][picked].pending = 1;
  retxb[minor][picked].nretxs  = 0;
  memcpy(&retxb[minor][picked].data, data, len);

  /* Now arrange pointers, for sorting */
  rtxMsg * head = retxbhd[minor];

  nretxb_bytes[minor] += len;
  nretxb[minor]++;

  if(head == NULL) { /* Nothing in list */
    vvprintk(KERN_INFO "dh_retx_buf_add: Adding first message to retx buffer w/ seqn=%hu"
	     ", nretxb[minor] = %d, nbytes=%d.\n",seqn, nretxb[minor], nretxb_bytes[minor]);
    retxbhd[minor] = &(retxb[minor][picked]);
    retxbhd[minor]->next = NULL; 
    return;
  }

  /* Move until head points to last existing element */
  while(head->next) head = head->next;
  head->next = &(retxb[minor][picked]); /* Add current after head */
  head->next->next = NULL;              /* and terminate it */

  vvprintk(KERN_INFO "dh_retx_buf_add: Adding %dth message to retx buffer w/ seqn=%hu (NB=%d).\n",
	   nretxb[minor], seqn, nretxb_bytes[minor]);
#endif
}

/********************************* Done w/ retx functions ***********************/


static void dh_drain_fifo(int icard, int ich, int mode) {
  u32 junk;
  int i, wc;
  wc = 0;
  char a,b,c,d;
  for(i=0;i<DH_MAX_HW_PKTSIZ * DH_MAX_FIFO_DEPTH;i++) {
    if(dh_rx_fifo_is_empty(icard, ich)) break;
    junk = dh_read_regi(icard, DH_RG_MBF0 + ich);
    wc++;
    if(mode == DH_LOUD) {
      a = (junk >>  0) & 0xFF; if(a < ' ' || a > '~') a = '.';
      b = (junk >>  8) & 0xFF; if(b < ' ' || b > '~') b = '.';
      c = (junk >> 16) & 0xFF; if(c < ' ' || c > '~') c = '.';
      d = (junk >> 24) & 0xFF; if(d < ' ' || d > '~') d = '.';
    
      printk(KERN_ERR "   RX FIFO(%02d) = 0x%08lx [%c%c%c%c]\n",wc,(unsigned long) junk,
	     a,b,c,d);
    }
    rmb(); /* in case printouts disappear */
  }
  if(mode == DH_LOUD) vprintk(KERN_INFO "   Drained %d words from FIFO.\n",wc);
}


static void dh_handle_recvd_ack(unsigned short seqn, int minor) {
  vvprintk(KERN_INFO "DH: Handling recv'd ACK seqn=%d minor=%d\n",seqn,minor);
  dh_ack_deletes_retx(minor, seqn);
}


static int dh_pkt_size_ok(int bytecnt) {
  if(bytecnt < 4 || (bytecnt % 4)) return 0;
  if(bytecnt > DH_MAX_HW_PKTSIZ) return 0;
  return 1;
}

static void dh_test_for_inc_pat(unsigned char * cb, int bytecnt, int where) {
  int iz, iw;
  if(bytecnt > 0) {
    unsigned char cur = cb[0];
    int curcnt = 0;
    for(iz=0; iz<bytecnt; iz++) {
      if(cb[iz] != cur) {
	printk(KERN_WARNING "!!!\n");
	printk(KERN_WARNING "dh_test_for_inc_pat(where=%d): "
	       "WARNING byte pattern violated in packet at iz=%d "
	       "bytecnt=%d cur=0x%02x cb=0x%02x!\n", where, iz, bytecnt,
	       (int) cur, (int) cb[iz]);
	printk(KERN_WARNING "!!!\n");
	break;
      }
      curcnt++;
      if(curcnt >= 4) { /* Wrap pattern byte cur */
	curcnt=0; 
	cur++; 
	if(cur==0) cur=1;
      }
    } /* loop over bytes */
  } /* if bytecnt > 0 */
}

static void dh_test_for_inc_pat_user(unsigned char * cb, int bytecnt, int where) {
  /* Slight kludge, should combine w/ prev. function; this one goes to userspace for cb */
  int iz, iw, bad=0;
  unsigned char cbiz;
  unsigned char cur;
  if(bytecnt > 0) {
    //unsigned char cur = cb[0];
    get_user(cur, cb);
    int curcnt = 0;
    for(iz=0; iz<bytecnt; iz++) {
      get_user(cbiz, cb+iz);
      //if(cb[iz] != cur) {
      if(cbiz != cur) {
	printk(KERN_WARNING "!!!\n");
	printk(KERN_WARNING "dh_test_for_inc_pat_user(where=%d): "
	       "WARNING byte pattern violated in packet at iz=%d "
	       "bytecnt=%d cur=0x%02x cb=0x%02x!\n", where, iz, bytecnt,
	       (int) cur, (int) cbiz);
	printk(KERN_WARNING "!!!\n");
	bad = 1;
	break;
      }
      curcnt++;
      if(curcnt >= 4) { /* Wrap pattern byte cur */
	curcnt=0; 
	cur++; 
	if(cur==0) cur=1;
      }
    } /* loop over bytes */
  } /* if bytecnt > 0 */
  if(bad) {
    int nwords = (bytecnt+3)/4;
    u32 w;
    u32 * cw = (u32 *) cb;
    for(iw=0; iw<nwords; iw++) {
      if(!get_user(w, cw+iw)) {
	printk(KERN_WARNING "word(%d)=%08lx\n",iw, (long) w);
      }
    }
  }
}

static void dh_clear_tx_dma(int icard) {
  /* Clears TX DMA after DMA timeout - should never happen; but Kalle suggests
     that if it does we reset the DMA counter and RD enable bit */
  dh_clear_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_MR_RD_EN);
  dh_write_regi_nosave(icard, DH_RG_MRTC, 0);
}

static void dh_clear_rx_dma(int icard) {
  /* Clears RX DMA after DMA timeout - should never happen; but Kalle suggests
     that if it does we reset the DMA counter and WR enable bit */
  dh_clear_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_MR_WR_EN);
  dh_write_regi_nosave(icard, DH_RG_MWTC, 0);
}

#ifdef DEBUG_BUFFER_PACKET_DETAILS
static void dumpd(int minor) {
  int iq, maxq=NDBGQENTRIES*2;
  for(iq=0; iq<maxq; iq++) {
    if(dbgQisEmpty(dbgq[minor])) break;
    char * sq = getDbgQ(dbgq[minor]);
    printk(KERN_INFO "%d: %s\n", minor, sq);
  }
  if(!dbgQisEmpty(dbgq[minor])) {
    printk(KERN_INFO "WARNING: debug queue for minor %d not empty (%d remain), "
	   "forceably emptying.\n", minor, nDbgQEntries(dbgq[minor]));
    emptyDbgQ(dbgq[minor]);
  }
}
#define TXMODE 0
#define RXMODE 1
static inline void showpkt(int mode, int minor, u32 *p) {
  printd(dbgq[minor], "%s %dB %d %s%s%s%s%s",
         mode==TXMODE?"TX":"  ", dhswpkt_len(p),
         dhswpkt_seqn(p), dhswpkt_isack(p)?"ACK ":"",
         dhswpkt_isic(p)?"IC ":"", dhswpkt_isci(p)?"CI ":"",
         dhswpkt_issynfin(p)?"SYN/FIN ":"",
         mode==RXMODE?"RX":"");
}
#else
#define dumpd(...)
#define showpkt(...)
#endif

void dh_dump_fpga(int icard) {
  printk(KERN_WARNING "FPGA registers:\n");
  printk(KERN_WARNING "CTRL  0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_CTRL));
  printk(KERN_WARNING "GSTAT 0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_GSTAT));
  printk(KERN_WARNING "DSTAT 0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_DSTAT));
  printk(KERN_WARNING "TTSIC 0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_TTSIC));
  printk(KERN_WARNING "RTSIC 0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_RTSIC));
  printk(KERN_WARNING "INTEN 0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_INTEN));
  printk(KERN_WARNING "DOMS  0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_DOMS));
  printk(KERN_WARNING "MRAR  0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_MRAR));
  printk(KERN_WARNING "MRTC  0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_MRTC));
  printk(KERN_WARNING "MWAR  0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_MWAR));
  printk(KERN_WARNING "MWTC  0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_MWTC));
  printk(KERN_WARNING "CURL  0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_CURL));
  printk(KERN_WARNING "DCUR  0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_DCUR));
  printk(KERN_WARNING "FLASH 0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_FLASH));
  printk(KERN_WARNING "DOMC  0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_DOMC));
  printk(KERN_WARNING "CERR  0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_CERR));
  printk(KERN_WARNING "DCREV 0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_DCREV));
  printk(KERN_WARNING "FREV  0x%08lx\n", (unsigned long) dh_read_regi(icard, DH_RG_FREV));
}


void dh_drain_and_dump_rx_fifo(int icard, int ich, int minor) {
  printk(KERN_ERR "DUMPING RX FIFO for card %d ch %d minor %d\n", icard, ich, minor);
  printk(KERN_ERR "jiffies=%ld nonemptyRXjif=%ld dt=%ld\n", jiffies, 
	 dh_domvars[minor].nonemptyRXjif, jiffies-dh_domvars[minor].nonemptyRXjif);
  printk(KERN_ERR "Displaying register dump prior to draining FIFO...\n");
  dh_dump_fpga(icard);
# define MAXRXDUMP (DH_MAX_HW_PKT_WORDS * 4) /* REALLY drain it */
  int idump;
  for(idump=0; idump<MAXRXDUMP; idump++) {
    u32 rtsic = dh_read_regi(icard, DH_RG_RTSIC);
    printk(KERN_ERR "idump=%04d rtsic=0x%08lx fifo=0x%08lx\n", idump, 
	   (unsigned long) rtsic,
	   (unsigned long) dh_read_regi(icard, DH_RG_MBF0 + ich)); 
    if(rtsic & DH_BITN(ich)) { /* FIFO is now empty - break */
      break;
    }
    
  }
}


static void dh_enqueue_message_from_rx_fifo(u32 * buffer, int buflen, int icard, int ich) {
  /** Take a message received in firmware's RX fifo 
      icard is the card number and ich is 
      from 0 .. 7.  Assumes that the RTSIC register has been checked to make sure that
      the RX FIFO for that DOM has a complete message. Also assumes there
      is room in the queue. */

  int ok;
  u32 word;
  int nwords,iword, bytesremain;
  unsigned char * dataptr;
  int ibyte;
  dma_addr_t target_addr;
# define MAX_DMA_POLLS 4000 /* assume 1 usec/bus read */
  int itrial;
  //u32 gstat;
  u32 mwtc;
  int sw_len;
  
  dh_bench_start(rxnqbench);
  unsigned char * cbuf = (unsigned char *) buffer;
  int minor = (icard * DH_DOMS_PER_CARD) + ich;

  vvvprintk(KERN_INFO "DH: dh_enqueue_message_from_rx_fifo: "\
	 "card %d, chan(0..7) %d, minor %d.\n", icard, ich, minor);

  /* Figure out where to stuff the bytes */
  //dataptr = scratch_msg->data;
  //dataptr = (unsigned char *) rxpkt[minor];

  /* First word has byte count -- convert to little-endian (portability) */

  buffer[0] = dh_read_regi(icard, DH_RG_MBF0 + ich);
  int bytecnt = dhswpkt_len(buffer); /* only 1st word exists, but ok */
  u16 seqn    = dhswpkt_seqn(buffer);
  int isack   = dhswpkt_isack(buffer);
  vvprintk(KERN_INFO "DH(%lu): RX, len=%dB seqn=%d %s\n",jiffies&0xFFFF, bytecnt,
	  seqn, isack?"ACK":"");

  vvprintk(KERN_INFO "DH: RX msg byte count is %d.\n", bytecnt);
  vvprintk(KERN_INFO "DH: RXed HDR 0x%08lx\n", (unsigned long) buffer[0]);

  if(((bytecnt+3)/4)*4+4 > buflen) {
    printk(KERN_ERR "dh_enqueue_message_from_rx_fifo: Corrupt bytecount (%d) on card %d, "
	   "channel %d.  Returning 0-length (bogus) packet.\n",
	   bytecnt, icard, ich);
    dh_drain_and_dump_rx_fifo(icard, ich, minor);
    dh_bench_end(&rxnqbench);
    return;
  }

  nwords = (bytecnt+3)/4;
  vvprintk(KERN_INFO "RX: will DMA %d words.\n", nwords);
  if(nwords > 0) {

    if(parm_rx_dma) {
      dh_bench_start(rxdmabench);
      target_addr = pci_map_single(dh_pci_dors[icard].dev, buffer+1, 
				   DH_MAX_HW_PKTSIZ-4, PCI_DMA_FROMDEVICE);
      vvprintk(KERN_INFO "dma target addr 0x%lx\n", (unsigned long) target_addr);
      if(target_addr == 0) {
	printk(KERN_ERR 
	       "dh_enqueue_message_from_rx_fifo ERROR! Couldn't get DMA target address!\n");
	dh_lasterr = DH_ERR_DMASETUPFAIL;
	dh_bench_end(&rxnqbench);
	return; /* No real way to handle error, this is in the interrupt handler */
      }
      vvprintk(KERN_INFO "byte count is %d\n", bytecnt);
      /* Load correct address into DOR register */
      dh_write_regi(icard, DH_RG_MWAR, target_addr); 
      /* Set byte count and DOM number (ASSUMES ich in 0..7) */
      dh_write_regi(icard, DH_RG_MWTC, (nwords<<2) | (ich<<24));
      wmb();
      //dmarx_bench
      //dh_bench_start(dmarx_bench);
      dh_set_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_MR_WR_EN);
      /* See if it finished */
      ok = 0;
      for(itrial=0; itrial<MAX_DMA_POLLS; itrial++) {
	mwtc = dh_read_regi(icard, DH_RG_MWTC);
	rmb();
	vvprintk(KERN_INFO "DMA polling... MWTC=0x%lx.\n", (unsigned long) mwtc);	
	if((mwtc & 0xFFFFF) == 0) {
	  ok = 1;
	  break;
	}
      }
      dh_clear_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_MR_WR_EN);
      /* Whether we succeeded or failed, release DMA mapping */
      pci_unmap_single(dh_pci_dors[icard].dev, target_addr, DH_MAX_HW_PKTSIZ-4, 
		       PCI_DMA_FROMDEVICE);
      dh_bench_end(&rxdmabench);      
      if(! ok) {
	printk(KERN_ERR "dh_enqueue_message_from_rx_fifo: DMA failed!\n");
	dh_clear_rx_dma(icard);
	dh_lasterr = DH_ERR_DMAFAIL;
	dh_bench_end(&rxnqbench);
	return;
      }
      vvprintk(KERN_INFO "dh_enqueue_message_from_rx_fifo: DMA succeeded!\n"); 
    } else { 
#define skipinsl 1
#undef skipinsl
#ifndef skipinsl
      dh_bench_start(inslbench);
      /* FIFO drain w/o DMA.  This is really compact though it breaks our philosophy slightly */
      insl(dh_pci_dors[icard].rstart + (DH_RG_MBF0 + ich)*4, buffer+1, nwords);
      dh_bench_end(&inslbench);
#else
#warning insl disabled, put it back
      int il;
      for(il = 0; il < nwords; il++) {
	int retry;
#define MAX_EMPTY_RETRIES 1000
	for(retry=0; retry<MAX_EMPTY_RETRIES; retry++) {
	  if(dh_rx_fifo_is_empty(icard, ich)) 
	    printk(KERN_WARNING "RX: FIFO empty midpacket! icard=%d ich=%d il=%d retry=%d.\n",
		   icard, ich, il, retry);
	  else break;
	}
	buffer[il+1] = dh_read_regi(icard, DH_RG_MBF0 + ich);
      }
#endif

#ifdef DH_TEST_INC_PAT
#warning DH_TEST_INC_PAT
      dh_test_for_inc_pat(cbuf+4, bytecnt, 0);
#endif

    } /* no DMA */
  } /* If data left in FIFO */

  /* At this point, I have some sort of hardware packet */
  dh_bench_end(&rxnqbench);
  showpkt(RXMODE,minor,buffer);
}

static inline int dh_cooked_len(int swlen) { return ((swlen+3)/4)*4; }

static void dh_cook(u32 *cooked, unsigned char *raw, unsigned short flags, 
		   unsigned short seqn, int len) {
  /* Convert raw packet (just a bunch of bytes, and a raw len) into a cooked packet 
     (header, contents (padded to word align at end), CRC) */
  int cooked_len;

  dhswpkt_setSeq(cooked, seqn);                 /* Set sequence number */
  dhswpkt_setLen(cooked, len);                  /* Set internal len */
  dhswpkt_setFlags(cooked, flags);

  vvvprintk(KERN_INFO "dh_cook: flags=%d\n",flags);
  /* Make sure we don't overflow */
  len = (len > DH_MAX_SW_PKTSIZ ? DH_MAX_SW_PKTSIZ : len);

  if(raw != NULL) {
    memcpy(dhswpkt_data(cooked), raw, len);       /* Copy contents */
  }
}

static void dh_dequeue_message_to_tx_fifo(u32 *dataptr, int icard, int ich) {
  /** Take a ** COOKED **
      message from the appropriate device's output queue and put it into the firmware's
      TX fifo. */
  u32 word, mrtc;
  int iword, ibyte, nwords, ok, fifo_almost_empty, itrial, ib;

  int minor = (icard * DH_DOMS_PER_CARD) + ich;
  showpkt(TXMODE,minor,dataptr);

  dh_bench_start(txdqbench);

  int pktlen = dhswpkt_len(dataptr);

#ifdef DH_DO_POLLUTE
  if(dh_pollute[minor] && pktlen > 0) {
    dataptr[0] ^= 0xFF000000; /* poison sequence number */
    dh_pollute[minor]--;
  }
#endif

  if(! dh_tx_fifo_hasroom(icard, ich, pktlen)) {
    printk(KERN_WARNING "TX: WARNING: TX FIFO too full for a message.... discarding message.\n");
    dh_lasterr = DH_ERR_TXFIFOFULL;
    dh_bench_end(&txdqbench);
    return;
  }
  
  /* Packet header is already formed up (cooked) -- barf it out and bye-bye */
  nwords = (pktlen+3)/4+1;


  if(!dhswpkt_isack(dataptr)) 
    vvprintk(KERN_INFO "DH (%lu): seqn=%d->TX\n", jiffies&0xFFFF, 
	     dhswpkt_seqn(dataptr));

  if(parm_tx_dma) {
    /* Send data using DMA.  Code adapted from dh_enqueue_message_from_rx_fifo */
    dh_bench_start(rxdmabench);

#define CHECK_DMA_DONE_BEFORE /* Toggle to change methodology */
#ifndef CHECK_DMA_DONE_BEFORE
#define CHECK_DMA_DONE_AFTER
#endif

#ifdef  CHECK_DMA_DONE_BEFORE
    ok = 0;
    unsigned long tlo0,tlo1;
    rdtscl(tlo0);
    for(itrial=0; itrial < MAX_DMA_POLLS; itrial++) {
      mrtc = dh_read_regi(icard, DH_RG_MRTC);
      rmb();
      vvvprintk(KERN_INFO "DMA polling... MRTC=0x%lx.\n", (unsigned long) mrtc);
      if((mrtc & 0xFFFFF) == 0) {
        ok = 1;
        break;
      }
    }
    if(!ok) {
      rdtscl(tlo1);
      unsigned long dt = tlo1-tlo0;
      unsigned long ttsic = dh_read_regi(icard, DH_RG_TTSIC);
      printk(KERN_ERR "DH: TX: DMA error, MRTC (0x%08lx) didn't clear on minor %d, "
	     "dt=%ld ticks, packet won't be transferred (TTSIC=0x%08lx)!\n", 
	     (unsigned long) mrtc, minor, dt, (unsigned long) ttsic);
    }
    /* Unmap DMA regardless of whether or not it succeeded... */
    pci_unmap_single(dh_pci_dors[icard].dev, dh_pci_dors[icard].tx_dma_target_addr, 
		     DH_MAX_HW_PKTSIZ, PCI_DMA_TODEVICE);

#else 
    ok = 1;
#endif
    
    if(ok) {
      dh_pci_dors[icard].tx_dma_target_addr
	= pci_map_single(dh_pci_dors[icard].dev, dataptr, DH_MAX_HW_PKTSIZ, PCI_DMA_TODEVICE);
      vvvprintk(KERN_INFO "dma target addr 0x%lx\n", 
		(unsigned long) dh_pci_dors[icard].tx_dma_target_addr);
      
      if(dh_pci_dors[icard].tx_dma_target_addr == 0) {
	printk(KERN_ERR "TX: ERROR! Couldn't get DMA target address!\n");
	dh_lasterr = DH_ERR_DMASETUPFAIL;
	dh_bench_end(&txdqbench);
	return;
      }

      /* Load correct address into DOR register */
      dh_write_regi(icard, DH_RG_MRAR, dh_pci_dors[icard].tx_dma_target_addr);
      /* Set byte count and DOM number (ASSUMES ich in 0..7) */
      dh_write_regi(icard, DH_RG_MRTC, (nwords<<2) | (ich<<24));
      
      wmb();
      dh_set_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_MR_RD_EN | DH_CTRL_MEM_RD_MULT);
      
#ifdef CHECK_DMA_DONE_AFTER
      ok = 0;
      for(itrial=0; itrial < MAX_DMA_POLLS; itrial++) {
	mrtc = dh_read_regi(icard, DH_RG_MRTC);
	rmb();
	vvvprintk(KERN_INFO "DMA polling... MRTC=0x%lx.\n", (unsigned long) mrtc);
	if((mrtc & 0xFFFFF) == 0) {
	  ok = 1;
	  break;
	}
      }

      /* Whether we succeeded or failed, release DMA mapping */
      pci_unmap_single(dh_pci_dors[icard].dev, dh_pci_dors[icard].tx_dma_target_addr, 
		       DH_MAX_HW_PKTSIZ, PCI_DMA_TODEVICE);
#endif
    }
    
    dh_bench_end(&txdmabench);

    if(! ok) {
      printk(KERN_ERR "DH: TX: DMA failed (timeout) on minor %d (MRTC=0x%08lx)!\n", minor,
	     (unsigned long) mrtc);
      dh_clear_tx_dma(icard);
      dh_lasterr = DH_ERR_DMAFAIL;
      dh_bench_end(&txdqbench);
      return;
    }
    
    /* Send remaining (non-word-aligned) bytes: done below, but update dataptr to reflect
       transfer. */
  } else { /* Non-DMA transfer */
    dh_bench_start(outslbench);
    outsl(dh_pci_dors[icard].rstart + (DH_RG_MBF0 + ich)*4, dataptr, nwords);
    dh_bench_end(&outslbench);
#ifdef DH_TEST_FOR_ZEROES
    unsigned char * cbuf = (unsigned char *) dataptr;
    int iz;
    for(iz=0; iz<pktlen; iz++)
      if(cbuf[iz+4] == 0) 
	printk(KERN_WARNING "TX(dequeue): got 0 at byte iz=%d nwords=%d pktlen=%d!\n",
	       iz, nwords, pktlen);
#endif
  }   /* else if not dma */
  
  if(vv()) {
    int emptystat = dh_tx_fifo_is_empty(icard, ich);
    vvprintk(KERN_INFO "DH: TX 0x%08lx%s <- minor %d (TX FIFO %s)\n", 
	     (unsigned long) dataptr[0], (dataptr[0]&0xFFF)?"...":".", minor, 
	     emptystat?"EMPTY!!!":"not empty");
  }

  dh_bench_end(&txdqbench);
}


static int dh_some_device_open(void) {
  /** @return 1 if one or more DOR cards has been opened, else 0 */
  int idev;
  for(idev=0;idev < DH_DOMHUB_DOMS; idev++) {
    if(dh_opened[idev]) {
      return 1;
    }
  }
  return 0;
}


static void dh_pair_pwr_off(int icard, int ipair) {
  /** Tell FPGA design on card icard to power off DOM pair ipair. */
  /** LOCKS card */
  unsigned long flags;

  if(ipair >= 0 && ipair <= 3) {
    vvprintk(KERN_INFO "dh: Card %d DOM pair %d OFF.\n",icard, ipair);

    /* Kalle now says wait up to 80 msec for DSTAT(WP_PON_RDY) before
       switching power state */
    int gotPwrOnRdy = 0;
    unsigned long startjif = jiffies;
    unsigned long maxjif   = startjif + (80*HZ)/1000 + 1; /* At least 80 msec */
    while(jiffies < maxjif) { 
      /* Check it */
      u32 dstat = dh_read_regi(icard, DH_RG_DSTAT);
      if(dstat & DH_BITN(ipair+4)) {
	gotPwrOnRdy = 1;
	break;
      }
      schedule();
    }
    if(!gotPwrOnRdy && (ipair < 2 || dh_pci_dors[icard].rev > 0)) { /* Skip for Rev 0, top wps */
      printk(KERN_WARNING 
	     "dh_pair_pwr_off: Card %d pair %d WP_PON_RDY bit never went high, "
	     "forceably switching off anyways...\n", icard, ipair);
    }

    dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
    /* Explicitly clear non-powered pairs: even though CTRL is off, bits
       in the software copy of the registers might be on, and we want to avoid
       toggling the actual wire pairs on */
    u32 ctrl      = dh_read_regi(icard, DH_RG_CTRL);
    u32 clearMask = (0x0000000F & ~ctrl) | DH_BITN(ipair);
    dh_clear_write_regi_bits(icard, DH_RG_CTRL, clearMask);
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/

    int minorA = dh_get_minor_from_card_pair_and_dom(icard, ipair, 0);
    int minorB = dh_get_minor_from_card_pair_and_dom(icard, ipair, 1);
    dh_conn_state[minorA] = DH_STATE_START;
    dh_conn_state[minorB] = DH_STATE_START;
  } else {
    dh_lasterr = DH_ERR_INVALIDARG;
    printk(KERN_WARNING "dh_pair_pwr_off: Invalid wire pair argument (%d).\n",ipair);
  }
}


static int dh_hardware_timeout_set(int icard, int ich) {
  u32 doms;
  doms = dh_read_regi(icard, DH_RG_DOMS);
  if(ich < 0 || ich > 7) { 
    return 0;
  }
  return (doms & DH_BITN(8+ich))!=0;
}


static int dh_is_pair_plugged(int icard, int ipair) {
  u32 dstat;
  dstat = dh_read_regi(icard, DH_RG_DSTAT);
  
  if(ipair < 0 || ipair > 3) {
    return 0;
  } 
  return (dstat >> ipair) & 0x01;
}



static void dh_channel_comms_reset(int icard, int channel_mask) {
  /* Reset Kalle's state machines for all cards on this channel. */
  pprintk(KERN_INFO "dh_channel_comms_reset: resetting card %d channel mask 0x%x.\n",icard,channel_mask);
  //if(ich < 0 || ich > DH_DOMS_PER_CARD) return;
  //iwp = ich/2;
  //dh_set_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_LOCAL_OSC_EN);
  //dh_set_write_regi_bits(icard, DH_RG_DOMC, DH_BITN(8 + ich));
  dh_write_regi_nosave(icard, DH_RG_DOMC, (channel_mask & 0xFF) << 8);
}



static int dh_dom_isready(int icard, int ich) {
  int icounter, loopmax = 1000;
  int comm_reset_done, dom_detected;
  u32 domc, doms;
  int false = 0, true = 1;
  comm_reset_done = false;

  if(ich < 0 || ich > 7) return false;
  for(icounter=0; icounter < loopmax; icounter++) {
    domc = dh_read_regi(icard, DH_RG_DOMC);
    if(!( domc & DH_BITN(8 + ich)) ) {
      comm_reset_done = true;
      break;
    }
    dh_jsleep(1);
  }
  if(!comm_reset_done) {
    printk(KERN_WARNING
	   "dh_dom_isready: WARNING: communications reset bit never went low.\n");
    domc = dh_read_regi(icard, DH_RG_DOMC);
    
    /* Don't need the following, since I write_regi_nosave: */
/*     domc &= ~(DH_BITN(8+ich)); /\* Clearing reset request bit *\/ */
/*     dh_write_regi(icard, DH_RG_DOMC, domc); */

    /* Skip for now: */
/*     domc = domc & ~( (u32) DH_BITN(8+ich) ); */
/*     dh_write_regi_nosave(icard, DH_RG_DOMC, domc); */
/*     return false; */
  }

  dom_detected = false;
  for(icounter=0; icounter < loopmax; icounter++) {
    doms = dh_read_regi(icard, DH_RG_DOMS);
    if(doms & DH_BITN(ich)) {
      dom_detected = true;
      break;
    }
    dh_jsleep(1);
  }

  if(!dom_detected) {
    printk(KERN_WARNING "dh_dom_isready: DOM %d was not detected in DOMS register.\n", ich);
    return false;
  }

  vprintk(KERN_INFO "dh_dom_isready: DOM %d is ready in DOMS.\n", ich);
  return true;
}


static int dh_dom_softboot(int icard, int ipair, int idom) {
  /** Issue softboot according to semantics agreed upon w/ Chuck,
      David and Kalle:
      1) Disable RX interrupts
      2) Drain inq
      3) Issue soft reset signal to DOMC
      4) Explicitly clear reset in DOMC via bit 20 in CTRL
      5) Re-enable RX interrupts */
  unsigned long flags;
  int ich, itry, minor;
#define SOFTBOOT_MAXTRIES 40
#define SOFTBOOT_TRIES_PER_SEC 10
  ich = ipair * DH_DOMS_PER_PAIR + idom;
  if(ich<0 || ich >= DH_DOMS_PER_CARD) return 1;

  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  /* Check for other channel empty / not empty */
  dh_write_regi_nosave(icard, DH_RG_DOMC, DH_BITN(ich));
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/

  int got_domc = 0;
  for(itry=0; itry<SOFTBOOT_MAXTRIES; itry++) {
    if(! (dh_read_regi(icard, DH_RG_DOMC) & DH_BITN(ich))) {
      got_domc = 1;
      break;
    } else if(itry == SOFTBOOT_MAXTRIES-1) {
      /* Don't reenable interrupts cause something is wrong */

      /* Skip this for now, can hose power on operations on other wire pairs */

/*       /\* Clear communications/softboot reset bits explicitly so that firmware doesn't waste */
/* 	 bandwidth (per Kalle) *\/ */
/*       dh_card_lock(icard, flags); /\*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*\/ */
/*       dh_set_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_CLR_COM_RES); */
/*       dh_clear_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_CLR_COM_RES); */
/*       dh_card_unlock(icard, flags); /\*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*\/ */
      return 1;
    } else {
      unsigned long jifdelay = HZ/SOFTBOOT_TRIES_PER_SEC+1;
      vvprintk(KERN_INFO "%d: Going to sleep for %ld jiffies...\n", itry, jifdelay);
      set_current_state(TASK_INTERRUPTIBLE);
      schedule_timeout(jifdelay);
      set_current_state(TASK_RUNNING);
      //dh_jsleep(1);
      if(signal_pending(current)) {
	printk(KERN_INFO "Got signal during softboot!\n");
	return 1;
      }
    }
  }

  return 0;
}


static int dh_get_pair_pwr_state(int icard, int ipair) {
  /* Get actual power state of (icard,ipair) from the hardware.
     return 1 if off, 2 if on, 3 if undefined/unknown/in between */

  u32 dstat = dh_read_regi(icard, DH_RG_DSTAT);
  u32 ctrl  = dh_read_regi(icard, DH_RG_CTRL);

  /* Get bits 4..7, shift so they are at positions 0..3. 
     Now mask w/ CTRL WP_PON since reload of FPGA seems to not clear DSTAT */
  u32 pwrbits = ((dstat >> 4) & 0x0F) & ctrl; 

  if(ipair < 0 || ipair >= DH_PAIRS_PER_CARD) {
    return 3;
  } else if((pwrbits >> ipair) & 0x01) {
    return 2;
  } else {
    return 1;
  }
}

static void dh_set_local_osc(int icard, int use_external_clock) {
  if(use_external_clock) {
    dh_clear_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_LOCAL_OSC_EN);
  } else {
    dh_set_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_LOCAL_OSC_EN);
  }
  /* Per Kalle - wait 10 msec */
  dh_jsleep( (HZ*10)/1000 + 1);
}

static void dh_set_timeout_enable(int icard) {
  dh_set_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_COM_TOUT_EN);
}


static void dh_clear_domcontrol_regi(int icard) {
  pprintk(KERN_INFO "dh_clear_domcontrol_regi: Zeroing DOMC...\n");
  dh_write_regi(icard, DH_RG_DOMC, 0);
}


static void dh_clear_rx_test_api_interrupt(int icard) {
  dh_clear_write_regi_bits(  icard, 29, DH_BITN(26) | DH_BITN(31)); /* INT_ON_FIFO_AFF and global */
  vprintk(KERN_INFO "Clearing interrupts on card %d...\n", icard);
  dh_write_regi(icard,29,0);
}


static int dh_get_comms_errors(int icard, int ich) {
  /* Get comms errors as given by CERR register 
     Must spinlock FPGA on card! */
  u32 cerr;
  cerr = (ich&07) << 28;
  dh_write_regi(icard, DH_RG_CERR, cerr);
  cerr = dh_read_regi(icard, DH_RG_CERR);
  return cerr & 0xFFFF; /* 16 bits only */
}


static int dh_get_dom_id(u64 * id, int icard, int ich) {
  /** Fill 64 bit DOM ID as returned by Kalle's firmware.
      return 1 if error. */
# define MAX_SECONDS 5
  int itry;
  u32 w0, w1;
  if(ich < 0 || ich > DH_DOMS_PER_CARD) return 1;
  if(icard < 0 || icard > DH_MAX_CARDS) return 1;

  /* Signal request for ID */
  unsigned long flags;

  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  dh_write_regi_nosave(icard, DH_RG_DOMC, DH_BITN(16+ich));
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
  unsigned long startjif = jiffies;
  unsigned long maxjif   = startjif + MAX_SECONDS * HZ;
  int gotid = 0;
  while(jiffies < maxjif) {
    if(dh_read_regi(icard, DH_RG_DOMC) & DH_BITN(16+ich)) {
      schedule();
    } else {
      gotid++;
      break;
    }
  }
  if(!gotid) {
    printk(KERN_WARNING "dh_get_dom_id: Timeout waiting for ID to arrive.\n");
    return 1; 
  }

  /* Read out results */
  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  dh_write_regi_nosave(icard, DH_RG_DOMID, (ich&0x07)<<25);
  w0 = dh_read_regi(icard, DH_RG_DOMID);
  dh_write_regi_nosave(icard, DH_RG_DOMID, ((ich&0x07)<<25) | DH_BITN(24));
  w1 = dh_read_regi(icard, DH_RG_DOMID);
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/

  *id = ( ((u64) w0) & 0xFFFFFF ) | ( ( ((u64) w1) & 0xFFFFFF) << 24 );
  return 0;
}


static void dh_dump_comm_debug_fifo(int icard) {
  int i;
  u32 gstat;
  vprintk(KERN_INFO "dh_dump_comm_debug_fifo: Contents of GSTAT comms. debug FIFO follow...\n");
  vprintk(KERN_INFO "TX:\n");
  for(i=0; i < DH_MAX_DEBUG_FIFO_READS; i++) {
    if(i != 0) {
      dh_write_regi_nosave(icard, DH_RG_GSTAT, DH_BITN(0)); /* Fetches next byte */
    }
    gstat = dh_read_regi(icard, DH_RG_GSTAT);
    vprintk(KERN_INFO "FIFO word %d GSTAT 0x%08lx (%s)\n", i,
	   (unsigned long) gstat, dh_comprot[ (gstat >> 24) & 0x0F ]);
    if(gstat & DH_BITN(22)) { /* FIFO empty */
      break;
    }
  }
  vprintk(KERN_INFO "RX:\n");
  for(i=0; i < DH_MAX_DEBUG_FIFO_READS; i++) {
    if(i != 0) {
      dh_write_regi_nosave(icard, DH_RG_GSTAT, DH_BITN(1)); /* Fetches next byte */
    }
    gstat = dh_read_regi(icard, DH_RG_GSTAT);
    vprintk(KERN_INFO "FIFO word %d GSTAT 0x%08lx (%s)\n", i,
           (unsigned long) gstat, dh_comprot[ (gstat >> 28) & 0x0F ]);
    if(gstat & DH_BITN(23)) { /* FIFO empty */
      break;
    }
  }


}

/***************** End of miscellaneous functions *****************/


/************* Interrupt Handler and Bottom Halves ****************/

static inline int dh_mrtc_clear(int icard) {
  u32 mrtc = dh_read_regi(icard, DH_RG_MRTC);
  return !(mrtc & 0xFFFFF);
}

static inline int dh_mwtc_clear(int icard) {
  u32 mwtc = dh_read_regi(icard, DH_RG_MWTC);
  return !(mwtc & 0xFFFFF);
}

static void dh_do_send_message(int icard) {
  /** This was the "transmit" bottom half for our interrupt handler.
      Dequeue the data from any output queues which have data into the
      appropriate TX FIFOs.  Re-enable "interrupt on message sendable"
      if more data needs to be sent.  wake_up_interruptible() any
      writing process which has blocked because the outgoing queue is
      full. */
  int ich;   /* Which channel on card, from 0..7 */
  int minor; /* Minor (device) number, from 0..63 */
  int msgcnt;
  int empty;
  int hwpktlen;
  u32 ttsic;

  dh_bench_start(txbench);
  vvprintk(KERN_INFO "dh_do_send_message.\n");
  
  int sentdata = 0;
  
  for(ich=0; ich < DH_PAIRS_PER_CARD*DH_DOMS_PER_PAIR; ich++) {
    
    minor = (icard * DH_PAIRS_PER_CARD*DH_DOMS_PER_PAIR) + ich;
    
    if(! dh_opened[minor]) continue; /* Skip unopened channels */
    /* Skip over channel if not connected -- let kernel timer handle it */
    if(dh_conn_state[minor] != DH_STATE_CONNECTED) continue;
    
    msgcnt = 0;
    empty = 0;
    vvprintk(KERN_INFO "DH(TX): Got interrupt, sending all pending messages on minor %d.\n",
	     minor);
    
    /* Send pending ACKs... */
    /*                                  ACKs are all DH_MIN_SW_PKTSIZ in length: */
    if(!ackQisEmpty(&ackq[minor]) 
       && dh_tx_fifo_hasroom(icard, ich, DH_MIN_SW_PKTSIZ)
       && dh_mrtc_clear(icard)) {

      aqType seqn = getAckQ(&ackq[minor]);
      dh_cook(txpkt[minor], NULL, FL_ACK, seqn, 0);
      vvprintk(KERN_INFO "DH(TX,%lu,%s): Send ACK seqn=%hd.\n", jiffies&0xFFFF,
	       dh_rx_fifo_has_message(icard, ich)?"RX+":"rx-", seqn);
      dh_dequeue_message_to_tx_fifo(txpkt[minor], icard, ich);
      dh_ackwritten[minor]++;
      dh_pktswritten[minor]++;
      txfree[minor] -= ( (dhswpkt_len(txpkt[minor])+3)/4*4 + 4 );
      wake_up_interruptible(&ackqwq[minor]);
      sentdata++;
      continue;
    }
    
    /* Send pending ReTXes */
    if(!rtxQisEmpty(&retxq[minor]) && dh_mrtc_clear(icard)) { 
      int next_retx_data_len = nextRetxQEntrySize(&retxq[minor]);
      int pktlen = dh_cooked_len(next_retx_data_len);
      if(dh_tx_fifo_hasroom(icard, ich, pktlen)) {      
	rtxMsg *m = getRetxQ(&retxq[minor]);
	dh_cook(txpkt[minor], m->data, m->flags, m->seqn, m->len);
	vvprintk(KERN_INFO "DH(TX,%lu,%s): Sending %d byte ReTX SEQN=%hd.\n", 
		 jiffies&0xFFFF, dh_rx_fifo_has_message(icard, ich)?"RX+":"rx-", m->len, m->seqn);
	dh_dequeue_message_to_tx_fifo(txpkt[minor], icard, ich);
	dh_resent[minor]++;
	dh_pktswritten[minor]++;
	txfree[minor] -= ( (dhswpkt_len(txpkt[minor])+3)/4*4 + 4 );
	sentdata++;
	continue;
      }
    }

    /* Send pending messages, if RETX buffer not full, and if RETXes haven't timed out
       (want to maintain sequence number) */
    if(!ioQisEmpty(outq[minor]) 
       && dh_retx_buf_not_full(minor) 
       && !retx_triggered[minor]
       && dh_mrtc_clear(icard)) {
      
      int next_tx_data_len = nextIOQEntrySize(outq[minor]);
      if(!dh_retxb_past_bytelimit(minor, next_tx_data_len)) { /* Bail if RETX already
								 has a lot of data, to
								 match DOM's read buffer. */
	int pktlen = dh_cooked_len(next_tx_data_len);
	if(dh_tx_fifo_hasroom(icard, ich, pktlen)) { /* Bail if not enough room */
	  ioqMsg * m    = getIOQ(outq[minor]);
	  int nbytes    = m->nbytes;
	  int is_synfin = m->synfin;
	  memcpy(dh_tx_scratch_msg[minor], m, sizeof(ioqMsg)); 
	  /* For optimization, do partial memcpy */

	  unsigned short flags = FL_DATA;
	  if(is_synfin) flags |= FL_SYNFIN;
	  vvprintk(KERN_INFO "DH(TX,%lu,%s): %d bytes, outq -> TX FIFO(flags=%d seqn=%d)\n", 
		   jiffies&0xFFFF, dh_rx_fifo_has_message(icard, ich)?"RX+":"rx-",
		   nbytes, flags, txseq[minor]);
	  dh_cook(txpkt[minor], dh_tx_scratch_msg[minor]->data, flags, 
		  txseq[minor], nbytes);
	  dh_dequeue_message_to_tx_fifo(txpkt[minor], icard, ich);
	  sentdata++;
	  msgcnt++;
      
	  /* Add to retx buffer */
	  dh_retx_buf_add(minor, dh_tx_scratch_msg[minor]->data, nbytes, txseq[minor], flags, 
			  jiffies);
	  dh_enable_msg_recvd_intr_chan(icard, ich); /* Must enable so ACK can be caught */
      
	  txseq[minor]++;
	  dh_pktswritten[minor]++;
	  /* Wake up any process which has blocked waiting for
	     room to open up in TX FIFO */
	  wake_up_interruptible(&outq[minor]->wq);
	  txfree[minor] -= ( (dhswpkt_len(txpkt[minor])+3)/4*4 + 4 );
	  continue;
	} /* If room in fifo */
      } /* If room in retxb */
    } /* OK to send pending messages */
    
  } /* Loop over channels of this card */

  /* reenable interrupts if have retx queue entries, ack queue entries */
  for(ich=0; ich < DH_PAIRS_PER_CARD*DH_DOMS_PER_PAIR; ich++) {
    minor = (icard * DH_PAIRS_PER_CARD*DH_DOMS_PER_PAIR) + ich;
    if( !rtxQisEmpty(&retxq[minor]) || !ackQisEmpty(&ackq[minor])) {
      dh_enable_msg_sndbl_intr_chan(icard, ich);
    }
    /* outq entries cause a reenable when an ACK is received, or by the kernel timer.
       Otherwise you get lots of spurious interrupts when retx is full. */
  }

  dh_bench_end(&txbench);
}


static void dh_fill_ctrlpkt(DomStatsPkt * dsp, unsigned char * src) {
  /* Show contents of received control packet */
  DomStatsPkt * srcPkt = (DomStatsPkt *) src;

  dsp->controlType = srcPkt->controlType;
  dsp->nBadFins = srcPkt->nBadFins;
  dsp->minRxQueueSize = srcPkt->minRxQueueSize;
  dsp->minRetxEntries = srcPkt->minRetxEntries;
  dsp->minAckQueueEntries = srcPkt->minAckQueueEntries;
  dsp->nTxPkts = srcPkt->nTxPkts;
  dsp->nTxAckPkts = srcPkt->nTxAckPkts;
  dsp->nTxReackPkts = srcPkt->nTxReackPkts;
  dsp->nTxDataPkts = srcPkt->nTxDataPkts;
  dsp->nTxResentPkts = srcPkt->nTxResentPkts;
  dsp->nTxControlPkts = srcPkt->nTxControlPkts;
  dsp->nRxPkts = srcPkt->nRxPkts;
  dsp->nRxDataPkts = srcPkt->nRxDataPkts;
  dsp->nRxGoodDataPkts = srcPkt->nRxGoodDataPkts;
  dsp->nRxAckPkts = srcPkt->nRxAckPkts;
  dsp->nRxDupDataPkts = srcPkt->nRxDupDataPkts;
  dsp->nRxDupAcks = srcPkt->nRxDupAcks;
  dsp->nRxGoodAcks = srcPkt->nRxGoodAcks;
  dsp->nRxControlPkts = srcPkt->nRxControlPkts;
  dsp->nRxDroppedPkts = srcPkt->nRxDroppedPkts;
  dsp->nRxBadPkts = srcPkt->nRxBadPkts;
  dsp->nInvalidPostIC = srcPkt->nInvalidPostIC;
}


static void dh_do_recv_message(int icard) {
  /** This was the "receive" bottom half for our interrupt handler.
      Now it's the workhorse for RX interrupts, called at interrupt
      time for performance reasons.
      It removes data from RX FIFOs and enqueue the messages.  Re-enables
      "interrupt on message received" if more data remains.
      wake_up_interruptible() any process blocking on read of the
      queue. */
  int ich, minor;
  u32 rtsic;
  int this_channel_had_data;
  unsigned short seqn;

  dh_bench_start(rxbench);
  vvvprintk(KERN_INFO "DH: Entered dh_do_recv_message (panic=%d).\n",dh_panic);

  int dataread = 0;
  /* Loop over each channel until there is no more data to be read out */
  for(ich=0; ich < DH_DOMS_PER_CARD; ich++) {
    minor = icard * DH_DOMS_PER_CARD  + ich;

    if(!dh_opened[minor]) continue; /* Skip this channel if not opened! */

    /* Skip over channel if not connected -- let kernel timer handle it */
    if(dh_conn_state[minor] != DH_STATE_CONNECTED) continue;
 
    this_channel_had_data = 0;

    int nacks = 0;
    int nrecv = 0;

    dh_bench_start(rxwhile);
    while(dh_mwtc_clear(icard) && dh_rx_fifo_has_message(icard, ich)) {
      vvprintk(KERN_INFO "Reading message on minor %d (RTSIC=0x%08lx)\n",minor,
	       (unsigned long) dh_read_regi(icard, DH_RG_RTSIC));
      /* If we got an interrupt for this channel, we reset the "dead dom" timer,
         keeping track of the largest delta-T value */
      unsigned long dt = jiffies - dh_domvars[minor].nonemptyRXjif;
      if(dh_domvars[minor].had_rx_data && dt > dh_domvars[minor].rxdt)
	dh_domvars[minor].rxdt = dt;

      dh_domvars[minor].nonemptyRXjif = jiffies;

      /* Find out if we have room for it in the input queue */
      if(ioQisFull(inq[minor])) { 
	vprintk(KERN_WARNING "DH: flow ctrl inq(minor=%d), RX stays disabled.\n", minor);
	break;
      } else if(ackQisFull(&ackq[minor])) {
        vprintk(KERN_WARNING "DH: flow ctrl ackq(minor=%d), RX stays disabled.\n", minor);
        break;
      } else {
	/* Read message from RX FIFO into input queue */
	dh_enqueue_message_from_rx_fifo(rxpkt[minor], DH_MAX_HW_PKTSIZ, icard, ich);
	dh_clear_message_received(icard, ich); /* Was below, now do immediately */
	dataread++;
	dh_pktsread[minor]++;
	int len = dhswpkt_len(rxpkt[minor]);
	if(len > DH_MAX_HW_PKTSIZ-4) { 
	  vprintk(KERN_WARNING "DH: RX HW pkt has bad size (%d bytes)\n", len);
	  dh_badhdr[minor]++;
	} 
	else if(dhswpkt_isctrl(rxpkt[minor])) {
	  vvprintk(KERN_INFO "DH: Got control packet.\n");
	  dh_fill_ctrlpkt(&domstats[minor].stats, dhswpkt_data(rxpkt[minor]));
	  domstats[minor].have_new = 1;
	  wake_up_interruptible(&domstatwq[minor]);
	  dh_ctrlrecv[minor]++;
	  /* And pitch it */
	} else if(dhswpkt_isic(rxpkt[minor])) {
	  vprintk(KERN_INFO "DH(%ld): RX (stale) IC!!! Going to UNCONNECTED "
		 "(minor %d)\n", jiffies, minor);
	  printd(dbgq[minor], "Got stale IC!\n");
	  dh_icrecv[minor]++;
	  dh_conn_state[minor] = DH_STATE_FLUSHIC; /* Kernel timer handles rest of resyn */
	  dh_conn_flush_jif[minor] = jiffies;
	  dh_disable_msg_sndbl_intr_chan(icard, ich); /* Disable TX: RX already disabled */
	  /* Blow away retx buffer, etc. */
	  if(dh_num_retxb(minor)) {
	    printk(KERN_WARNING "DH: Warning: nuking retxb with %d pending packets!\n",
		   dh_num_retxb(minor));
	    dumpd(minor);
	  }
	  dh_empty_retxb(minor);
	  retx_triggered[minor] = 0;
	  emptyRetxQ(&retxq[minor]); /* Requeue ReTXes 'cause seqn's have changed */
	  emptyAckQ(&ackq[minor]);
	  break;
	} else if(dhswpkt_isci(rxpkt[minor])) {
	  dh_cirecv[minor]++;
	  printd(dbgq[minor], "RX CI - DROPPING");
	  vvprintk(KERN_INFO "DH: minor=%d, RX CI.... DROPPING....\n", minor);
	  /* Pitch it */
	} else {
	  seqn = dhswpkt_seqn(rxpkt[minor]);
	  vvprintk(KERN_INFO "DH: RX pkt len=%d seqn=%d\n", dhswpkt_len(rxpkt[minor]), seqn);
	  if(dhswpkt_isack(rxpkt[minor])) {
	    printd(dbgq[minor], "RXed ACK");
	    dh_handle_recvd_ack(seqn, minor);
	    /* retx has drained, so we can enable more interrupts to flow */
	    if(!ioQisEmpty(outq[minor]))
	      dh_enable_msg_sndbl_intr_chan(icard, ich);
	    vvprintk(KERN_INFO "DH(RX(%d),%lu): ACK (seqn=%d)\n", minor,jiffies&0xFFFF,seqn);
	    dh_acksread[minor]++;
	    nacks++;
	  } else {
	    /* Check sequence number */
	    signed short dels = seqn - rxseq[minor];
	    if(dels > 1) { /* Packet was in future -- drop it */
	      printd(dbgq[minor], "pkt dels=%d "
		     "seqn=%d, rxseq=%d, packet in future, dropping pkt.\n", 
		     dels, seqn, rxseq[minor]);
	      dh_badseq[minor]++;
	    } else if(dels < 1) { /* ACK old packet but do not add to queue */
	      if(! ackQisFull(&ackq[minor])) {
		putAckQ(&ackq[minor], seqn);
		dh_enable_msg_sndbl_intr_chan(icard, ich);
	      }
	      printd(dbgq[minor], "pkt dels=%d, in past, "
		     "ACKing and dropping pkt.\n", dels);
	      dh_badseq[minor]++;
	    } else { /* Last (default) case: dels == 1 */
	      if(len == 0) {
		printd(dbgq[minor], "RX HW pkt length=0, dropping.\n");
		dh_badhdr[minor]++;
	      } else {
		nrecv++;
		if(! ackQisFull(&ackq[minor])) {
		  dh_bench_start(rxuncook);
		  memcpy(dh_rx_scratch_msg[minor]->data, 
			 dhswpkt_data(rxpkt[minor]), dhswpkt_len(rxpkt[minor]));
		  vvprintk(KERN_INFO"DH(RX) saved data portion 0x%08lx 0x%08lx...\n",
			   *(unsigned long *) dh_rx_scratch_msg[minor]->data,
			   *(((unsigned long *) dh_rx_scratch_msg[minor]->data)+1));

		  dh_rx_scratch_msg[minor]->nbytes = dhswpkt_len(rxpkt[minor]);
		  dh_rx_scratch_msg[minor]->synfin = dhswpkt_issynfin(rxpkt[minor]);
		  putAckQ(&ackq[minor], seqn);
		  dh_enable_msg_sndbl_intr_chan(icard, ich); /* For ack */
		  putIOQ(inq[minor], dh_rx_scratch_msg[minor]);
		  this_channel_had_data = 1;
		  vvprintk(KERN_INFO "DH(RX,%lu) %dB (%s,seqn=%d,flags=%d)-> minor %d\n", 
			   jiffies&0xFFFF, dh_rx_scratch_msg[minor]->nbytes,
			   dh_rx_scratch_msg[minor]->synfin?"SYN/FIN":"no syn/fin", seqn, 
			   dhswpkt_flags(rxpkt[minor]),minor);
		  rxseq[minor]++;
		  dh_bench_end(&rxuncook);
		} /* ackq not full */
	      } /* nonzero length */
	    } /* correct seqn */
	  } /* non-ack */
	} /* Packet not corrupt */
      } /* if queue not full */
#ifdef DH_SHORT_INTERRUPTS
      continue; /* Let further messages be handled by other calls to interrupt handler */
#endif
    }   /* looping to drain data */
    dh_bench_end(&rxwhile);

	/* Wake up any reading processes that had data */
    if(this_channel_had_data) {
      wake_up_interruptible(&inq[minor]->wq);
    }
    
    /* Only reenable interrupt if input queue for channel wasn't full; if polling mode, 
       user will call function again to get next message.  Also leave disabled if DOM asked
       to reconnect; this will be handled by the kernel timer, and interrupts reenabled. */
    
    if(! ioQisFull(inq[minor]) && !ackQisFull(&ackq[minor])) {
      dh_enable_msg_recvd_intr_chan(icard, ich);
      vvprintk(KERN_INFO "RX: reenabling RX interrupts on card %d chan %d (acks=%d recv=%d)\n",
	       icard, ich, nacks, nrecv);
    } else {
      vvprintk(KERN_INFO "RX: NOT reenabling RX interrupts on card %d chan %d: "
	       "state=%d\n", icard, ich,
	       dh_conn_state[minor]);
    }
  } /* Loop over channels in card */
  if(vv() && !dataread) {
    printk(KERN_WARNING "RX: Got interrupt on card %d, but no data actually read out.\n",
	   icard);
  }
  dh_bench_end(&rxbench);
}

#ifdef DH_SIMULATE_TIMEOUTS
static inline int dh_do_simulate_timeout(int minor) {
  return dh_msgsread[minor] % DH_MSGS_PER_TIMEOUT == 0 ? 1 : 0;
}
#else
#define dh_do_simulate_timeout(x) (0)
#endif

#define TO_DOM 1
#define TO_DOR 2

#define ICTYPE 1
#define CITYPE 2
#define NEITHER 0
#define NODATA 3
int dh_get_icci(int icard, int ich, int minor) {
  /* Fetch packet and categorize by type */
  unsigned long flags;
  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  if(dh_rx_fifo_has_message(icard, ich)) {
    dh_enqueue_message_from_rx_fifo(rxpkt[minor], DH_MAX_HW_PKTSIZ, icard, ich);
    dh_clear_message_received(icard, ich);
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    if(dhswpkt_isic(rxpkt[minor])) {
      dh_icrecv[minor]++;
      return ICTYPE;
    } else if(dhswpkt_isci(rxpkt[minor])) {
      dh_cirecv[minor]++;
      return CITYPE;
    } else {
      return NEITHER;
    }
  } else {
      dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
      return NODATA;
  }
}

int dh_send_ic(int icard, int ich, int minor) {
  unsigned long flags;
  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  if(! dh_mrtc_clear(icard) || !dh_tx_fifo_almost_empty(icard, ich)) {
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    return 1;
  }
  dh_cook(txpkt[minor], NULL, FL_IC, 0, 0);
  dh_dequeue_message_to_tx_fifo(txpkt[minor], icard, ich);
  dh_icsent[minor]++;
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
  return 0;
}


int dh_send_ci(int icard, int ich, int minor) {
  unsigned long flags;
  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  if(! dh_mrtc_clear(icard) || !dh_tx_fifo_almost_empty(icard, ich)) {
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    return 1;
  }
  dh_cook(txpkt[minor], NULL, FL_CI, 0, 0);
  dh_dequeue_message_to_tx_fifo(txpkt[minor], icard, ich);
  dh_cisent[minor]++;
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
  return 0;
}


int dh_send_ctrl(int icard, int ich, int minor) {
  unsigned long flags;
  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  if(! dh_mrtc_clear(icard) || !dh_tx_fifo_almost_empty(icard, ich)) {
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    return 1;
  }
  dh_cook(txpkt[minor], NULL, FL_CTRL, 0, 0);
  dh_dequeue_message_to_tx_fifo(txpkt[minor], icard, ich);
  dh_ctrlsent[minor]++;
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
  return 0;
}

void dh_handle_connection(int icard, int ich, int minor, unsigned long * next_del) {
  /* Handles connection state changes on appropriate minor number.
     Change next_del if appropriate so that scheduler comes back more quickly. */
  int ptyp;

  while(1) {
    switch(dh_conn_state[minor]) {
    case DH_STATE_START: 
      printd(dbgq[minor], "START");
      ptyp = dh_get_icci(icard, ich, minor);
      if(ptyp == ICTYPE) { /* Got IC -- flush any pending ICs */
	dh_conn_state[minor] = DH_STATE_FLUSHIC;
        dh_conn_flush_jif[minor] = jiffies;
	printd(dbgq[minor], "START->FLUSHIC");
	break;
      }
      /* Handle case where no IC has arrived */
      printd(dbgq[minor], "START->WAITIC");
      dh_conn_state[minor] = DH_STATE_WAITIC;
      dh_conn_startic_jif[minor] = jiffies;
      break; /* Any other data means we return and wait for IC */
    case DH_STATE_WAITIC: /* Wait for first IC to arrive; if timeout exceeded, TX IC */
      ptyp = dh_get_icci(icard, ich, minor);
      if(ptyp == ICTYPE) {
	dh_conn_state[minor] = DH_STATE_FLUSHIC;
	printd(dbgq[minor], "WAITIC->FLUSHIC");
	dh_conn_flush_jif[minor] = jiffies;
	break;
      } else if(ptyp != NODATA) {
	break; /* Do it again */
      }
      if(jiffies >= dh_conn_startic_jif[minor] 
	 && jiffies-dh_conn_startic_jif[minor]<MIN_WAITIC_JIF)
	return; /* No ICs arrived and not timed out - just wait */
      if(dh_send_ic(icard, ich, minor)) return;
      dh_conn_startic_jif[minor] = jiffies;
      return; /* Nothing to do, try back later. */
    case DH_STATE_FLUSHIC: /* Wait for any extra ICs to arrive */
      { 
	int gotci = 0;
	while(1) {
	  ptyp = dh_get_icci(icard, ich, minor);
	  if(ptyp == NODATA) {
	    break;
	  } else if(ptyp == CITYPE) {
	    dh_conn_state[minor] = DH_STATE_SENDCI;
	    gotci = 1; 
	    break;
	  } else {
	    dh_conn_flush_jif[minor] = jiffies;
	  }
	}
	if(gotci) { 
	  printd(dbgq[minor], "FLUSH_IC->SENDCI_NOWAIT");
	  dh_conn_state[minor] = DH_STATE_SENDCI_NOWAIT;
	  break;
	}
      }
      if(jiffies >= dh_conn_flush_jif[minor] && jiffies-dh_conn_flush_jif[minor]<MIN_FLUSH_JIFS)
	return;
      /* Got no data, and timer has expired -- we're good to go to next state. */
      dh_conn_state[minor] = DH_STATE_SENDCI;
      printd(dbgq[minor], "FLUSH_IC->SENDCI");
      break;
    case DH_STATE_SENDCI_NOWAIT: /* Already got CI, send CI and finish up */
      if(dh_send_ci(icard, ich, minor)) return;
      dh_conn_state[minor] = DH_STATE_CONNECTING;
      printd(dbgq[minor], "SENDCI_NOWAIT->CONNECTING");
      break;
    case DH_STATE_SENDCI: /* Send 10 CIs, wait for CI */
      {
	int i;
	for(i=0;i<10;i++) {
	  if(dh_send_ci(icard, ich, minor)) return;
	}
      }
      dh_conn_state[minor] = DH_STATE_WAITCI;
      dh_conn_waitci_jif[minor] = jiffies;
      printd(dbgq[minor], "SENDCI->WAITCI");
      break; /* Just wait for CI to arrive */
    case  DH_STATE_WAITCI: /* Wait for CI */
      ptyp = dh_get_icci(icard, ich, minor);
      if(ptyp == ICTYPE) {
	dh_conn_flush_jif[minor] = jiffies;
	dh_conn_state[minor] = DH_STATE_FLUSHIC;
	printd(dbgq[minor], "WAITCI->FLUSHIC");
	break;
      }
      if(ptyp==NODATA) {
	if(jiffies-dh_conn_waitci_jif[minor]>MIN_WAITCI_JIF
	   || jiffies < dh_conn_waitci_jif[minor]) {
	  dh_conn_startic_jif[minor] = jiffies;
	  dh_conn_state[minor] = DH_STATE_WAITIC;
	  printd(dbgq[minor], "WAITCI->WAITIC");
	  break;
	} else { /* No data, still happy to wait for a CI... */
	  return;
	}
      }
      /* Other choice is CI or data packet (implicit CI) */
      printd(dbgq[minor], "WAITCI->CONNECTING");
      dh_conn_state[minor] = DH_STATE_CONNECTING;
      break;
    case DH_STATE_CONNECTING:
      txseq[minor] = 0;
      rxseq[minor] = (unsigned short) (-1); /* Cause Arthur's first pkt will be seq# 0 */
      unsigned long flags;
      dh_card_lock(icard,flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
      dh_enable_msg_recvd_intr_chan(icard, ich);
      dh_enable_msg_sndbl_intr_chan(icard, ich);
      dh_domvars[minor].had_rx_data = 0; /* Reset dead dom timer */
      printd(dbgq[minor], "CONNECTING->CONNECTED");
      dh_conn_state[minor] = DH_STATE_CONNECTED;
      wake_up_interruptible(&connectwq[minor]);
      wake_up_interruptible(&inq[minor]->wq);
      wake_up_interruptible(&outq[minor]->wq);
      dh_card_unlock(icard,flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
      break;
    case DH_STATE_CONNECTED:
      return; /* We're good */
    default:
      printk(KERN_ERR "dh_handle_connection: Unknown state (%d).\n", dh_conn_state[minor]);
      return;
    }
  }
}


void dh_check_for_dead_doms(int icard, int ich, int minor) {
  /* Look for "dead DOMs": RX FIFO has data but hasn't gotten an interrupt/msg recvd
     in DH_MAX_FIFO_HOLD_TIME jiffies.  Interrupt handler can reset timer value but let
     this function (called by dh_timer) reset the "had_rx_data" state explicitly
     if no data is seen in FIFO.

     Also we reset the timer if inq or ackq are full, or interrupts not enabled */

  if(!dh_rx_interrupt_is_enabled(icard, ich) ||
     dh_rx_fifo_is_empty(icard, ich) ||
     ioQisFull(inq[minor]) ||
     ackQisFull(&ackq[minor])) {

    if(dh_domvars[minor].had_rx_data) { 
      /* Interrupt handler probably cleared it, save new state */
      dh_domvars[minor].had_rx_data = 0;
    } else { 
      /* Nothing to do, everyone's happy */
    }
  } else { 
    if(dh_domvars[minor].had_rx_data) { 
      /* Had data before, still have it... check for timeouts: */
      if(dh_mwtc_clear(icard) &&
	 jiffies > dh_domvars[minor].nonemptyRXjif && 
	 jiffies - dh_domvars[minor].nonemptyRXjif > DH_MAX_FIFO_HOLD_TIME) {
	printk(KERN_ERR "Holy dead DOM, Batman!  "
	       "Got 'dead DOM' state for card=%d ch=%d minor=%d\n",
	       icard, ich, minor);
	dh_drain_and_dump_rx_fifo(icard, ich, minor);
	dh_domvars[minor].had_rx_data = 0;
      } else {
	/* Don't update timer unless jiffies overflows */
	if(jiffies < dh_domvars[minor].nonemptyRXjif) 
	  dh_domvars[minor].nonemptyRXjif = jiffies;
      }
    } else { // fifo not empty, but was empty before; save last timer value
      dh_domvars[minor].had_rx_data = 1;
      dh_domvars[minor].nonemptyRXjif = jiffies; /* Remember when this occurred */
    }
  }
}

void dh_timer(unsigned long icard) {
  /* Kernel timer for our card, which does a lot of things.  See new protocol spec.
     document per JJ and Arthur */

  unsigned long flags;
  int ich, minor;
  static int ntimers=0;

  unsigned long now = jiffies;
  unsigned long next_del = DH_TIMER_RESTART_JIF;

  vvprintk(KERN_INFO "dh_timer: Timer for card %ld triggered at %ld jiffies.\n",icard, now);
  ntimers++;
  if(!(ntimers%100)) {
    vvprintk(KERN_INFO "dh_timer:card %ld trigger=%d jiffies=%ld INTEN=0x%08lx TTSIC=0x%08lx.\n",
	     icard, ntimers, now, 
	     (unsigned long) dh_read_regi(icard, DH_RG_INTEN), 
	     (unsigned long) dh_read_regi(icard, DH_RG_TTSIC));
  }

  if(dh_pci_dors[icard].found) {  

    for(ich=0; ich < DH_DOMS_PER_CARD; ich++) {
      minor = icard * DH_DOMS_PER_CARD  + ich;
      int ipair = ich/2;
      if(!dh_opened[minor] || dh_get_pair_pwr_state(icard, ipair)!= 2) continue;
      

      /* Make sure close doesn't get stuck (can time out) */
      wake_up_interruptible(&outq[minor]->wq);
      wake_up_interruptible(&retxbwq[minor]);
      wake_up_interruptible(&ackqwq[minor]);

      /* Make sure open doesn't get stuck (can time out) */
      wake_up_interruptible(&comms_reset_wq[minor]);
      wake_up_interruptible(&connectwq[minor]);
      wake_up_interruptible(&device_in_use_wq[minor]);

      dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/

      /* (re-)enable TX interrupts if data needs to be sent and there is nothing in the retx
	 queue.  Print diagnostic info because this should happen rarely or never. */
      if(!ioQisEmpty(outq[minor]) && dh_num_retxb(minor) == 0) {
	dh_enable_msg_sndbl_intr_chan(icard, ich);
	vvprintk(KERN_WARNING "dh_timer: (re-)enabled TX interrupts on minor %d! "
		"NOUTQ=%d NRETXB=%u RETXB_BYTES=%d NRETXQ=%u\n", minor,
		nIOQEntries(outq[minor]), dh_num_retxb(minor), 
		dh_retxb_bytes(minor), nRetxQEntries(&retxq[minor]));
      }

      /* Handle hardware timeouts */
      if(dh_comms_state[minor] == DH_COMMS_READY 
	 && (dh_hardware_timeout_set(icard, ich) || dh_do_sim_timeout[minor])) {
	dh_comms_reset_in_progress[minor] = 1;
	dh_do_sim_timeout[minor] = 0;
	dh_nhdtouts[minor]++;
	if(parm_auto_reset) {
	  dh_write_regi_nosave(icard, DH_RG_DOMC, DH_BITN(ich+8));
	  printk(KERN_WARNING "dh: HARDWARE TIMEOUT on minor %d, did comms. reset\n", minor);
	} else {
	  printk(KERN_WARNING "dh: HARDWARE TIMEOUT on minor %d, but skipping reset\n", minor);
	}
	dh_comms_state[minor] = DH_COMMS_W_DOMC;
	dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
	continue;
      } else if(dh_comms_state[minor] == DH_COMMS_W_DOMC) {
	if(parm_auto_reset && !(dh_read_regi(icard, DH_RG_DOMC) & DH_BITN(ich+8))) { /* Done */
	  vvprintk(KERN_INFO "dh_timer: minor %d, DOMC clear, waiting for DOMS...\n", minor);
	  dh_comms_state[minor] = DH_COMMS_W_DOMS;
	  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
	  continue;
	} else {
	  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
	  continue;
	}
      } else if(dh_comms_state[minor] == DH_COMMS_W_DOMS) {
	if(dh_read_regi(icard, DH_RG_DOMS) & DH_BITN(ich)) {
	  vprintk(KERN_INFO "dh: minor %d, DOMS ok, comms reset succeeded.\n", minor);
	  dh_comms_reset_in_progress[minor] = 0;
	  dh_comms_state[minor] = DH_COMMS_READY;
	  wake_up_interruptible(&comms_reset_wq[minor]);
	  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
	  continue;
	} else {
	  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
	  continue;
	}
      } else {
	dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
      }

      /* Make SURE if we get here we don't have a timeout. */
      if(dh_comms_state[minor] != DH_COMMS_READY) {
	printk(KERN_ERR "dh_timer: ERROR: comms state != READY, inconsistent state, call John.\n");
	continue;
      } 

      /* Handle DOM connection (TX IC/CI) state here.  Wait for no DMAs to be in progress. */
      if(dh_mrtc_clear(icard) && dh_mwtc_clear(icard)) 
	dh_handle_connection(icard, ich, minor, &next_del);

      /* Check for / clear out "dead DOMs" (stale RX data with no msg received bit set) */
      dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
      dh_check_for_dead_doms(icard, ich, minor);
      dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/

      
    } /* loop over channels */
    
    /* If no hardware timeout, and connected to DOM, handle timed out, pending ReTX messages */
    for(ich=0; ich < DH_DOMS_PER_CARD; ich++) {
      minor = icard * DH_DOMS_PER_CARD  + ich;
      int ipair = ich/2;
      if(!dh_opened[minor] 
	 || dh_conn_state[minor] != DH_STATE_CONNECTED
	 || dh_get_pair_pwr_state(icard, ipair)!= 2) continue;

      /* Queue up stale messages in order of seqn */
      dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/

      if(!dh_retx_buf_isempty(minor))
	dh_queue_from_retx(minor, DH_TIMER_RETX_JIF, now, icard, ich);

      /* If inq is full, RX FIFO will be blocked, but we'll be ReTXing like crazy.
	 ACKs may be stuck in the RX FIFO; in this case, we have to discard the contents
	 in RX, and process any pending ACKs.  DOM will retx the discarded data. */
      if(ioQisFull(inq[minor])) {
	vprintk(KERN_INFO "dh_timer: inq queue is full, checking for stuck acks\n");
	if(dh_retx_have_timeout(minor, DH_TIMER_UNSTICK_JIF, now)) {
	  while(dh_mwtc_clear(icard) && dh_rx_fifo_has_message(icard, ich)) {
	    vprintk(KERN_INFO "dh_timer: ...Aggressive draining of RX FIFO to clear RETX...\n");
	    dh_enqueue_message_from_rx_fifo(rxpkt[minor], DH_MAX_HW_PKTSIZ, icard, ich);
	    dh_clear_message_received(icard, ich);
	    if(dhswpkt_isack(rxpkt[minor]) && dhswpkt_len(rxpkt[minor])==0) {
	      vprintk(KERN_INFO "dh_timer(RX(%d),%lu): ACK (seqn=%d)\n", minor,jiffies&0xFFFF,dhswpkt_seqn(rxpkt[minor]));
	      dh_handle_recvd_ack(dhswpkt_seqn(rxpkt[minor]), minor);
	      dh_acksread[minor]++;
	    } /* Pitch everything else */
	  } /* While data in RX FIFO */
	}   /* If have timed out data in ReTXb */
      }     /* If inq[minor] is full */

      dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    } /* Loop over channels */

    /* Reschedule timer */
    mod_timer(&dh_timer_struct[(int)icard], jiffies+next_del);
    vvprintk("dh_timer: sched restart thisjif=%lu del=%lu nextjif=%lu\n",
	    jiffies, next_del, jiffies+next_del);
  } /* if card exists */
  return;
}

/**************** INTERRUPT HANDLER ******************/
#ifdef KERNEL2_4
#define RETURN(a) return
void dh_intr_hdlr(int irq, void *dev_id, struct pt_regs *regs) {
#else
#define RETURN(a) return a
int dh_intr_hdlr(int irq, void *dev_id, struct pt_regs *regs) {
#endif
  /** Interrupt handler for DOM Hub device driver.  See ref 1. pg. 267
      for description of arguments. */
  int * pcard = dev_id; /* Passed in when you request_irq(irq, dh_intr_hdlr, 0, 
			                                  "dh", &icard, ...); */
  unsigned long flags;

  const int cpu = smp_processor_id();

  int icard = *pcard;
  
  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  dh_bench_start(cpuint_bench[cpu]);

  pprintk(KERN_INFO "DH: Got an interrupt on IRQ %d; card is %d.\n",irq,icard);

  /* See if this interrupt belongs to us - if not, quit immediately */
  u32 intr0 = dh_read_regi(icard, DH_RG_INTEN);
  u32 gstat = dh_read_regi(icard, DH_RG_GSTAT); 
  if(! (gstat & intr0 & DH_GSTAT_GLOBAL_INT)) { /* Definitely not ours - bye bye */
    dh_bench_end(&cpuint_bench[cpu]);
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    RETURN(IRQ_NONE);
  }

  dh_disable_global_intr(icard);
  if(dh_panic) {
    dh_bench_end(&cpuint_bench[cpu]);
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    printk(KERN_ERR 
	   "DH: Driver panic condition caught (card=%d), "
	   "leaving interrupts DISABLED.  Call John.\n", icard);
    RETURN(IRQ_HANDLED);
  }

  /* Get a little more specific to determine interrupt type */
  u32 ttsic = dh_read_regi(icard, DH_RG_TTSIC);
  u32 rtsic = dh_read_regi(icard, DH_RG_RTSIC);
  u32 mwtc  = dh_read_regi(icard, DH_RG_MWTC);
  u32 intr1 = dh_read_regi(icard, DH_RG_INTEN); /* Read for diagnostics - could remove later */
  u32 dstat = dh_read_regi(icard, DH_RG_DSTAT); /* ditto */
  
  vvprintk(KERN_INFO "DH: GSTAT 0x%08lx, TTSIC 0x%08lx, RTSIC 0x%08lx, INTR1 0x%08lx.\n",\
	   (unsigned long) gstat,\
	   (unsigned long) ttsic, (unsigned long) rtsic, (unsigned long) intr1);

  /* Now that we know it's one of ours, handle the various sources of interrupts... */

  /* Handle message received interrupt - complement rtsic because the
     appropriate bits are high when the FIFO is empty, not when FIFO has
     data */

  if((intr0 & gstat & DH_GSTAT_MBFn_MRCVD) && dh_mwtc_clear(icard)
                 && 
    /* This line is subtle: masks bits 24-31 of rtsic, then ANDs them bit-wise with
       bits 16-23. If any of the resulting bits are high, the RX interrupt is enabled
       and a complete message has been received. */
     ((rtsic & DH_RTSIC_INT_EN_0__7) >> 8) & rtsic) {

    /* If we get here, one or more cards has data to be queued in the driver.
       Disable interrupts, recv message. */

    vvprintk(KERN_INFO "DH(RX): GSTAT 0x%08lx, TTSIC 0x%08lx, RTSIC 0x%08lx, INTEN 0x%08lx.\n",\
            (unsigned long) gstat,\
            (unsigned long) ttsic, (unsigned long) rtsic, (unsigned long) intr0);

    dh_disable_msg_recvd_intr(icard); /* All channels on the card */
      
    /* Do EXPLICITLY now rather than tasklet_schedule(&dh_recv_message_tasklet) : */
    dh_do_recv_message(icard);

    dh_enable_global_intr(icard);

    dh_bench_end(&cpuint_bench[cpu]);
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    RETURN(IRQ_HANDLED);
  }

  /* Handle message sendable interrupt */
  if((intr0 & gstat & DH_GSTAT_MBFn_AEF) && dh_mrtc_clear(icard)
     && 
     /* Similar to RX case: require both TX buf almost empty AND interrupt enabled */
     ((ttsic & DH_TTSIC_INT_EN_0__7) >> 16) & ttsic) {
    /* An interrupt occurred because one of the 8 channels on this card had room in 
       its RX FIFO.  Disable the interrupt dequeue the 
       data from the output queue into the RX FIFO.  Will re-enable
       the interrupt if more data needs to be sent.  And it will wake_up_interruptible
       any writing process which has blocked because the outgoing queue is full. */


    vvprintk(KERN_INFO "DH(TX): GSTAT 0x%08lx, TTSIC 0x%08lx, RTSIC 0x%08lx, INTEN 0x%08lx.\n",\
	    (unsigned long) gstat,\
	    (unsigned long) ttsic, (unsigned long) rtsic, (unsigned long) intr0);

    dh_disable_msg_sndbl_intr(icard); /* All channels on the card */
    dh_do_send_message(icard);
    dh_enable_global_intr(icard);
    dh_bench_end(&cpuint_bench[cpu]);
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    RETURN(IRQ_HANDLED);
  }


  /* If control gets to this point, something strange is going on. */

  u32 intr2 = dh_read_regi(icard, DH_RG_INTEN);
  printk(KERN_ERR "dh_intr_hdlr: unhandled, spurious/inconsistent interrupt condition on card %d. "
	 "Proceeding undaunted... call John and/or Kalle if you get this often.\n", icard);
  printk(KERN_ERR 
	 "dh_intr_hdlr: FYI: "
	 "INTR0 (value at handler entry) 0x%lx\n"
	 "INTR1 (value just after ttsic, rtsic read) 0x%lx\n"
	 "INTR2 (value before this print statement) 0x%lx\n"
	 "GSTAT 0x%lx, DSTAT 0x%lx, TTSIC 0x%lx, RTSIC 0x%lx\n",
	 (unsigned long) intr0, 
         (unsigned long) intr1,
         (unsigned long) intr2,
	 (unsigned long) gstat, 
	 (unsigned long) dstat, 
	 (unsigned long) ttsic, 
	 (unsigned long) rtsic);
  dh_lasterr = DH_ERR_UNHDLEDINT;
  dh_enable_global_intr(icard);
  dh_bench_end(&cpuint_bench[cpu]);
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
  RETURN(IRQ_HANDLED);
}

/***************** End Interrupt Handling Code ********************/


/******************* PCI Interface Functions **********************/

static void dh_release_pci_region(int icard, struct dh_pci_dev * dhdev) {
  int l_len;
  l_len = dhdev->rend - dhdev->rstart + 1;
  vprintk(KERN_INFO "dh_release_pci_region: Releasing I/O region for card %d.\n", icard);
  release_region(dhdev->rstart, l_len);
}


static int dh_save_pci_config_space(struct dh_pci_dev *dor) {
  /** Save a copy of the PCI configuration space for device dor.
      Both the Linux PCI pci_dev pointer and the buffer space are
      members of dor (struct dh_pci_dev). */
  int ibyte;

  pprintk(KERN_INFO "Saving PCI configuration space.\n");

  for(ibyte=0;ibyte<PCI_CONFIG_SPACE_SIZE;ibyte++) {
    if(pci_read_config_byte(dor->dev, ibyte, &(dor->config_space[ibyte]))) {
      return 1;
    }
    pprintk(KERN_INFO "PCI config byte %d : 0x%x.\n",ibyte,\
	   (int) dor->config_space[ibyte]);
  }
  return 0;
}


static int dh_restore_pci_config_space(struct dh_pci_dev *dor) {
  int ibyte;

  pprintk(KERN_INFO "dh_restore_pci_config_space - skipping actual writes for now.\n");
  for(ibyte=0;ibyte<PCI_CONFIG_SPACE_SIZE;ibyte++) {

    if(pci_write_config_byte(dor->dev, ibyte, dor->config_space[ibyte])) {
      return 1;
    }

    pprintk(KERN_INFO "PCI config byte %d : 0x%x.\n",ibyte,\
           (int) dor->config_space[ibyte]);
  }

  return 0;
}


static void dh_set_comms_params(int icard) {
  /** No lock used because this is done before device is ready before interrupts
      are enabled */
  vvprintk(KERN_INFO 
	 "dh_set_comms_params(card=%d): autodac=%d rdelay=%u sdelay=%u "
	 "dacmax=%u thresh=%u minclev=%u maxclev=%u\n", icard,
	 dh_pci_dors[icard].autodac, dh_pci_dors[icard].rdelay, dh_pci_dors[icard].sdelay,
	 dh_pci_dors[icard].dacmax, dh_pci_dors[icard].thresh, dh_pci_dors[icard].minclev, 
	 dh_pci_dors[icard].maxclev);
  if(! dh_pci_dors[icard].autodac) {
    dh_set_write_regi_bits(icard, DH_RG_CTRL, DH_BITN(15));
    vprintk(KERN_INFO "dh_set_comms_params: no autodac, setting CTRL bit 15...\n");
  }
  /* DCREV */
  u32 dcrev = (dh_pci_dors[icard].rdelay << 16) | (dh_pci_dors[icard].sdelay << 24) 
    | (dh_pci_dors[icard].dacmax << 12) | dh_pci_dors[icard].thresh;
  dh_write_regi(icard, DH_RG_DCREV, dcrev);
  vprintk(KERN_INFO "dh_set_comms_params: DCREV <- 0x%08lx\n", (unsigned long) dcrev);
  /* FREV */
  u32 frev = (dh_pci_dors[icard].maxclev << 16) | dh_pci_dors[icard].minclev;
  vprintk(KERN_INFO "dh_set_comms_params: FREV <- 0x%08lx\n", (unsigned long) frev);
}

static void dh_rev1_fpga_global_res(int icard) {
  dh_set_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_GLOBAL_RES);
  dh_clear_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_GLOBAL_RES);
  dh_jsleep(HZ/1000+1); /* Kalle says sleep 1 msec to wait for op. to complete */
}


void dh_write_maxcur(int icard, int maxcur) {
  u32 curl = dh_read_regi(icard, DH_RG_CURL);
  dh_write_regi(icard, DH_RG_CURL, (curl & 0xFFFF00FF) | ((maxcur & 0xFF) << 8));
  vprintk(KERN_INFO "dh_write_maxcur: set minimum current to %d mA.\n", maxcur);
}

void dh_write_mincur(int icard, int mincur) {
  u32 curl = dh_read_regi(icard, DH_RG_CURL);
  dh_write_regi(icard, DH_RG_CURL, (curl & 0xFFFFFF00) | (mincur & 0xFF));
  vprintk(KERN_INFO "dh_write_mincur: set minimum current to %d mA.\n", mincur);
}

void dh_write_maxvolts(int icard, int maxvolts) {
  int maxvolts_fpga = (maxvolts * 225 + 50) / 100;
  u32 curl = dh_read_regi(icard, DH_RG_CURL);
  dh_write_regi(icard, DH_RG_CURL, (curl & 0x00FFFFFF) | ((maxvolts_fpga & 0xFF) << 24));
  vprintk(KERN_INFO "dh_write_maxvolts: set minimum voltage to %d V (%d FPGA units).\n",
	  maxvolts, maxvolts_fpga);
}

void dh_write_minvolts(int icard, int minvolts) {
  int minvolts_fpga = (minvolts * 225 + 50) / 100;
  u32 curl = dh_read_regi(icard, DH_RG_CURL);
  dh_write_regi(icard, DH_RG_CURL, (curl & 0xFF00FFFF) | ((minvolts_fpga & 0xFF) << 16));
  vprintk(KERN_INFO "dh_write_minvolts: set minimum voltage to %d V (%d FPGA units).\n", 
	  minvolts,minvolts_fpga);
}

static int dh_get_min_current(int icard) { 
  u32 curl = dh_read_regi(icard, DH_RG_CURL);
  return (curl & 0xFF);
}

static int dh_get_max_current(int icard) {
  u32 curl = dh_read_regi(icard, DH_RG_CURL);
  return ((curl>>8) & 0xFF);
}

static int dh_get_min_voltage(int icard) {
  u32 curl = dh_read_regi(icard, DH_RG_CURL);
  return ((((curl>>16)&0xFF)*100 + 112)/225);
}

static int dh_get_max_voltage(int icard) { 
  u32 curl = dh_read_regi(icard, DH_RG_CURL);
  return ((((curl>>24)&0xFF)*100 + 112)/225);
}

static void dh_write_fpga_voltage_current_limits(int icard) {
  dh_write_minvolts(icard, dh_pci_dors[icard].minvolts);
  dh_write_maxvolts(icard, dh_pci_dors[icard].maxvolts);
  dh_write_mincur(icard,   dh_pci_dors[icard].mincur);
  dh_write_maxcur(icard,   dh_pci_dors[icard].maxcur);
}

#define WAIT_CFG_DONE    1
#define NO_WAIT_CFG_DONE 0

static int dh_fpga_regi_init(int icard, int w_cfg_done) {
  /* Perform register init per Kalle, at time of dor card
     discovery or FPGA reload */

  if(w_cfg_done && dh_pci_dors[icard].rev >= 1) {
    /* Make sure / wait for FPGA CFG_DONE */
    int timeout_msec = 2000;
    unsigned long timeout_jiffies = (timeout_msec * HZ)/1000LL + 1;
    unsigned long j0 = jiffies;
    int gstat_ok = 0;
    int keepgoing = 1;
    while(keepgoing || WITHIN_TOUT(timeout_jiffies, j0)) {
      if(! WITHIN_TOUT(timeout_jiffies, j0)) keepgoing = 0;
      u32 gstat = dh_read_regi(icard, DH_RG_GSTAT);
      if(gstat & DH_BITN(1)) {
	gstat_ok = 1;
	break;
      }
      schedule();
    }
    if(!gstat_ok) {
      printk(KERN_ERR "dh_fpga_regi_init: timeout on card %d waiting for GSTAT CFG_DONE!\n",
	     icard);
      return 1;
    } 
  }

  dh_zero_write_regi_copies_card(icard);

  dh_set_local_osc(icard, dh_clksel[icard] == DH_CLKSEL_EXTERNAL ? 1 : 0);
     /* Now includes 10 msec wait */

  if(dh_pci_dors[icard].rev >= 1)
    dh_rev1_fpga_global_res(icard);

  dh_set_timeout_enable(icard);
  dh_set_comms_params(icard);

  dh_write_fpga_voltage_current_limits(icard);

  /* Zero DMA counters */
  dh_write_regi(icard, DH_RG_MRTC, 0);
  dh_write_regi(icard, DH_RG_MWTC, 0);

  /* Disable all interrupt masks except global. */
  dh_write_regi(icard, DH_RG_INTEN, DH_INTEN_GLOBAL_INT_EN);

  /* Uncomment the enable, below to enable software-selected LEDs, 
     then use e.g. dh_set_led_value(icard, 0x04, SET);     <- turn individual leds on
                   dh_set_led_value(icard, 0x04, CLEAR);   <-                      off
               or  dh_set_led_value(icard, 0x04, REPLACE); <- replace all 8 leds
     Otherwise, Kalle's firmware sets the LED values. */
  /* dh_enable_leds(icard); */
  return 0;
}

static int dh_get_pci_devices(struct dh_pci_dev *dh_devs) {
  /** Initialize PCI devices (DOR cards).  \return number of devices found,
      or a negative error code. 
      Procedure for initializing PCI devices:
      Get the device structures.  Store where each card lives (start, end, memory type).
      Keep memory copies of all write registers. */
  int icard;
  int found;
  int result;
  unsigned long l_len;
  unsigned long rstart, rend, rflags;
  int is_io;
  struct pci_dev *dev;
  u32 dstat;
  u16 pci_config_reg;
  u8 interrupt_pin, irq, rev;
  int problem;

#ifdef KERNEL2_4
  if(!pci_present()) {
    return -ENODEV;
  }
#else
  // FIXME: find out the right way to do this in 2.6
  //if(list_empty(&pci_devices)) {
  //  return -ENODEV;
  //}
#endif

  dev = NULL;

  /* Tag all as not found, and having no device files open */
  for(icard=0; icard < DH_MAX_CARDS; icard++) {
    dh_devs[icard].found    = 0;
    dh_devs[icard].num_open = 0;
    init_MUTEX(&dh_devs[icard].open_close_sem);
  }

  /* Find them */
  for(found=0; found < DH_MAX_CARDS;) {

    dev = pci_find_device(DH_VENDOR, DH_DEVICEID, dev);
    if(!dev) {
      break;
    }

    if(parm_rx_dma || parm_tx_dma) {
      if(pci_dma_supported(dev, 0xFFFFFFFF)) {
	dev->dma_mask = 0xFFFFFFFF;
	pprintk(KERN_INFO "dh_get_pci_devices: DMA is supported by this device.\n");
      } else {
	printk(KERN_ERR "dh_get_pci_devices: DMA is NOT supported by this device.\n");
	dh_lasterr = DH_ERR_DMANOTSUPP;
	continue;
      }
    }

    problem = 0;
    if(pci_enable_device(dev)) {
      printk(KERN_ERR "dh_get_pci_devices: Couldn't enable %d'th found card.\n",found);
      dh_lasterr = DH_ERR_PCIENABFAIL;
      continue;
    }

    /* rev0 or rev1? check pci revision id byte... */
    result = pci_read_config_word(dev, PCI_REVISION_ID, &pci_config_reg);
    if(result) {
      dh_lasterr = DH_ERR_PCIENABFAIL;
      printk(KERN_ERR "dh_get_pci_devices: "
             "Couldn't read PCI configuration space for %dth card.\n", found);
      continue;
    }
    rev = pci_config_reg;

    /* Show configuration register */
    result = pci_read_config_word(dev, PCI_COMMAND, &pci_config_reg);
    if(result) {
      dh_lasterr = DH_ERR_PCIENABFAIL;
      printk(KERN_ERR "dh_get_pci_devices: "
	     "Couldn't read PCI configuration space for %dth card.\n",found);
      continue;
    }

    vvprintk(KERN_INFO "dh_get_pci_devices: PCI command register is 0x%04lx.\n",\
	     (unsigned long) pci_config_reg);

    if((    ! (pci_config_reg & PCI_COMMAND_IO) 
	 || ! (pci_config_reg & PCI_COMMAND_MASTER))) {
      printk(KERN_WARNING "dh: PCI command register 0x%04x does not have both"
	     " PCI_COMMAND_IO and PCI_COMMAND_MASTER set; asserting these now.\n",
	     (unsigned short) pci_config_reg);
      pci_config_reg |= (PCI_COMMAND_IO | PCI_COMMAND_MASTER);
      pci_write_config_word(dev, PCI_COMMAND, pci_config_reg);
    }


    /* Get IRQ info */
    result = pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &interrupt_pin);
    if(result) {
      printk(KERN_ERR "dh_get_pci_devices: "
	     "Couldn't read PCI configuration space for %dth card.\n",found);
      dh_lasterr = DH_ERR_PCIENABFAIL;
      continue;
    }

    result = pci_read_config_byte(dev, PCI_INTERRUPT_LINE, &irq);
    if(result) {
      printk(KERN_ERR "dh_get_pci_devices: "
	     "Couldn't read PCI configuration space for %dth card.\n",found);
      dh_lasterr = DH_ERR_PCIENABFAIL;
      continue;
    }
    
    if(interrupt_pin == 0) {
	  printk(KERN_ERR "dh_get_pci_devices: %dth card doesn't seem to support interrupts!\n",
		 found);
	  dh_lasterr = DH_ERR_PCIENABFAIL;
	  continue;
    }

    /* Get IRQ explicitly from Kernel now, not PCI config space: */
    irq = dev->irq;
    vprintk(KERN_INFO "dh_get_pci_devices: Using IRQ %d for card %d.\n", irq, found);

    rstart = pci_resource_start(dev, 0);
    rend   = pci_resource_end(dev,0);
    rflags = pci_resource_flags(dev,0);
    l_len  = rend - rstart + 1;
    is_io  = ((rflags & IORESOURCE_IO) ? 1 : 0);	
    
    /* Get IO regions */

#ifdef KERNEL2_4
    int result = check_region(rstart, l_len);
    if(result != 0) {
      printk(KERN_ERR "dh_get_pci_devices: Can't get I/O region for DOR card "
	     "from %lx to %lx (error %d).\n", rstart, rend, result);
      dh_lasterr = DH_ERR_PCIALLOCFAIL;
      continue;
    }
    request_region(rstart, l_len, "dh");
#else
    /* FIXME: Does res need to persist & be freed? */
    struct resource * res = request_region(rstart, l_len, "dh");
    if(!res) {
      printk(KERN_ERR "dh_get_pci_devices: Can't get I/O region for DOR card "
             "from %lx to %lx (error %d).\n", rstart, rend, result);
      dh_lasterr = DH_ERR_PCIALLOCFAIL;
      continue;
    }
#endif

    /* Now that we have mem regions, get card ID */
    dstat = inl(rstart + DH_RG_DSTAT*4);
    icard = (dstat >> 12) & 0x0F; /* 0..15 -- see DOR FPGA API */
    if(icard < 0 || icard >= DH_MAX_CARDS) {
      printk(KERN_ERR "dh: Found card %d with ID out of range (DSTAT = %lx)!\n", icard, 
	     (unsigned long) dstat);
      release_region(rstart, l_len);
      dh_lasterr = DH_ERR_BADCARDID;
      continue;
    }
    if(dh_excluded[icard]) {
      vprintk(KERN_INFO "dh: Found DOR card %d but EXCLUDING it from driver.\n",
		    icard);
      release_region(rstart, l_len);
      continue;
    }
    if (dh_devs[icard].found) {
      vprintk(KERN_INFO "dh: card %d already found, duplicate card id?\n", 
	      icard);
      release_region(rstart, l_len);
      continue;
    }
    vvprintk(KERN_INFO "dh: dstat is 0x%lx.\n", (unsigned long) dstat);

    /* Save needed pointers/indices/flags */
    dh_devs[icard].rstart    = rstart;
    dh_devs[icard].rend      = rend;
    dh_devs[icard].rflags    = rflags;
    dh_devs[icard].is_io     = is_io;
    dh_devs[icard].dev       = dev;
    dh_devs[icard].irq       = irq;
    dh_devs[icard].rev       = rev;
    dh_devs[icard].icard     = icard; /* This is to store icard and pass as 
			 		 ptr into request_irq, so each interrupt handler
				 	 knows which card it belongs to */
    dh_devs[icard].autodac   = DH_DEF_AUTODAC;
    dh_devs[icard].dacmax    = DH_DEF_DACMAX;
    dh_devs[icard].minclev   = DH_DEF_MINCLEV;
    dh_devs[icard].maxclev   = DH_DEF_MAXCLEV;
    dh_devs[icard].rdelay    = DH_DEF_RDELAY;
    dh_devs[icard].sdelay    = DH_DEF_SDELAY;
    dh_devs[icard].thresh    = DH_DEF_THRESH;
    
    dh_devs[icard].mincur    = DH_DEF_MINCUR;
    dh_devs[icard].maxcur    = DH_DEF_MAXCUR;
    dh_devs[icard].minvolts  = DH_DEF_MINVOLTS;
    dh_devs[icard].maxvolts  = DH_DEF_MAXVOLTS;

    spin_lock_init(&dh_devs[icard].card_lock);

    /* Save a copy of the PCI configuration space */
    if(dh_save_pci_config_space(&dh_devs[icard])) {
      dh_lasterr = DH_ERR_CANTSAVPCI;
      printk(KERN_ERR "dh: Can't save PCI configuration space for card %d\n", icard);
      release_region(rstart, l_len);
      continue;
    }

    /* Regi init.  Don't wait for CFG_DONE when working off of page 0. */
    if(dh_fpga_regi_init(icard, NO_WAIT_CFG_DONE)) { 
      printk(KERN_WARNING "dh: FPGA register init for card %d failed!\n", icard);
      release_region(rstart, l_len);
      continue;
    }

    dh_devs[icard].found = 1;    
    found++;

    vprintk(KERN_INFO "dh: Got DOR PCI card w/ ID=%d, IRQ=%d, start=%lx, end=%lx\n",
	    icard, dh_devs[icard].irq, dh_devs[icard].rstart, dh_devs[icard].rend);
    vprintk(KERN_INFO "                        flags=%lx, is_io=%d.\n", 
	    dh_devs[icard].rflags, dh_devs[icard].is_io);
  } /* Loop over pci_find_device */

  if(!found) {
    printk(KERN_WARNING "dh: No DOR cards found!\n");
    return -ENODEV;
  }
  return found;
}
 

static void dh_release_pci_devices(struct dh_pci_dev *dh_devs, int max_devs) {
  /** Release DOR PCI card resources.  */
  int icard;
  unsigned long l_len;

  for(icard=0; icard <  DH_MAX_CARDS; icard++) {
    if(dh_devs[icard].found) {
      l_len = dh_devs[icard].rend - dh_devs[icard].rstart + 1;
      dh_release_pci_region(icard, &dh_devs[icard]);
      dh_devs[icard].found = 0;
    }
  }
}

/***************** End PCI Interface Functions ********************/


/**************** /proc File Handling Functions *******************/

static void dh_mark_all_proc_entries_as_unregistered(void) {
/** Tags every fancy_proc_dir_entry as not yet registered */
  int icard;
  int ipair;
  int idom;

  fpde_top_dir.regi     = 0;
  fpde_driverstat.regi  = 0;
  fpde_revision.regi    = 0;
  fpde_pwrall.regi      = 0;
  fpde_bufsiz.regi      = 0;
  fpde_blocking.regi    = 0;
  fpde_bench.regi       = 0;
  fpde_verbose.regi     = 0;
  fpde_lasterr.regi     = 0;
  fpde_pseudo_intr.regi = 0;
  for(icard=0;icard < DH_MAX_CARDS; icard++) {
    fpde_card_dirs[icard].regi = 0;
    fpde_card_fpga[icard].regi = 0;
    fpde_card_pwr_check[icard].regi = 0;
    fpde_card_autodac[icard].regi = 0;
    fpde_card_dacmax[icard].regi = 0;
    fpde_card_minclev[icard].regi = 0;
    fpde_card_maxclev[icard].regi = 0;
    fpde_card_rdelay[icard].regi = 0;
    fpde_card_sdelay[icard].regi = 0;
    fpde_card_thresh[icard].regi = 0;
    fpde_card_minvolts[icard].regi = 0;
    fpde_card_maxvolts[icard].regi = 0;
    fpde_card_mincur[icard].regi = 0;
    fpde_card_maxcur[icard].regi = 0;
    fpde_card_flprotect[icard].regi = 0;
    fpde_card_pci[icard].regi = 0;
    fpde_card_1pps[icard].regi = 0;
    fpde_card_clksel[icard].regi = 0;
    fpde_card_syncgps[icard].regi = 0;
    fpde_card_rev[icard].regi = 0;
    fpde_card_test_log[icard].regi = 0;
    fpde_card_fpga_regs[icard].regi = 0;
    for(ipair=0;ipair < DH_PAIRS_PER_CARD; ipair++) {
      fpde_pair_dirs[icard][ipair].regi = 0;
      fpde_pair_pwr[icard][ipair].regi = 0;
      fpde_pair_plugged[icard][ipair].regi = 0;
      fpde_pair_current[icard][ipair].regi = 0;
      fpde_pair_voltage[icard][ipair].regi = 0;
      fpde_pair_pwr_check[icard][ipair].regi = 0;
      for(idom=0;idom < DH_DOMS_PER_PAIR; idom++) {
	fpde_dom_dirs[icard][ipair][idom].regi = 0;
	fpde_dom_tcalib[icard][ipair][idom].regi = 0;
	fpde_dom_comstat[icard][ipair][idom].regi = 0;
        fpde_dom_domstat[icard][ipair][idom].regi = 0;
	fpde_dom_id[icard][ipair][idom].regi = 0;
        fpde_dom_bench[icard][ipair][idom].regi = 0;
        fpde_dom_pollute[icard][ipair][idom].regi = 0;
	fpde_dom_rxdt[icard][ipair][idom].regi = 0;
        fpde_dom_softboot[icard][ipair][idom].regi = 0;
        fpde_dom_iscomm[icard][ipair][idom].regi = 0;
        fpde_dom_isconn[icard][ipair][idom].regi = 0;
        fpde_dom_isnotcb[icard][ipair][idom].regi = 0;
      }
    }
  }
}

static void dh_rm_free_proc(struct fancy_pde_entry *f, 
			    struct proc_dir_entry *parent) {
  if(f->regi) {
    if(f->pde->data != NULL) {
      kfree(f->pde->data);
      f->pde->data = NULL;
    }
    remove_proc_entry(f->name, parent);
    f->regi = 0;
  }
}

static void dh_cleanup_proc_entries(void) {
  /** Removes proc entries if needed, "backing out" so that directories are
      emptied of their contents before the directories themselves are removed.
      Takes care not to kfree null pointers. */
  int icard;
  int ipair;
  int idom;
  int iflash;

  remove_proc_entry("root", fpde_top_dir.pde);

  /* Clean up /proc entries */
  if(fpde_top_dir.regi) {
    for(icard=0; icard < DH_MAX_CARDS; icard++) {
      
      /* Make sure card was marked as "registered" (in the proc dir entry sense)
	 and, if running in normal mode, that the PCI device for this card was found. */
      if(fpde_card_dirs[icard].regi && dh_pci_dors[icard].found) {

	dh_rm_free_proc(&fpde_card_1pps[icard],      fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_clksel[icard],    fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_syncgps[icard],   fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_rev[icard],       fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_test_log[icard],  fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_pci[icard],       fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_flprotect[icard], fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_fpga_regs[icard], fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_pwr_check[icard], fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_autodac[icard],   fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_dacmax[icard],    fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_minclev[icard],   fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_maxclev[icard],   fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_rdelay[icard],    fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_sdelay[icard],    fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_thresh[icard],    fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_minvolts[icard],  fpde_card_dirs[icard].pde);
        dh_rm_free_proc(&fpde_card_maxvolts[icard],  fpde_card_dirs[icard].pde);
        dh_rm_free_proc(&fpde_card_mincur[icard],    fpde_card_dirs[icard].pde);
        dh_rm_free_proc(&fpde_card_maxcur[icard],    fpde_card_dirs[icard].pde);
	dh_rm_free_proc(&fpde_card_fpga[icard],      fpde_card_dirs[icard].pde);

	for(iflash = 0; iflash < DH_FLASH_PAGES; iflash++) {
	  dh_rm_free_proc(&fpde_card_flash[icard][iflash], fpde_card_dirs[icard].pde);
	}

	for(ipair=0;ipair < DH_PAIRS_PER_CARD; ipair++) {
	  if(fpde_pair_dirs[icard][ipair].regi) {

	    dh_rm_free_proc(&fpde_pair_current[icard][ipair],   fpde_pair_dirs[icard][ipair].pde);
	    dh_rm_free_proc(&fpde_pair_voltage[icard][ipair],   fpde_pair_dirs[icard][ipair].pde);
	    dh_rm_free_proc(&fpde_pair_plugged[icard][ipair],   fpde_pair_dirs[icard][ipair].pde);
	    dh_rm_free_proc(&fpde_pair_pwr[icard][ipair],       fpde_pair_dirs[icard][ipair].pde);
	    dh_rm_free_proc(&fpde_pair_pwr_check[icard][ipair], fpde_pair_dirs[icard][ipair].pde);

	    for(idom=0;idom < DH_DOMS_PER_PAIR; idom++) {
	      if(fpde_dom_dirs[icard][ipair][idom].regi) {

		dh_rm_free_proc(&fpde_dom_tcalib[icard][ipair][idom],
			       fpde_dom_dirs[icard][ipair][idom].pde);
		dh_rm_free_proc(&fpde_dom_comstat[icard][ipair][idom],
                               fpde_dom_dirs[icard][ipair][idom].pde);
		dh_rm_free_proc(&fpde_dom_domstat[icard][ipair][idom],
                               fpde_dom_dirs[icard][ipair][idom].pde);
		dh_rm_free_proc(&fpde_dom_id[icard][ipair][idom],
				fpde_dom_dirs[icard][ipair][idom].pde);
		dh_rm_free_proc(&fpde_dom_bench[icard][ipair][idom],
                               fpde_dom_dirs[icard][ipair][idom].pde);
		dh_rm_free_proc(&fpde_dom_pollute[icard][ipair][idom],
                               fpde_dom_dirs[icard][ipair][idom].pde);
		dh_rm_free_proc(&fpde_dom_rxdt[icard][ipair][idom],
				fpde_dom_dirs[icard][ipair][idom].pde);
		dh_rm_free_proc(&fpde_dom_softboot[icard][ipair][idom],
                               fpde_dom_dirs[icard][ipair][idom].pde);
		dh_rm_free_proc(&fpde_dom_iscomm[icard][ipair][idom],
                               fpde_dom_dirs[icard][ipair][idom].pde);
		dh_rm_free_proc(&fpde_dom_isconn[icard][ipair][idom],
                               fpde_dom_dirs[icard][ipair][idom].pde);
		dh_rm_free_proc(&fpde_dom_isnotcb[icard][ipair][idom],
                               fpde_dom_dirs[icard][ipair][idom].pde);

		dh_rm_free_proc(&fpde_dom_dirs[icard][ipair][idom], 
				fpde_pair_dirs[icard][ipair].pde);
	      }
	    }

	    dh_rm_free_proc(&fpde_pair_dirs[icard][ipair], fpde_card_dirs[icard].pde);
	  }
	}

	dh_rm_free_proc(&fpde_card_dirs[icard], fpde_top_dir.pde);
      }
    }

    dh_rm_free_proc(&fpde_pseudo_intr, fpde_top_dir.pde); 
    dh_rm_free_proc(&fpde_driverstat,  fpde_top_dir.pde);
    dh_rm_free_proc(&fpde_revision,    fpde_top_dir.pde);
    dh_rm_free_proc(&fpde_pwrall,      fpde_top_dir.pde);
    dh_rm_free_proc(&fpde_bench,       fpde_top_dir.pde);
    dh_rm_free_proc(&fpde_blocking,    fpde_top_dir.pde);
    dh_rm_free_proc(&fpde_bufsiz,      fpde_top_dir.pde);
    dh_rm_free_proc(&fpde_verbose,     fpde_top_dir.pde);
    dh_rm_free_proc(&fpde_lasterr,     fpde_top_dir.pde);
    dh_rm_free_proc(&fpde_top_dir,     NULL); 
  }
}

static unsigned char dh_pairmask(unsigned char dmask) {
  unsigned char b=0;
  int iwp;
  for(iwp = 0; iwp<DH_PAIRS_PER_CARD; iwp++) {
    if(dmask & DH_BITN(iwp*2) || dmask & DH_BITN(iwp*2+1)) 
      b |= DH_BITN(iwp);
  }
  return b;
}

static int dh_get_valid_wps(unsigned char * pmask, unsigned char *vwp) { 
  int icard;
  int count=0;

  for(icard=0; icard < DH_MAX_CARDS; icard++) {
    /* Eliminate firmware mismatches or missing dor cards */
    if(! dh_pci_dors[icard].found || dh_firmware_mismatch[icard]) {
      vwp[icard] = 0;
      continue;
    }
    vwp[icard] = pmask[icard] & 0x0F;
    if(! vwp[icard]) continue; /* Nothing requested for this card */

    int ip;

    for(ip=0; ip<DH_PAIRS_PER_CARD; ip++) {
      if(vwp[icard] & DH_BITN(ip)) {
	if(dh_get_pair_pwr_state(icard, ip)==2) {
	  printk(KERN_WARNING "dh: attempt to power on already-on wire pair "
		 "(card %d pair %d)!\n", icard, ip);
	  dh_lasterr = DH_ERR_ALLNOTOFF;
	  vwp[icard]  &= ~(0x0F & DH_BITN(ip));
	  continue;
	}
	if(! dh_is_pair_plugged(icard, ip)) {
	  vvprintk(KERN_INFO "card %d pair %d is not plugged in, excluding\n", icard, ip);
	  vwp[icard] &= ~(0x0F & DH_BITN(ip));
	  continue;
	}
      }
    }
    if(vwp[icard]) count++;
  }
  return count;
}

static unsigned char dh_dommask(unsigned char wpmask) {
  int ipair;
  unsigned char it;
  it = 0;
  //vprintk(KERN_INFO "wpmask=0x%02x\n", wpmask); 
  for(ipair=0; ipair<DH_PAIRS_PER_CARD; ipair++) {
    if(DH_BITN(ipair) & wpmask) {
      //printk(KERN_INFO "will turn on wp %d\n", ipair);
      it |= DH_BITN(ipair*2) | DH_BITN(ipair*2+1);
    } //printk(KERN_INFO "will NOT turn on wp %d\n", ipair);
  }
  return it;
}

static void dh_wait_for_communicating(unsigned char * wmask, unsigned char * iscom) {
  int icard, ipair;
  unsigned long flags;
  unsigned long j0 = jiffies;
  //  unsigned long TOUT = HZ*3; /* 3 sec */
  unsigned long TOUT = HZ*3+1;
  int outstanding = 0;

  for(icard=0; icard < DH_MAX_CARDS; icard++) iscom[icard] = 0;

  int keepgoing = 1;
  while(keepgoing || WITHIN_TOUT(TOUT, j0)) {
    if(!WITHIN_TOUT(TOUT,j0)) keepgoing = 0; /* Bail after this iteration */
    outstanding = 0;
    /* See if DOMS shows everyone we want */
    for(icard=0; icard < DH_MAX_CARDS; icard++) {
      if(wmask[icard]) {
	unsigned char wanted = dh_dommask(wmask[icard]);
	iscom[icard] = (dh_read_regi(icard, DH_RG_DOMS)&0xFF);
	if(wanted == (iscom[icard]&wanted)) {
	  vvprintk(KERN_INFO "got both DOMs: iscom=0x%02x wanted=0x%02x\n",
		   iscom[icard], wanted);
	  continue;
	} else {
	  outstanding = 1;
	}
      }
    }
    if(outstanding) {
      schedule();
    } else {
      vvprintk(KERN_INFO "DOMs are ready for action\n");
      break;
    }
  }
  if(outstanding) {
    vprintk(KERN_INFO "some DOMs on requested wire pairs on were not communicating\n");
    for(icard=0; icard < DH_MAX_CARDS; icard++) {
      if(wmask[icard]) {
        unsigned char wanted = dh_dommask(wmask[icard]);
        if(wanted != (iscom[icard]&wanted)) {
	  vprintk(KERN_INFO "did NOT get all requested DOMs on card %d: "
		  "iscom=0x%02x wanted=0x%02x\n",
		  icard, iscom[icard], wanted);
	}
      }
    }
  }
}


static void dh_wp_resets_pwron_mult(unsigned char * wmask) {
  int icard, ipair;
  unsigned long flags;
  for(icard=0; icard < DH_MAX_CARDS; icard++) {
    if(wmask[icard]) {
      dh_card_lock(icard, flags); /* LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL */
      dh_set_write_regi_bits(icard, DH_RG_CTRL, wmask[icard] | wmask[icard]<<4);
      dh_card_unlock(icard, flags); /* UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU */
    }
  }
}

static void dh_check_for_pon_rdy(unsigned char * wmask) {
  int icard, ipair;
  
  unsigned long j0 = jiffies;
  unsigned long TOUT = (HZ*500)/1000 + 1; /* 500 msec */
  unsigned char rdy[DH_MAX_CARDS];
  int outstanding = 0;

  for(icard=0; icard < DH_MAX_CARDS; icard++) 
    rdy[icard] = 0;

  int keepgoing = 1;

  while(keepgoing || WITHIN_TOUT(TOUT,j0)) {
    if(!WITHIN_TOUT(TOUT,j0)) keepgoing = 0; /* Bail after this iteration */
    outstanding = 0;
    /* See if power-on-ready bits have come on */
    for(icard=0; icard < DH_MAX_CARDS; icard++) {
      if(wmask[icard]) {
	u32 dstat = dh_read_regi(icard, DH_RG_DSTAT);
	rdy[icard] = (dstat>>4) & 0x0F;
	/* See if all the desired bits have come on */
	if((rdy[icard] & wmask[icard]) == wmask[icard]) {
	  vprintk(KERN_INFO "pon_rdy -- card %d done: rdy=0x%02x DSTAT=0x%08x wmask=0x%02x\n",
		  icard, rdy[icard], dstat, wmask[icard]);
	  continue;
	} else {
	  vprintk(KERN_INFO "pon_rdy -- card %d not done: rdy=0x%02x DSTAT=0x%08x wmask=0x%02x\n",
		  icard, rdy[icard], dstat, wmask[icard]);
	  outstanding = 1;
	}
      }
    }
    if(outstanding) {
      vvprintk(KERN_INFO "j=%lu, scheduling.\n", jiffies);
      schedule();
    } else {
      vvprintk(KERN_INFO "all WP PON RDY bits arrived\n");
      break;
    }
  } 
  if(outstanding) {
    printk(KERN_WARNING "dh_check_for_pon_rdy: Some wire pairs did not have WP PON RDY!\n");    
    for(icard=0; icard < DH_MAX_CARDS; icard++)
      if(wmask[icard])
	printk(KERN_WARNING "card %d rdy=0x%02x DSTAT=0x%08lx wmask=0x%02x\n",
	       icard, rdy[icard], (unsigned long) dh_read_regi(icard, DH_RG_DSTAT), wmask[icard]);
  }
  /* Save state to remember wire pairs which are ready to be powered on */
  for(icard=0; icard < DH_MAX_CARDS; icard++) 
    wmask[icard] &= rdy[icard];
}


static void dh_clear_wp_resets_check_pon_rdy(unsigned char * wmask) {
  /* Almost the same as dh_check_for_pon_rdy, except timeout is longer
     and we clear the wire pair resets */
  int icard, ipair;
  unsigned long flags;
  unsigned long j0 = jiffies;
  unsigned long TOUT = (HZ*500)/1000 + 1; /* Wait 500 msec before giving up */
  unsigned char rdy[DH_MAX_CARDS];
  int outstanding = 0;

  for(icard=0; icard < DH_MAX_CARDS; icard++) rdy[icard] = 0;
  int keepgoing = 1;
  while(keepgoing || WITHIN_TOUT(TOUT,j0)) { /* Make sure this executes at least once, even if 
						heavily loaded */
    if(!WITHIN_TOUT(TOUT,j0)) keepgoing = 0; /* Bail after this iteration */
    outstanding = 0;
    /* See if power-on-ready bits have come on */
    for(icard=0; icard < DH_MAX_CARDS; icard++) {
      if(!wmask[icard]) continue;
      if((rdy[icard] & wmask[icard]) != wmask[icard]) { /* Prevent unneeded extra reads */
	rdy[icard] = (dh_read_regi(icard, DH_RG_DSTAT)>>4) & 0x0F;
	/* See if all the desired bits have come on */
	if((rdy[icard] & wmask[icard]) == wmask[icard]) {
	  vvprintk(KERN_INFO "this guy's good: card %d, rdy=0x%02x wmask=0x%02x\n",
		  icard, rdy[icard], wmask[icard]);
	} else {
	  vvprintk(KERN_INFO "this guy's not done: card %d, rdy=0x%02x wmask=0x%02x\n",
		  icard, rdy[icard], wmask[icard]);
	  outstanding = 1;
	}
      }
    }
    if(outstanding) {
      schedule();
    } else {
      vvprintk(KERN_INFO "all wire pair resets have completed\n");
      break;
    }
  } 
  if(outstanding) {
    vprintk(KERN_INFO "some wp resets didn't complete:\n");
    for(icard=0; icard < DH_MAX_CARDS; icard++) {
      if((rdy[icard] & wmask[icard]) != wmask[icard]) {
	vprintk(KERN_INFO "Card %d, rdy=0x%02x wmask=0x%02x\n",
		icard, rdy[icard], wmask[icard]);
      }
    }
  }
  /* Find wire pairs which failed to power on */
  for(icard=0; icard < DH_MAX_CARDS; icard++) {
    if(wmask[icard]) {
      u32 ctrl = dh_read_regi(icard, DH_RG_CTRL);
      /* Eliminate failed wire pairs from active set and from control write register*/
      rdy[icard] &= (ctrl & 0x0F);
      u8 to_clear = 0x0F & (~rdy[icard]);
      dh_card_lock(icard, flags); /* LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL */
      dh_clear_write_regi_bits(icard, DH_RG_CTRL, (u32) to_clear);
      dh_card_unlock(icard, flags); /* UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU */
    }
  }
  /* Apply wire pair resets */
  for(icard=0; icard < DH_MAX_CARDS; icard++) {
    if(wmask[icard] && rdy[icard]) {
      dh_card_lock(icard, flags); /* LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL */
      dh_clear_write_regi_bits(icard, DH_RG_CTRL, rdy[icard]<<4);
      dh_card_unlock(icard, flags); /* UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU */
    } 
    wmask[icard] &= rdy[icard]; /* Save state to remember responsive DOMs */
  }
}

static void dh_comms_reset_mult(unsigned char * wp) {
  int icard, ipair;
  unsigned long flags;
  for(icard=0; icard < DH_MAX_CARDS; icard++) {
    if(wp[icard]) {
      dh_card_lock(icard, flags); /* LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL */
      unsigned char dommask = dh_dommask(wp[icard]);
      vvprintk(KERN_INFO "wp 0x%02x dommask 0x%02x\n",wp[icard], dommask);
      dh_write_regi_nosave(icard, DH_RG_DOMC, (dommask & 0xFF)<<8);
      dh_card_unlock(icard, flags); /* UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU */
    }
  }
}

static void dh_clear_enables(unsigned char * wp) {
  int icard, idom;
  unsigned long flags;
  for(icard=0; icard < DH_MAX_CARDS; icard++) {
    if(wp[icard]) {
      unsigned char dommask = dh_dommask(wp[icard]);
      dh_card_lock(icard, flags); /* LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL */
      vvprintk(KERN_INFO "Clearing enables for card %d\n", icard);

      /* Clear interrupt enable on RX */
      dh_clear_write_regi_bits(icard, DH_RG_TTSIC, dommask<<24);

      /* Clear interrupt enable on TX */
      dh_clear_write_regi_bits(icard, DH_RG_RTSIC, dommask<<24);

      /* Clear message received for RX */
      dh_set_write_regi_bits_nosave(icard, DH_RG_RTSIC, dommask<<16);

      /* Clear TCAL interrupt bits (should already be zero, since we don't 
	 use this [yet]) */
      dh_clear_write_regi_bits(icard, DH_RG_TCSIC, dommask<<24);

      /* Clear TCAL packet received */
      dh_set_write_regi_bits_nosave(icard, DH_RG_TCSIC, dommask<<16);

      /* Clear any pending DOMCs (really only affects regi. copies) */
      dh_clear_write_regi_bits(icard, DH_RG_DOMC, 
			       (dommask<<24)|(dommask<<16)|(dommask<<8)|dommask);

      /* Clear msg recv'd w/ error */
      dh_set_write_regi_bits_nosave(icard, DH_RG_DOMS, dommask<<24);

      /* Clear DSTAT conditions */
      dh_set_write_regi_bits_nosave(icard, DH_RG_DSTAT, 
				     (wp[icard]<<16)|(wp[icard]<<20)
				    |(wp[icard]<<24)|(wp[icard]<<28));

      /* Clear GSTAT conditions: FIXME -- this isn't 100% clean,
         as other errors could arise while this set of WPs is powering on... */
      dh_set_write_regi_bits_nosave(icard, DH_RG_GSTAT, 0xC0C);

      /* Set is-communicating, connection state */
      for(idom=0; idom < DH_DOMS_PER_CARD; idom++) {
	if(dommask & DH_BITN(idom)) { 
	  int m = dh_get_minor_from_card_and_channel(icard, idom);
	  dh_is_communicating[m] = 0;
	  dh_conn_state[m]       = DH_STATE_START;
	}
      }
      dh_card_unlock(icard, flags); /* UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU */
    }
  }
}

static int dh_pwr_on_doms_mult(unsigned char * pmask) {
  /* pmask is an array of bytes specifying pairs, one for each card */
  int ic, id, ip;
  unsigned char valid_wp_masks[DH_MAX_CARDS];
  unsigned char iscom[DH_MAX_CARDS];
  int somevalid = 0;
  unsigned long flags;

  int wpcards = dh_get_valid_wps(pmask, valid_wp_masks);
  if(!wpcards) {
    printk(KERN_WARNING "Requested wire pairs were invalid or already powered on, sorry.\n");
    return 0;
  }

  for(ic=0; ic<DH_MAX_CARDS; ic++) {
    vvprintk(KERN_INFO "WP masks for card %d: requested=0x%02x, valid=0x%02x\n",
	     ic, pmask[ic], valid_wp_masks[ic]);
  }

  dh_clear_enables(valid_wp_masks);
  dh_check_for_pon_rdy(valid_wp_masks); /* valid_wp_masks can change here ! */
  dh_wp_resets_pwron_mult(valid_wp_masks);
  dh_clear_wp_resets_check_pon_rdy(valid_wp_masks); /* valid_wp_masks can change here ! */
  dh_comms_reset_mult(valid_wp_masks);
  dh_wait_for_communicating(valid_wp_masks, iscom);

  for(ic=0; ic<DH_MAX_CARDS; ic++) {
    int had_communicating = 0;
    /* This is not clean - other processes could be powering other wire pairs
       on the same channel */
/*     /\* Explicitly clear pending comms reset to save bandwidth when only one channel */
/*        is communicating: *\/ */
/*     if(valid_wp_masks[ic]) { */
/*       dh_card_lock(ic, flags); /\* LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL *\/ */
/*       dh_set_write_regi_bits(ic, DH_RG_CTRL, DH_CTRL_CLR_COM_RES); */
/*       dh_clear_write_regi_bits(ic, DH_RG_CTRL, DH_CTRL_CLR_COM_RES);     */
/*       dh_card_unlock(ic, flags); /\* UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU *\/ */
/*     } */
    /* Report on communicating DOMs */
    for(id=0; id<DH_DOMS_PER_CARD; id++) {
      int m = dh_get_minor_from_card_and_channel(ic, id);
      dh_comms_reset_in_progress[m] = 0;
      if(DH_BITN(id) & iscom[ic] & dh_dommask(valid_wp_masks[ic])) {
	dh_is_communicating[m] = 1;
	vprintk(KERN_INFO "dh: DOM %d on card %d (minor %d) is responsive\n", id, ic, m);
      }
    }
  }
  return 1;
}

static void dh_pair_pwr_on(int icard, int ipair) {
  /** Tell FPGA design on card icard to power on DOM pair ipair. */
  //u32 pair_select_flags;
#define DH_POWER_ON_RDY_TIMEOUT 1000
  u32 dstat;
  int itrial, irdy;
  unsigned long flags;
  unsigned char pair_masks[DH_MAX_CARDS]; /* DOM Power vector for pwrall, etc. */

  int ic;
  for(ic=0; ic < DH_MAX_CARDS; ic++) pair_masks[ic] = 0;
  pair_masks[icard] = DH_BITN(ipair);
  dh_pwr_on_doms_mult(pair_masks);
}


static int dh_pwrall_write_proc(struct file* file, const char *buffer,
				unsigned long count, void *data) {
  /** Power ALL plugged-in wire pairs on.  This is needed for the moment because
      the act of powering on one channel can disturb adjacent channels. */
  char tmpbuf[REQUEST_STR_LEN];
  int icard, ipair, idom, compare_len, minor;
  int icounter, loopmax = 300, idel;
  unsigned long flags;
  u32 dstat, doms;
  unsigned char pair_masks[DH_MAX_CARDS]; /* DOM Power vector for pwrall, etc. */

  /* Make sure we don't blow buffer */
  compare_len = count > REQUEST_STR_LEN ? REQUEST_STR_LEN : count;

  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_pwrall_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  /* Handle "on" command */
  if(compare_len >= 2 && !strncmp(tmpbuf, "on", 2)) {
    for(icard=0; icard < DH_MAX_CARDS; icard++) pair_masks[icard] = 0x0F;
    dh_pwr_on_doms_mult(pair_masks);
    return count;
  } 

  /* Handle power off request */
  if(compare_len >= 3 && !strncmp(tmpbuf, "off", 3)) {
    vprintk(KERN_INFO "dh: Powering off all DOMs.\n");
    for(icard=0; icard < DH_MAX_CARDS; icard++) {
      if(! dh_pci_dors[icard].found) continue;
      for(ipair=0; ipair < DH_PAIRS_PER_CARD; ipair++) {
	vvprintk(KERN_INFO "dh_pwrall_write_proc: Will power off card %d pair %d.\n",
		 icard, ipair);
	dh_pair_pwr_off(icard, ipair);
      }
    }
    vvprintk(KERN_INFO "dh_pwrall_write_proc: Sleeping for %ld jiffies...\n", (long) HZ);
    dh_jsleep(HZ); /* wait one second */
    return count;
  }

  printk(KERN_ERR "dh_pwrall_write_proc: Got invalid control string, use \"on\" or \"off\".\n");
  dh_lasterr = DH_ERR_INVALIDARG;
  return -EINVAL;
}


static int dh_pseudo_intr_write_proc(struct file* file, const char *buffer,
				     unsigned long count, void *data) {
  /** Get pseudo interrupts for code testing/development.  \return bytes read from buffer. */
  struct pt_regs pt_dummy;
  int idummy;
  
  vprintk(KERN_INFO "dh_pseudo_intr_write_proc - got a pseudo interrupt.\n");
  dh_intr_hdlr(0, &idummy, &pt_dummy);
  return count;
}


static int dh_pair_pwr_check_write_proc(struct file* file, const char *buffer,
					unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int ipair = (d->pairnum&0x03);
  vprintk(KERN_INFO "dh_pair_pwr_check_write_proc: Resetting power check for card %d pair %d.\n",
	  icard, ipair);
  unsigned long flags;
  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  dh_set_write_regi_bits_nosave(icard, DH_RG_DSTAT,
				DH_BITN(8+ipair)  | 
				DH_BITN(16+ipair) | DH_BITN(20+ipair) | 
				DH_BITN(24+ipair) | DH_BITN(28+ipair));
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
  return count;
}


static int dh_pair_pwr_check_read_proc(char *page, char **start, off_t off, int count,
                                       int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard  = d->cardnum;
  *eof = 1;
  int len = 0;
  if(dh_pci_dors[icard].rev < 1) {
    len += snprintf(page+len, count-len, "Card %d is DOR Rev. 0, power check not supported.\n",
		    icard);
    return len;
  }
  int ipair  = d->pairnum&0x03;
  u32 dstat  = dh_read_regi(icard, DH_RG_DSTAT);
  u32 unpl   = dstat & DH_BITN(8+ipair);
  u32 cur_bl = dstat & DH_BITN(16+ipair);
  u32 cur_al = dstat & DH_BITN(20+ipair);
  u32 vol_bl = dstat & DH_BITN(24+ipair);
  u32 vol_al = dstat & DH_BITN(28+ipair);
  len += snprintf(page+len, count-len, "Card %d pair %d pwr check: "
		  "plugged(%s) current(%s,%s) voltage(%s,%s)\n", icard, ipair,
		  unpl ? "ERR_UNPLUGGED" : "ok",
		  cur_bl ? "ERR_CURRENT_BELOW_LIMITS" : "ok",
		  cur_al ? "ERR_CURRENT_ABOVE_LIMITS" : "ok",
		  vol_bl ? "ERR_VOLTAGE_BELOW_LIMITS" : "ok",
		  vol_al ? "ERR_VOLTAGE_ABOVE_LIMITS" : "ok");
  return len;
}


static int dh_pwr_write_proc(struct file* file, const char *buffer,
			     unsigned long count, void *data) {
  /** Get power on/off command from buffer, take appropriate action.  \return bytes read from buffer. */
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int compare_len;


  u32 ctrl  = dh_read_regi(d->cardnum, DH_RG_CTRL);

  if(d->cardnum < 0 || d->cardnum >= DH_MAX_CARDS || ! dh_pci_dors[d->cardnum].found) {
    printk(KERN_ERR "dh_pwr_write_proc: card number %d is not valid.\n",d->cardnum);
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  if(d->pairnum < 0 || d->pairnum >= DH_PAIRS_PER_CARD) {
    printk(KERN_ERR "dh_pwr_write_proc: pair number %d is out of bounds.\n",d->pairnum);
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  if(dh_firmware_mismatch[d->cardnum]) {
    printk(KERN_ERR "dh_pwr_write_proc: can't power on card %d, PCI or comms "
	   "firmware is unsupported.\n", d->cardnum);
    dh_lasterr = DH_ERR_UNSUPPFWARE;
    return -EINVAL;
  }

  /* Make sure we don't blow buffer */
  compare_len = count > REQUEST_STR_LEN ? REQUEST_STR_LEN : count;

  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_pwr_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  if(compare_len >= 2 && !strncmp(tmpbuf, "on", 2)) {
    dh_pair_pwr_on(d->cardnum, d->pairnum);
    return count;
  }

  if(compare_len >= 3 && !strncmp(tmpbuf, "off", 3)) {
    if(dh_mode != DH_NORMAL) {
      dh_sim_pwr_state[d->cardnum][d->pairnum] = 0;
      return count; /* Just don't do anything if simulation mode */
    } else {
      dh_pair_pwr_off(d->cardnum, d->pairnum);
      dh_jsleep(HZ); /* wait one second */
      return count;
    }
  }

  printk(KERN_ERR "dh_pwr_write_proc: Got invalid control string, use \"on\" or \"off\".\n");
  dh_lasterr = DH_ERR_INVALIDARG;
  return -EINVAL;
}

static int dh_rxdt_write_proc(struct file* file, const char *buffer,
			      unsigned long count, void *data) {
  /** Reset rxdt variable (max delta-T per message in RX FIFO) */
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int minor = dh_get_minor_from_card_pair_and_dom(d->cardnum,
                                                  d->pairnum,
                                                  d->domnum);
  /* Make sure we don't blow buffer */
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;

  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_rxdt_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  if(compare_len >= 5 && !strncmp(tmpbuf, "reset", 5)) {
    vprintk(KERN_INFO 
	    "dh_rxdt_write_proc: Resetting rxdt max. value for card %d pair %d dom %c.\n",
	    d->cardnum, d->pairnum, d->domnum?'B':'A');
    dh_domvars[minor].rxdt = 0;
  } else {
    printk(KERN_WARNING "dh_rxdt_write_proc: Invalid control string, use \"reset\".\n");
  }
  return count;
}

static int dh_pollute_write_proc(struct file* file, const char *buffer,
				  unsigned long count, void *data) {
  /** Pollutes N TX hardware packets, where N is written in decimal to proc file */
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];

  int minor = dh_get_minor_from_card_pair_and_dom(d->cardnum,
						  d->pairnum,
						  d->domnum);
  /* Make sure we don't blow buffer */
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_pollute_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  tmpbuf[compare_len] = '\0';
  int npollute;
  if(!sscanf(tmpbuf, "%d", &npollute)) {
    printk(KERN_ERR "dh_pollute_write_proc: Need to write an integer to this proc file.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
#ifdef DH_DO_POLLUTE
  printk(KERN_INFO "dh_pollute_write_proc: Will poison %d packets.\n", npollute);
  dh_pollute[minor] += npollute;
#else
  printk(KERN_INFO "dh_pollute_write_proc: DH_DO_POLLUTE not defined, ignoring request to pollute "
         "proc file (%d packets).\n", npollute);
#endif
  return count;
}



static int dh_softboot_write_proc(struct file* file, const char *buffer,
				  unsigned long count, void *data) {
  /** Initiates soft boot in DOM.
  \return bytes read from buffer. */

  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int compare_len;

  /* Check for valid argument */
  if(d->cardnum < 0 || d->cardnum >= DH_MAX_CARDS ||
     d->pairnum < 0 || d->pairnum >= DH_PAIRS_PER_CARD ||
     d->domnum < 0  || d->domnum  >= DH_DOMS_PER_PAIR) {
    printk(KERN_ERR "dh_softboot_write_proc: invalid argument (card %d pair %d dom %d).\n",
	   d->cardnum, d->pairnum, d->domnum);
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  int minor = dh_get_minor_from_card_pair_and_dom(d->cardnum,
						  d->pairnum,
						  d->domnum);
  /* Make sure DOM is plugged in */
  if(! dh_is_pair_plugged(d->cardnum, d->pairnum)) {
    printk(KERN_WARNING "dh_softboot_write_proc: Can't softboot unplugged wire pair"
	   " (card %d pair %d dom %d).\n",
           d->cardnum, d->pairnum, d->domnum);
    dh_lasterr = DH_ERR_SOFTBNOTPLGGD;
    return -EFAULT;
  }

  /* Make sure DOM is powered on */
  if(dh_get_pair_pwr_state(d->cardnum, d->pairnum) != 2) {  
    printk(KERN_WARNING "dh_softboot_write_proc: Can't softboot unpowered wire pair"
           " (card %d pair %d dom %d).\n",
           d->cardnum, d->pairnum, d->domnum);
    dh_lasterr = DH_ERR_SOFTBNOTPWRD;
    return -EFAULT;
  }

  /* Make sure DOM is communicating */
  if(! dh_is_communicating[minor]) {
    printk(KERN_WARNING "dh_softboot_write_proc: Can't softboot, DOM is not communicating"
           " (card %d pair %d dom %d).\n",
           d->cardnum, d->pairnum, d->domnum);
    dh_lasterr = DH_ERR_SOFTBNOTCOMM;
    return -EFAULT;
  }

  /* Make sure we don't blow buffer */
  compare_len = count > REQUEST_STR_LEN ? REQUEST_STR_LEN : count;

  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_softboot_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  if(compare_len >= 5 && !strncmp(tmpbuf, "reset", 5)) {
    if(dh_mode != DH_NORMAL) {
      vprintk(KERN_INFO "dh_softboot_write_proc: softboot (sim)"
	     " succeeded (card %d, pair %d, dom %d).\n",
             d->cardnum, d->pairnum, d->domnum);
      return count; /* Just don't do anything if simulation mode */
    } else {

      /* Acquire semaphore on wire pair - unfortunately firmware doesn't support 
	 A/B softbooting at same time */
      // acquire
      if(down_interruptible(&dh_wirepairs[d->cardnum][d->pairnum].softboot_sem)) {  /* DDDDDDDDDDD */
	printk(KERN_ERR "dh_softboot_write_proc: ABORT semaphore acquire for pair %d %d.\n",
	       d->cardnum, d->pairnum);
	dh_lasterr = DH_ERR_SEMAINT;
	return -EFAULT;
      }

      /* DOM is now by definition uncommunicating, even if softboot fails. */
      /* Empty retxb and retxq as well */
      unsigned long flags;
      dh_card_lock(d->cardnum, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
      dh_empty_retxb(minor);
      emptyRetxQ(&retxq[minor]);
      emptyAckQ(&ackq[minor]);
      emptyIOQ(inq[minor]);
      emptyIOQ(outq[minor]);
      txseq[minor] = 0;
      retx_triggered[minor] = 0;
      dh_conn_state[minor]  = DH_STATE_START;
      dh_comms_state[minor] = DH_COMMS_READY; /* Kernel timer will look for timeouts */
      dh_card_unlock(d->cardnum, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/

      if(dh_dom_softboot(d->cardnum, d->pairnum, d->domnum)) {
	printk(KERN_WARNING "dh: softboot (card %d, pair %d, dom %c) "
	       "was interrupted, or timed out.\n", d->cardnum, d->pairnum, d->domnum?'B':'A');
	up(&dh_wirepairs[d->cardnum][d->pairnum].softboot_sem);
	dh_lasterr = DH_ERR_SOFTBOOTFAIL;
	return -EAGAIN;
      }

      // release
      up(&dh_wirepairs[d->cardnum][d->pairnum].softboot_sem);

      vprintk(KERN_INFO "dh: softboot %d%d%c succeeded\n",
	     d->cardnum, d->pairnum, d->domnum?'B':'A');

      return count;
    }
  }

  printk(KERN_WARNING "dh_softboot_write_proc: Got invalid control string, use \"reset\".\n");
  return -EINVAL;
}


static int dh_pwr_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data) {
/** Method for reading power state for a card.
    @return number of characters "read," or put into the page/buffer */
  int len = 0;
  int ic,iw;
  int ipwr_state; /* 1: off 2: on 3: unknown/undefined */
  char *statestr;
  char onstate_str[] = "on";
  char offstate_str[] = "off";
  char undefstate_str[] = "undefined, unknown or in between";

  struct fpde_data *d;
  static int mycount=0;


  len = 0;
  mycount++;
  d = data;
  ic = d->cardnum;
  iw = d->pairnum;

  ipwr_state = dh_get_pair_pwr_state(ic, iw);
  
  switch(ipwr_state) {
  case 1: 
    statestr = offstate_str; break;
  case 2: 
    statestr = onstate_str; break;
  default:
    statestr = undefstate_str; break;
  }

  len += snprintf(page+len, count-len, "Card %d Pair %d power status is %s.\n",
		  ic,iw, statestr);
  *eof = 1;
  
  return len;
}


static int dh_domid_read_proc(char *page, char **start, off_t off, int count, 
				  int *eof, void *data) {
  /** Method for reading DOM ID.
      @return number of characters "read," or put into the page/buffer */
  struct fpde_data *d = (struct fpde_data *) data;

  int len;
  unsigned long flags;
  u64 id;

  len = 0;
  *eof = 1;

  /* Make sure arguments are valid and offset is zero (first entry) */
  if(off != 0 
     || d->cardnum < 0 || d->pairnum < 0 || d->domnum < 0
     || d->cardnum >= DH_MAX_CARDS 
     || d->pairnum >= DH_PAIRS_PER_CARD
     || d->domnum  >= DH_DOMS_PER_PAIR) {
    return 0;
  }

  /* Handle simulation */
  if(dh_mode != DH_NORMAL) {
    len += snprintf(page+len, count-len,
		    "Card %d Pair %d DOM %c ID is %012llx\n",
		    d->cardnum, d->pairnum,
		    d->domnum == 0 ? 'A' : 'B',
		    (unsigned long long) 0x01234567);
    return len;
  }

  /* Make sure DOM is communicating */
  if(! dh_get_pair_pwr_state(d->cardnum, d->pairnum) == 2     /* Powered on */
     || ! dh_is_pair_plugged(d->cardnum, d->pairnum)             /* Plugged in */
     || ! dh_is_communicating[dh_get_minor_from_card_pair_and_dom(d->cardnum,
								  d->pairnum,
								  d->domnum)] ) {
    len += snprintf(page+len, count-len,
		    "Can't get ID for non-communicating DOM (card %d Pair %d DOM %c).\n",
		    d->cardnum, d->pairnum,
		    d->domnum == 0 ? 'A' : 'B');
    dh_lasterr = DH_ERR_CANTGETDOMID;
    return len;
  }

  /* Get DOM ID */
  if(dh_get_dom_id(&id, d->cardnum, d->pairnum*2 + d->domnum)) {
    len += snprintf(page+len, count-len,
		    "Can't get DOM ID for card %d Pair %d DOM %c.\n",
		    d->cardnum, d->pairnum,
		    d->domnum == 0 ? 'A' : 'B');
    dh_lasterr = DH_ERR_CANTGETDOMID;
    return len;
  }

  /* Report DOM ID */
  len += snprintf(page+len, count-len,
		  "Card %d Pair %d DOM %c ID is %012llx\n",
		  d->cardnum, d->pairnum,
		  d->domnum == 0 ? 'A' : 'B',		    
		  (unsigned long long) id);

  return len;
}

static void dh_clear_cerr(int icard, int ipair, int idom) {
  int ich = ipair * 2 + idom;
  u32 cerr;
  cerr = ((ich&07) << 28) | DH_CERR_COMM_ERR_CLR;
  vvprintk(KERN_INFO "dh_clear_cerr: dom %d%d%c CERR=0x%08lx\n",
	   icard,ipair,idom?'B':'A',(unsigned long) cerr);
  dh_write_regi(icard, DH_RG_CERR, cerr);
}

static int dh_comstat_write_proc(struct file* file, const char *buffer,
                                  unsigned long count, void *data) {
  /** Reset communications statistics for DOM */
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int compare_len;

  /* Check for valid argument */
  if(d->cardnum < 0 || d->cardnum >= DH_MAX_CARDS ||
     d->pairnum < 0 || d->pairnum >= DH_PAIRS_PER_CARD ||
     d->domnum < 0  || d->domnum  >= DH_DOMS_PER_PAIR) {
    printk(KERN_ERR "dh_comstat_write_proc: invalid argument (card %d pair %d dom %d).\n",
           d->cardnum, d->pairnum, d->domnum);
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  /* Make sure we don't blow buffer */
  compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;

  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_comstat_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  if(compare_len >= 5 && !strncmp(tmpbuf, "reset", 5)) {
    vprintk(KERN_INFO "dh_comstat_write_proc: Resetting counters for card %d pair %d dom %c.\n",
	   d->cardnum, d->pairnum, d->domnum?'B':'A');
    unsigned long flags;
    dh_card_lock(d->cardnum, flags); /* LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL */
    int minor = dh_get_minor_from_card_pair_and_dom(d->cardnum,d->pairnum,d->domnum);
    dh_bytesread[minor]      = 
      dh_msgsread[minor]     =
      dh_pktsread[minor]     =
      dh_acksread[minor]     = 
      dh_badhdr[minor]       =
      dh_badseq[minor]       =
      dh_byteswritten[minor] = 
      dh_msgswritten[minor]  = 
      dh_resent[minor]       =
      dh_pktswritten[minor]  = 
      dh_ctrlsent[minor]     =
      dh_ctrlrecv[minor]     =
      dh_connects[minor]     =
      dh_nhdtouts[minor]     =
      dh_icsent[minor]       =
      dh_icrecv[minor]       =
      dh_cisent[minor]       =
      dh_cirecv[minor]       =
      dh_ackwritten[minor]   = 0;
    dh_clear_cerr(d->cardnum,d->pairnum,d->domnum);
    dh_card_unlock(d->cardnum, flags); /* UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU */
    
  } else {
    printk(KERN_WARNING "dh_comstat_write_proc: Invalid control string, use \"reset\".\n");
  }
  return count;
}

static int dh_domstat_read_proc(char *page, char **start, off_t off, int count,
				int *eof, void *data) {
  /** FIXME: finish the rest of this 
      Method for reading counters from DOM.
      Wanted to use wait_event_interruptible but for some reason that hangs
      and triggers the NMI watchdog. */
  struct fpde_data *d = (struct fpde_data *) data;
  int len = 0;
  *eof = 1; 
  int icard = d->cardnum;
  int ipair = d->pairnum;
  int idom  = d->domnum;
  int ich   = ipair*2 + idom;
  int minor = dh_get_minor_from_card_pair_and_dom(icard,ipair,idom);
  unsigned long flags;

  if(icard < 0 || ipair < 0 || idom < 0
     || icard >= DH_MAX_CARDS || ipair >= DH_PAIRS_PER_CARD
     || idom  >= DH_DOMS_PER_PAIR) {
    /* nop */
    printk(KERN_WARNING "dh_domstat_read_proc: invalid argument.\n"); 
  } else if(off != 0) {
    return len;
  } else if(!dh_opened[minor] ||
	    dh_hardware_timeout_set(icard, ipair*2 + idom) ||
	    dh_get_pair_pwr_state(icard, ipair) != 2 ||
	    ! dh_is_pair_plugged(icard, ipair) ||
	    ! dh_is_communicating[minor]) {
    len += snprintf(page+len, count-len,
		    "/dev/dhc%dw%dd%c is not ready (not powered, communicating and/or opened)\n",
		    icard, ipair, idom == 0 ? 'A' : 'B');
    return len;
  } else {
    /* Do it: queue packet */
#define DH_MAX_CTRL_REQUESTS 5
#define DH_CTRL_TIMEOUT 500
    int itrial;
    int gotresp = 0;
    for(itrial = 0; itrial < DH_MAX_CTRL_REQUESTS; itrial++) {
      vvprintk(KERN_INFO "dh: Sending control packet request %d.\n", itrial);
      domstats[minor].have_new = 0;
      if(dh_send_ctrl(icard, ich, minor)) {
        dh_jsleep(HZ/100);
	continue;
      }
      dh_ctrlsent[minor]++;
      interruptible_sleep_on_timeout(&domstatwq[minor], DH_CTRL_TIMEOUT);
      if(signal_pending(current)) {
	printk(KERN_WARNING "dh: Aborted wait for DOM control packet on minor %d.\n", minor);
        len += snprintf(page+len, count-len, "/dev/dhc%dw%dd%c control packet interrupted.\n",
			icard, ipair, idom == 0 ? 'A' : 'B');
	break;
      } else if(domstats[minor].have_new) {
	gotresp++;
	len += snprintf(page+len, count-len, "/dev/dhc%dw%dd%c\n"
			"CTYP=%d BADFINS=%d MINRXQ=%d MINRETX=%d MINACKQ=%d\n\n"
			"TX: PKTS=%d ACKS=%d REACKS=%d DATA=%d RESENT=%d CTRL=%d\n"
			"RX: PKTS=%d DATA=%d GOOD=%d ACKS=%d DUPDATA=%d DUPACKS=%d\n"
			"    GOODACKS=%d CTRL=%d DROPPED=%d BAD=%d INVALPOSTIC=%d\n",
			icard, ipair, idom == 0 ? 'A' : 'B',
			domstats[minor].stats.controlType,
			domstats[minor].stats.nBadFins,
			domstats[minor].stats.minRxQueueSize, 
			domstats[minor].stats.minRetxEntries, 
			domstats[minor].stats.minAckQueueEntries,
			domstats[minor].stats.nTxPkts,
			domstats[minor].stats.nTxAckPkts,
			domstats[minor].stats.nTxReackPkts,
			domstats[minor].stats.nTxDataPkts,
			domstats[minor].stats.nTxResentPkts,
			domstats[minor].stats.nTxControlPkts,
			domstats[minor].stats.nRxPkts,
			domstats[minor].stats.nRxDataPkts,
			domstats[minor].stats.nRxGoodDataPkts,
			domstats[minor].stats.nRxAckPkts,
			domstats[minor].stats.nRxDupDataPkts,
			domstats[minor].stats.nRxDupAcks,
			domstats[minor].stats.nRxGoodAcks,
			domstats[minor].stats.nRxControlPkts,
			domstats[minor].stats.nRxDroppedPkts,
			domstats[minor].stats.nRxBadPkts,
			domstats[minor].stats.nInvalidPostIC);
	break;
      } else {
	vvprintk(KERN_WARNING "dh: No CTRL response yet for minor %d.\n", minor);
      }
    }   // loop over ctrl trials
    if(!gotresp) {
      len += snprintf(page+len, count-len, 
		      "/dev/dhc%dw%dd%c: Timeout, no control packets returned.\n",
		      icard, ipair, idom == 0 ? 'A' : 'B');
    }
  } 
  return len;
}



static int dh_comstat_read_proc(char *page, char **start, off_t off, int count, 
				  int *eof, void *data) {
  /** Method for reading messagecount, etc. for DOM.
      @return number of characters "read," or put into the page/buffer */
  struct fpde_data *d = (struct fpde_data *) data;
  int len = 0;
  int icard = d->cardnum;
  int ipair = d->pairnum;
  int idom  = d->domnum;
  int ich   = ipair*2 + idom;
  *eof = 1;
  if(off != 0) return 0;
  if(icard < 0 || ipair < 0 || idom < 0
     || icard >= DH_MAX_CARDS || ipair >= DH_PAIRS_PER_CARD
     || idom  >= DH_DOMS_PER_PAIR) {
    /* nop */
  } else {  
    int minor = dh_get_minor_from_card_pair_and_dom(icard,ipair,idom);
    unsigned long flags;
    dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
    int nBadHwPkts = dh_get_comms_errors(icard, ipair*2+idom);
    dh_card_unlock(icard,flags);/*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    len += snprintf(page+len, count-len, 
		    "/dev/dhc%dw%dd%c\n"
		    "RX: %lluB, MSGS=%lu NINQ=%d PKTS=%lu ACKS=%lu \n"
		    "    BADPKT=%d BADHDR=%lu BADSEQ=%lu NCTRL=%lu NCI=%lu NIC=%lu\n",
		    icard, ipair, idom == 0 ? 'A' : 'B',
		    dh_bytesread[minor], dh_msgsread[minor], nIOQEntries(inq[minor]),
		    dh_pktsread[minor], dh_acksread[minor], 
		    nBadHwPkts, dh_badhdr[minor], dh_badseq[minor],
		    dh_ctrlrecv[minor], dh_cirecv[minor], dh_icrecv[minor]);
    len += snprintf(page+len, count-len,
		    "TX: %lluB, MSGS=%lu NOUTQ=%d RESENT=%lu PKTS=%lu ACKS=%lu\n"
		    "    NACKQ=%d NRETXB=%u RETXB_BYTES=%d NRETXQ=%u "
		    "NCTRL=%lu NCI=%lu NIC=%lu\n\n",
		    dh_byteswritten[minor], dh_msgswritten[minor], nIOQEntries(outq[minor]),
		    dh_resent[minor], dh_pktswritten[minor], 
		    dh_ackwritten[minor], nAckQEntries(&ackq[minor]), dh_num_retxb(minor),
		    dh_retxb_bytes(minor),
		    nRetxQEntries(&retxq[minor]), dh_ctrlsent[minor], dh_cisent[minor],
		    dh_icsent[minor]);
    len += snprintf(page+len, count-len,
                    "    NCONNECTS=%lu NHDWRTIMEOUTS=%lu OPEN=%s CONNECTED=%s\n"
		    "    RXFIFO=%s TXFIFO=%s,%s DOM_RXFIFO=%s\n",
		    dh_connects[minor], dh_nhdtouts[minor],
		    (dh_opened[minor]?"true":"FALSE"),
		    ((dh_conn_state[minor]==DH_STATE_CONNECTED)?"true":"FALSE"),
		    (dh_rx_fifo_is_empty(icard, ich)?"empty":"notempty"),
		    (dh_tx_fifo_almost_empty(icard, ich)?"almost empty":"NOT ALMOST EMPTY"),
		    (dh_tx_fifo_is_empty(icard, ich)?"empty":"notempty"),
		    (dh_dom_rx_fifo_isfull(icard, ich)?"FULL":"notfull"));
		    
  }
  return len;
}


static int dh_is_time_calibration_ready(int icard, int ipair, int idom) {
  /** Check to see if time calibration data is available */
  u32 tcsic, domc;
  int ich = ipair*DH_DOMS_PER_PAIR + idom;
  if(ich<0 || ich >= DH_DOMS_PER_CARD) return 0;  
  domc = dh_read_regi(icard, DH_RG_DOMC);
  if( domc & DH_BITN(24+ich) ) { /* Initiation bit still high / didn't self-clear 
				    -- not done  */
    pprintk(KERN_INFO "dh_is_time_calibration_ready: DOMC is 0x%08lx.\n",(unsigned long) domc);
    return 0;
  }

  tcsic = dh_read_regi(icard, DH_RG_TCSIC);
  pprintk(KERN_INFO "dh_is_time_calibration_ready: TCSIC is 0x%08lx.\n",(unsigned long) tcsic);
  if( (tcsic & DH_BITN(ich) ) 
         || 
      !(tcsic & DH_BITN(ich+16) )) { /* buffer is empty or packet not recv'd */
    return 0;
  }
  return 1;
}


static void dh_initiate_single_time_calibration(int icard, int ipair, int idom) {
  /** Initiate time calibration.  Set bit 16..23 of DOMC.  */
  int ich = ipair*DH_DOMS_PER_PAIR + idom;
  /* Per ALJ, add a single PCI read cycle to increase the delay to avoid nasty
     little race in the firmware -- only necessary if spinlock comes immediately
     before this function is called */
  dh_read_regi(icard, DH_RG_DOMC);
  dh_write_regi_nosave(icard, DH_RG_DOMC, DH_BITN(24+ich) & 0xFF000000);
}


static int dh_fill_tcalib_data(int icard, int ich, int minor) {
  /** Fill time calibration record from FIFO.  Use dh_is_time_calibration_ready first
      to make sure you have a time calibration record to read out of the TCBUF FIFO.
      Protect against races by using card lock in the caller. */

  u64 lo, hi;
  int nwf_words; int iword;

  dh_clear_write_regi_bits(icard, DH_RG_CTRL, 0x7<<12); /* Clear bits 12..14 */
  dh_set_write_regi_bits(icard, DH_RG_CTRL, (ich&0x7)<<12);

  /* Get header */
  u32 tcbuf = dh_read_regi(icard, DH_RG_TCBUF);
  if( (tcbuf >> 16) != 0x01) { /* Missing header info */
    printk(KERN_WARNING 
	   "dh_fill_tcalib_data: Header missing or corrupt (TCBUF==0x%lx).\n",\
	   (unsigned long) tcbuf);
    return 1;
  }

  unsigned numbytes = tcbuf & 0xFFFF; /* Low 16 bits */

  if(numbytes > 4*sizeof(u64) + DH_MAX_TCAL_WF_LEN*sizeof(u32)) {
    printk(KERN_WARNING "dh_fill_tcalib_data: Record has too many (%d) bytes; expected %d.\n",
	   numbytes, sizeof(struct dh_tcalib_t)); 
    return 1;
  }

  dh_tcalib_recs[minor]->hdr = tcbuf;

  /* DOR TX time */
  lo = dh_read_regi(icard, DH_RG_TCBUF);
  hi = dh_read_regi(icard, DH_RG_TCBUF);
  dh_tcalib_recs[minor]->dor_t0 = lo | hi<<32;
  pprintk(KERN_INFO "dor_t0 = 0x%llx\n", (unsigned long long) dh_tcalib_recs[minor]->dor_t0);

  /* DOR RX time */
  lo = dh_read_regi(icard, DH_RG_TCBUF);
  hi = dh_read_regi(icard, DH_RG_TCBUF);
  dh_tcalib_recs[minor]->dor_t3 = lo | hi<<32;
  pprintk(KERN_INFO "dor_t3 = 0x%llx\n", (unsigned long long) dh_tcalib_recs[minor]->dor_t3);


  /* nwf_words = DH_TCAL_WF_LEN/2; */
  nwf_words = (numbytes - 4*sizeof(u64))/8; /* Factor of 2 for bytes->samples, factor of 2 because
					       we have two waveforms; factor 2 'cause of two
					       samples per word */

  /* Get dorwf here */

  pprintk(KERN_INFO "nwf_words %d\n",nwf_words);
  for(iword = 0; iword < nwf_words; iword++) {
    tcbuf = dh_read_regi(icard, DH_RG_TCBUF);
    dh_tcalib_recs[minor]->dorwf[iword*2] = tcbuf & 0x3FF;
    dh_tcalib_recs[minor]->dorwf[iword*2+1] = (tcbuf >> 16) & 0x3FF;
    pprintk(KERN_INFO "tcbuf(%d) = 0x%lx.\n",iword, (unsigned long) tcbuf);
    pprintk(KERN_INFO "dorwf[%d] = %d.\n",iword*2, dh_tcalib_recs[minor]->dorwf[iword*2]);
    pprintk(KERN_INFO "dorwf[%d] = %d.\n",iword*2+1, dh_tcalib_recs[minor]->dorwf[iword*2+1]);
  }
  /* Zero out high samples */
  for(iword = nwf_words; iword < DH_MAX_TCAL_WF_LEN/2; iword++) {
    dh_tcalib_recs[minor]->dorwf[iword*2]   = 0;
    dh_tcalib_recs[minor]->dorwf[iword*2+1] = 0;
  }


  /* DOM RX time */
  lo = dh_read_regi(icard, DH_RG_TCBUF);
  hi = dh_read_regi(icard, DH_RG_TCBUF);
  dh_tcalib_recs[minor]->dom_t1 = lo | hi<<32;
  pprintk(KERN_INFO "dom_t1 = 0x%llx\n", (unsigned long long) dh_tcalib_recs[minor]->dom_t1);

  /* DOM TX time */
  lo = dh_read_regi(icard, DH_RG_TCBUF);
  hi = dh_read_regi(icard, DH_RG_TCBUF);
  dh_tcalib_recs[minor]->dom_t2 = lo | hi<<32;
  pprintk(KERN_INFO "dom_t2 = 0x%llx\n", (unsigned long long) dh_tcalib_recs[minor]->dom_t2);


  /* fill domwf here */
  for(iword = 0; iword < nwf_words; iword++) {
    tcbuf = dh_read_regi(icard, DH_RG_TCBUF);
    dh_tcalib_recs[minor]->domwf[iword*2] = tcbuf & 0x3FF;
    dh_tcalib_recs[minor]->domwf[iword*2+1] = (tcbuf >> 16) & 0x3FF;
    pprintk(KERN_INFO "tcbuf(%d) = 0x%lx.\n",iword, (unsigned long) tcbuf);
    pprintk(KERN_INFO "domwf[%d] = %d.\n",iword*2, dh_tcalib_recs[minor]->domwf[iword*2]);
    pprintk(KERN_INFO "domwf[%d] = %d.\n",iword*2+1, dh_tcalib_recs[minor]->domwf[iword*2+1]);
  }

  /* Zero out high samples */
  for(iword = nwf_words; iword < DH_MAX_TCAL_WF_LEN/2; iword++) {
    dh_tcalib_recs[minor]->domwf[iword*2]   = 0;
    dh_tcalib_recs[minor]->domwf[iword*2+1] = 0;
  }

  dh_tcalib_avail[minor] = 1;
  return 0;
}

static int dh_get_tcalib_data(int minor, struct dh_tcalib_t * dest) {
  /** Get a time calibration record into <dest>, if available.
      Set err if not.  
      Assumes minor is checked and that page has at least sizeof(struct dh_tcalib_t) bytes. 
      Must protect against races by locking in the caller. */

  if(! dh_tcalib_avail[minor]) return 1;
  memcpy(dest, dh_tcalib_recs[minor], sizeof(struct dh_tcalib_t));
  dh_tcalib_avail[minor] = 0;
  return 0;
}


static int dh_tcalib_read_proc(char *page, char **start, off_t off, int count,
			       int *eof, void *data) {
  /** Method for reporting time calibration results.  Uses the protected function
      dh_get_tcalib_data to fetch the record into a temporary buffer before copying
      to user space.  Got to do this in case <page> would page fault. */
#define TCAL_MAXTRIES 1000
  struct fpde_data *d = (struct fpde_data *) data;
  struct dh_tcalib_t this_tcal;
  unsigned long flags;

  *eof = 1;

  pprintk(KERN_INFO "dh_tcalib_read_proc: Asked for %d bytes.\n", count);

  if(count != sizeof(struct dh_tcalib_t)) {
    printk(KERN_WARNING "dh_tcalib_read_proc: Invalid time calibration read request (you have"
	   " to ask for %d bytes, you asked for %d).\n",sizeof(struct dh_tcalib_t), count);
    dh_lasterr = DH_ERR_TCALBADBUF;
    return 0;
  }

  int minor = dh_get_minor_from_card_pair_and_dom(d->cardnum,d->pairnum,d->domnum);
  int ich   = d->pairnum * DH_DOMS_PER_PAIR + d->domnum;


  if(dh_mode != DH_NORMAL) { /* No lock in sim mode */
    dh_get_tcalib_data(minor, &this_tcal); /* Guaranteed to succeed, filled data in write_proc */
    memcpy(page, &this_tcal, count);
    pprintk(KERN_INFO "dh_tcalib_read_proc: Returning record of %d bytes.\n", count);
    return count;
  }
    
  /* Normal mode, have to extract the data from the hardware */  

  dh_card_lock(d->cardnum, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/


  /* Watch for hardware timeouts */
  if(dh_hardware_timeout_set(d->cardnum, d->pairnum * 2 + d->domnum) || dh_do_sim_timeout[minor]) {
    vprintk(KERN_WARNING "dh_tcalib_read_proc: Hardware timeout on card %d, "
            "pair %d, DOM %d.\n", d->cardnum, d->pairnum, d->domnum);
    dh_card_unlock(d->cardnum, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    return -EAGAIN;
  }

  if(dh_comms_reset_in_progress[minor]) {
    vprintk(KERN_WARNING "dh_tcalib_read_proc: Comms reset in progress on card %d, "
            "pair %d, DOM %d.\n", d->cardnum, d->pairnum, d->domnum);
    dh_card_unlock(d->cardnum,flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    return -EAGAIN;
  }

  if(! dh_is_time_calibration_ready(d->cardnum, d->pairnum, d->domnum)) {
    dh_card_unlock(d->cardnum,flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    pprintk(KERN_WARNING "dh_tcalib_read_proc: Time calibration not ready on card %d, "
	    "pair %d, DOM %d.\n", d->cardnum, d->pairnum, d->domnum);
    return -EAGAIN; /* Tell user he/she needs to retry */
  }
  
  pprintk(KERN_INFO "dh_tcalib_read_proc: Time calibration ready on card %d, "\
	  "pair %d, DOM %d.\n", d->cardnum, d->pairnum, d->domnum);
  
  int iresult = dh_fill_tcalib_data(d->cardnum, ich, minor);

  if(iresult) {
    dh_card_unlock(d->cardnum, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    printk(KERN_WARNING "dh_tcalib_read_proc: Time calibration failed on card %d, "
	   "pair %d, DOM %d.\n", d->cardnum, d->pairnum, d->domnum);
    dh_lasterr = DH_ERR_TCALMISSORBAD;
    return -EAGAIN;
  }

  if(dh_get_tcalib_data(minor, &this_tcal)) {
    dh_card_unlock(d->cardnum, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    return 0; /* No data available -- probably time calibration event didn't happen. */
  }

  dh_card_unlock(d->cardnum, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
  
  memcpy(page, &this_tcal, count);
  pprintk(KERN_INFO "dh_tcalib_read_proc: Returning record of %d bytes.\n", count);
  return count;
}


static int dh_tcalib_write_proc(struct file* file, const char *buffer,
				unsigned long count, void *data) {
  /** Get time calib command from buffer, take appropriate action.  
      \return bytes read from buffer. */
#define MAXDIGITS 10
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int idigit, iwd;
  long loopmsec;
  unsigned long flags;

  /* Make sure we don't blow buffer */
  int compare_len = count > REQUEST_STR_LEN ? REQUEST_STR_LEN : count;

  int minor = dh_get_minor_from_card_pair_and_dom(d->cardnum,d->pairnum,d->domnum);

  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_tcalib_write_proc: Invalid pointer in user space.\n");
    return -EFAULT;
  }

  if(compare_len >= 6 && !strncmp(tmpbuf, "single", 6)) {

    if(dh_mode != DH_NORMAL) {
      pprintk(KERN_INFO "dh_tcalib_write_proc: got 'single' command (sim mode).\n");
      /* Set up a single time calibration */
      dh_tcalib_avail[minor] = 1; 
      dh_tcalib_recs[minor]->hdr    = 152 | (1 << 16);
      dh_tcalib_recs[minor]->dor_t0 = 0x12345678;
      dh_tcalib_recs[minor]->dor_t3 = 0x87654321;
      dh_tcalib_recs[minor]->dom_t1 = 0x11223344;
      dh_tcalib_recs[minor]->dom_t2 = 0x44332211;
      for(iwd = 0; iwd < DH_MAX_TCAL_WF_LEN; iwd++) {
	dh_tcalib_recs[minor]->domwf[iwd] = 0x100+iwd;
	dh_tcalib_recs[minor]->dorwf[iwd] = 0x200+iwd;
      }
      return count; /* Just don't do anything if simulation mode */
    } else {
      pprintk(KERN_INFO "dh_tcalib_write_proc: got 'single' command (normal mode).\n");

      /* No lock on pair_plugged or pwr_state because they are single register reads */
      if(!dh_is_pair_plugged(d->cardnum, d->pairnum)) {
	printk(KERN_WARNING "dh_tcalib_write_proc: Attempt to initiate tcalib on unplugged"
               " channel (card %d, pair %d).\n",d->cardnum, d->pairnum);
        dh_lasterr = DH_ERR_TCALCHNINVAL;
        return -EINVAL;
      }

      if(dh_get_pair_pwr_state(d->cardnum, d->pairnum) != 2) {
	printk(KERN_WARNING "dh_tcalib_write_proc: Attempt to calibrate unpowered"
               " channel (card %d, pair %d).\n",d->cardnum, d->pairnum);
        dh_lasterr = DH_ERR_TCALCHNINVAL;
        return -EINVAL;
      }

      dh_card_lock(d->cardnum, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
      if(dh_comms_reset_in_progress[minor]) {
	dh_card_unlock(d->cardnum,flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
	return -EAGAIN;
      }

      dh_initiate_single_time_calibration(d->cardnum, d->pairnum, d->domnum);
      dh_card_unlock(d->cardnum,flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/

      return count;
    }
  }

  if(compare_len >= 4 && !strncmp(tmpbuf, "stop", 4)) {
    if(dh_mode != DH_NORMAL) {
      vprintk(KERN_INFO "dh_tcalib_write_proc: got 'stop' command (sim mode).\n");
      return count; /* Just don't do anything if simulation mode */
    } else {
      vprintk(KERN_INFO "dh_tcalib_write_proc: got 'stop' command (normal mode).\n");
      /* Just don't do anything until loop code is implemented (if it ever is). */
      return count;
    }
  }

#define LOOPSTR "loop "
#define LOOPLEN strlen(LOOPSTR)
  if(compare_len >= LOOPLEN+1
     && !strncmp(tmpbuf, LOOPSTR, LOOPLEN) 
     && tmpbuf[LOOPLEN] >= '0' 
     && tmpbuf[LOOPLEN] <= '9') {
    loopmsec = 0;
    for(idigit = 0; idigit < MAXDIGITS; idigit++) {
      if(tmpbuf[LOOPLEN+idigit] >= '0' && tmpbuf[LOOPLEN+idigit] <= '9') {
	loopmsec *= 10;
	loopmsec += (tmpbuf[LOOPLEN+idigit]-'0');
      } else {
	break;
      }
    }
    if(dh_mode != DH_NORMAL) {
      vprintk(KERN_INFO "dh_tcalib_write_proc: got 'loop %ld' command (sim mode).\n", loopmsec);
      return count; /* Just don't do anything if simulation mode */
    } else {
      printk(KERN_WARNING "dh_tcalib_write_proc: got 'loop %ld' command"
	     " (normal mode, NOT IMPLEMENTED).\n", loopmsec);
      dh_lasterr = DH_ERR_TCALCMDINVAL;
      return count;
    }
  }


  printk(KERN_WARNING "dh_tcalib_write_proc: Got invalid control string, use \"single\", "
	 "\"stop\", or \"loop X\" (X in msec).\n");
  dh_lasterr = DH_ERR_TCALCMDINVAL;
  return -EINVAL;

}


static int dh_iscomm_write_proc(struct file* file, const char *buffer,
				  unsigned long count, void *data) {
  /** Initiate communications reset, sometimes needed after powering on all DOMs
      if voltage ramping not supported by hardware/firmware */
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int compare_len, minor;

  if(d->cardnum < 0 || d->pairnum < 0 || d->domnum < 0
     || d->cardnum >= DH_MAX_CARDS
     || d->pairnum >= DH_PAIRS_PER_CARD
     || d->domnum  >= DH_DOMS_PER_PAIR) {
    return -EFAULT;
  }

  minor = dh_get_minor_from_card_pair_and_dom(d->cardnum,d->pairnum,d->domnum);

  if(! dh_is_pair_plugged(d->cardnum, d->pairnum) ||
     (dh_get_pair_pwr_state(d->cardnum, d->pairnum) != 2)) {
    printk(KERN_WARNING "dh_iscomm_write_proc: Can't reset communications on unplugged"
	   " or unpowered wire pair (card %d pair %d dom %d).\n",
           d->cardnum, d->pairnum, d->domnum);
    dh_lasterr = DH_ERR_RESETINVAL;
    return -EFAULT;
  }

  /* Make sure we don't blow buffer */
  compare_len = count > REQUEST_STR_LEN ? REQUEST_STR_LEN : count;

  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_iscomm_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  if(compare_len >= 5 && !strncmp(tmpbuf, "reset", 5)) {
    if(dh_mode != DH_NORMAL) {
      vprintk(KERN_INFO "dh_iscomm_write_proc: Reset (sim)"  
	     " succeeded (card %d, pair %d, dom %d).\n",
             d->cardnum, d->pairnum, d->domnum);
      return count; /* Just don't do anything if simulation mode */
    }

    /* FIXME -- this should get rewritten, comms reset takes some time */

    dh_channel_comms_reset(d->cardnum, 0xFF & DH_BITN(d->pairnum*2 + d->domnum));

    if(dh_dom_isready(d->cardnum,d->pairnum*2 + d->domnum)) {
      vprintk(KERN_INFO "dh_iscomm_write_proc: Reset succeeded "
	     "(card %d, pair %d, dom %d).\n",
             d->cardnum, d->pairnum, d->domnum);
      dh_is_communicating[minor] = 1;
    } else {
      printk(KERN_WARNING "dh_iscomm_write_proc: Reset failed, DOM not communicating "
             "(card %d, pair %d, dom %d).\n",
             d->cardnum, d->pairnum, d->domnum);
      dh_is_communicating[minor] = 0;
    }
    /* Take this out, comms reset could be in progress on other DOMs */
    /* Clear communications reset bits explicitly so that firmware doesn't waste
       bandwidth (per Kalle) */
/*     dh_set_write_regi_bits(d->cardnum, DH_RG_CTRL, DH_CTRL_CLR_COM_RES); */
/*     dh_clear_write_regi_bits(d->pairnum, DH_RG_CTRL, DH_CTRL_CLR_COM_RES); */
    return count; 
  }

  printk(KERN_WARNING "dh_iscomm_write_proc:  Got invalid control string, use \"reset\".\n");
  return -EINVAL;

}


static int dh_iscomm_read_proc(char *page, char **start, off_t off, int count, 
				  int *eof, void *data) {
  /** Method for reporting if DOM is communicating or no (is firmware communications
      machinery working) */
  struct fpde_data *d = (struct fpde_data *) data;
  int len, minor, ok;

  len = 0;
  ok  = 0;

  if(d->cardnum < 0 || d->pairnum < 0 || d->domnum < 0
     || d->cardnum >= DH_MAX_CARDS 
     || d->pairnum >= DH_PAIRS_PER_CARD
     || d->domnum  >= DH_DOMS_PER_PAIR) {
    /* nop */
  } else {
    minor = dh_get_minor_from_card_pair_and_dom(d->cardnum,d->pairnum,d->domnum);
    if(dh_mode == DH_NORMAL 
              && !(dh_hardware_timeout_set(d->cardnum, d->pairnum * 2 + d->domnum))
		                                                /* No hardware timeout */
	      &&  dh_get_pair_pwr_state(d->cardnum, d->pairnum) == 2     /* Powered on */
	      &&  dh_is_pair_plugged(d->cardnum, d->pairnum)             /* Plugged in */
	      &&  dh_is_communicating[dh_get_minor_from_card_pair_and_dom(d->cardnum,
									  d->pairnum, 
									  d->domnum)] ) { 
      ok = 1;
    } else if(dh_mode != DH_NORMAL 
	  && dh_sim_pwr_state[d->cardnum][d->pairnum]) { /* All simulation modes are ok. */
      ok = 1; 
    }
      
    len += snprintf(page+len, count-len,
		    "Card %d Pair %d DOM %c %s\n",
		    d->cardnum, d->pairnum, 
		    d->domnum == 0 ? 'A' : 'B',
		    ok ? "is communicating" : "is NOT communicating");
  }
  *eof = 1;
  return len;
}


static int dh_isconn_read_proc(char *page, char **start, off_t off, int count, 
				  int *eof, void *data) {
  /** Method for reporting if DOM is in connected state.
   Requirements: 
        dh_conn_state == DH_STATE_CONNECTED
	no hardware timeout set
	pair powered on
	pair plugged in
	pair was communicating at power on */
  struct fpde_data *d = (struct fpde_data *) data;
  int len, minor, ok;

  len = 0;
  ok  = 0;

  if(d->cardnum < 0 || d->pairnum < 0 || d->domnum < 0
     || d->cardnum >= DH_MAX_CARDS 
     || d->pairnum >= DH_PAIRS_PER_CARD
     || d->domnum  >= DH_DOMS_PER_PAIR) {
    /* nop */
  } else {
    minor = dh_get_minor_from_card_pair_and_dom(d->cardnum,d->pairnum,d->domnum);
    if(dh_mode == DH_NORMAL 
       && dh_conn_state[minor] == DH_STATE_CONNECTED
       && !(dh_hardware_timeout_set(d->cardnum, d->pairnum * 2 + d->domnum))
       /* No hardware timeout */
       &&  dh_get_pair_pwr_state(d->cardnum, d->pairnum) == 2     /* Powered on */
       &&  dh_is_pair_plugged(d->cardnum, d->pairnum)             /* Plugged in */
       &&  dh_is_communicating[dh_get_minor_from_card_pair_and_dom(d->cardnum,
								   d->pairnum,
								   d->domnum)] ) {
      ok = 1;
    } else if(dh_mode != DH_NORMAL) {
      ok = 1; 
    }      
    len += snprintf(page+len, count-len,
		    "Card %d Pair %d DOM %c %s\n",
		    d->cardnum, d->pairnum, 
		    d->domnum == 0 ? 'A' : 'B',
		    ok ? "is connected" : "is NOT connected");
  }
  *eof = 1;
  return len;
}


int dh_dom_out_of_configboot(int icard, int ipair, int idom) {
  /* Kalle doesn't check for power-on state, so we do */
  u32 doms = dh_read_regi(icard, DH_RG_DOMS);
  if(dh_get_pair_pwr_state(icard, ipair) == 2 && dh_is_pair_plugged(icard, ipair)
     && (doms & (0x01 << (ipair*2+idom+24)))) {
    return 1;
  } else {
    return 0;
  }
}

static int dh_isnotcb_read_proc(char *page, char **start, off_t off, int count, 
				  int *eof, void *data) {
  /** Method for reporting if DOM is in configboot */

  struct fpde_data *d = (struct fpde_data *) data;
  int len = 0;
  len += snprintf(page+len, count-len,
		     "Card %d Pair %d DOM %c %s out of configboot\n", d->cardnum, d->pairnum, 'A' + d->domnum, 
		     dh_dom_out_of_configboot(d->cardnum, d->pairnum, d->domnum) ? "is" : "is not");
  *eof = 1;
  return len;
}


static int dh_plugged_read_proc(char *page, char **start, off_t off, int count, 
				int *eof, void *data) {
  /** Method for reading plugged in / not plugged in state of wire pair.
      @return number of characters "read," or put into the page/buffer */
  struct fpde_data *d = (struct fpde_data *) data;
  static char is_not[] = "is not";
  static char is[] = "is";
  static char *is_is_not;
  unsigned int plugged_in;
  int len;

  len = 0;

  if(d->cardnum < 0 || d->pairnum < 0 
     || d->cardnum >= DH_MAX_CARDS 
     || d->pairnum >= DH_PAIRS_PER_CARD) {
    is_is_not = is_not;
  } else if(dh_mode == DH_NORMAL) {
    /* Handle real hardware */
    plugged_in = dh_is_pair_plugged(d->cardnum, d->pairnum);
    is_is_not = plugged_in ? is : is_not;
  } else {
    is_is_not = is;
  }

  len += snprintf(page+len, count-len, "Card %d Pair %d %s plugged in.\n",
                  d->cardnum, d->pairnum, is_is_not);
  *eof = 1;
  return len;
}

static int dh_voltage_read_proc(char *page, char **start, off_t off, int count,
                                int *eof, void *data) {
  /** Method for reading voltage across wire pair. */
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int ipair = d->pairnum;
  int len = 0;
  if(off != 0) return 0;
  unsigned short V = 64;
  if(dh_pci_dors[icard].rev == 0) {
    len += snprintf(page+len, count-len, "Card %d Pair %d voltage is not available: "
		    "card %d is DOR Rev. 0.\n", icard, ipair, icard);
  } else {
    unsigned long flags;
    dh_card_lock(icard, flags); /* LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL */
    u32 dcur = dh_read_regi(icard, DH_RG_DCUR);
    vvprintk(KERN_INFO "rd dcur 0x%08lx\n", (unsigned long) dcur);
    dcur &= ~(0x03 << 28);
    dcur |= ((ipair & 0x03) << 28); /* Pair to read from */
    vvprintk(KERN_INFO "wr dcur 0x%08lx\n", (unsigned long) dcur);
    dh_write_regi(icard, DH_RG_DCUR, dcur);
    dcur = dh_read_regi(icard, DH_RG_DCUR);
    dh_card_unlock(icard, flags); /* UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU */
    unsigned long intvolts_raw = (dcur & 0xFFF0000) >> 16;
    unsigned long V000 = intvolts_raw * 28; /* Volts * 1000 / 35.64 */
    unsigned int Vhi  = V000/1000;
    unsigned int Vlo  = V000%1000;
    len += snprintf(page+len, count-len, "Card %d Pair %d voltage is %d.%03d Volts.\n",
		    icard, ipair, Vhi, Vlo);
  }
  *start = 0;
  *eof   = 1;
  return len;
}

static int dh_current_read_proc(char *page, char **start, off_t off, int count,
                                int *eof, void *data) {
  /** Method for reading current draw of wire pair.
      @return number of characters "read," or put into the page/buffer */
  struct fpde_data *d = (struct fpde_data *) data;
  int len;
  int intcur;
  u32 dcur;

  len = 0;
  if(off != 0) return 0;
  if(d->cardnum < 0 || d->pairnum < 0
     || d->cardnum >= DH_MAX_CARDS
     || d->pairnum >= DH_PAIRS_PER_CARD) {

    len += snprintf(page+len, count-len, "Card %d Pair %d current is not available"
		    ": bad card or pair ID.\n",
		    d->cardnum, d->pairnum);
    dh_lasterr = DH_ERR_INVALIDARG;
    *eof = 1;
    return len;
  }

  /* Handle real hardware */
  /* Prime DCUR with correct bits; protect existing regi. values */
  /* For DOR Rev 0, "prime the pump" with Bit 15 */
  unsigned long flags;
  dh_card_lock(d->cardnum, flags); /* LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL */
  dcur = dh_read_regi(d->cardnum, DH_RG_DCUR);
  dcur &= ~(0x03 << 12);
  dcur |= ((d->pairnum & 0x03) << 12); /* Pair to read from */
  if(dh_pci_dors[d->cardnum].rev == 0) 
    dcur |= DH_BITN(15);                /* Initiate read */
  dh_write_regi(d->cardnum, DH_RG_DCUR, dcur);
  dh_card_unlock(d->cardnum, flags); /* UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU */

  /* Wait for result (Rev. 0 only) */
  int done=0;
  if(dh_pci_dors[d->cardnum].rev == 0) {
    int iloop;
#define MAXCURRENTTRIES 500
    for(iloop = 0; iloop < MAXCURRENTTRIES; iloop++) {
      dcur = dh_read_regi(d->cardnum, DH_RG_DCUR);
      if(!(dcur & DH_BITN(15))) { /* Done */
	done = 1;
	vvprintk(KERN_INFO "dh_current_read_proc: done after %d tries: 0x%08lx\n", iloop,
		 (unsigned long) dcur);
	break;
      }
      schedule();
    }
  } else {
    dcur = dh_read_regi(d->cardnum, DH_RG_DCUR);
  }
  if(!done && dh_pci_dors[d->cardnum].rev == 0) {
    vprintk(KERN_INFO "dh_current_read_proc: timeout reading current on card %d pair %d.\n",
	    d->cardnum, d->pairnum);
    dh_lasterr = DH_ERR_CURRDTOUT;
    len += snprintf(page+len, count-len, "Card %d Pair %d current read FAILED.\n",
		    d->cardnum, d->pairnum);
  } else {
    intcur = dcur & 0x0FFF; /* bits 0..11 */

    if(dh_pci_dors[d->cardnum].rev == 0) {
      short shortcur = (short) (intcur << 4); /* pick up sign */
      short divcur   = shortcur / 16;         /* get rid of 0s from LSH */
      short curmA    = divcur / 13;           /* Convert to proper units */
      len += snprintf(page+len, count-len, "Card %d Pair %d current is %hd mA.\n",
		      d->cardnum, d->pairnum,
		      curmA);
    } else {
      unsigned short uscur = intcur/2;
      len += snprintf(page+len, count-len, "Card %d Pair %d current is %hu mA.\n",
		      d->cardnum, d->pairnum, uscur);
    }
  }
  *start = 0;
  *eof   = 1;
  return len;
}



static int dh_blocking_read_proc(char *page, char **start, off_t off,
				int count, int *eof, void *data) {
  /** Report parm_block */
  *eof = 1;
  if(off == 0) {
    return snprintf(page, count, "Default blocking mode is %s.  Write %d to change.\n", 
		    parm_block ? "BLOCKING" : "NONBLOCK", !parm_block);
  } else {
    return 0;
  }
}


static int dh_blocking_write_proc(struct file* file, const char *buffer,
				  unsigned long count, void *data) {
  /** Set default blocking/non-blocking mode */
  char mode;
  if(get_user(mode, buffer)) {
    printk(KERN_ERR "dh_blocking_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  vprintk(KERN_INFO "mode is %c.\n", mode);
  parm_block = (mode != '0');
  return count;
}

static int dh_verbose_read_proc(char *page, char **start, off_t off,
                                int count, int *eof, void *data) {
  /** Report buffer size to use when calling read/write */
  *eof = 1;
  if(off == 0) {
    return snprintf(page, count, "Verbose level is %d.\n", parm_verbose);
  } else {
    return 0;
  }
}

static int dh_verbose_write_proc(struct file* file, const char *buffer,
				 unsigned long count, void *data) {
  /** Set output level of driver */
  char level;
  if(get_user(level, buffer)) {
    printk(KERN_ERR "dh_verbose_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  if(level < '0' || level > '9') {
    printk(KERN_ERR "dh_verbose_write_proc: Verbose level must be 0..9.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EINVAL;
  }
  vprintk(KERN_INFO "verbose level is now %c.\n", level);
  parm_verbose = level - '0';
  return count;
}


static int dh_bufsiz_read_proc(char *page, char **start, off_t off,
				int count, int *eof, void *data) {
  /** Report buffer size to use when calling read/write */
  *eof = 1;
  if(off == 0) {
    return snprintf(page, count, "%d\n", DH_MAX_MESSAGE_SIZE);
  } else {
    return 0;
  }
}


static int dh_revision_read_proc(char *page, char **start, off_t off, int count, 
				 int *eof, void *data) {
  /** Report driver version/revision.
      @return number of characters "read," or put into the page/buffer */
  int len;

  *eof = 1;
  if(off == 0) {
    return snprintf(page, count, DH_DELIVERY_TAG " (dh.c $Revision: 1.381 $)\n");
  } else {
    return 0;
  }
}


static int dh_lasterr_read_proc(char *page, char **start, off_t off, int count, 
				int *eof, void *data) {
  int len;

  len  = 0;
  if(off == 0) {
    len += snprintf(page+len, count-len, "%d: %s\n", dh_lasterr, 
		    dh_errs[dh_lasterr]);
  }
  pprintk(KERN_INFO "dh_lasterr_read_proc: len is %d, count %d, off %d.\n", len, count, off);
  dh_lasterr = DH_ERR_NONE;
  *eof = 1;
  return len;
}


static int dh_driverstat_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data) {
  /** Proc file callback to report general driver status info. 
      Uses the record selector hack described in Rubini et. al., vol.2, p.105 
      (or see /usr/src/linux-2.4/fs/proc/generic.c).
      This routine still has the problem that sometimes count is too small
      to accommodate the proc file output.
      @return number of characters "read," or put into the page/buffer */
  int len;
  int nopened;
  int idev;

  len     = 0;
  nopened = 0;

  idev = (int) off;

  *eof = 0;

  if(idev >= DH_DOMHUB_DOMS) { /* Last record done */
    pprintk(KERN_INFO "status proc file: idev == %d.\n", idev);
    *start = 0;
    *eof = 1;
    return 0;
  } else {
    (*start)++;
  }

  if(idev == 0) { // && dh_some_device_open()) {
    len += snprintf(page+len, count-len, "FILE    OUTQ     INQ    MSGS    MSGS\n"
		                         " NUM     TOT     TOT    READ WRITTEN\n"
		                         "====    ====    ====    ====    ====\n");
    pprintk(KERN_INFO "status proc file: idev == %d.\n", idev);
  }

  if(dh_opened[idev]) {
    pprintk(KERN_INFO "count %d len %d\n",count,len);
    if(count-len < 72) {
      len += snprintf(page+len,count-len,"BUFFER TOO SMALL!\n"); 
    } else {
      len += snprintf(page+len, count-len,"\n%4d %7d %7d %7lu %7lu\n",
		      idev, nIOQEntries(inq[idev]), nIOQEntries(outq[idev]), 
		      dh_msgsread[idev], dh_msgswritten[idev]);
    }
  } else {
    //len += snprintf(page+len, count-len,"%4d not open\n", idev);
    len += snprintf(page+len, count-len," ");
  }

  if((int) off == DH_DOMHUB_DOMS-1) {
    len += snprintf(page+len, count-len,
		    "\n%lu msgs read from and %lu msgs written to "
		    "the driver (%lu total).\n", 
		    dh_total_msgs_read,
		    dh_total_msgs_written,
		    dh_total_msgs_written + dh_total_msgs_read);
  }

  pprintk(KERN_INFO "returning len %d, count is %d.  idev is %d, eof %d.\n",\
	 len, count, idev, *eof);
  return len < count ? len : count;
}



static int dh_flash_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data) {
  /** Read data from flash
      @return number of characters "read," or put into the page/buffer */
  int ipos;
  struct flash_data *d = (struct flash_data *) data;
  unsigned char b;

  if(dh_mode != DH_NORMAL) { /* Do nothing if running as simulation */
    return 0;
  }

  pprintk(KERN_INFO "This is the contents of flash page %d on card %d, from position %ld.\n",\
	 d->flashnum, d->cardnum, off);

  *start = page; /* Causes function to be called again to get more data */

  for(ipos = 0; ipos < count; ipos++) {
    if(off + ipos >= dh_flash_page_size(d->cardnum, d->flashnum)) {
      *start = 0;
      *eof   = 1;
      return ipos;
    }

    /* Copy byte here, from flash position off+ipos to userspace position page+ipos */
    b = dh_read_flash_byte(d->cardnum, d->flashnum, off+ipos);
    pprintk(KERN_INFO "dh_flash_read_proc - Read byte 0x%x from position %ld.\n",(int) b,(long)(off+ipos));
    *page++ = b;

  }

  pprintk(KERN_INFO "returning %d bytes\n", ipos);
  return ipos;
}

static int dh_flprotect_write_proc(struct file* file, const char *buffer,
				   unsigned long count, void *data) {
  /** Implement chip-wide flash protect/unprotect a la Arthur's dt.c program.
      See e-mail correspondence March-April '05. */
  char * tmpbuf = vmalloc(count+1);
  struct fpde_data * d = data;
  int icard = d->cardnum;

  if(tmpbuf == NULL) {
    printk(KERN_ERR "dh_flprotect_write_proc: Couldn't get space for buffer copy.\n");
    return -ENOMEM;
  }

  if(! parm_unlock_pages) {
    printk(KERN_ERR "Experts only -- you can't lock/unlock flash with the driver in this mode.\n");
    return -EINVAL;
  }

  /* Copy the buffer */
  if(copy_from_user(tmpbuf,buffer,count)) {
    printk(KERN_ERR "dh_flprotect_write_proc: Copy_from_user failed (bad buffer arg?).\n");
    vfree(tmpbuf);
    return -EFAULT;
  }
  tmpbuf[count] = '\0';

  int uplen = strlen("unprotect");
  int prlen = strlen("protect");
  if(!strncmp(tmpbuf, "unprotect", (count>uplen?uplen:count))) {
    printk(KERN_INFO "Ok, doing unprotect of DOR card %d.\n", icard);
    if(flashUnprotectChip(icard)) {
      printk(KERN_INFO "Flash unprotect of DOR card %d failed!\n", icard);
      vfree(tmpbuf);
      return -EIO;
    }
    printk(KERN_INFO "Unprotect of card %d succeeded.\n", icard);
  } else if(!strncmp(tmpbuf, "protect", (count>prlen?prlen:count))) {
    printk(KERN_INFO "Ok, doing protect of DOR card %d.\n", icard);
    if (flashProtectChip(icard)) {
      printk(KERN_INFO "Flash protect of DOR card %d failed!\n", icard);
      vfree(tmpbuf);
      return -EIO;
    }
    printk(KERN_INFO "Protect of card %d succeeded.\n", icard);
  } else {
    printk(KERN_INFO "dh_flprotect_write_proc: Unknown command \"%s\", use \"protect\""
	   " or \"unprotect\".\n", tmpbuf);
    vfree(tmpbuf);
    return -EINVAL;
  }

  vfree(tmpbuf);
  return count;
}

static int dh_flash_write_proc(struct file* file, const char *buffer,
			       unsigned long count, void *data) {
  /** load data into flash.  Look for primer command.  If present, reset flash
      position to zero.  Otherwise, move bytes from buffer into flash.  Make
      sure you don't go past the page boundary.
      @return number of bytes written. */

  struct flash_data *d = (struct flash_data *) data;
  int ibyte;
  char to_write, read_back;
  char * tmpbuf; /* Allocate internally to copy buffer from userspace */
# define PROGRAM_STRING_LEN 15
  char program_string[]   = "flash X program"; /* X will be set by snprintf */

  if(dh_mode != DH_NORMAL) { /* Do nothing if simulation mode */
    return count;
  }

  pprintk(KERN_INFO "dh_flash_write_proc, card %d, flash page %d: to write %ld bytes.\n",
         d->cardnum, d->flashnum, count);

  if(d->cardnum < 0 || d->cardnum >= DH_MAX_CARDS) {
    printk(KERN_ERR "dh_flash_write_proc: card number %d is out of bounds.\n",d->cardnum);
    return -EFAULT;
  }

  if(d->flashnum < 0 || d->flashnum  >= DH_FLASH_PAGES) {
    printk(KERN_ERR "dh_flash_write_proc: flash page number %d is out of bounds.\n",d->flashnum);
    return -EFAULT;
  }

  /* Allocate the buffer.  */
  tmpbuf = vmalloc(count);
  if(tmpbuf == NULL) {
    printk(KERN_ERR "dh_flash_write_proc: Couldn't get space for buffer copy.\n");
    return -ENOMEM;
  }

  /* Copy the buffer */
  if(copy_from_user(tmpbuf,buffer,count)) {
    printk(KERN_ERR "dh_flash_write_proc: Copy_from_user failed (bad buffer arg?).\n");
    vfree(tmpbuf);
    return -EFAULT;
  }
  
  pprintk(KERN_INFO "dh_flash_write_proc: Buffer to write consists of the following bytes:\n");
  /* dh_printbuf(tmpbuf,count); */

  /* Prepare flash page for read/write */

  /* Look for primer command */
  snprintf(program_string, PROGRAM_STRING_LEN+1, "flash %d program", d->flashnum);

  if(!strncmp(tmpbuf, program_string, 
              ((PROGRAM_STRING_LEN < count) ? PROGRAM_STRING_LEN : count))) {
    vprintk(KERN_INFO "dh_flash_write_proc: rewinding fpos on flash page %d, card %d.\n",
	    d->flashnum, d->cardnum);
    d->fpos = 0;

    vprintk(KERN_INFO "dh_flash_write_proc: issuing flash write enable - don't forget to clear!!!.\n");

    vprintk(KERN_INFO "dh_flash_write_proc: issuing flash reset.\n");
    dh_flash_reset(d->cardnum);

    if(dh_flash_erase_page(d->cardnum, d->flashnum)) {
      printk(KERN_WARNING "dh_flash_write_proc: Attempt to clear flash page failed.\n");
      vfree(tmpbuf);
      return -EFAULT;
    }

    /* Check it to be sure it worked */
    if(!dh_flash_page_isempty(d->cardnum, d->flashnum)) {
      printk(KERN_WARNING "dh_flash_write_proc: Attempt to clear flash page failed.\n");
      //dh_clear_write_regi_bits(d->cardnum, DH_RG_CTRL, DH_CTRL_FLASH_ACC_EN);
      vfree(tmpbuf);
      return -EFAULT;
    } else {
      vprintk(KERN_INFO "dh_flash_write_proc: Page %d clear.\n", d->flashnum);
    }

  } else { /* Data to actually write to flash must be given in a separate write command */

    /* Make sure you're not trying to write past page bdry */
    if(d->fpos + count > dh_flash_page_size(d->cardnum, d->flashnum)) {
      printk(KERN_ERR "dh_flash_write_proc: Attempt to write too many bytes (fpos %ld, count %ld)"
	     " to flash (max size %ld).\n",
	     d->fpos,count,(long) dh_flash_page_size(d->cardnum, d->flashnum));
      //dh_clear_write_regi_bits(d->cardnum, DH_RG_CTRL, DH_CTRL_FLASH_ACC_EN);
      vfree(tmpbuf);
      return -EFAULT;
    }

    vprintk(KERN_INFO "dh_flash_write_proc: %ld bytes -> flash(%ld,...).\n",
	   count, d->fpos);

    /* Send it on in... */
    for(ibyte=0;ibyte<count;ibyte++) {
      to_write = tmpbuf[ibyte];
      read_back = dh_write_flash_byte(d->cardnum,d->flashnum,d->fpos+ibyte, to_write);
      if(read_back != to_write) {
	printk(KERN_ERR 
	       "dh_flash_write_proc: ERROR - couldn't write byte 0x%x('%c') to flash location %ld"
	       " on page %d, flash addr (in page) = %lx, card %d (read back byte value 0x%2x('%c')).\n", 
	       (unsigned int) to_write, DH_PR_BYTE(to_write), d->fpos, d->flashnum, d->fpos+ibyte,
	       d->cardnum, (unsigned int) read_back & 0xFF, DH_PR_BYTE(read_back));
	//dh_clear_write_regi_bits(d->cardnum, DH_RG_CTRL, DH_CTRL_FLASH_ACC_EN);
	vfree(tmpbuf);
	return -EFAULT;
      }
    }

    d->fpos += count;
  }

  vfree(tmpbuf);
  return count;
}


static int dh_pci_write_proc(struct file* file, const char *buffer,
			     unsigned long count, void *data) {
  /** Restore the PCI configuration space that was saved at device init time */
  struct fpde_data * d = data;
  int icard = d->cardnum;

  if(dh_restore_pci_config_space(&dh_pci_dors[icard])) {
    printk(KERN_ERR "dh_pci_write_proc: Couldn't restore PCI configuration space for card %d.\n",
	   icard);
    dh_lasterr = DH_ERR_CANTRESTRPCI;
    return -EFAULT;
  }
  vprintk(KERN_INFO "dh_pci_write_proc: Restored PCI configuration space for card %d.\n",icard);
  return count;
} 

static int dh_1pps_write_proc(struct file* file, const char *buffer,
			      unsigned long count, void *data) {
  /** Restore the PCI configuration space that was saved at device init time */
  struct fpde_data * d = data;
  char tmpbuf[REQUEST_STR_LEN];
  int compare_len;
  u32 gstat;
  int icard = d->cardnum;
  unsigned long flags;

  compare_len = count > REQUEST_STR_LEN ? REQUEST_STR_LEN : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_1pps_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  if(compare_len < 4 || strncmp(tmpbuf, "init", 4)) {
    printk(KERN_ERR "dh_1pps_write_proc: Invalid command.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  if(dh_mode != DH_NORMAL) return count;

  dh_utc_timer_status[icard] = DH_UTC_TIMER_IDLE;
  gstat = dh_read_regi(icard, DH_RG_GSTAT);
  if(gstat & DH_GSTAT_PPS_TEST) {
    printk(KERN_WARNING "dh_1pps_write_proc: PPS_TEST already high.\n");
    dh_utc_timer_status[icard] = DH_UTC_TIMER_ERR;
    dh_lasterr = DH_ERR_1PPSINITFAIL;
    return count;
  }

  /* Set PPS_TEST */
  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  gstat = dh_read_regi(icard, DH_RG_GSTAT);
  dh_write_regi_nosave(icard, DH_RG_GSTAT, gstat | DH_GSTAT_PPS_TEST);
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/

  /* Verify that PPS_TEST is still high */
  gstat = dh_read_regi(icard, DH_RG_GSTAT);
  if(!(gstat & DH_GSTAT_PPS_TEST)) {
    printk(KERN_WARNING "dh_1pps_write_proc: PPS_TEST didn't go high.\n");
    dh_utc_timer_status[icard] = DH_UTC_TIMER_ERR;
    dh_lasterr = DH_ERR_1PPSINITFAIL;
    return count;
  }

  dh_utc_timer_status[icard] = DH_UTC_TIMER_START;
  vprintk(KERN_INFO "dh_1pps_write_proc: Initialized timer load/read sequence.\n");
  return count;
}


static int dh_1pps_read_proc(char *page, char **start, off_t off,
			     int count, int *eof, void *data) {
  struct fpde_data * d = data;
  int icard = d->cardnum;
  int len;
  *eof = 1;

  len = 0;
  /* Make sure timer load completed */
  if((dh_utc_timer_status[icard] != DH_UTC_TIMER_LOADED) 
     &&
     (dh_utc_timer_status[icard] != DH_UTC_TIMER_FETCHED)) {
    printk(KERN_ERR "dh_1pps_read_proc: 1pps timer was not loaded (or fetched) successfully.\n");
    dh_lasterr = DH_ERR_TIMERNOTLOAD;
    len += snprintf(page+len, count-len, "error: timer operation for card %d did not complete.\n",
		    d->cardnum);
    return len;
  }
  dh_utc_timer_status[icard] = DH_UTC_TIMER_IDLE;
  len += snprintf(page+len, count-len, "ok");
  return len;
}


static int dh_rev_read_proc(char *page, char **start, off_t off,
			    int count, int *eof, void *data) {
  struct fpde_data * d = data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  len += snprintf(page+len, count-len, "%d\n", dh_pci_dors[d->cardnum].rev);
  return len;
}

static int dh_test_log_read_proc(char *page, char **start, off_t off,
                            int count, int *eof, void *data) {
  const struct fpde_data *d = data;
  const int icard = d->cardnum;
  int ipos;

  if(dh_mode != DH_NORMAL) { /* Do nothing if running as simulation */
    *eof=1;
    return 0;
  }

  if(dh_pci_dors[icard].rev==0) { /* rev 0 cards do not have the test log... */
    *eof=1;
    return 0;
  }

  /** Read data from flash log sector -- until end of sector or 0xff...
      @return number of characters "read," or put into the page/buffer */
  *start = page; /* Causes function to be called again to get more data */
  for(ipos = 0; ipos < count; ipos++) {
    const int flash_page = DH_LOG_SECTOR/flash_sectors_per_page(icard);
    const int sector = DH_LOG_SECTOR - flash_page*flash_sectors_per_page(icard);
    const int addr = sector*dh_flash_sector_size(icard);
    if(off + ipos >= dh_flash_sector_size(icard) ||
       (unsigned char)(page[ipos]=dh_read_flash_byte(icard, flash_page, addr + off + ipos))==0xff) {
      *start = 0;
      *eof   = 1;
      break;
    }
  }

  return ipos;
}

static int dh_syncgps_read_proc(char *page, char **start, off_t off,
				int count, int *eof, void *data) {
  struct fpde_data * d = data;
  unsigned char tbuf[DH_TSBYTES];
  int icard = d->cardnum;
  *eof = 1;
#define SOH 1
#ifdef SIM_GPS
  char timestr[14] = { SOH, '1','2','3',':','4','5',':','6','7',
		       ':','8','9','?' };
  unsigned long long tsim = 0x7766554433221100LL;
  memcpy(page, timestr, 14);
  memcpy(page+14, &tsim, 8);
  return 22;
#else
  u32 gstat;
  u32 tstrg;
  unsigned char cts;
#define MAX_TSRG_TRIES 100
#define MAX_CTS_TRIES  1000
  int itstrg;
  for(itstrg=0; itstrg<MAX_TSRG_TRIES; itstrg++) {
    int icts;
    for(icts=0; icts<MAX_CTS_TRIES; icts++) {
      /* Move forward through TSTRG buffer until SOH
	 found or GSTAT says no more data */
      gstat = dh_read_regi(icard, DH_RG_GSTAT);
      if(! (gstat & DH_GSTAT_TSTRG_10) && 
	 ! (gstat & DH_GSTAT_TSTRG_1)  &&
	 ! (gstat & DH_GSTAT_TSTRG_F_FF)) {
	return 0;
      }
      cts = (unsigned char) (dh_read_regi(icard, DH_RG_TSTRG) & 0xFF);
      if(cts == SOH) break;
      if(cts == CR || cts == LF) continue;
      printk(KERN_WARNING "dh_syncgps_read_proc: Skipping non-SOH start byte "
	     "(0x%02x) on card %d.\n", cts, icard);
    }
    if(icts >= MAX_CTS_TRIES) {
      printk(KERN_WARNING "dh_syncgps_read_proc: Read %d bytes from FPGA, no SOH found!\n",
	     icts);
      return 0;
    }
    tbuf[0] = cts;
    int i;
    int goodstring = 1;
    for(i=1; i<DH_TSBYTES; i++) {
      cts = (unsigned char) (dh_read_regi(icard, DH_RG_TSTRG) & 0xFF);
      vvprintk(KERN_INFO "dh: read byte 0x%02x\n", (unsigned char) cts);
      /* Check for valid digits - should probably write a real parser */
      if((( i >= 1 && i <= 3) || /* Day */
	  ( i == 5 || i == 6) || /* Hr */
	  ( i == 8 || i == 9)) 
	 && /* Enforce 0..9 */
	 (cts < '0' || cts > '9')) {
	goodstring = 0; break;
      } else if((i == 4 || i == 7 || i == 10) && cts != ':') {
	goodstring = 0; break;
      }
      tbuf[i] = cts;
    }
    if(!goodstring) {
      printk(KERN_WARNING "Bad time string character 0x%02x at position %d; retrying.\n",
	     cts, i);
      continue;
    }
    memcpy(page, tbuf, DH_TSBYTES);
    return DH_TSBYTES; 
  } /* Loop over attempts to get valid time string */
  return 0;
#endif
}


static int dh_clksel_read_proc(char *page, char **start, off_t off,
                             int count, int *eof, void *data) {
  struct fpde_data * d = data;
  int icard = d->cardnum;
  int len = 0;
  *start = 0;
  *eof = 1;    

  vprintk(KERN_INFO "dh_clksel_read_proc: entered.\n");

  if(dh_clksel[d->cardnum] == DH_CLKSEL_EXTERNAL) {
    len += snprintf(page+len, count-len, "external\n");
  } else if(dh_clksel[d->cardnum] == DH_CLKSEL_INTERNAL) {
    len += snprintf(page+len, count-len, "internal\n");
  } else {
    len += snprintf(page+len, count-len, "error\n");
  }
  return len;
}

static int dh_clksel_write_proc(struct file* file, const char *buffer,
                              unsigned long count, void *data) {
  /** Restore the PCI configuration space that was saved at device init time */
  struct fpde_data * d = data;
  int icard = d->cardnum;
  char localbuf[80];

  unsigned long nbytes = count > 80 ? 80 : count;
  unsigned long flags;
  if(copy_from_user((char *) localbuf, buffer, nbytes)) {
    printk(KERN_ERR "dh_clksel_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }

  if(!strncmp(localbuf, "external", 8)) {
    vprintk(KERN_INFO "dh_clksel_write_proc: Setting clock source for card %d to EXTERNAL.\n", 
	   icard);
    dh_clksel[icard] = DH_CLKSEL_EXTERNAL;
    return count;
  }

  if(!strncmp(localbuf, "internal", 8)) {
    vprintk(KERN_INFO "dh_clksel_write_proc: Setting clock source for card %d to INTERNAL.\n",
	    icard);
    dh_clksel[icard] = DH_CLKSEL_INTERNAL;
    return count;
  }

  printk(KERN_WARNING "dh_clksel_write_proc: Invalid argument, must be \"external\" or"
	 " \"internal\" (card=%d).\n", icard);
  return count;
}


static int dh_pci_read_proc(char *page, char **start, off_t off, 
			    int count, int *eof, void *data) {
  /** Show PCI configuration space */
  int len = 0;
  struct fpde_data * d = data;
  u16 vendor, device, command, status;
  u8 revision;
  u8 class0, class1, class2;
  u8 cache_line, latency_timer, header_type, BIST;
  u32 bar0, bar1, bar2, bar3, bar4, bar5, CIS;
  u16 ss_vendor, ss_dev;
  u32 rom_addr;
  u8 irq_line, irq_pin, min_gnt, max_lat;
  struct pci_dev * dev;

  dev = dh_pci_dors[d->cardnum].dev;

  if(!dh_pci_dors[d->cardnum].found) {
    len += snprintf(page+len, count-len, "No PCI information available for card %d.\n",d->cardnum);
  } else {
    len += snprintf(page+len, count-len, 
		    "PCI configuration space for card %d:\n", d->cardnum);
    
    pci_read_config_word(dev, 0, &vendor);
    pci_read_config_word(dev, 2, &device);
    pci_read_config_word(dev, 4, &command);
    pci_read_config_word(dev, 6, &status);
    pci_read_config_byte(dev, 8, &revision);
    pci_read_config_byte(dev, 9, &class0);
    pci_read_config_byte(dev, 0xa, &class1);
    pci_read_config_byte(dev, 0xb, &class2);
    pci_read_config_byte(dev, 0xc, &cache_line);
    pci_read_config_byte(dev, 0xd, &latency_timer);
    pci_read_config_byte(dev, 0xe, &header_type);
    pci_read_config_byte(dev, 0xf, &BIST);
    pci_read_config_dword(dev,0x10, &bar0);
    pci_read_config_dword(dev,0x14, &bar1);
    pci_read_config_dword(dev,0x18, &bar2);
    pci_read_config_dword(dev,0x1C, &bar3);
    pci_read_config_dword(dev,0x20, &bar4);
    pci_read_config_dword(dev,0x24, &bar5);
    pci_read_config_dword(dev,0x28, &CIS);
    pci_read_config_word(dev,0x2c, &ss_vendor);
    pci_read_config_word(dev,0x2e, &ss_dev);
    pci_read_config_dword(dev,0x30, &rom_addr);
    pci_read_config_byte(dev, 0x3c, &irq_line);
    pci_read_config_byte(dev, 0x3d, &irq_pin);
    pci_read_config_byte(dev, 0x3e, &min_gnt);
    pci_read_config_byte(dev, 0x3f, &max_lat);

    len += snprintf(page+len, count-len, "%26s: 0x%04x\n", 
		    "Vendor", (unsigned int) vendor);
    len += snprintf(page+len, count-len, "%26s: 0x%04x\n", 
		    "Device", (unsigned int) device);
    len += snprintf(page+len, count-len, "%26s: 0x%04x\n", 
		    "Command", (unsigned int) command);
    len += snprintf(page+len, count-len, "%26s: 0x%04x\n", 
		    "Status", (unsigned int) status);
    len += snprintf(page+len, count-len, "%26s: 0x%02x\n", 
		    "Revision", (unsigned int) revision);
    len += snprintf(page+len, count-len, "%26s: 0x%02x%02x%02x\n", 
		    "Class", (unsigned int) class2, (unsigned int) class1, (unsigned int) class0);
    len += snprintf(page+len, count-len, "%26s: 0x%02x\n", 
		    "Cache line", (unsigned int) cache_line);
    len += snprintf(page+len, count-len, "%26s: 0x%02x\n", 
		    "Latency timer", (unsigned int) latency_timer);
    len += snprintf(page+len, count-len, "%26s: 0x%02x\n", 
		    "Header type", (unsigned int) header_type);
    len += snprintf(page+len, count-len, "%26s: 0x%02x\n", 
		    "BIST", (unsigned int) BIST);

    len += snprintf(page+len, count-len, "%26s: 0x%08x\n", 
		    "Base address register 0", (unsigned int) bar0);
    len += snprintf(page+len, count-len, "%26s: 0x%08x\n", 
		    "Base address register 1", (unsigned int) bar1);
    len += snprintf(page+len, count-len, "%26s: 0x%08x\n", 
		    "Base address register 2", (unsigned int) bar2);
    len += snprintf(page+len, count-len, "%26s: 0x%08x\n", 
		    "Base address register 3", (unsigned int) bar3);

    len += snprintf(page+len, count-len, "%26s: 0x%08x\n", 
		    "CardBus CIS Pointer", (unsigned int) CIS);

    len += snprintf(page+len, count-len, "%26s: 0x%04x\n", 
		    "Subsystem vendor ID", (unsigned int) ss_vendor);

    len += snprintf(page+len, count-len, "%26s: 0x%04x\n", 
		    "Subsystem Device ID", (unsigned int) ss_dev);

    len += snprintf(page+len, count-len, "%26s: 0x%08x\n", 
		    "Expansion ROM Base Address", (unsigned int) rom_addr);
    len += snprintf(page+len, count-len, "%26s: 0x%02x\n", 
		    "IRQ Line", (unsigned int) irq_line);
    len += snprintf(page+len, count-len, "%26s: 0x%02x\n", 
		    "IRQ Pin", (unsigned int) irq_pin);
    len += snprintf(page+len, count-len, "%26s: 0x%02x\n", 
		    "Min_Gnt", (unsigned int) min_gnt);
    len += snprintf(page+len, count-len, "%26s: 0x%02x\n", 
		    "Max_Lat", (unsigned int) max_lat);

  }
  *eof = 1;
  return len;
}


static void dh_calibrate_benchstat(void) {
  vprintk(KERN_INFO "dh: calibrating benchmark statistics, see /proc/driver/domhub/bench...\n");

  dh_bench_start(calbench0);
  dh_bench_end(&calbench0);

  int i;
  for(i=0;i<100000;i++) {
    dh_bench_start(calbench1);
    dh_bench_end(&calbench1);
  }

  dh_bench_start(jifbench);
  unsigned long jif0 = jiffies;
  while((dtbench = jiffies-jif0) < JIFFIES_TO_CALIB) {
    schedule();
  }
  dh_bench_end(&jifbench);
  
  /* This one takes too long, but could use it as a proof of concept once in a while
     for(i=0;i<100000000;i++) {
     dh_bench_start(calbench2);
     dh_bench_end(&calbench2);
     }
  */
}


static void dh_reset_benchstat(dh_bench_rec_t * cb) {
  cb->t1 = cb->t2 = cb->sumdt = cb->ndt = cb->mindt = cb->maxdt = cb->lastdt = 0;
}


static int dh_pr_benchstat(char *buf, int count, char * name, dh_bench_rec_t * cb) {
  return snprintf(buf, count, 
		  "%s: t1=%lld t2=%lld sumdt=%lld ndt=%lld mindt=%lld maxdt=%lld lastdt=%lld\n",
		  name, cb->t1, cb->t2, cb->sumdt, cb->ndt, cb->mindt,
		  cb->maxdt, cb->lastdt);
}


static void dh_init_benchstat(void) {
  dh_reset_benchstat(&calbench0);
  dh_reset_benchstat(&calbench1);
  dh_reset_benchstat(&jifbench);
  dh_reset_benchstat(&gread_bench);
  dh_reset_benchstat(&gwrite_bench);
  dh_reset_benchstat(&rxbench);
  dh_reset_benchstat(&rxnqbench);
  dh_reset_benchstat(&inslbench);
  dh_reset_benchstat(&txbench);
  dh_reset_benchstat(&txdqbench);
  dh_reset_benchstat(&outslbench);
  dh_reset_benchstat(&rxwhile);
  dh_reset_benchstat(&rxuncook);
  dh_reset_benchstat(&rxdmabench);
  dh_reset_benchstat(&txdmabench);
  int i;

  for(i=0; i<NR_CPUS; i++) 
    dh_reset_benchstat(&cpuint_bench[i]);

  for(i=0; i<DH_DOMHUB_DOMS; i++) {
    dh_reset_benchstat(&dmarx_bench[i]);
    dh_reset_benchstat(&dmatx_bench[i]);
    dh_reset_benchstat(&intrx_bench[i]);
    dh_reset_benchstat(&inttx_bench[i]);
    dh_reset_benchstat(&read_bench[i]);
    dh_reset_benchstat(&write_bench[i]);
  }
  dh_calibrate_benchstat(); /* Changes calbench0 & calbench1 as side effects 
			       (not passed as pointers for speed) */
}


static int dh_bench_write_proc(struct file* file, const char *buffer,
			       unsigned long count, void *data) {
  /** Reset benchmarking/time profiling statistics */

  vprintk(KERN_INFO "dh_bench_write_proc: Resetting counters.\n");
  dh_init_benchstat();
  return count;
} 

static int report_dom_proc_value(char *page, int count, int *eof, int val) {
  /* Generic proc file implementation for gettable values from driver */
  *eof = 1;
  return snprintf(page, count, "%d\n", val);
}

static int dh_rxdt_read_proc(char *page, char **start, off_t off,
			 int count, int *eof, void *data) {
  /** Show max time for the F/W to fill a message in the RX FIFO */
  struct fpde_data *d = (struct fpde_data *) data;
  int minor = dh_get_minor_from_card_pair_and_dom(d->cardnum,d->pairnum,d->domnum);
  return report_dom_proc_value(page, count, eof, dh_domvars[minor].rxdt);
}

static int dh_bench_read_proc(char *page, char **start, off_t off, 
			      int count, int *eof, void *data) {
  /** Show benchmarking/time profiling statistics */

  int len = 0;
  len += snprintf(page+len, count-len, "dtbench=%hd jiffies, HZ=%d\n", dtbench, HZ);
  len += dh_pr_benchstat(page+len, count-len, "Calib 0", &calbench0);
  len += dh_pr_benchstat(page+len, count-len, "Calib 1", &calbench1);
  len += dh_pr_benchstat(page+len, count-len, "dtbench jiffies", &jifbench);
  len += dh_pr_benchstat(page+len, count-len, "CPU 0 interrupts", &cpuint_bench[0]);
  if(NR_CPUS > 0) { /* Kludge, but assume 0 or 1 CPUs */
    len += dh_pr_benchstat(page+len, count-len, "CPU 1 interrupts", &cpuint_bench[1]);
  }
  len += dh_pr_benchstat(page+len, count-len, "Read", &gread_bench);
  len += dh_pr_benchstat(page+len, count-len, "Write", &gwrite_bench);
  len += dh_pr_benchstat(page+len, count-len, "RX(do_recv_message)", &rxbench);
  len += dh_pr_benchstat(page+len, count-len, "RX(while loop)", &rxwhile);
  len += dh_pr_benchstat(page+len, count-len, "RX(uncook)", &rxuncook);
  len += dh_pr_benchstat(page+len, count-len, "RX(enqueue)", &rxnqbench);
  len += dh_pr_benchstat(page+len, count-len, "RX(DMA)", &rxdmabench);
  len += dh_pr_benchstat(page+len, count-len, "RX(insl)", &inslbench);
  len += dh_pr_benchstat(page+len, count-len, "TX(do_send_message)", &txbench);
  len += dh_pr_benchstat(page+len, count-len, "TX(dequeue)", &txdqbench);
  len += dh_pr_benchstat(page+len, count-len, "TX(DMA)", &txdmabench);
  len += dh_pr_benchstat(page+len, count-len, "RX(outsl)", &outslbench);
  *eof = 1;
  return len;
}



static int dh_dombench_read_proc(char *page, char **start, off_t off, int count, 
				  int *eof, void *data) {
  /** Read time profiling values for this DOM */
  struct fpde_data *d = (struct fpde_data *) data;

  int len;
  len = 0;
  *eof = 1;

  int minor = dh_get_minor_from_card_pair_and_dom(d->cardnum,d->pairnum,d->domnum);

  /* Report DOM ID */
  len += dh_pr_benchstat(page+len, count-len, "DMA RX",  &dmarx_bench[minor]);
  len += dh_pr_benchstat(page+len, count-len, "DMA TX",  &dmatx_bench[minor]);
  len += dh_pr_benchstat(page+len, count-len, "Int. RX", &intrx_bench[minor]);
  len += dh_pr_benchstat(page+len, count-len, "Int. TX", &inttx_bench[minor]);
  len += dh_pr_benchstat(page+len, count-len, "Read",    &read_bench[minor]);
  len += dh_pr_benchstat(page+len, count-len, "Write",   &write_bench[minor]);

  return len;
}


static int dh_card_pwr_check_read_proc(char *page, char **start, off_t off, int count,
				       int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  if(dh_pci_dors[icard].rev < 1) {
    len += snprintf(page+len, count-len, "Card %d is DOR Rev. 0, power check not supported.\n",
		    icard);
  } else {
    u32 status = dh_read_regi(icard, DH_RG_GSTAT) & DH_GSTAT_WP_POW_FAIL;
    len += snprintf(page+len, count-len, "Card %d power check: %s\n",
		    icard, status ? "FAIL" : "ok");
  }
  return len;
}

static int dh_card_autodac_read_proc(char *page, char **start, off_t off, int count,
				      int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  len += snprintf(page+len, count-len, "Card %d autodac: %d\n", icard,
		  dh_pci_dors[icard].autodac);
  return len;
}

static int dh_card_dacmax_read_proc(char *page, char **start, off_t off, int count,
				     int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  len += snprintf(page+len, count-len, "Card %d dacmax: %d\n", icard,
                  dh_pci_dors[icard].dacmax);
  return len;
}

static int dh_card_minclev_read_proc(char *page, char **start, off_t off, int count,
				    int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  len += snprintf(page+len, count-len, "Card %d minclev: %d\n", icard,
                  dh_pci_dors[icard].minclev);
  return len;
}

static int dh_card_maxclev_read_proc(char *page, char **start, off_t off, int count,
				    int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  len += snprintf(page+len, count-len, "Card %d maxclev: %d\n", icard,
                  dh_pci_dors[icard].maxclev);
  return len;
}

static int dh_card_rdelay_read_proc(char *page, char **start, off_t off, int count,
				    int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  len += snprintf(page+len, count-len, "Card %d rdelay: %d\n", icard,
                  dh_pci_dors[icard].rdelay);
  return len;
}

static int dh_card_sdelay_read_proc(char *page, char **start, off_t off, int count,
				    int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  len += snprintf(page+len, count-len, "Card %d sdelay: %d\n", icard,
                  dh_pci_dors[icard].sdelay);
  return len;
}

static int dh_card_thresh_read_proc(char *page, char **start, off_t off, int count,
				    int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  len += snprintf(page+len, count-len, "Card %d thresh: %d\n", icard,
                  dh_pci_dors[icard].thresh);
  return len;
}

static int dh_card_minvolts_read_proc(char *page, char **start, off_t off, int count,
				      int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  u32 curl = dh_read_regi(icard, DH_RG_CURL);
  len += snprintf(page+len, count-len, "Card %d minimum voltage limit: %d V\n", icard,
		  dh_get_min_voltage(icard));
  return len;
}

static int dh_card_maxvolts_read_proc(char *page, char **start, off_t off, int count,
                                      int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  u32 curl = dh_read_regi(icard, DH_RG_CURL);
  len += snprintf(page+len, count-len, "Card %d maximum voltage limit: %d V\n", icard,
                  dh_get_max_voltage(icard));
  return len;
}

static int dh_card_mincur_read_proc(char *page, char **start, off_t off, int count,
				    int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  len += snprintf(page+len, count-len, "Card %d minimum current limit: %d mA\n", icard,
                  dh_get_min_current(icard));
  return len;
}

static int dh_card_maxcur_read_proc(char *page, char **start, off_t off, int count,
                                    int *eof, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  int len = 0;
  *eof = 1;
  len += snprintf(page+len, count-len, "Card %d maximum current limit: %d mA\n", icard,
                  dh_get_max_current(icard));
  return len;
}

static int dh_card_autodac_write_proc(struct file* file, const char *buffer,
				      unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int icard = d->cardnum;
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_card_autodac_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  int autodac_val;
  tmpbuf[compare_len] = '\0';
  if(!sscanf(tmpbuf, "%d", &autodac_val)) {
    printk(KERN_ERR "dh_card_autodac_write_proc: Need to write an integer to this proc file.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  dh_pci_dors[icard].autodac = autodac_val;
  vprintk(KERN_INFO "dh_card_autodac_write_proc: set autodac comms param to %d.\n", autodac_val);
  dh_set_comms_params(icard);
  return count;
}

static int dh_card_dacmax_write_proc(struct file* file, const char *buffer,
                                      unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int icard = d->cardnum;
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_card_dacmax_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  int dacmax_val;
  tmpbuf[compare_len] = '\0';
  if(!sscanf(tmpbuf, "%d", &dacmax_val)) {
    printk(KERN_ERR "dh_card_dacmax_write_proc: Need to write an integer to this proc file.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  dh_pci_dors[icard].dacmax = dacmax_val;
  vprintk(KERN_INFO "dh_card_dacmax_write_proc: set dacmax comms param to %d.\n", dacmax_val);
  dh_set_comms_params(icard);
  return count;
}

static int dh_card_minclev_write_proc(struct file* file, const char *buffer,
                                      unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int icard = d->cardnum;
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_card_minclev_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  int minclev_val;
  tmpbuf[compare_len] = '\0';
  if(!sscanf(tmpbuf, "%d", &minclev_val)) {
    printk(KERN_ERR "dh_card_minclev_write_proc: Need to write an integer to this proc file.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  dh_pci_dors[icard].minclev = minclev_val;
  vprintk(KERN_INFO "dh_card_minclev_write_proc: set minclev comms param to %d.\n", minclev_val);
  dh_set_comms_params(icard);
  return count;
}

static int dh_card_maxclev_write_proc(struct file* file, const char *buffer,
                                      unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int icard = d->cardnum;
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_card_maxclev_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  int maxclev_val;
  tmpbuf[compare_len] = '\0';
  if(!sscanf(tmpbuf, "%d", &maxclev_val)) {
    printk(KERN_ERR "dh_card_maxclev_write_proc: Need to write an integer to this proc file.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  dh_pci_dors[icard].maxclev = maxclev_val;
  vprintk(KERN_INFO "dh_card_maxclev_write_proc: set maxclev comms param to %d.\n", maxclev_val);
  dh_set_comms_params(icard);
  return count;
}

static int dh_card_rdelay_write_proc(struct file* file, const char *buffer,
                                      unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int icard = d->cardnum;
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_card_rdelay_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  int rdelay_val;
  tmpbuf[compare_len] = '\0';
  if(!sscanf(tmpbuf, "%d", &rdelay_val)) {
    printk(KERN_ERR "dh_card_rdelay_write_proc: Need to write an integer to this proc file.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  dh_pci_dors[icard].rdelay = rdelay_val;
  vprintk(KERN_INFO "dh_card_rdelay_write_proc: set rdelay comms param to %d.\n", rdelay_val);
  dh_set_comms_params(icard);
  return count;
}

static int dh_card_sdelay_write_proc(struct file* file, const char *buffer,
                                      unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int icard = d->cardnum;
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_card_sdelay_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  int sdelay_val;
  tmpbuf[compare_len] = '\0';
  if(!sscanf(tmpbuf, "%d", &sdelay_val)) {
    printk(KERN_ERR "dh_card_sdelay_write_proc: Need to write an integer to this proc file.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  dh_pci_dors[icard].sdelay = sdelay_val;
  vprintk(KERN_INFO "dh_card_sdelay_write_proc: set sdelay comms param to %d.\n", sdelay_val);
  dh_set_comms_params(icard);
  return count;
}

static int dh_card_thresh_write_proc(struct file* file, const char *buffer,
				     unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int icard = d->cardnum;
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_card_thresh_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  tmpbuf[compare_len] = '\0';
  int thresh_val;
  if(!sscanf(tmpbuf, "%d", &thresh_val)) {
    printk(KERN_ERR "dh_card_thresh_write_proc: Need to write an integer to this proc file.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  dh_pci_dors[icard].thresh = thresh_val;
  vprintk(KERN_INFO "dh_card_thresh_write_proc: set thresh comms param to %d.\n", thresh_val);
  dh_set_comms_params(icard);
  return count;
}


static int dh_card_minvolts_write_proc(struct file* file, const char *buffer,
				       unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int icard = d->cardnum;
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_card_minvolts_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  int minvolts_val;
  tmpbuf[compare_len] = '\0';
  if(!sscanf(tmpbuf, "%d", &minvolts_val)) {
    printk(KERN_ERR "dh_card_minvolts_write_proc: Need to write an integer to this proc file.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  if(minvolts_val < 0 || minvolts_val > 113) {
    printk(KERN_ERR "dh_card_minvolts_write_proc: Value out of bounds, write 0..113.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  dh_pci_dors[icard].minvolts = minvolts_val;
  dh_write_fpga_voltage_current_limits(icard);
  return count;
}


static int dh_card_maxvolts_write_proc(struct file* file, const char *buffer,
                                       unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int icard = d->cardnum;
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_card_maxvolts_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  int maxvolts_val;
  tmpbuf[compare_len] = '\0';
  if(!sscanf(tmpbuf, "%d", &maxvolts_val)) {
    printk(KERN_ERR "dh_card_maxvolts_write_proc: Need to write an integer to this proc file.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  if(maxvolts_val < 0 || maxvolts_val > 113) {
    printk(KERN_ERR "dh_card_maxvolts_write_proc: Value out of bounds, write 0..113.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  dh_pci_dors[icard].maxvolts = maxvolts_val;
  dh_write_fpga_voltage_current_limits(icard);
  return count;
}

static int dh_card_mincur_write_proc(struct file* file, const char *buffer,
				       unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int icard = d->cardnum;
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_card_mincur_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  int mincur_val;
  tmpbuf[compare_len] = '\0';
  if(!sscanf(tmpbuf, "%d", &mincur_val)) {
    printk(KERN_ERR "dh_card_mincur_write_proc: Need to write an integer to this proc file.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  if(mincur_val < 0 || mincur_val > 255) {
    printk(KERN_ERR "dh_card_mincur_write_proc: Value out of bounds, write 0..255.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  dh_pci_dors[icard].mincur = mincur_val;
  dh_write_fpga_voltage_current_limits(icard);
  return count;
}


static int dh_card_maxcur_write_proc(struct file* file, const char *buffer,
                                       unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  char tmpbuf[REQUEST_STR_LEN];
  int icard = d->cardnum;
  int compare_len = count > (REQUEST_STR_LEN-1) ? (REQUEST_STR_LEN-1) : count;
  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_card_maxcur_write_proc: Invalid pointer in user space.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  int maxcur_val;
  tmpbuf[compare_len] = '\0';
  if(!sscanf(tmpbuf, "%d", &maxcur_val)) {
    printk(KERN_ERR "dh_card_maxcur_write_proc: Need to write an integer to this proc file.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  if(maxcur_val < 0 || maxcur_val > 255) {
    printk(KERN_ERR "dh_card_maxcur_write_proc: Value out of bounds, write 0..255.\n");
    dh_lasterr = DH_ERR_INVALIDARG;
    return -EFAULT;
  }
  dh_pci_dors[icard].maxcur = maxcur_val;
  dh_write_fpga_voltage_current_limits(icard);
  return count;
}


static int dh_fpga_read_proc(char *page, char **start, off_t off, int count, 
			     int *eof, void *data) {
/** Read FPGA status 
    @return number of characters "read," or put into the page/buffer */
  int len = 0;
  int ic;
  struct fpde_data *d;

  d = data;
  ic = d->cardnum;

  len += snprintf(page+len, count-len, "FPGA registers:\n");

  len += snprintf(page+len, count-len, "CTRL  0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_CTRL));
  len += snprintf(page+len, count-len, "GSTAT 0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_GSTAT));
  len += snprintf(page+len, count-len, "DSTAT 0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_DSTAT));
  len += snprintf(page+len, count-len, "TTSIC 0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_TTSIC));
  len += snprintf(page+len, count-len, "RTSIC 0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_RTSIC));
  len += snprintf(page+len, count-len, "INTEN 0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_INTEN));
  len += snprintf(page+len, count-len, "DOMS  0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_DOMS));
  len += snprintf(page+len, count-len, "MRAR  0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_MRAR));
  len += snprintf(page+len, count-len, "MRTC  0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_MRTC));
  len += snprintf(page+len, count-len, "MWAR  0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_MWAR));
  len += snprintf(page+len, count-len, "MWTC  0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_MWTC));
  len += snprintf(page+len, count-len, "CURL  0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_CURL));
  len += snprintf(page+len, count-len, "DCUR  0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_DCUR));
  len += snprintf(page+len, count-len, "FLASH 0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_FLASH));
  len += snprintf(page+len, count-len, "DOMC  0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_DOMC));
  len += snprintf(page+len, count-len, "CERR  0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_CERR));
  len += snprintf(page+len, count-len, "DCREV 0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_DCREV));
  len += snprintf(page+len, count-len, "FREV  0x%08lx\n", (unsigned long) dh_read_regi(ic, DH_RG_FREV));
  *eof = 1;

  return len;
}



static int dh_reload_fpga_from_flash(int icard, int ipage) {
  /** Cause the actual FPGA reload sequence to occur, reloading from flash.
      Assume already checked arguments icard and ipage.
      See Kalle's API Document for a description of the steps here.
      @return 1 if failure occurs, else 0 */
  u32 flash_select_flags;
  u16 saved_id, new_id;
  int icount;
  int ifound_good_device;

  pprintk(KERN_INFO "dh_reload_fpga_from_flash: card %d page %d.\n",icard,ipage);

  if(! dh_pci_dors[icard].found) {
    printk(KERN_ERR "dh_reload_fpga_from_flash - no PCI information available for card %d.\n",icard);
    return 1;
  }

  /* Save the 256 bytes of PCI configuration space for the device */
  if(dh_save_pci_config_space(&dh_pci_dors[icard])) {
    printk(KERN_ERR "dh_reload_fpga_from_flash - couldn't save PCI configuration on card %d.\n",icard);
    return 1;
  }

  /* Also save device_id for comparison */
  if(pci_read_config_word(dh_pci_dors[icard].dev, 2, &saved_id)) {
    printk(KERN_ERR "dh_reload_fpga_from_flash - couldn't get PCI device word on card %d.\n",icard);
    return 1;
  }

  pprintk(KERN_INFO "dh_reload_fpga_from_flash - saved PCI dev ID is 0x%x.\n", (int) saved_id);

  if(ipage < 0 || ipage >= DH_FLASH_PAGES) {
    printk(KERN_ERR "dh_reload_fpga_from_flash - invalid page selected (%d).\n", ipage);
    return 1;
  }

  flash_select_flags = (ipage & 0x03) << 24; /* Bits 24 and 25 */

  pprintk(KERN_INFO "dh_reload_fpga_from_flash: flash_select_flags 0x%x.\n",
	 flash_select_flags);
  /* Set CTRL bit 26 high to cause an FPGA reload; also select which flash page to load from */
  dh_set_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_FPGA_RELOAD | flash_select_flags);
  /* dh_clear_write_regi_bits(icard, DH_RG_CTRL, DH_CTRL_FPGA_RELOAD | flash_select_flags); */

  /* Now look for device_id (bytes 2-3) in PCI configuration space and see if it matches
     what we saved - if it does, we're good to go. */
  ifound_good_device = 0;
  dh_jsleep(HZ/2);

  for(icount = 0; icount < 1000; icount++) {
    /* Sleep a tiny bit to allow FPGA to change state */
    dh_jsleep(5);
    
    /* Poll on Device_ID in configuration space */
    pci_read_config_word(dh_pci_dors[icard].dev, 2, &new_id);
    pprintk(KERN_INFO "dh_reload_fpga_from_flash - new PCI dev ID is 0x%x.\n", (int) new_id);
    if(new_id == saved_id) {
      ifound_good_device = 1;
      break;
    } 
  }

  if(!ifound_good_device) {
    printk(KERN_ERR "dh_reload_fpga_from_flash - FPGA on card %d didn't fill PCI device"
	   " register correctly (got device 0x%x).  Your FPGA design may be hosed.  Sorry.\n", 
	   icard, (int) new_id);
    return 1;
  }

  /* Restore the last-used configuration space */
  pprintk(KERN_INFO "dh_reload_fpga_from_flash - about to restore PCI configuration on card %d.\n",icard);
  if(dh_restore_pci_config_space(&dh_pci_dors[icard])) {
    printk(KERN_ERR "dh_reload_fpga_from_flash - couldn't restore PCI configuration on card %d.\n",icard);
    return 1;
  }

  if(dh_fpga_regi_init(icard, WAIT_CFG_DONE)) {
    printk(KERN_ERR "dh_reload_fpga_from_flash: card %d register init failed\n", icard);
    return 1;
  }

  vprintk(KERN_INFO "dh_reload_fpga_from_flash: Loaded FPGA from page %d on card %d.\n",
	 ipage, icard);

  return 0;
}


static int dh_card_pwr_check_write_proc(struct file* file, const char *buffer,
					unsigned long count, void *data) {
  struct fpde_data *d = (struct fpde_data *) data;
  int icard = d->cardnum;
  vprintk(KERN_INFO "dh_card_pwr_check_write_proc: Resetting power check for card %d.\n",
	  icard);
  unsigned long flags;
  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  dh_set_write_regi_bits_nosave(icard, DH_RG_GSTAT, DH_GSTAT_WP_POW_FAIL);
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
  return count;
}


static int dh_fpga_write_proc(struct file* file, const char *buffer, 
			      unsigned long count, void *data) {
  /** get FPGA load command from stdin.  @return count or a negative error value */
  struct fpde_data *d = (struct fpde_data *) data;
  int pagenum;
  char *tmpbuf;
  int compare_len;
  int ich, minor;

  pprintk(KERN_INFO "dh_fpga_write_proc, card %d: to write %ld bytes.\n",\
	 d->cardnum, count);

  if(d->cardnum < 0 || d->cardnum >= DH_MAX_CARDS) {
    printk(KERN_ERR "dh_fpga_write_proc: card number %d is out of bounds.\n",d->cardnum);
    return -EFAULT;
  }

  tmpbuf = vmalloc(REQUEST_STR_LEN);
  if(tmpbuf == NULL) {
    printk(KERN_ERR "dh_fpga_write_proc: Couldn't get scratch buffer space.\n");
    return -ENOMEM;
  }

  /* Make sure we don't blow buffer */
  compare_len = count > REQUEST_STR_LEN ? REQUEST_STR_LEN : count;

  if(copy_from_user(tmpbuf, buffer, compare_len)) {
    printk(KERN_ERR "dh_fpga_write_proc: Invalid pointer in user space.\n");
    vfree(tmpbuf);
    dh_lasterr = DH_ERR_BADUSERBUFPTR;
    return -EFAULT;
  }

  /* Look for correct command */
  if(compare_len < 8) {
    printk(KERN_ERR "dh_fpga_write_proc: Invalid command written to proc file.\n");
    vfree(tmpbuf);
    dh_lasterr = DH_ERR_BADPROCCMD;
    return -EFAULT;
  }

  if(strncmp(tmpbuf, "reload ", 7)) {
    printk(KERN_ERR "dh_fpga_write_proc: Invalid command written to proc file.\n");
    vfree(tmpbuf);
    dh_lasterr = DH_ERR_BADPROCCMD;
    return -EFAULT;
  }

  /* At this point, tmpbuf[7] MUST be equal to the flash page number */
  pagenum = tmpbuf[7] - '0';
  if(pagenum < 0 || pagenum > 3) {
    printk(KERN_ERR "dh_fpga_write_proc: Invalid page number given.\n");
    vfree(tmpbuf);
    dh_lasterr = DH_ERR_BADPROCARG;
    return -EFAULT;
  }

  /* Make sure communications aren't already in progress somewhere on this card */  

  if(down_interruptible(&dh_pci_dors[d->cardnum].open_close_sem)) { /* DDDDDDDDDDDDDDDDD */
    printk(KERN_ERR "dh_fpga_write_proc:  ABORTED semaphore acquire on card %d, "
	   "won't reload FPGA.\n", d->cardnum);
    dh_lasterr = DH_ERR_SEMAINT;
    return -EFAULT;
  }

  for(ich = 0; ich < DH_DOMS_PER_CARD; ich++) {
    minor = dh_get_minor_from_card_and_channel(d->cardnum, ich);
    if(dh_opened[minor]) {
      printk(KERN_WARNING "dh_fpga_write_proc: reload attempted while device is open."
	     " Please close all device files on this card first.\n");
      vfree(tmpbuf);
      up(&dh_pci_dors[d->cardnum].open_close_sem); /*PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP*/
      dh_lasterr = DH_ERR_FPGA_COMMPROG;
      return -EFAULT;
    }
  }

  if(dh_reload_fpga_from_flash(d->cardnum, pagenum)) {
    printk(KERN_ERR "dh_fpga_write_proc: reload of card %d's FPGA "
	   "from flash page %d failed.\n", d->cardnum,pagenum);
    vfree(tmpbuf);
    dh_lasterr = DH_ERR_FPGARELODFAIL;
    up(&dh_pci_dors[d->cardnum].open_close_sem); /*PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP*/
    return -EFAULT;
  }

  up(&dh_pci_dors[d->cardnum].open_close_sem); /*PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP*/
  
  vprintk(KERN_INFO "dh: reload of card %d's FPGA "
	 "from flash page %d succeeded.\n", d->cardnum,pagenum);

  /* Check firmware version -- disallow open or power up if bad version */
  dh_firmware_mismatch[d->cardnum] = dh_firmware_version_check(d->cardnum);

  vfree(tmpbuf);
  return count;
}



static int dh_fpga_regs_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data) {
/** Read FPGA register, primed first by the write proc function fpga_regs_write_proc 
    @return number of characters "read," or put into the page/buffer */
  static int len;
  struct proc_fpga_data *pfd = (struct proc_fpga_data *) data;
  int icard;
  u32 reg;

  len = 0;
  icard = pfd->cardnum;

  if(off != 0) {
    *eof = 1;
    return 0; /* Exit right away if function has already been called */
  }

  pprintk(KERN_INFO "begin: len %d, off %d.\n", len, (int) off);

  if(pfd->read_request_given) {
    pfd->read_request_given = 0;
    if(pfd->register_to_read >= 0
       && 
       pfd->register_to_read < DH_NUM_REGIS) {
      switch (dh_mode) {
      case DH_NORMAL:
        if(icard >= 0 && icard < DH_MAX_CARDS && dh_pci_dors[icard].found) {
	  reg = dh_read_regi(icard, pfd->register_to_read);
	  len += snprintf(page+len, count-len,
			  "FPGA on card %d: register %d = %ld (0x%lx)\n",
			  pfd->cardnum,
			  pfd->register_to_read,
			  (unsigned long) reg, (unsigned long) reg);
	} else {
          printk(KERN_ERR "dh: ERROR: Invalid card number %d in dh_fpga_regs_write_proc.\n",icard);
	  len += snprintf(page+len, count-len, "ERROR: Invalid card number %d.\n", icard);
	  dh_lasterr = DH_ERR_BADPROCARG;
	}
	break;
      case DH_SIM_COPY:
      case DH_SIM_NOCOPY:
      case DH_SIM_QUEUE:
	pprintk(KERN_INFO "1st sprintf: len %d.\n", len);
	len += snprintf(page+len, count-len,
			"FPGA on card %d: register %d = %d (0x%x) (simulation)\n",
			pfd->cardnum,
			pfd->register_to_read,
			(int) pfd->sim_regs[pfd->register_to_read],
			(int) pfd->sim_regs[pfd->register_to_read]);
	break;
      default:
        printk(KERN_ERR "dh: ERROR: Unknown mode %d in dh_fpga_regs_read_proc.\n",dh_mode);
	dh_lasterr = DH_ERR_BADPROCARG;
	break;
      }

    } else {
      len += snprintf(page+len, count-len,
                      "register number %d is out of range (must be in 0 .. %d)\n",
		      pfd->register_to_read,
		      DH_NUM_REGIS);
      dh_lasterr = DH_ERR_BADPROCARG;
    }
  } else {
    len += snprintf(page+len, count-len, 
		    "no data for card %d - prepare it by sending 'r N' "
		    "where N is register number (simulation).\n",
		    pfd->cardnum);
    dh_lasterr = DH_ERR_BADPROCCMD;
  }
  pprintk(KERN_INFO "end: len %d.\n", len);
  *eof = 1;
  return len;
}



static int dh_fpga_regs_write_proc(struct file* file, const char *buffer, unsigned long count, void *data) {
/** write an FPGA register or prime it for reading using dh_fpga_regs_read_proc 
    \return number of characters actually read from buffer */
  struct proc_fpga_data *pfd = (struct proc_fpga_data *) data;
  int len;
  int regi;
  int icard;
  u32 value;
#define U32_MAX_CHARS 9 /* Includes \0 */
  char valuestr[U32_MAX_CHARS];

  icard = pfd->cardnum;

  if(count > REQUEST_STR_LEN-1) {
    len = REQUEST_STR_LEN-1;
  } else {
    len = count;
  }

  if(copy_from_user(pfd->request_str,buffer,len)) {
    return -EFAULT;
  }

  pfd->request_str[len] = '\0';

  if(sscanf(pfd->request_str, "r %d", &regi)) {
    pfd->read_request_given = 1;
    pfd->register_to_read = regi;
    return len;
  } else if(sscanf(pfd->request_str, "w %d %8s", &regi, valuestr)) {
    pprintk(KERN_INFO "dh: Got request to write '%s' into register %d on card %d (len is %d).\n",\
	    valuestr, regi, icard, len);
    pprintk(KERN_INFO "Original request string was '%s'.\n", pfd->request_str);
    value = dh_hex_str_to_u32(valuestr, U32_MAX_CHARS);
    pprintk(KERN_INFO "Value is %lx.\n",(unsigned long) value);
    if(regi >= 0 && regi < DH_NUM_REGIS) {
      switch (dh_mode) {
      case DH_NORMAL:
	if(icard >= 0 && icard < DH_MAX_CARDS && dh_pci_dors[icard].found) {
	  dh_write_regi(icard, regi, value);
	  pprintk(KERN_INFO "dh: Read back card %d regi %d value 0x%lx.\n",icard,regi,\
		 (unsigned long) dh_read_regi(icard, regi));
	} else {
	  printk(KERN_ERR "dh: ERROR: Invalid card number %d in dh_fpga_regs_write_proc.\n",icard);
	  dh_lasterr = DH_ERR_BADPROCARG;
	}
	break;
      case DH_SIM_COPY:
      case DH_SIM_NOCOPY:
      case DH_SIM_QUEUE:
	pfd->sim_regs[regi] = (unsigned char) value;
	break;
      default:
	printk(KERN_ERR "dh: ERROR: Unknown mode %d in dh_fpga_regs_write_proc.\n",dh_mode);
	dh_lasterr = DH_ERR_BADPROCARG;
	break;
      }
     
    } else {
      printk(KERN_WARNING "dh: Bad request string '%s'.\n",pfd->request_str);
      dh_lasterr = DH_ERR_BADPROCCMD;
    }
    return len;
  }
  return len;
  /* Incorrect format string leaves no side effects */
}


static int dh_make_fpde_as_file(struct fancy_pde_entry *fpde,
				char                   *name_string,
				struct proc_dir_entry  *parent,
				int                    mode,
				read_proc_t            *read_proc,
				write_proc_t           *write_proc,
				int                    data_size) {
  /** Creates/initializes a fancy proc dir entry as a file, installing
      the appropriate callbacks.  Allocate data_size bytes for associated 
      data, if appropriate (otherwise caller passes data_size as 0 and no
      allocation is performed).
      \return 0 if success, else the error 
  */
  void *f;

  snprintf(fpde->name,PDE_NAME_LEN,name_string);
  fpde->pde = create_proc_entry(fpde->name,mode,parent);
  if(fpde->pde == NULL) {
    return -ENOMEM;
  }
  fpde->pde->owner      = THIS_MODULE;
  fpde->regi            = 1;
  fpde->pde->read_proc  = read_proc;
  fpde->pde->write_proc = write_proc;
  
  if(data_size > 0) { /* Allocate data for the /proc file */
    f = kmalloc(data_size, GFP_KERNEL);
    if(f == NULL) {
      fpde->pde->data = NULL;
      return -ENOMEM;
    } else {
      fpde->pde->data = f;
    }
  }
  return 0;
}


static int dh_make_fpde_as_dir(struct fancy_pde_entry *fpde, 
			       char *name_string,
			       struct proc_dir_entry *parent) {
  /** Creates/initializes a fancy proc dir entry as a directory;
      \return 0 if success, else the error.
  */
  snprintf(fpde->name, PDE_NAME_LEN, name_string);
  fpde->pde = proc_mkdir(fpde->name, parent);
  if(fpde->pde == NULL) {
    return -ENOMEM;
  }
  fpde->pde->owner = THIS_MODULE;
  fpde->regi = 1; /* Mark this entry as registered, for dh_cleanup_proc_entries() */
  return 0;
}


void dh_fill_card_data(void * pde_data, int icard) {
  struct fpde_data * f = (struct fpde_data *) pde_data;
  f->cardnum = icard;
}


void dh_fill_card_pair_data(void * pde_data, int icard, int ipair) {
  struct fpde_data * f = (struct fpde_data *) pde_data;
  f->cardnum = icard;
  f->pairnum = ipair;
}


void dh_fill_card_pair_dom_data(void * pde_data, int icard, int ipair, int idom) {
  struct fpde_data * f = (struct fpde_data *) pde_data;
  f->cardnum = icard;
  f->pairnum = ipair;
  f->domnum = idom;
}


static int dh_initialize_proc_files(void) {
  /** Initialize the entries in the /proc filesystem which we use to
      control and access various features of the DOMs and the DOM Hub
      cards.

      \return 0 if success, else the relevant error (generally, -ENOMEM).
  */

  int icard;
  int ipair;
  int idom;
  struct fpde_data *f; 
  struct flash_data *fl;
  struct proc_fpga_data *pfd;
  int ireg;
  int result;
  int iflash;

  char tmpname[PDE_NAME_LEN];

  /** Initialize proc files.  Return zero if ok, else an error. */

  /* Create top level directory for domhub proc files */

  /* First, mark all as unregistered to enable cleanup at any time */
  dh_mark_all_proc_entries_as_unregistered();

  /* Do top level directory (/proc/driver/domhub) */
  if((result = dh_make_fpde_as_dir(&fpde_top_dir,DH_TOP_DIR,NULL)) != 0) {
    dh_cleanup_proc_entries();
    return result;
  }

  /* Create system-wide proc files */

  /* Buffer size read proc */
  if((result = dh_make_fpde_as_file(&fpde_bufsiz, DH_BUFSIZ,
				    fpde_top_dir.pde, 0444,
                                    dh_bufsiz_read_proc, NULL, 0)) != 0) {
    dh_cleanup_proc_entries();
    return result;
  }

  /* Verbose proc file */
  if((result = dh_make_fpde_as_file(&fpde_verbose, DH_VERBOSE,
                                    fpde_top_dir.pde, 0666,
                                    dh_verbose_read_proc, dh_verbose_write_proc, 0)) != 0) {
    dh_cleanup_proc_entries();
    return result;
  }

  /* Blocking size read proc */
  if((result = dh_make_fpde_as_file(&fpde_blocking, DH_BLOCKING,
				    fpde_top_dir.pde, 0666,
                                    dh_blocking_read_proc, dh_blocking_write_proc, 0)) != 0) {
    dh_cleanup_proc_entries();
    return result;
  }


  /* Time profiling code */
  if((result = dh_make_fpde_as_file(&fpde_bench, DH_BENCH,
				    fpde_top_dir.pde, 0666,
                                    dh_bench_read_proc, dh_bench_write_proc, 0)) != 0) {
    dh_cleanup_proc_entries();
    return result;
  }

  /* ALL Channels power on/off */
  if((result = dh_make_fpde_as_file(&fpde_pwrall, DH_PWRALL,
                                    fpde_top_dir.pde, 0222,
				    NULL, dh_pwrall_write_proc, 0)) != 0) {
    dh_cleanup_proc_entries();
    return result;
  }

  /* Driver revision file */
  if((result = dh_make_fpde_as_file(&fpde_revision, DH_REVISION,
                                    fpde_top_dir.pde, 0444,
                                    dh_revision_read_proc, NULL, 0)) != 0) {
    dh_cleanup_proc_entries();
    return result;
  }

  /* Last error message file */
  if((result = dh_make_fpde_as_file(&fpde_lasterr, DH_LASTERR,
				    fpde_top_dir.pde, 0444,
				    dh_lasterr_read_proc, NULL, 0)) != 0) {
    dh_cleanup_proc_entries();
    return result;
  }

  /* Driver status file */
  if((result = dh_make_fpde_as_file(&fpde_driverstat, DH_DRIVERSTAT,
				    fpde_top_dir.pde, 0444,
				    dh_driverstat_read_proc, NULL, 0)) != 0) {
    dh_cleanup_proc_entries();
    return result;
  }

  /* Pseudo-interrupt proc file for testing */
  if((result = dh_make_fpde_as_file(&fpde_pseudo_intr, DH_PSEUDO_INTR,
                                    fpde_top_dir.pde, 0200,
				    NULL, dh_pseudo_intr_write_proc, 0)) != 0) {
    dh_cleanup_proc_entries();
    return result;
  }

  /* Create card-wise proc files */
  for(icard=0; icard < DH_MAX_CARDS; icard++) {

    if(dh_mode != DH_NORMAL || dh_pci_dors[icard].found) {

      /* Create the directory for this domhub card */

      snprintf(tmpname,PDE_NAME_LEN,"card%d",icard);
      if((result = dh_make_fpde_as_dir(&fpde_card_dirs[icard],
				       tmpname,fpde_top_dir.pde)) != 0) {
	dh_cleanup_proc_entries();
	return result;
      }
      
      /* Create the flash entries for this card */
      for(iflash=0;iflash<DH_FLASH_PAGES;iflash++) {
	snprintf(tmpname,PDE_NAME_LEN,"flash%d",iflash);
	if((result = dh_make_fpde_as_file(&fpde_card_flash[icard][iflash], tmpname,
					  fpde_card_dirs[icard].pde, 0644,
					                             /* only root can write */
					  dh_flash_read_proc, dh_flash_write_proc,
					  sizeof(struct flash_data))) != 0) {
	  dh_cleanup_proc_entries();
	  return result;
	}
	fl = (struct flash_data *) fpde_card_flash[icard][iflash].pde->data;
	fl->cardnum  = icard;
	fl->flashnum = iflash;
	fl->fpos     = 0; /* Writes start at beginning of flash */
      }

      /* Create the flash protect/unprotect file for this card */
      if((result = dh_make_fpde_as_file(&fpde_card_flprotect[icard], DH_CARD_FLPROTECT,
                                        fpde_card_dirs[icard].pde, 0222,
					NULL, dh_flprotect_write_proc,
					sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_flprotect[icard].pde->data, icard);

      /* Create the FPGA load / status file for this card */
      if((result = dh_make_fpde_as_file(&fpde_card_fpga[icard], DH_CARD_FPGA,
					fpde_card_dirs[icard].pde, 0666,
					dh_fpga_read_proc, dh_fpga_write_proc,
					sizeof(struct fpde_data))) != 0) {
	dh_cleanup_proc_entries();
	return result;
      }
      dh_fill_card_data(fpde_card_fpga[icard].pde->data, icard);
      
      /* "Power check" (current or voltage over/under limits) for entire card */
      if((result = dh_make_fpde_as_file(&fpde_card_pwr_check[icard], DH_CARD_PWR_CHECK,
                                        fpde_card_dirs[icard].pde, 0666,
                                        dh_card_pwr_check_read_proc, dh_card_pwr_check_write_proc,
                                        sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_pwr_check[icard].pde->data, icard);

      /* Comms parameters for card: autodac */
      if((result = dh_make_fpde_as_file(&fpde_card_autodac[icard], DH_CARD_AUTODAC,
					fpde_card_dirs[icard].pde, 0666,
                                        dh_card_autodac_read_proc, dh_card_autodac_write_proc,
                                        sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_autodac[icard].pde->data, icard);

      /* Comms parameters for card: dacmax */
      if((result = dh_make_fpde_as_file(&fpde_card_dacmax[icard], DH_CARD_DACMAX,
					fpde_card_dirs[icard].pde, 0666,
                                        dh_card_dacmax_read_proc, dh_card_dacmax_write_proc,
                                        sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_dacmax[icard].pde->data, icard);

      /* Comms parameters for card: minclev */
      if((result = dh_make_fpde_as_file(&fpde_card_minclev[icard], DH_CARD_MINCLEV,
                                        fpde_card_dirs[icard].pde, 0666,
                                        dh_card_minclev_read_proc, dh_card_minclev_write_proc,
                                        sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_minclev[icard].pde->data, icard);

      /* Comms parameters for card: maxclev */
      if((result = dh_make_fpde_as_file(&fpde_card_maxclev[icard], DH_CARD_MAXCLEV,
                                        fpde_card_dirs[icard].pde, 0666,
                                        dh_card_maxclev_read_proc, dh_card_maxclev_write_proc,
                                        sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_maxclev[icard].pde->data, icard);

      /* Comms parameters for card: rdelay */
      if((result = dh_make_fpde_as_file(&fpde_card_rdelay[icard], DH_CARD_RDELAY,
                                        fpde_card_dirs[icard].pde, 0666,
                                        dh_card_rdelay_read_proc, dh_card_rdelay_write_proc,
                                        sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_rdelay[icard].pde->data, icard);

      /* Comms parameters for card: sdelay */
      if((result = dh_make_fpde_as_file(&fpde_card_sdelay[icard], DH_CARD_SDELAY,
                                        fpde_card_dirs[icard].pde, 0666,
                                        dh_card_sdelay_read_proc, dh_card_sdelay_write_proc,
                                        sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_sdelay[icard].pde->data, icard);

      /* Comms parameters for card: thresh */
      if((result = dh_make_fpde_as_file(&fpde_card_thresh[icard], DH_CARD_THRESH,
                                        fpde_card_dirs[icard].pde, 0666,
                                        dh_card_thresh_read_proc, dh_card_thresh_write_proc,
                                        sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_thresh[icard].pde->data, icard);

      /* Min/max current/voltage limits */
      if((result = dh_make_fpde_as_file(&fpde_card_minvolts[icard], DH_CARD_MINVOLTS,
					fpde_card_dirs[icard].pde, 0666,
					dh_card_minvolts_read_proc, dh_card_minvolts_write_proc,
					sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_minvolts[icard].pde->data, icard);

      if((result = dh_make_fpde_as_file(&fpde_card_maxvolts[icard], DH_CARD_MAXVOLTS,
					fpde_card_dirs[icard].pde, 0666,
					dh_card_maxvolts_read_proc, dh_card_maxvolts_write_proc,
					sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_maxvolts[icard].pde->data, icard);

      if((result = dh_make_fpde_as_file(&fpde_card_mincur[icard], DH_CARD_MINCUR,
                                        fpde_card_dirs[icard].pde, 0666,
                                        dh_card_mincur_read_proc, dh_card_mincur_write_proc,
                                        sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_mincur[icard].pde->data, icard);

      if((result = dh_make_fpde_as_file(&fpde_card_maxcur[icard], DH_CARD_MAXCUR,
                                        fpde_card_dirs[icard].pde, 0666,
                                        dh_card_maxcur_read_proc, dh_card_maxcur_write_proc,
                                        sizeof(struct fpde_data))) != 0) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_maxcur[icard].pde->data, icard);

      /* Create the PCI status file for this card */
      if((result = dh_make_fpde_as_file(&fpde_card_pci[icard], DH_CARD_PCI,
					fpde_card_dirs[icard].pde, 0666,
					dh_pci_read_proc, dh_pci_write_proc,
                                        sizeof(struct fpde_data))) != 0) {
	dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_pci[icard].pde->data, icard);
      
      /* Create the FPGA register r/w proc file for this card */
      if((result = dh_make_fpde_as_file(&fpde_card_fpga_regs[icard],DH_CARD_FPGA_REGS,
					fpde_card_dirs[icard].pde, 0600,
					dh_fpga_regs_read_proc, dh_fpga_regs_write_proc,
					sizeof(struct proc_fpga_data))) != 0) {
	dh_cleanup_proc_entries();
	return result;
      }
      pfd = (struct proc_fpga_data *) fpde_card_fpga_regs[icard].pde->data;
      pfd->cardnum            = icard;
      pfd->register_to_read   = 0;
      pfd->read_request_given = 0;
      for(ireg=0; ireg<DH_NUM_REGIS; ireg++) {
	pfd->sim_regs[ireg] = 0xFF;
      }

      /* Create the 1pps timer init proc file for this card */
      if((result = dh_make_fpde_as_file(&fpde_card_1pps[icard],DH_CARD_1PPS,
                                        fpde_card_dirs[icard].pde, 0666,
                                        dh_1pps_read_proc, dh_1pps_write_proc,
					sizeof(struct fpde_data)) != 0)) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_1pps[icard].pde->data, icard);

      /* Create the DOR clock select (internal clock/external clock) */
      if((result = dh_make_fpde_as_file(&fpde_card_clksel[icard],DH_CARD_CLKSEL,
                                        fpde_card_dirs[icard].pde, 0666,
                                        dh_clksel_read_proc, dh_clksel_write_proc,
                                        sizeof(struct fpde_data)) != 0)) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_clksel[icard].pde->data, icard);

      /* Create the gps<->dor clock sync. function proc file */
      if((result = dh_make_fpde_as_file(&fpde_card_syncgps[icard],DH_CARD_SYNCGPS,
                                        fpde_card_dirs[icard].pde, 0444,
                                        dh_syncgps_read_proc, NULL,
                                        sizeof(struct fpde_data)) != 0)) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_syncgps[icard].pde->data, icard);
      
      if((result = dh_make_fpde_as_file(&fpde_card_rev[icard],DH_CARD_REV,
                                        fpde_card_dirs[icard].pde, 0444,
                                        dh_rev_read_proc, NULL,
                                        sizeof(struct fpde_data)) != 0)) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_rev[icard].pde->data, icard);

      /* Create the dor-test logfile interface */
      if((result = dh_make_fpde_as_file(&fpde_card_test_log[icard],DH_CARD_TEST_LOG,
                                        fpde_card_dirs[icard].pde, 0444,
                                        dh_test_log_read_proc, NULL,
                                        sizeof(struct fpde_data)) != 0)) {
        dh_cleanup_proc_entries();
        return result;
      }
      dh_fill_card_data(fpde_card_test_log[icard].pde->data, icard);

      /* Create the pair-wise directories for this card */
      for(ipair=0;ipair < DH_PAIRS_PER_CARD;ipair++) {
	
	snprintf(tmpname,PDE_NAME_LEN,"pair%d",ipair);
	if((result = dh_make_fpde_as_dir(&fpde_pair_dirs[icard][ipair],
					 tmpname,fpde_card_dirs[icard].pde)) != 0) {
	  dh_cleanup_proc_entries();
	  return result;
	}
	
	/* Create pair power proc file */
	if((result = dh_make_fpde_as_file(&fpde_pair_pwr[icard][ipair],DH_PAIR_PWR,
					  fpde_pair_dirs[icard][ipair].pde, 0666,
					  dh_pwr_read_proc, dh_pwr_write_proc,
					  sizeof(struct fpde_data)) != 0)) {
	  dh_cleanup_proc_entries();
	  return result;
	}
        dh_fill_card_pair_data(fpde_pair_pwr[icard][ipair].pde->data, icard, ipair);

	/* Create pair proc file which shows whether we're plugged in or not */
        if((result = dh_make_fpde_as_file(&fpde_pair_plugged[icard][ipair],DH_PAIR_IS_PLUGGED,
					  fpde_pair_dirs[icard][ipair].pde, 0444,
                                          dh_plugged_read_proc, NULL,
					  sizeof(struct fpde_data)) != 0)) {
	  dh_cleanup_proc_entries();
          return result;
        }
        dh_fill_card_pair_data(fpde_pair_plugged[icard][ipair].pde->data, icard, ipair);

        /* Create pair proc file which shows wire pair current */
        if((result = dh_make_fpde_as_file(&fpde_pair_current[icard][ipair],DH_PAIR_CURRENT,
                                          fpde_pair_dirs[icard][ipair].pde, 0444,
                                          dh_current_read_proc, NULL,
                                          sizeof(struct fpde_data)) != 0)) {
          dh_cleanup_proc_entries();
          return result;
        }
	dh_fill_card_pair_data(fpde_pair_current[icard][ipair].pde->data, icard, ipair);
	
	/* Create voltage proc file */
        if((result = dh_make_fpde_as_file(&fpde_pair_voltage[icard][ipair],DH_PAIR_VOLTAGE,
                                          fpde_pair_dirs[icard][ipair].pde, 0444,
                                          dh_voltage_read_proc, NULL,
                                          sizeof(struct fpde_data)) != 0)) {
          dh_cleanup_proc_entries();
          return result;
        }
	dh_fill_card_pair_data(fpde_pair_voltage[icard][ipair].pde->data, icard, ipair);

	/* Power check proc file */
	if((result = dh_make_fpde_as_file(&fpde_pair_pwr_check[icard][ipair],DH_PAIR_PWR_CHECK,
					  fpde_pair_dirs[icard][ipair].pde, 0666,
					  dh_pair_pwr_check_read_proc,
					  dh_pair_pwr_check_write_proc,
					  sizeof(struct fpde_data)) != 0)) {
          dh_cleanup_proc_entries();
          return result;
        }
        dh_fill_card_pair_data(fpde_pair_pwr_check[icard][ipair].pde->data, icard, ipair);

	/* Create the DOM-wise directories for this pair - for tcalib, status */
	for(idom=0;idom < DH_DOMS_PER_PAIR;idom++) {
	  
	  snprintf(tmpname,PDE_NAME_LEN,"dom%s",(idom==0 ? "A" : "B"));
	  if((result = dh_make_fpde_as_dir(&fpde_dom_dirs[icard][ipair][idom],
					   tmpname,
					   fpde_pair_dirs[icard][ipair].pde)) != 0) {
	    dh_cleanup_proc_entries();
	    return result;
	  }
	  
	  /* Create tcalib file */
	  if((result = dh_make_fpde_as_file(&fpde_dom_tcalib[icard][ipair][idom],DH_DOM_TCALIB,
					    fpde_dom_dirs[icard][ipair][idom].pde, 0666,
					    dh_tcalib_read_proc,
					    dh_tcalib_write_proc,
					    sizeof(struct fpde_data))) != 0) {
	    dh_cleanup_proc_entries();
	    return result;
	  } /* if dh_make_fpde_as_file not ok */
	  dh_fill_card_pair_dom_data(fpde_dom_tcalib[icard][ipair][idom].pde->data,
				     icard, ipair, idom);

	  /* Create DOM ID file */
          if((result = dh_make_fpde_as_file(&fpde_dom_id[icard][ipair][idom],DH_DOM_ID,
					    fpde_dom_dirs[icard][ipair][idom].pde, 0444,
                                            dh_domid_read_proc, NULL,
                                            sizeof(struct fpde_data)) != 0)) {
            dh_cleanup_proc_entries();
            return result;
          } /* if dh_make_fpde_as_file not ok */
          dh_fill_card_pair_dom_data(fpde_dom_id[icard][ipair][idom].pde->data,
                                     icard, ipair, idom);

	  /* Create proc file for poisoning TX data */
          if((result = dh_make_fpde_as_file(&fpde_dom_pollute[icard][ipair][idom],DH_DOM_POLLUTE,
					    fpde_dom_dirs[icard][ipair][idom].pde, 0222,
                                            NULL, dh_pollute_write_proc,
                                            sizeof(struct fpde_data)) != 0)) {
            dh_cleanup_proc_entries();
            return result;
          } /* if dh_make_fpde_as_file not ok */
          dh_fill_card_pair_dom_data(fpde_dom_pollute[icard][ipair][idom].pde->data,
                                     icard, ipair, idom);

	  if((result = dh_make_fpde_as_file(&fpde_dom_rxdt[icard][ipair][idom],DH_DOM_RXDT,
                                            fpde_dom_dirs[icard][ipair][idom].pde, 0666,
                                            dh_rxdt_read_proc, dh_rxdt_write_proc,
                                            sizeof(struct fpde_data)) != 0)) {
            dh_cleanup_proc_entries();
            return result;
          } /* if dh_make_fpde_as_file not ok */
          dh_fill_card_pair_dom_data(fpde_dom_rxdt[icard][ipair][idom].pde->data,
                                     icard, ipair, idom);

	  /* Create DOM time profile proc file */
          if((result = dh_make_fpde_as_file(&fpde_dom_bench[icard][ipair][idom],DH_DOM_BENCH,
					    fpde_dom_dirs[icard][ipair][idom].pde, 0444,
                                            dh_dombench_read_proc, NULL,
                                            sizeof(struct fpde_data)) != 0)) {
            dh_cleanup_proc_entries();
            return result;
          } /* if dh_make_fpde_as_file not ok */
	  dh_fill_card_pair_dom_data(fpde_dom_bench[icard][ipair][idom].pde->data,
                                     icard, ipair, idom);

	  /* Create DOM softboot file */
          if((result = dh_make_fpde_as_file(&fpde_dom_softboot[icard][ipair][idom],DH_DOM_SOFTBOOT,
                                            fpde_dom_dirs[icard][ipair][idom].pde, 0222,
                                            NULL, dh_softboot_write_proc,
                                            sizeof(struct fpde_data)) != 0)) {
            dh_cleanup_proc_entries();
            return result;
          } /* if dh_make_fpde_as_file not ok */
          dh_fill_card_pair_dom_data(fpde_dom_softboot[icard][ipair][idom].pde->data,
                                     icard, ipair, idom);

	  /* Create DOM comms statistics file */
          if((result = dh_make_fpde_as_file(&fpde_dom_comstat[icard][ipair][idom],
					    DH_DOM_COMSTAT,
					    fpde_dom_dirs[icard][ipair][idom].pde, 0666,
					    dh_comstat_read_proc, dh_comstat_write_proc,
					    sizeof(struct fpde_data)) != 0)) {
	    dh_cleanup_proc_entries();
	    return result;
	  } /* if dh_make_fpde_as_file not ok */
          dh_fill_card_pair_dom_data(fpde_dom_comstat[icard][ipair][idom].pde->data,
                                     icard, ipair, idom);

	  /* Create domstat proc file to collect comms stats data from DOM */
	  if((result = dh_make_fpde_as_file(&fpde_dom_domstat[icard][ipair][idom],
                                            DH_DOM_DOMSTAT,
                                            fpde_dom_dirs[icard][ipair][idom].pde, 0444,
                                            dh_domstat_read_proc, NULL,
                                            sizeof(struct fpde_data)) != 0)) {
	    dh_cleanup_proc_entries();
	    return result;
	  }
	  dh_fill_card_pair_dom_data(fpde_dom_domstat[icard][ipair][idom].pde->data,
                                     icard, ipair, idom);

	  /* Create DOM communications ready status file */
          if((result = dh_make_fpde_as_file(&fpde_dom_iscomm[icard][ipair][idom],DH_DOM_ISCOMM,
					    fpde_dom_dirs[icard][ipair][idom].pde, 0666,
					    dh_iscomm_read_proc, dh_iscomm_write_proc,
					    sizeof(struct fpde_data)) != 0)) {
	    dh_cleanup_proc_entries();
	    return result;
	  } /* if dh_make_fpde_as_file not ok */
	  dh_fill_card_pair_dom_data(fpde_dom_iscomm[icard][ipair][idom].pde->data,
                                     icard, ipair, idom);

          /* Create proc file to indicate whether DOM is out of configboot */
          if((result = dh_make_fpde_as_file(&fpde_dom_isnotcb[icard][ipair][idom],DH_DOM_ISNOTCB,
					    fpde_dom_dirs[icard][ipair][idom].pde, 0444,
					    dh_isnotcb_read_proc, NULL,
					    sizeof(struct fpde_data)) != 0)) {
	    dh_cleanup_proc_entries();
            return result;
          } /* if dh_make_fpde_as_file not ok */
          dh_fill_card_pair_dom_data(fpde_dom_isnotcb[icard][ipair][idom].pde->data,
                                     icard, ipair, idom);

	  /* Create DOM is-connected proc file */
          if((result = dh_make_fpde_as_file(&fpde_dom_isconn[icard][ipair][idom],DH_DOM_ISCONN,
                                            fpde_dom_dirs[icard][ipair][idom].pde, 0444,
					    dh_isconn_read_proc, NULL,
					    sizeof(struct fpde_data)) != 0)) {
            dh_cleanup_proc_entries();
            return result;
          } /* if dh_make_fpde_as_file not ok */
          dh_fill_card_pair_dom_data(fpde_dom_isconn[icard][ipair][idom].pde->data,
                                     icard, ipair, idom);

	} /* loop over doms */
      } /* loop over pairs */
    } /* if card found */
  } /* loop over cards */

  return 0; /* Good to go */
}

/************** End /proc File Handling Functions *****************/


/*********************** "unit" tests *****************************/

#define ioq_assert(x, str) if(!(x)) { printk(KERN_ERR "dh: ioq self test: "str"!\n"); return 1; }

int dh_test_ioq(void) {
  vprintk(KERN_INFO "dh: Testing I/O queues ...\n");

  ioQueue * q = newIOQ();

  ioq_assert(q != NULL, "newIOQ failed");
  ioq_assert(nIOQEntries(q) == 0, "nIOQEntries(q) != 0");
  ioq_assert(nIOQSlots(q) == NIOQENTRIES, "nIOSlots(q) != NIOQENTRIES");

  ioqMsg * g = getIOQ(q);
  ioq_assert(g == NULL, "getIOQ test 1");

  ioqMsg * m = (ioqMsg *) kmalloc(sizeof(ioqMsg), GFP_ATOMIC);
  ioq_assert(m != NULL, "kmalloc of ioqMsg failed");

  putIOQ(q, m);
  ioq_assert(nIOQEntries(q) == 1, "test 5");

  g = getIOQ(q);
  ioq_assert(nIOQEntries(q) == 0, "test 6");

  int i;
  for(i=0; i<NIOQENTRIES; i++) {
    m->nbytes = i;
    ioq_assert(! ioQisFull(q), "test 7");
    putIOQ(q, m);
  }

  ioq_assert(ioQisFull(q), "test 8");

  for(i=0; i<NIOQENTRIES; i++) {
    g = getIOQ(q);
    ioq_assert(g != NULL, "test 9");
    ioq_assert(g->nbytes == i, "test 10");
  }

  ioq_assert(ioQisEmpty(q), "test 11");

  m->nbytes = 66;
  putIOQ(q, m);
  g = getIOQ(q);
  ioq_assert(m->nbytes == g->nbytes, "test 12");

  kfree(m);

  kfree(q);
  return 0;
}

/************* Principle driver callbacks (dh_fops) ***************/

static unsigned int dh_poll(struct file *filp, poll_table *wait) {

  /** Implement polling for poll() and select() on dh device file.  In
      sim_copy or sim_nocopy mode, poll flags for read or write are
      always true. */

  int minor;
  struct inode * da_inode;
  unsigned int poll_flags;

  da_inode = filp->f_dentry->d_inode;

  minor = MINOR(da_inode->i_rdev);

  vvprintk(KERN_INFO "dh: Called poll (minor %d).\n",minor);

  poll_wait(filp, &inq[minor]->wq,  wait);
  poll_wait(filp, &outq[minor]->wq, wait);
  
  vvprintk(KERN_INFO "dh: passed poll wait (minor %d).\n",minor);

  poll_flags = 0; 
  int icard  = dh_minor_to_cardno(minor);
  unsigned long flags;

  /* Lock card to guarantee consistent state */
  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  if(ioQhasSynfin(inq[minor])) {
    vvprintk(KERN_INFO "dh_poll: minor %d is readable\n", minor);
    poll_flags |= POLLIN | POLLRDNORM;
  }

  /* Make sure a whole, maximum message can fit in the buffer: */
  if(nIOQSlots(outq[minor]) >= DH_PKTS_PER_MAXIMAL_MESSAGE) {
    //if(!ioQisFull(outq[minor])) {
    poll_flags |= POLLOUT | POLLWRNORM;
  }

  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
  
  vvprintk(KERN_INFO "dh: returning poll_flags=(%d: %s,%s,%s,%s)\n", poll_flags,
	   (poll_flags&POLLIN)?"POLLIN":"!pollin",
	   (poll_flags&POLLRDNORM)?"POLLRDNORM":"!pollrdnorm",
	   (poll_flags&POLLOUT)?"POLLOUT":"!pollout",
	   (poll_flags&POLLWRNORM)?"POLLWRNORM":"!pollwrnorm");
  
  return poll_flags;
}

static void dh_cleanup_before_close(int minor, int icard, int ich) {
  /* Was in dh_close but dh_open has to call this if open fails late in the call */
  unsigned long flags;

  /* down_interruptible doesn't help us here, we just need to get the sem. and cleanup */
  down(&dh_pci_dors[icard].open_close_sem);
  dh_pci_dors[icard].num_open--;
  if(dh_pci_dors[icard].num_open == 0) {
    del_timer_sync(&dh_timer_struct[icard]);
    dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
    dh_write_regi(dh_pci_dors[icard].icard, DH_RG_INTEN, 0);
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    free_irq(dh_pci_dors[icard].irq, &(dh_pci_dors[icard].icard));
  }
  dh_opened[minor]     = 0;
  dh_conn_state[minor] = DH_STATE_START;

  wake_up_interruptible(&device_in_use_wq[minor]);
  up(&dh_pci_dors[icard].open_close_sem);
}

static int dh_open(struct inode *inode, struct file *filp) {

/** Open communications device file.
    \return 0 if success, else:
    - -EINVAL : bad minor number in device file
    - -EBUSY  : Device is already open
    - -ENOMEM : Can't get buffer space */

  unsigned long flags=0;

#ifdef KERNEL2_4
#define USE_MOD_USE_COUNT
#endif

#ifdef USE_MOD_USE_COUNT
  MOD_INC_USE_COUNT;
#endif

  /* Check minor number for range & availability */
  int minor = MINOR(inode->i_rdev);
  int icard = dh_minor_to_cardno(minor);    
  int ipair = dh_minor_to_wireno(minor);
  int idom  = dh_minor_to_domno(minor);
  int ich   = idom + ipair*2;

  int block = 1;
  if(!parm_block || (filp->f_flags & O_NONBLOCK)) block=0;
  vprintk(KERN_INFO "dh: opening %d%d%c (minor %d)\n", 
	  icard, ipair, idom?'B':'A', minor);

  if(minor < 0 || minor > (DH_DOMHUB_DOMS-1)) { 
    printk(KERN_ERR "dh: Bad minor number (%d) for device file.\n",minor);
    dh_lasterr = DH_ERR_BADMINOR;
#ifdef USE_MOD_USE_COUNT
    MOD_DEC_USE_COUNT;
#endif
    return -EINVAL;
  }

  printd(dbgq[minor], "Open Device");

  if(down_interruptible(&dh_pci_dors[icard].open_close_sem)) { /* DDDDDDDDDDDDDDDDD */
    printk(KERN_ERR "dh_open: ABORTED semaphore acquire, open of minor %d failed.\n", minor);
#ifdef USE_MOD_USE_COUNT
    MOD_DEC_USE_COUNT;
#endif
    return -ERESTARTNOHAND;
  }

  while(dh_opened[minor]) {
  /* Block until other processes close file */
    if(dh_opened[minor]) 
      vprintk(KERN_WARNING "dh_open: Blocking until other process closes minor %d.\n", minor);
    up(&dh_pci_dors[icard].open_close_sem); /* PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP */
    wait_event_interruptible(device_in_use_wq[minor], ! dh_opened[minor]);
    if(signal_pending(current)) {
#ifdef USE_MOD_USE_COUNT
      MOD_DEC_USE_COUNT;
#endif
      return -ERESTARTNOHAND;
    }
    if(down_interruptible(&dh_pci_dors[icard].open_close_sem)) { /* DDDDDDDDDDDD */
      printk(KERN_ERR "dh_open: ABORTED semaphore acquire, open of minor %d failed.\n", minor);
#ifdef USE_MOD_USE_COUNT
      MOD_DEC_USE_COUNT;
#endif
      return -ERESTARTNOHAND; /* No up() here, didn't get semaphore */
    }
  }

  /* Make sure PCI info for DOR card is really available */

  if(! dh_pci_dors[icard].found) {
    up(&dh_pci_dors[icard].open_close_sem); /*PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP*/
    printk(KERN_ERR "dh: Device %d (card %d) had no PCI information found.\n", minor, icard);
    dh_lasterr = DH_ERR_NODORPCIINFO;
#ifdef USE_MOD_USE_COUNT
    MOD_DEC_USE_COUNT;
#endif
    return -ENODEV;
  }

  /* Make sure firmware version is correct */

  if(dh_firmware_mismatch[icard]) {
    up(&dh_pci_dors[icard].open_close_sem); /*PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP*/
    printk(KERN_ERR "dh: Can't open device %d, PCI or comms. firmware is unsupported.\n", minor);
    dh_lasterr = DH_ERR_UNSUPPFWARE;
#ifdef USE_MOD_USE_COUNT
    MOD_DEC_USE_COUNT;
#endif
    return -EINVAL;
  }

  /* Check that DOM was communicating before & is powered up */
  
  if(dh_get_pair_pwr_state(icard, ipair) != 2) {
    up(&dh_pci_dors[icard].open_close_sem); /*PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP*/
    printk(KERN_ERR "dh_open: minor %d is not powered on\n", minor);
    dh_lasterr = DH_ERR_OPENNOTPWR;
#ifdef USE_MOD_USE_COUNT
    MOD_DEC_USE_COUNT;
#endif
    return -EAGAIN;
  }
  
  if(! dh_is_communicating[minor]) {
    up(&dh_pci_dors[icard].open_close_sem); /*PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP*/
    printk(KERN_WARNING "dh_open: minor %d didn't communicate at power on\n", minor);
    dh_lasterr = DH_ERR_DOMNOTAVAIL;
#ifdef USE_MOD_USE_COUNT
    MOD_DEC_USE_COUNT;
#endif
    return -EAGAIN;
  }

  dh_empty_retxb(minor);
  retx_triggered[minor] = 0;

  emptyRetxQ(&retxq[minor]);
  init_waitqueue_head(&retxbwq[minor]);
  init_waitqueue_head(&ackqwq[minor]);
  init_waitqueue_head(&connectwq[minor]);
  init_waitqueue_head(&comms_reset_wq[minor]);
  init_waitqueue_head(&domstatwq[minor]);
  init_waitqueue_head(&device_in_use_wq[minor]);
  init_waitqueue_head(&domstatwq[minor]);
  domstats[minor].have_new = 0;
  emptyIOQ(outq[minor]);
  emptyIOQ(inq[minor]);
  
  txseq[minor]          = 0;
  dh_conn_state[minor]  = DH_STATE_START;
  dh_comms_state[minor] = DH_COMMS_READY; /* Kernel timer will look for timeouts */
  
  if(dh_pci_dors[icard].num_open == 0) {
    /* Do IRQ allocation and register initialization */
    dh_init_intr(icard); /* Disable all interrupts */
    vvprintk(KERN_INFO "dh_open: requesting IRQ for minor %d.\n", minor);
    if(request_irq(dh_pci_dors[icard].irq, /* Can SLEEP (does GFP_KERNEL kmalloc)*/
		   dh_intr_hdlr, SA_SHIRQ, 
		   "dh", &(dh_pci_dors[icard].icard))) {
      up(&dh_pci_dors[icard].open_close_sem); /*PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP*/
      printk(KERN_ERR "dh_open: request_irq failed.\n");
      dh_lasterr = DH_ERR_CANTGETIRQ;
#ifdef USE_MOD_USE_COUNT
      MOD_DEC_USE_COUNT;
#endif
      return -EBUSY;
    }
    
    /* Have to disable all TX and RX interrupts first, in case they were left
       enabled last time the file was opened */
    dh_write_regi(icard, DH_RG_TTSIC, 0);
    dh_write_regi(icard, DH_RG_RTSIC, 0);
    dh_enable_global_intr(icard); /* Nothing enabled yet, except global. */
    
    /* Start kernel timer here -- one per card */
    init_timer(&dh_timer_struct[icard]);
    dh_timer_struct[icard].expires  = jiffies+1;
    dh_timer_struct[icard].data     = (unsigned long) icard;
    dh_timer_struct[icard].function = &dh_timer;
    add_timer(&dh_timer_struct[icard]);
    vvprintk(KERN_INFO "dh_open: Started kernel timer for card %d.\n", icard);
    vvprintk(KERN_INFO "dh_open: Blocking minor %d while checking for hardware timeouts...\n",
	     minor);      
  }

  dh_pci_dors[icard].num_open++;
  dh_opened[minor]      = 1; /* Need to do this before looking for CI/IC, so kernel timer
				will look */
  unsigned long j0 = jiffies;
#define OTIMEDOUT() (jiffies-j0 > DH_OPENTIMEOUT)
  /* Don't sleep while holding semaphore: critical section is over */
  up(&dh_pci_dors[icard].open_close_sem); /*PPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPPP*/

  wait_event_interruptible(comms_reset_wq[minor],
			   OTIMEDOUT() || dh_comms_state[minor] == DH_COMMS_READY);
  if(signal_pending(current)) {
    vprintk(KERN_INFO "dh_open: Aborting open on minor %d "
	    "before interrupt enable, cross your fingers...\n", minor);
    dh_cleanup_before_close(minor, icard, ich);
    dh_lasterr = DH_ERR_CRESETINTR;
#ifdef USE_MOD_USE_COUNT
    MOD_DEC_USE_COUNT;
#endif
    return -EFAULT;
  } else if(dh_comms_state[minor] != DH_COMMS_READY) {
    vprintk(KERN_WARNING "dh_open: Timeout (comms. reset) on minor %d.\n", minor);
    dh_cleanup_before_close(minor, icard, ich);
    dh_lasterr = DH_ERR_OCRESETTOUT;
#ifdef USE_MOD_USE_COUNT
    MOD_DEC_USE_COUNT;
#endif
    return -EAGAIN;
  }
  
  vvprintk(KERN_INFO "dh_open: Blocking minor %d until DOM agrees to resequence...\n", minor);
  /* Interrupts get enabled as soon as we go to state C */
  wait_event_interruptible(connectwq[minor], 
			   OTIMEDOUT() || dh_conn_state[minor] == DH_STATE_CONNECTED);
  vvprintk(KERN_INFO "dh_open: After wait_event_interruptible dt=%ld\n", jiffies-j0);
  if(signal_pending(current)) {
    vprintk(KERN_INFO "dh_open: Aborting open on minor %d "
	   "before interrupt enable, cross your fingers...\n", minor);
    dh_cleanup_before_close(minor, icard, ich);
    dh_lasterr = DH_ERR_RESEQINTR;
#ifdef USE_MOD_USE_COUNT
    MOD_DEC_USE_COUNT;
#endif
    return -EFAULT;
  } else if(dh_conn_state[minor] != DH_STATE_CONNECTED) {
    printk(KERN_WARNING "dh_open: Timeout (synchronize failed) on minor %d.\n", minor);
    dh_dump_fpga(icard);
    printd(dbgq[minor], "Open failed!\n");
    dumpd(minor);
    dh_cleanup_before_close(minor, icard, ich);
    dh_lasterr = DH_ERR_OCONNTOUT;
#ifdef USE_MOD_USE_COUNT
    MOD_DEC_USE_COUNT;
#endif
    return -EAGAIN;
  } else{
    vvprintk(KERN_INFO "dh: minor %d connected.\n", minor);
  }

  printd(dbgq[minor], "Open succeeded!\n");

  /* Handle interrupt enable in kernel timer now */

  /* Wake up any processes waiting for this file to open */
  wake_up_interruptible(&file_open_waiter[minor]);

  vvprintk(KERN_INFO "dh: Successfully opened dom hub device with minor number %d.\n",minor);
  return 0;
}


static int dh_close(struct inode *inode, struct file *file) {
  /** Close communications device file.
      \return 0 unless device file has a bad minor number, 
      in which case -EINVAL is returned */
  int minor, jiffdelays;

  minor = MINOR(inode->i_rdev); 
  if(minor < 0 || minor > (DH_DOMHUB_DOMS-1)) {
    printk(KERN_ERR "dh: Bad minor number (%d) for device file.\n",minor);
    dh_lasterr = DH_ERR_BADMINOR;
    return -EINVAL;
  }

  int icard = dh_minor_to_cardno(minor);
  int ipair = dh_minor_to_wireno(minor);
  int idom  = dh_minor_to_domno(minor);
  int ich   = idom + ipair*2;
  unsigned long flags;

  unsigned long jif0 = jiffies; /* Compare jiffies-jif0 with DH_CLOSETIMEOUT */

  if(dh_mode == DH_NORMAL) {

    /* This is a little kludgy but reenable TX interrupts here if there is 
       stuff left over to send */

    vvprintk(KERN_INFO "dh_close(minor=%d): Waiting for output queues to drain...\n",
	     minor);

#define CTO() (jiffies-jif0 >= DH_CLOSETIMEOUT)

    while(!CTO() && !signal_pending(current) && !ioQisEmpty(outq[minor])) {
      vvprintk(KERN_INFO "dh_close: sending last %d outq messages on minor %d...\n", 
	       nIOQEntries(outq[minor]), minor);
      dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
      dh_enable_global_intr(icard);
      dh_enable_msg_sndbl_intr_chan(icard, ich);
      dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
      if(wait_event_interruptible(outq[minor]->wq, CTO() || ioQisEmpty(outq[minor]))) 
	break;
    }

    if(CTO() && !ioQisEmpty(outq[minor])) { /* Show it loud and clear; DOM probably stuck */
      vprintk(KERN_INFO "!!!\n");
      vprintk(KERN_INFO "!!! dh_close: Drain of last messages on minor %d "
	     "failed after timeout (%d remain) !!!\n", minor, nIOQEntries(outq[minor]));
      vprintk(KERN_INFO "!!!\n");
    }
	     

    while(!CTO() && !signal_pending(current)) {
      dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
      if(dh_retx_buf_isempty(minor)) {
        dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
        break;
      }
      vprintk(KERN_INFO "dh_close: sending last %d ReTX messages on minor %d...\n", 
	     dh_num_retxb(minor), minor);
      dh_enable_global_intr(icard);
      dh_enable_msg_sndbl_intr_chan(icard, ich);
      dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
      if(wait_event_interruptible(retxbwq[minor], CTO() || dh_retx_buf_isempty(minor))) 
	break;
    }
    
    if(CTO() && (!dh_retx_buf_isempty(minor) || nRetxQEntries(&retxq[minor])>0)) { 
      vprintk(KERN_INFO "!!!\n");
      vprintk(KERN_INFO "!!! dh_close: Drain of last ReTXes on minor %d "
	     "failed after timeout (%d remain in retxb, %d in retxq)!!!\n", minor, 
	     dh_num_retxb(minor), nRetxQEntries(&retxq[minor]));
      vprintk(KERN_INFO "!!!\n");
    }

    while(!CTO() && !signal_pending(current)) {
      dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
      if(ackQisEmpty(&ackq[minor])) {
        dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
        break;
      }
      vprintk(KERN_INFO "dh_close: sending last %d ackq messages on minor %d...\n",
	     nAckQEntries(&ackq[minor]), minor);    
      dh_enable_global_intr(icard);
      dh_enable_msg_sndbl_intr_chan(icard, ich);
      dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
      if(wait_event_interruptible(ackqwq[minor], CTO() || ackQisEmpty(&ackq[minor])))
	break;
    }

    if(CTO() && !ackQisEmpty(&ackq[minor])) {
      vprintk(KERN_INFO "!!!\n");
      vprintk(KERN_INFO "!!! dh_close: Drain of last ACKs on minor %d "
             "failed after timeout (%d remain) !!!\n", minor, nAckQEntries(&ackq[minor]));
      vprintk(KERN_INFO "!!!\n");
    }

    /* Disable interrupts, release IRQ */
    unsigned long flags;
    dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/

    dh_domvars[minor].had_rx_data = 0;
    printd(dbgq[minor], "CLOSE: disable interrupts");
    dh_disable_msg_sndbl_intr_chan(icard, ich);
    dh_disable_msg_recvd_intr_chan(icard, ich);

    //if(signal_pending(current)) {
    //vprintk(KERN_WARNING "dh_close(%d): Signal pending, clearing out all queues.\n",
    //        minor);
    dh_empty_retxb(minor);
    retx_triggered[minor] = 0;
    emptyRetxQ(&retxq[minor]);
    emptyAckQ(&ackq[minor]);
    emptyIOQ(inq[minor]);
    emptyIOQ(outq[minor]);
    //}

    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/

  } /* if DH_NORMAL */
  dh_cleanup_before_close(minor, icard, ich);

  vprintk(KERN_INFO "dh: closed %d%d%c (minor %d)\n", 
	            icard, ipair, idom?'B':'A', minor);

#ifdef USE_MOD_USE_COUNT
  MOD_DEC_USE_COUNT;
#endif

  emptyDbgQ(dbgq[minor]);
  return 0;
}


static ssize_t dh_read(struct file *filp, char *buf, size_t count, loff_t *fpos) {

/** Read a message from message queue, put there by interrupt handler
    (hardware) or by dh_write (simulation).  In nocopy mode, just
    return count (for benchmarking).  In copy mode, copy buffer from
    kernel space buffer.  In queue mode, do the actual message queue
    handling.

    Note that count should ALWAYS equal the maximum message size, 
    DH_MAX_MAX_MESSAGE_SIZE.  The call may return less, however.

    \return number of bytes transferred from device to user buffer buf,
    or a negative error value. */

  int minor, icard, ich;
  unsigned long flags=0;
  static int dh_show_timeout = 1;
  u32 rtsic;
  int size   = 0;

  dh_bench_start(gread_bench);

  if(count < DH_MAX_MESSAGE_SIZE) {
    printk(KERN_ERR
  	   "dh: Read callback requires space for at least %d bytes in user space "
  	   "(your buffer is only %d bytes).\n",
  	   DH_MAX_MESSAGE_SIZE, count);
    dh_lasterr = DH_ERR_BADUSERBUFSIZ;
    dh_bench_end(&gread_bench);
    return -EFAULT;
  }

  
  minor = MINOR(filp->f_dentry->d_inode->i_rdev);
  icard = dh_minor_to_cardno(minor);

  printd(dbgq[minor], "READ");

  unsigned char * pktcpy = kmalloc(DH_MAX_SW_PKTSIZ, GFP_KERNEL); /* CAN SLEEP! */
  if(!pktcpy) return -ENOMEM;

  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
  
  ich   = dh_minor_to_wireno(minor)*2 + dh_minor_to_domno(minor);

  /* Block or -EAGAIN if channel is not "connected" */
  if(dh_conn_state[minor] != DH_STATE_CONNECTED) {
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    if(!parm_block || filp->f_flags & O_NONBLOCK) {
      vprintk(KERN_INFO "dh_read: minor %d went unconnected, sorry, come EAGAIN.\n", minor);
      kfree(pktcpy);
      return -EAGAIN;
    } else { /* Blocking case: wait for kernel timer to reestablish connection, then re-lock */
      wait_event_interruptible(connectwq[minor], dh_conn_state[minor] == DH_STATE_CONNECTED);
      if(signal_pending(current)) {
        printk(KERN_INFO "dh_read: Caught signal while waiting to get connected.\n");
	kfree(pktcpy);
	return -EAGAIN;
      }
      dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
    }
  }

  /* Wait for appropriate queue states */
  if(!parm_block || filp->f_flags & O_NONBLOCK) {
    if(!ioQhasSynfin(inq[minor])) {
      vvvprintk("dh_read: no syn/fin\n");
      dh_bench_end(&gread_bench);
      dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
      kfree(pktcpy);
      return -EAGAIN;
    }
  } else { /* Blocking I/O case */
    vvprintk(KERN_INFO "dh: blocking on read from dev %d.\n",minor);
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    wait_event_interruptible(inq[minor]->wq, ioQhasSynfin(inq[minor]));
    vvprintk(KERN_INFO "dh: woke up from read!  signal_pending=%d\n", signal_pending(current));
    if(signal_pending(current)) {
      dh_bench_end(&gread_bench);
      kfree(pktcpy);
      //return -ERESTARTSYS; /* fs layer handles signals */
      return -ERESTARTNOHAND;
    }
    dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
    vvprintk(KERN_INFO "dh: woke up from blocked read on dev %d!\n",minor);
  }
  
  /* Get the actual message to user space, assembled from packets */
  int synfin = 0;
  int ofs    = 0;
  int pktlen = 0;
  int ihwpkt = 0;
  unsigned char c1st, clast=0;
  while(!synfin && size < count && !ioQisEmpty(inq[minor])) { 
    /* Second condition should always be true */

    if(ihwpkt >= NIOQENTRIES) {
      printk(KERN_WARNING "dh: ERROR: got too many packets before SYN/FIN!\n");
      dh_lasterr = DH_ERR_TOOMANYPKTS;
      dh_bench_end(&gread_bench);
      dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
      kfree(pktcpy);
      return -EFAULT; 
    }

    ioqMsg * m = getIOQ(inq[minor]);
    pktlen = m->nbytes;
    if(pktlen > DH_MAX_SW_PKTSIZ) {
      printk(KERN_WARNING "dh_read(minor=%d): pktlen %d is too large!  Truncating.\n",
	     minor,pktlen);
      pktlen = DH_MAX_SW_PKTSIZ;
    }
    synfin = m->synfin;
    /* Copy packet payload while we have the lock so the interrupt handler doesn't nuke it */
    memcpy(pktcpy, m->data, pktlen);

#ifdef DH_TEST_FOR_ZEROES
    int iz;
    for(iz=0; iz<pktlen;iz++) {
      if(pktcpy[iz] == 0) 
	printk(KERN_WARNING "dh_read: got 0 at iz=%d ihwpkt=%d pktlen=%d synfin=%d!\n",
	       iz, ihwpkt, pktlen, synfin);
    }
#endif

#ifdef DH_TEST_INC_PAT
    dh_test_for_inc_pat(pktcpy, pktlen, 1);
#endif

    /* Need to unlock here because copy_to_user can page fault, which can deadlock if
       interrupt handler kicks in */

    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    if(copy_to_user(buf+ofs, pktcpy, pktlen < DH_MAX_SW_PKTSIZ ? pktlen : DH_MAX_SW_PKTSIZ)) {
      printk(KERN_WARNING "dh read(minor=%d): bad userspace buffer!\n", minor);
      dh_lasterr = DH_ERR_BADUSERBUFPTR;
      dh_bench_end(&gread_bench);
      kfree(pktcpy);
      return -EFAULT; 
    }
    dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/

    vvprintk(KERN_INFO "dh_read: got %d bytes, synfin=%d\n", pktlen, synfin);
    /* If RX interrupts are currently disabled because of a full FIFO, 
       reenable them on this channel */
    dh_enable_msg_recvd_intr_chan(icard, ich);

    ofs += pktlen;
    size += pktlen;
    ihwpkt++;
  }

#ifdef DH_TEST_INC_PAT
  dh_test_for_inc_pat_user(buf, size, 2);
#endif

  vvprintk(KERN_INFO "dh_read(pre-RX-enable): inq: %d entries; outq: %d entries.\n",
	   nIOQEntries(inq[minor]), nIOQEntries(outq[minor]));

  /* Turn off LED */
    
  vvprintk(KERN_INFO "dh_read: Got %d bytes from dh_queue_get.\n",size);
  dh_bytesread[minor]+=size;
  dh_msgsread[minor]++;
  if(dh_do_simulate_timeout(minor)) dh_do_sim_timeout[minor] = 1; /* Generate fake hardware 
								     timeout in kernel timer */
  dh_total_msgs_read++;
      
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
  vvprintk(KERN_INFO "dh_read %dB <- minor %d\n",size, minor);  
  //dh_bench_end(&read_bench);
  dh_bench_end(&gread_bench);

  kfree(pktcpy);
  return size;
}
  

static ssize_t dh_write(struct file *filp, const char *buf, size_t count, loff_t *fpos) {
/** Write to communications device file. \return number of bytes written, or a negative
    error if there is one.  Message size is given by count.  If count bytes not 
    written, then a negative error value is returned. */

  ssize_t ret; 
  int    minor, icard, ich;
  unsigned long flags=0;
  size_t nbytes;
  int    i;
  int    empty;

  dh_bench_start(gwrite_bench);

  ret    = 0;
  minor  = MINOR(filp->f_dentry->d_inode->i_rdev);
  icard  = dh_minor_to_cardno(minor);
  ich    = dh_minor_to_wireno(minor)*2 + dh_minor_to_domno(minor);


  int npkts = (count+DH_MAX_SW_PKTSIZ-1)/(DH_MAX_SW_PKTSIZ);

  int room;

  vvprintk(KERN_INFO "dh_write(%d): writing %d packets\n",minor, npkts);
  if(npkts > NIOQENTRIES) {
    printk(KERN_ERR "dh_write: too many packets to fit in outq\n");
    dh_lasterr = DH_ERR_TOOMANYPKTS;
    dh_bench_end(&gwrite_bench);
    return -EFAULT;
  }

  if(count > DH_MAX_MESSAGE_SIZE) {
    printk(KERN_ERR "dh_write: message too big, must be <= %d bytes.\n", DH_MAX_MESSAGE_SIZE);
    dh_lasterr = DH_ERR_MSGTOOBIG;
    dh_bench_end(&gwrite_bench);
    return -EFAULT;
  }

  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/

  /* Block or -EAGAIN if channel is not "connected" */
  if(dh_conn_state[minor] != DH_STATE_CONNECTED) {
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    if(!parm_block || filp->f_flags & O_NONBLOCK) {
      vprintk(KERN_INFO "dh_write: minor %d went unconnected, sorry, come EAGAIN.\n", minor);
      dh_bench_end(&gwrite_bench);
      return -EAGAIN;
    } else { /* Blocking case: wait for kernel timer to reestablish connection, then re-lock */
      wait_event_interruptible(connectwq[minor], dh_conn_state[minor] == DH_STATE_CONNECTED);
      if(signal_pending(current)) {
	printk(KERN_INFO "dh_write: Caught signal while waiting to get connected.\n");
	dh_bench_end(&gwrite_bench);
	return -EAGAIN;
      }
      dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/
    }
  }

  /* Block or -EAGAIN if this device's output queue is full */
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
  if( (!parm_block || filp->f_flags & O_NONBLOCK)) { /* Nonblock case */
    if(nIOQSlots(outq[minor]) < npkts) {
      dh_bench_end(&gwrite_bench);
      return -EAGAIN;
    }
  } else { /* Blocking case */
    if(wait_event_interruptible(outq[minor]->wq, nIOQSlots(outq[minor]) >= npkts)) {
      dh_bench_end(&gwrite_bench);
      return -ERESTARTSYS; /* Tell fs layer to handle signals */
    }
  }
  dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/

  printd(dbgq[minor], "WRITE");
  //dumpd(minor);

  int ipkt;
  int off = 0;
  int remain = count;
  for(ipkt = 0; ipkt < npkts && remain > 0; ipkt++) {
    int pktlen = remain > DH_MAX_SW_PKTSIZ ? DH_MAX_SW_PKTSIZ : remain;
    nbytes = pktlen;
    vvprintk(KERN_INFO "dh_write: ipkt=%d off=%d remain=%d nbytes=%d\n", ipkt, off, remain,nbytes);
    int synfin = 0;
    if(ipkt == npkts-1) {
      vvprintk(KERN_INFO "dh_write: last packet, setting SYN/FIN.\n");
      synfin = 1;
    }
    
    dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
    
    if(putIOQ_from_user(outq[minor], buf+off, nbytes, synfin)) {
      printk(KERN_ERR "dh write(minor=%d): invalid user space pointer!\n", minor);
      dh_lasterr = DH_ERR_BADUSERBUFPTR;
      dh_bench_end(&gwrite_bench);
      return -EINVAL; 
    }
    dh_card_lock(icard, flags); /*LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL*/

    vvprintk(KERN_INFO "dh_write %dB -> minor %d\n", nbytes, minor);

    if(nbytes != pktlen) {
      printk(KERN_WARNING "dh: dh_write WARNING: wanted to write %d but only wrote %d bytes.\n",
	     pktlen, nbytes);
      dh_lasterr = DH_ERR_QSHORTWRITE;
      dh_bench_end(&gwrite_bench);
      dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
      return -EAGAIN;
    }
    
    off += pktlen;
    remain -= pktlen;
    
    vvprintk(KERN_INFO "dh_write: enabling interrupts on minor %d.\n", minor);
    dh_enable_msg_sndbl_intr_chan(icard, ich);
  } /* Looping over npkts */

  dh_msgswritten[minor]++;
  dh_total_msgs_written++;
  dh_byteswritten[minor]+=count;
  dh_card_unlock(icard, flags); /*UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU*/
  dh_bench_end(&gwrite_bench);
  return count;
}

/**************** End principle driver callbacks ******************/


/*************** Module init and cleanup routines *****************/

static void __exit dh_cleanup(void) {
/** Handle unregistration of the device driver; clean up / remove proc_dir_entries */
  /* Free dummy message allocated earlier for simulation modes */
  int minor;

  vprintk(KERN_INFO "dh: Releasing input and output message queues.\n");

  for(minor=0; minor < DH_DOMHUB_DOMS; minor++) {
    if(inq[minor] != NULL) {
      kfree(inq[minor]);
      inq[minor] = NULL;
    }
    
    if(outq[minor] != NULL) {
      kfree(outq[minor]);
      outq[minor] = NULL;
    }

    if(dbgq[minor] != NULL) {
      kfree(dbgq[minor]);
      dbgq[minor] = NULL;
    }

    if(txpkt[minor] != NULL) {
      kfree(txpkt[minor]);
      txpkt[minor] = NULL;
    }
    
    if(rxpkt[minor] != NULL) {
      kfree(rxpkt[minor]);
      rxpkt[minor] = NULL;
    }
    
    if(dh_tx_scratch_msg[minor] != NULL) {
      kfree(dh_tx_scratch_msg[minor]);
      dh_tx_scratch_msg[minor] = NULL;
    }
    
    if(dh_rx_scratch_msg[minor] != NULL) {
      kfree(dh_rx_scratch_msg[minor]);
      dh_rx_scratch_msg[minor] = NULL;
    }
   
    if(dh_tcalib_recs[minor] != NULL) {
      vfree(dh_tcalib_recs[minor]);
      dh_tcalib_recs[minor] = NULL;
    }
  }

  unregister_chrdev(DH_MAJOR, "dh"); 

  dh_cleanup_proc_entries();

  if(dh_mode == DH_NORMAL) {
    dh_release_pci_devices(dh_pci_dors, DH_MAX_CARDS);
  }

  vprintk(KERN_INFO "Removed module dh... thanks for playing!\n");
}

/** dh_fops tells register_chrdev where the driver callback functions
    are.  It has to be a global for some reason, or an oops message
    results on open */
struct file_operations dh_fops = {
  owner:          THIS_MODULE,
  open:           dh_open,
  release:        dh_close,
  read:           dh_read,
  write:          dh_write,
  poll:           dh_poll,
};


static int __init dh_init(void) { 
/** Kernel module initialization.  Check and store the driver mode,
    zero the appropriate counters, register the device major number,
    create the /proc control files.  */

  int result;
  int idev;
  int mode_ok;
  int minor;
  int icard;

  vvprintk(KERN_INFO "DH_TIMER_MSEC=%d DH_TIMER_JIF=%d "
	   "DH_TIMER_JIFFIES=%d DH_TIMER_RESTART_JIF=%d HZ=%d\n",
	   DH_TIMER_MSEC, DH_TIMER_JIF, DH_TIMER_JIFFIES, DH_TIMER_RESTART_JIF, HZ);

  printk(KERN_INFO "DOR driver (dh) "DH_DELIVERY_TAG" by jacobsen@npxdesigns.com, 2002-2006\n");

  if(dh_test_ioq()) {
    vprintk(KERN_ERR "dh: IOQ self test FAILED.\n");
    return -EIO;
  } 
  vprintk(KERN_INFO "dh: IOQ self test passed.\n");

  dh_calibrate_benchstat();

  /* Check/report verbose mode */
  vprintk(KERN_INFO "DOR Driver verbosity level is %d.\n", parm_verbose);

  vprintk(KERN_INFO "Dead DOM debugging: DH_MAX_FIFO_HOLD_TIME=%d DH_MAX_HW_PKTSIZ=%d HZ=%d\n", 
	  DH_MAX_FIFO_HOLD_TIME,DH_MAX_HW_PKTSIZ,HZ);

  if(parm_tx_dma) {
    vprintk(KERN_INFO "dh: Will transmit messages to DOR using DMA.\n");
  }

  if(parm_rx_dma) {
    vprintk(KERN_INFO "dh: Will receive messages from DOR using DMA.\n");
  }

  dh_determine_excluded_cards(parm_exclude, dh_excluded, DH_MAX_CARDS);

  /* Initialize software copies of write registers */
  dh_zero_write_regi_copies();

  /* Zero global counters */
  dh_total_msgs_read    = 0;
  dh_total_msgs_written = 0;
  
  /* Initialize softboot semaphores (forbid a/b simultaneous softboot) */
  int ipair;
  for(icard=0; icard < DH_MAX_CARDS; icard++) 
    for(ipair=0; ipair < DH_PAIRS_PER_CARD; ipair++)
      init_MUTEX(&dh_wirepairs[icard][ipair].softboot_sem);

  /* Find PCI devices */
  if(dh_get_pci_devices(dh_pci_dors) <= 0) {
    printk(KERN_ERR "dh: I couldn't find any valid DOR cards on this system!\n");
    return -ENODEV;
  }

  /* Register the major number */
  result = register_chrdev( DH_MAJOR, "dh", &dh_fops );
  if(result < 0) {
      printk(KERN_WARNING "dh: Sorry, I can't get major number %d (result %d)\n", 
	     DH_MAJOR, result);
      return result;
  }

  /* Mark all files as unopened; init file-open wait queues; reset RX timers */
  for(idev=0; idev < DH_DOMHUB_DOMS; idev++) {
    dh_opened[idev] = 0;
    dh_is_communicating[idev] = 0;
    dh_domvars[idev].nonemptyRXjif = jiffies;
    dh_domvars[idev].rxdt = 0;
    dh_domvars[idev].had_rx_data = 0;
    init_waitqueue_head(&file_open_waiter[idev]);
  }

  vprintk(KERN_INFO "dh: Allocating message buffers for all minor numbers...\n");
  
  int gtg = 1; /* Good to go */

  for(minor=0; minor < DH_DOMHUB_DOMS; minor++) {

    /* Try to allocate message buffers */
    if((inq[minor] = newIOQ()) == NULL) {
      printk(KERN_ERR "dh: Couldn't create input message queue for minor %d.\n",minor);
      dh_lasterr = DH_ERR_QCREATFAIL;
      gtg = 0; break;
    }
    
    if( (outq[minor] = newIOQ()) == NULL) {
      printk(KERN_ERR "dh: Couldn't create output message queue for minor %d.\n",minor);
      dh_lasterr = DH_ERR_QCREATFAIL;
      gtg = 0; break;
    }
    
    if( (dbgq[minor] = newDbgQ()) == NULL) {
      printk(KERN_ERR "dh: Couldn't create debug message queue for minor %d.\n", minor);
      dh_lasterr = DH_ERR_QCREATFAIL;
      gtg = 0; break;
    }
    emptyDbgQ(dbgq[minor]);

    txpkt[minor] = (u32 *) kmalloc(DH_MAX_HW_PKTSIZ, GFP_ATOMIC);
    if(txpkt[minor] == NULL) {
      printk(KERN_ERR "dh: Can't allocate space for scratch message.\n");
      dh_lasterr = DH_ERR_BUFALLOCFAIL;
      gtg = 0; break;
    }
    
    //rxpkt[minor] = (SWpkt *) kmalloc(sizeof(SWpkt), GFP_KERNEL);
    rxpkt[minor] = (u32 *) kmalloc(DH_MAX_HW_PKTSIZ, GFP_ATOMIC);
    if(rxpkt[minor] == NULL) {
      printk(KERN_ERR "dh: Can't allocate space for scratch message.\n");
      dh_lasterr = DH_ERR_BUFALLOCFAIL;
      gtg = 0; break;
    }
    
    dh_tx_scratch_msg[minor] = (ioqMsg *) kmalloc(sizeof(ioqMsg), GFP_ATOMIC);
    if(dh_tx_scratch_msg[minor] == NULL) {
      printk(KERN_ERR "dh: Can't allocate space for scratch message.\n");
      dh_lasterr = DH_ERR_BUFALLOCFAIL;
      gtg = 0; break;
    }
    
    dh_rx_scratch_msg[minor] = (ioqMsg *) kmalloc(sizeof(ioqMsg), GFP_ATOMIC);
    if(dh_rx_scratch_msg[minor] == NULL) {
      printk(KERN_ERR "dh: Can't allocate space for scratch message.\n");
      dh_lasterr = DH_ERR_BUFALLOCFAIL;
      gtg = 0; break;
    }
    
    dh_tcalib_recs[minor] = (struct dh_tcalib_t *) vmalloc(sizeof(struct dh_tcalib_t));
    if(dh_tcalib_recs[minor] == NULL) {
      printk(KERN_ERR "dh: Can't allocate space for time calib record.\n");
      dh_lasterr = DH_ERR_BUFALLOCFAIL;
      gtg = 0; break;
    }

    dh_tcalib_avail[minor] = 0;
    dh_w_cis_reqd[minor]   = NUM_W_CIS; /* Configurable via proc file */
  }

  if(!gtg) {
    dh_release_pci_devices(dh_pci_dors, DH_MAX_CARDS);
    unregister_chrdev(DH_MAJOR, "dh");
    return -ENOMEM;
  }

  /* Set (default) clock select value (external or internal) */
  for(icard=0; icard < DH_MAX_CARDS; icard++) {
    vvprintk(KERN_INFO "dh: Setting clock select for card %d to INTERNAL by default"
	     " (not actually setting bits yet).\n", icard);
    dh_clksel[icard] = parm_clock_source;
  }

  /* Load default FPGA pages where appropriate */
  if(parm_fpga_page > 0 && parm_fpga_page < DH_FLASH_PAGES) {
    for(icard=0; icard < DH_MAX_CARDS; icard++) {
      if(dh_pci_dors[icard].found) {
 	if(dh_reload_fpga_from_flash(icard, parm_fpga_page)) {
	  printk(KERN_ERR "dh: Couldn't load FPGA on card %d from flash page %d.  You'll"
		 " probably have to reboot to see this card at all.\n",
		 icard, parm_fpga_page);
	  /* Release PCI resources, etc. */
	  dh_release_pci_devices(dh_pci_dors, DH_MAX_CARDS);
	  unregister_chrdev(DH_MAJOR, "dh");
	  dh_lasterr = DH_ERR_FPGARELODFAIL;
	  return -EINVAL;
	} else {
	  vprintk(KERN_INFO "dh: Loaded FPGA on card %d from flash page %d.\n",
		  icard, parm_fpga_page);
	}
      }
    }
  } else {
    vprintk(KERN_INFO "dh: Keeping default (currently loaded) FPGA design"
	    " on all cards.\n");
  }

  
  result = dh_initialize_proc_files();
  if(result) {
    printk(KERN_WARNING "dh: Had problem initializing proc files.  Giving up.\n");
    return result;
  }

  /* Note any cards with firmware not supported by current driver */
  for(icard=0; icard < DH_MAX_CARDS; icard++) {
    if(dh_pci_dors[icard].found) {
      dh_firmware_mismatch[icard] = dh_firmware_version_check(icard);
    } else {
      dh_firmware_mismatch[icard] = 1;
    }
  }

  if(parm_unlock_pages) {
    printk(KERN_WARNING "dh: WARNING: DOR Flash page lock/unlock ENABLED.  "
	   "If page 0 is corrupted with non-functioning FPGA design, "
	   "the DOR flash will have to be programmed using a physical "
	   "(byteblaster) cable.  DO NOT USE THIS FEATURE UNLESS YOU "
	   "KNOW WHAT YOU ARE DOING.\n");
  }

  vprintk(KERN_INFO "Module dh successfully installed.\n");
  return 0;
}

/* Tell the kernel to use our initialization and cleanup routines for this module: */
module_init(dh_init);
module_exit(dh_cleanup);


/************* End module init and cleanup routines ***************/

/*                          end dh.c                              */
